#ifndef _EXAMPLESDEFINES_H
#define _EXAMPLESDEFINES_H

#ifdef WIN32
#include <sys/timeb.h>
#else
#include <unistd.h>
#include <sys/time.h>
#endif
#include <time.h>

// general defines
// use realtime for speed measurement
// (cppvmspeed, pvmspeed). Otherwise 
// processor time is used.
#define REALTIME
#define CPPVM_GET_PROCESSOR_TIME(SEC) \
  SEC = ((double)clock()) / CLOCKS_PER_SEC
#define CPPVM_DIFF_TIME(START_SEC, START_USEC, END_SEC, END_USEC, RESULT) \
{ long resultSec;                                                         \
  long resultUSec;                                                        \
  resultSec = END_SEC - START_SEC;                                        \
  /* which interval? */                                                   \
  if(END_USEC<START_USEC){                                                \
    /* new interval */                                                    \
    resultUSec = (1000000 - START_USEC) + END_USEC;                       \
    resultSec = resultSec -1;                                             \
  }                                                                       \
  else{                                                                   \
    /* old interval */                                                    \
    resultUSec = END_USEC - START_USEC;                                   \
  }                                                                       \
  RESULT = ((double)(resultUSec) / 1000000.0) + (double) resultSec;} 


// Platform specific defines
#ifdef WIN32
#define CPPVM_SLEEP(seconds) Sleep(seconds*1000) 
#define CPPVM_GET_REAL_TIME(SEC, USEC) \
{ struct _timeb timebuffer;       \
  _ftime( &timebuffer );           \
  SEC = timebuffer.time;           \
  USEC = timebuffer.millitm * 1000;}
#else
#define CPPVM_SLEEP(seconds) sleep(seconds) 
#define CPPVM_GET_REAL_TIME(SEC, USEC) \
{  struct timeval  tv;            \
  struct timezone tz;             \
  gettimeofday(&tv, &tz);         \
  SEC = tv.tv_sec;                \
  USEC = tv.tv_usec;}
#endif

#endif // _EXAMPLESDEFINES_H
