// cppvm 'adminis'
// IMPORTANT: Multicasted (array of tids) and broadcasted
//            (to group members) messages are NOT send to
//            the caller even if the caller is in the
//            array of tids resp. member of the group

#include "cppvm.h"

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  ///////////////////////
  // start servers     //
  // every server gets //
  // its own context!  //
  ///////////////////////
  
  cout << "starting servers...";
  cout.flush();
  cppvmSpawnMultiConnection servers;
  servers.add(new cppvmSpawnConnection("gimme5", "",
                                       PvmTaskDefault, "", CPPvmNewContext));
  servers.add(new cppvmSpawnConnection("sayhi", "",
                                       PvmTaskDefault, "", CPPvmNewContext));
  
  // collection of server objects
  cppvmObjectCollection *serverCollection;
  // get the pvm task descriptions of all servers
  serverCollection = servers.getAllChildTasks();
  // servers collection as distributed object
  serverCollection->cppvmDistObj("servers");
  cout << "...done." << endl;

  //////////////////////////////////////
  // start applications               //
  // all cout/cerr of the application //
  // will be redirected to stdout     //
  //////////////////////////////////////
  cppvmCatchout catchOut(CPPvmStdout);
  // catchout in file
  //cppvmCatchout catchOut(CPPvmFile, "out.txt");
  
  cout << "starting applications...";
  cout.flush();
  // string object
  cppvmString resultStr("");
  int numApps = 3;
  int numResults = 0;
  cppvmSpawnMultiConnection apps("app", numApps);
  cout << "...done." << endl;

  // receive a string from any process
  cppvmReceiveStream recStrm(CPPvmRNBany);

  // distributed object to indicate
  // server/application errors
  cppvmInt errorNotify(0);
  errorNotify.cppvmDistObj("error notify");

  bool error = false;
  bool result;
  // leave the loop if
  // 1) all results are sent back
  // 2) a server or an application
  //    has set the "error notify" != 0
  while(numResults<numApps && 
        error == false){
    recStrm >> resultStr;
    result = recStrm.received();
    if(result==true){
      numResults++;
    }
    if(errorNotify.read()!=0){
      cout << "ADMIN: critical error in server/application occurred!" << endl;
      error = true;
    }
  }

  if(error==true){
    // an application or an server has
    // caused an error
    // get the error message from message box
    // distributed objects are not restricted to contexts
    cppvmString errStr("");
    errStr.cppvmDistObj("exception error message");
    // and write it to error out
    cerr << "ADMIN: error occurred: "
         << errStr.read() << endl;
  }
  
  // kill all servers and applications
  servers.killAllChildren();
  apps.killAllChildren();
  
//   unsigned int i;
//   for(i=0; i<servers.numElems(); i++){
//     servers[i].killChild();
//   }
//   for(i=0; i<apps.numElems(); i++){
//     apps[i].killChild();
//   }

  cout << "admin: The End." << endl;

  // halt the virtual machine
  pvmConn.halt();
  
  return 0;
}







