// cppvm 'app'

#include "cppvm.h"

void
app(void)
{
  // connect to pvm
  cppvmConnection pvmConn;

  // collection of server objects
  cppvmObjectCollection serverCollection;
  // servers collection as distributed object
  serverCollection.cppvmDistObj("servers");

  cout << serverCollection << endl;

  int i;
  cppvmString orderStr("Your result, please:");
  cppvmString resStr("");
  cppvmSendStream sendStrmGimme5, sendStrmSayhi;  
  cppvmReceiveStream recStrmGimme5, recStrmSayhi;
 
  // send to 'gimme5'
  sendStrmGimme5.set(serverCollection.getTaskWithName("gimme5"),
                     CPPvmSNBtask);
  // receive blocking from any process
  // with context of 'gimme5'
  recStrmGimme5.set((serverCollection.getTaskWithName("gimme5"))
                    .getContextId(),
                    CPPvmRBany);

  // send to 'sayhi'
  sendStrmSayhi.set(serverCollection.getTaskWithName("sayhi"),
                    CPPvmSNBtask);
  // receive blocking from any process
  // with context of 'sayhi'
  recStrmSayhi.set((serverCollection.getTaskWithName("sayhi"))
                    .getContextId(),
                    CPPvmRBany);

  for(i=0; i<3; i++){
    /////////////////////////////
    // contact server 'gimme5' //
    /////////////////////////////

    // send order
    sendStrmGimme5 << orderStr;

    /////////////////////////////
    // contact server 'sayhi'  //
    /////////////////////////////

    // send order
    sendStrmSayhi << orderStr;

    // although the receive descriptors
    // accepts sent strings from any
    // processes, the strings sent by the servers
    // to the application cannot be received in
    // a false order here!
    // This is because the servers are started
    // in different contexts and message passing
    // is only possible within one context.

    // receiveFromAny result from "sayhi" server
    recStrmSayhi >> resStr;
    cout << "Result from sayhi: " << resStr << endl;

    // receiveFromAny result from "gimme5" server
    recStrmGimme5 >> resStr;
    cout << "Result from gimme5: " << resStr << endl;

  }

  ///////////////////////////////////////////////////
  // uncomment the next line to cause an exception //
  //cppvmString((char*) NULL);
  ///////////////////////////////////////////////////

  // blocking send to parent process
  cppvmSendStream sendStrmParent(pvmConn, CPPvmSBparent);

  sendStrmParent << resStr;
}

void errorHandling(void){
  // an exception occurred
  cppvmConnection pvmConn;
  // fill distributed error message
  // distributed objects are not restricted to contexts
  cppvmString errStr("");
  errStr.cppvmDistObj("exception error message");
  errStr.write("exception occurred in app");
  // signal exception to administrator
  cppvmInt indicateError;
  indicateError.cppvmDistObj("error notify");
  // set error notifier
  indicateError.write(1);
  
  // endless wait (until admin has
  // killed all processes)
  int counter=0;
  while(1){
    counter++;
  }
}

int
main(){
  try {
    app();
  }
  catch (cppvmException& exception) {                      
    cerr << exception;                                   
    errorHandling();
  }
  catch (...) {
    cerr << "exception occurred (not part of cppvm)" << endl;
    errorHandling();
  }
  return 0;
}

