// cppvm 'array test'
#include "cppvm.h"
// include array type ("../h/array.h")
#include "array.h"

void fillArray(ArrayType& array)
{
  unsigned int i;
  for(i=0;i<array.numElems();i++){
    array[i] = i+65;
  }
}

void testAndShowArray(ArrayType& array,
                      unsigned int correctSize,
                      unsigned int correctLength)
{
  cout << "Array: size: " << array.getSize()
       << " length: " << array.getLength()
       << " data: " << array << endl;
  if(array.getSize()==correctSize){
    cout << "Array size ok.";
  }
  else{
    cout << "Array size FALSE.";
  }
  if(array.getLength()==correctLength){
    cout << " Array length ok." << endl;
  }
  else{
    cout << " Array length FALSE." << endl;
  }
}


int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  // spawn child process 'array2'
  cppvmSpawnConnection child("array2");

  // receive descriptor	  
  // (blocking receive from child process)
  cppvmReceiveStream recStrm(child, CPPvmRBchild);

  // array
  // 'ArrayType' is defined in '../h/array.h'
  ArrayType array;

  testAndShowArray(array, 0, 0);

  recStrm >> array;
  testAndShowArray(array, 5, 5);

  array.setCPPvmTransferMode(CPPvmTransferModeAppend);
  recStrm >> array;
  testAndShowArray(array, 10, 10);

  ArrayType array2(3);
  fillArray(array2);
  array.append(array2);
  testAndShowArray(array, 13, 13);

  array = array2;
  testAndShowArray(array, 13, 3);
  
  array.resize(4);
  testAndShowArray(array, 13, 4);

  fillArray(array);
  testAndShowArray(array, 13, 4);

  // clean only new elements 
  // (parameter 'true')
  array.resizeClean(7, true);
  testAndShowArray(array, 13, 7);

  array.setFitSizeToLength(true);
  testAndShowArray(array, 7, 7);

  array.resizeClean(8);
  testAndShowArray(array, 8, 8);

  array.resize(6);
  fillArray(array);
  testAndShowArray(array, 6, 6);

  array = array2;
  testAndShowArray(array, 3, 3);

  array.setFitSizeToLength(false);
  array.resize(2);
  fillArray(array);
  testAndShowArray(array, 3, 2);

  array.resize(9);
  testAndShowArray(array, 9, 9);

  // halt the virtual machine
  pvmConn.halt();

  return 0;
}
