// cppvm 'send/rec' with object message tags

#include "cppvm.h"
// sleep
#include "examplesdefines.h"
// include new class 'chameleon'
#include "chameleon.h"

void
printIncomingMsgTag(cppvmReceiveStream& recStrm)
{
  int msgTag;
  recStrm.probeAny(msgTag);
  cout << "Message tag of current message in receive buffer: "
       << msgTag << endl;
}

void
printReceivedMessage(cppvmReceiveStream& recStrm, chameleon& chamObj)
{
  if(recStrm.received()){
    cout << "received object with message tag: "
         << chamObj.cppvmMsgTag() << endl;
  }
  else{
    cout << "receive for object with message tag: "
         << chamObj.cppvmMsgTag() << " failed." << endl;
  }
}

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  // object with object 
  // message tag 3
  chameleon chamObj(3);

  // spawn child process 'cham2'
  cppvmSpawnConnection child("cham2");
  
  // non-blockin receive stream
  cppvmReceiveStream recStrm(child, CPPvmRNBchild);

  // wait for message
  CPPVM_SLEEP(2);
  printIncomingMsgTag(recStrm);

  // try to receive object
  // this will fail, wrong message tag
  recStrm >> chamObj;
  printReceivedMessage(recStrm, chamObj);

  // change object message tag
  chamObj.setCPPvmObjectMsgTag(4);

  // receive object
  recStrm >> chamObj;
  printReceivedMessage(recStrm, chamObj);

  // use the class message tag
  chamObj.cppvmDeactivateObjectMsgTag();
  
  // wait for message
  CPPVM_SLEEP(2);
  printIncomingMsgTag(recStrm);

  // receive object
  recStrm >> chamObj;
  printReceivedMessage(recStrm, chamObj);

  // halt the virtual machine
  pvmConn.halt();

  return 0;
}
