#include "cppvm.h"
// CPPVM_SLEEP define
#include "examplesdefines.h"

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  unsigned int i;
  cppvmObjectCollection myColl;
  myColl.add(new cppvmString("Dummy"));

  //////////////////////////////////
  // receive collection           //
  //////////////////////////////////

  // (nested) collections can be sent/received
  // as usual

  // receive descriptor
  // (blocking receive from parent process)
  cppvmReceiveStream recStrm(pvmConn, CPPvmRBparent);

  recStrm >> myColl;

  cout << "received from parent:" << endl
       << myColl << endl;

  recStrm >> myColl;

  //////////////////////////////////
  // distributed collection       //
  //////////////////////////////////

  // register the object as a distributed object
  myColl.cppvmDistObj("distColl");

  // the (received from parent) collection has as member 
  // a string
  // add two strings
  myColl.add(new cppvmString("  >  >  >  >  >"));
  myColl.add(new cppvmString("<  <  <  <  <  "));
  // modify the collection order by swapping elements
  myColl.swap(0,1);
  // write current collection into the mailbox
  myColl.cppvmWriteMboxInfo();

  // modify the collection by adding chars to the
  // second elements (element is of class 'cppvmString')
  char msg[] = "Hello parent!";
  char *chPtr = msg;
  for(i=0;i<13;i++){
    // append a single char to the string in the collection
    ((cppvmString&)myColl[1]).append((*chPtr));
    // update distributed object
    myColl.cppvmWriteMboxInfo();
    // increase char pointer and wait a while
    chPtr++;
    CPPVM_SLEEP(1);
  }
  
  // scroll the first and last 'cppvmString' elements
  for(i=0;i<3000;i++){
    // shift strings
    ((cppvmString&)myColl[0]).cycleShiftRight();
    ((cppvmString&)myColl[2]).cycleShiftLeft();
    // update distributed object
    myColl.cppvmWriteMboxInfo();
  }

  //////////////////////////////////
  // receive collection           //
  // in different modes           //
  //////////////////////////////////

  // there are two modes to receive a collection
  // 1) clean the current collection, receive the new one (default)
  // 2) append the received collection to the current collection
  // the modes can be switched by the 'setCPPvmTransferMode' method

  myColl.setCPPvmTransferMode(CPPvmTransferModeAppend);
  recStrm >> myColl;
  cout << "received from parent process append mode:" << endl
       << myColl << endl;

  myColl.setCPPvmTransferMode(CPPvmTransferModeClean);
  recStrm >> myColl;
  cout << "received from parent process clean mode:" << endl
       << myColl << endl;

  // send descriptor
  // (nonblocking send to parent process)
  cppvmSendStream sendStrm(pvmConn, CPPvmSNBparent);

  sendStrm << myColl;
  
  return 0;
}
