#include "cppvm.h"
#include "examplesdefines.h"

#ifdef CPPvmUseStandardCIncludeFiles
#include <fstream.h>
#else
#ifdef CPPvmUseStandardCIncludeFileIostream
#include <fstream.h>
#else
#include <fstream>
#ifndef CPPvmDoNotUseStdNamespace
using std::ofstream;
#endif
#endif
#endif

// global variables
cppvmConnection *pvmConn;

int minMsgSize, maxMsgSize, msgSizeStep, intMsgMode, initSend;
cppvmInt numMsg, msgMode, route, msgEncoding;
double numMsgFactor;
cppvmString hostSpeed2, outFile;
cppvmInt handshake;

cppvmSendStream      *sendStrm;
cppvmSendStream      *paramsSendStrm;
cppvmSendStream      *handshakeSendStrm;
cppvmReceiveStream   *recStrm;
cppvmReceiveStream   *notifyRecStrm;
cppvmReceiveStream   *handshakeRecStrm;
cppvmArrayChar       *msg;
cppvmSpawnConnection *speed2;

// timer variables
#ifdef REALTIME
long startSec;
long endSec;
long startUSec;
long endUSec;
#else
double startTimeSec;
double endTimeSec;
#endif 

void
printUsageAndExit()
{
  cout << "Usage: cppvmspeed <MinMsgSize> <MaxMsgSize> <MsgSizeStep> <NumMsg> " << endl
       << "                  <MsgMode> <MsgEncoding> <Route> <HostSpeed2> <OutFile>" 
       << endl;
  cout << endl;
  cout << "       <MinMsgSize>: beginning messag size (bytes)";
  cout << endl;
  cout << "       <MaxMsgSize>: ending message size (bytes)";
  cout << endl;
  cout << "       <MsgSizeStep>: message size step (bytes)";
  cout << endl;
  cout << "       <NumMsg>: number of messages for each measurement";
  cout << endl;
  cout << "       <MsgMode>: 0: explicit message passing using non-blocking send";
  cout << endl;
  cout << "                  1: explicit message passing using blocking send";
  cout << endl;
  cout << "                  2: message mailbox using messages for synchronization";
  cout << endl;
  cout << "       <MsgEncoding>: 0: PvmDefault: XDR encoding";
  cout << endl;
  cout << "                      1: PvmDataInPlace: data left in place";
  cout << endl;
  cout << "                      2: PvmDataRaw: no encoding;" << endl;
  cout << "                                     only possible on homogeneous hosts!";
  cout << endl;
  cout << "       <Route>:   0: route messages via the pvm daemon";
  cout << endl;
  cout << "                  1: route messages directly between cppvmspeed and cppvmspeed2 (faster)";
  cout << endl;
  cout << "       <HostSpeed2>: host of the child process cppvmspeed2";
  cout << endl;
  cout << "       <OutFile>: protocol file name";
  cout << endl;

  pvmConn->halt();
}

int
convertStringToInt(char * str)
{
  int val;
  
  cppvmString string(str);
  val = (int) string.readValDouble();
  return val;
}

void
extractParameter(int argc, char *argv[])
{
  bool parseError=false;

  if(argc!=10){
    printUsageAndExit();
  }
  minMsgSize  = convertStringToInt(argv[1]);
  maxMsgSize  = convertStringToInt(argv[2]);
  msgSizeStep = convertStringToInt(argv[3]);
  numMsg      = convertStringToInt(argv[4]);
  msgMode     = convertStringToInt(argv[5]);
  intMsgMode  = (int) msgMode;
  msgEncoding = convertStringToInt(argv[6]);
  if((int)msgEncoding==1){
    initSend = PvmDataInPlace;
  }
  else if((int)msgEncoding==2){
    initSend = PvmDataRaw;
  }
  else{
    initSend = PvmDataDefault;
  }
  route = convertStringToInt(argv[7]);
  hostSpeed2  = argv[8];
  outFile     = argv[9];
  
  // check some values
  if(minMsgSize<=0||maxMsgSize<=0||
     msgSizeStep<=0||(int)numMsg<=0){
    parseError=true;
  }
  if((int)msgMode<0||(int)msgMode>2){
    parseError=true;
  }
  if((int)msgEncoding<0||(int)msgEncoding>2){
    parseError=true;
  }
  if((int)route<0||(int)route>1){
    parseError=true;
  }
  if(minMsgSize>maxMsgSize){
    parseError=true;
  }

  // exit if necessary
  if(parseError==true){
    printUsageAndExit();
  }
}

void
initMeasurement()
{
  // set the global variables
  // paramsStrm, sendStrm, recStrm, msg, speed2

  // hand over parameters to child, receive notification
  paramsSendStrm = new cppvmSendStream((*speed2), CPPvmSNBchild);
  notifyRecStrm  = new cppvmReceiveStream((*speed2),  CPPvmRBchild);

  // send route init and msgEncoding
  (*paramsSendStrm) << route << msgEncoding;
  
  msg = new cppvmArrayChar(0);

  if((int)msgMode==0){
    sendStrm = new cppvmSendStream((*speed2), CPPvmSNBchild, initSend);
    recStrm  = new cppvmReceiveStream((*speed2),  CPPvmRBchild);
    // number of messages * 2
    numMsgFactor = 2;
  }
  else if((int)msgMode==1){
    sendStrm = new cppvmSendStream((*speed2), CPPvmSBchild, initSend);
    recStrm  = new cppvmReceiveStream((*speed2),  CPPvmRBchild);
    // number of messages * 2
    numMsgFactor = 2;
  }
  else if((int)msgMode==2){
    msg->cppvmDistObj("m");
    // only number of messages
    numMsgFactor = 1;
  }
}

void
dataTransfer()
{
  if(intMsgMode==2){
    // signal speed2 to put message in data base
    (*handshakeSendStrm) << handshake;
    // wait for speed2 finished
    (*handshakeRecStrm) >> handshake;
    // read message from data base
    msg->cppvmReadMboxInfo();
  }
  else{
    (*sendStrm) << (*msg);
    (*recStrm) >> (*msg);
  }
}

void
startTimer()
{
#ifdef REALTIME
  CPPVM_GET_REAL_TIME(startSec, startUSec);
#else
  CPPVM_GET_PROCESSOR_TIME(startTimeSec);
#endif
}

double 
stopTimer()
{
#ifdef REALTIME
  double result;
  CPPVM_GET_REAL_TIME(endSec, endUSec);
  CPPVM_DIFF_TIME(startSec, startUSec, endSec, endUSec, result);
  return result;
#else
  CPPVM_GET_PROCESSOR_TIME(endTimeSec);
  return endTimeSec - startTimeSec;
#endif
}

void
measurement()
{
  int cycles;
  // calculate cycle number
  cycles = (int)((float)(maxMsgSize - minMsgSize) / (float)(msgSizeStep));
  if((cycles*msgSizeStep)+minMsgSize!=maxMsgSize){
    cycles = cycles +2;
  }
  else{
    cycles++;
  }
  

  // pass parameter to child
  (*paramsSendStrm) << msgMode << numMsg;

  // open out file
  ofstream of(outFile.getPtr());
  if(!of){
    cerr << "can not open outfile" << endl;
    printUsageAndExit();
  }
  // print header into file
  of << "# CPPvm transfer rate measurement" << endl;
  of << "# called with: " << endl;
  of << "# cppvmspeed <MinMsgSize> <MaxMsgSize> <MsgSizeStep> <NumMsg> <MsgMode>" 
     << endl;
  of << "#       <MsgEncoding> <Route> <HostSpeed2> <OutFile>" << endl;
  of << "# cppvmspeed " << minMsgSize << " " << maxMsgSize << " "
     << msgSizeStep << " " << numMsg << " " 
     << msgMode << " " << msgEncoding << " "
     << route << " " << hostSpeed2 << " " << outFile 
     << endl;
  of << "# format:" << endl;
  of << "# msgSize[byte] transferRate[MByte]" << endl;

  int i, j;
  cppvmInt msgSize;
  cppvmInt notify;
  msgSize = minMsgSize;
  double time, rate;
  int loopNum = (int)numMsg;

  for(i=0;i<cycles;i++){
    // set the current message size
    msg->resize((unsigned int)((int)msgSize));
    // send current message size to speed2
    (*paramsSendStrm) << msgSize;
    // wait for notifiy of speed2
    (*notifyRecStrm) >> notify;
    // begin time measurement

    // get start time
    startTimer();
    // transfer data
    for(j=0;j<loopNum;j++){
       dataTransfer();
    }
    // get end time
    time = stopTimer();

    if(time<=0.011){
      rate = 0.0;
      cerr << "!!! The processor time given by the function 'clock()' is probably to" 
           << endl
           << "!!! unprecise for the measurement! Use a higher number of messages" 
           << endl
           << "!!! (e.g. 100 times larger than current 'NumMsg') for better results." 
           << endl;
    }
    else{
      rate = (((double)((int)msgSize)/ (double)(1024*1024)) * 
              ((double)((int)numMsg)) * numMsgFactor)
        / time;
    }

    cout << "message size (byte): " << msgSize;
    cout << " time (s): " << time;
    cout << " rate (MByte/s): " << rate << endl;

    // write data in file
    of << (int)msgSize << " " << rate << endl;
    // calculate new msgSize
    msgSize = (int)msgSize + msgSizeStep;
    if((int)msgSize>maxMsgSize){
      msgSize = maxMsgSize;
    }
  }
}

void
main2(int argc, char *argv[])
{
  // connect to pvm
  pvmConn = new cppvmConnection();

  // convert command line parameters
  extractParameter(argc, argv);

  if((int)route==1){
    // direct connection between two tasks
    pvmConn->setOpt(PvmRoute, PvmRouteDirect);
  }
  
  // start speed2
  speed2 = new cppvmSpawnConnection("cppvmspeed2", "",
                                    PvmTaskHost, hostSpeed2.getPtr());

  // create handshake streams 
  if((int)msgMode==2){
    handshakeSendStrm = new cppvmSendStream((*speed2), CPPvmSNBchild);
    handshakeRecStrm  = new cppvmReceiveStream((*speed2),  CPPvmRBchild);
  }
  
  // init evaluation
  initMeasurement();

  // evaluate transfer rate
  measurement();
  
  // halt the virtual machine
  pvmConn->halt();
}

int
main(int argc, char *argv[])
{
  // catch all possible exceptions
  CPPvmTryAll(main2(argc, argv))
  return 0;
}

