#include "cppvm.h"

#ifdef CPPvmUseStandardCIncludeFiles
#include <fstream.h>
#else
#ifdef CPPvmUseStandardCIncludeFileIostream
#include <fstream.h>
#else
#include <fstream>
#ifndef CPPvmDoNotUseStdNamespace
using std::ofstream;
using std::ifstream;
#endif
#endif
#endif

cppvmString pvmInFile, cppvmInFile, outFile;

void
printUsageAndExit()
{
  cout << "Usage: cppvmvspvmmerge <PvmMeasuredSpeedInFile> <CPPvmMeasuredSpeedInFile> <OutFile>" << endl;

  exit(1);
}

void
extractParameter(int argc, char *argv[])
{
  if(argc!=4){
    printUsageAndExit();
  }
  pvmInFile  = argv[1];
  cppvmInFile  = argv[2];
  outFile = argv[3];
}

bool
getNextData(ifstream &inFile, int &msgSize, float& rate)
{
  char line[1024];
  line[0]='\0';

  if(inFile.eof()){
    return false;
  }
  while(true){
    inFile.getline(&line[0], 1024);
    
    if(line[0]!='#' && line[0]!='\0'){
      int num;
      num = sscanf(line, "%d %f", &msgSize, &rate);
      if(num!=2){
        return false;
      }
      else{
        return true;
      }
    }
    if(inFile.eof()){
      return false;
    }
  }
}

void mergeMeasurementResults()
{
  int msgSizePvm, msgSizeCPPvm;
  float ratePvm, rateCPPvm, ratio;
  double accRatio=0.0;
  int numRatio=0;
  
  // open pvm in file
  ifstream ifPvm(pvmInFile.getPtr());
  if(!ifPvm.is_open()){
    cerr << "can not open file: " << pvmInFile << endl;
    exit(1);
  }
  // open cppvm in file
  ifstream ifCPPvm(cppvmInFile.getPtr());
  if(!ifCPPvm.is_open()){
    cerr << "can not open file: " << cppvmInFile << endl;
    exit(1);
  }

  // open out file
  ofstream of(outFile.getPtr());
  if(!of.is_open()){
    cerr << "can not open outfile" << endl;
    exit(1);
  }
  // print header into file
  of << "# CPPvm vs PVM transfer rate measurement" << endl;
  of << "# called with: " << endl;
  of << "# cppvmvspvmmerge <PvmMeasuredSpeedInFile> "
     << "<CPPvmMeasuredSpeedInFile> <OutFile>" << endl;
  of << "# cppvmvspvmmerge " << pvmInFile << " " 
     << cppvmInFile << " " << outFile << endl;
  of << "# format:" << endl;
  of << "# msgSize[byte] transferRatePvm[MByte] transferRateCPPvm[MByte] " 
     << "transferRateCPPvm/transferRatePvm" << endl;

  bool readPvmOk=true;
  bool readCPPvmOk=true;
  bool writeOk=true;
  accRatio = 0.0;
  accRatio = 0.0;

  while(readPvmOk && readCPPvmOk && writeOk){
    
    readPvmOk = getNextData(ifPvm, msgSizePvm, ratePvm);
    if(readPvmOk){
      readCPPvmOk = getNextData(ifCPPvm, msgSizeCPPvm, rateCPPvm);
      if(readCPPvmOk){
        if(msgSizePvm!=msgSizeCPPvm){
          cerr << "cppvmvspvmmerge: message size in PVM file ("
               << msgSizePvm << ") differs from CPPvm file ("
               << msgSizeCPPvm << ")! Sorry, no merge possible." << endl;
          exit(1);
        }
        // write data in file
        if(ratePvm>0.0){
          ratio = (rateCPPvm/ratePvm);
        }
        else{
          ratio = 0.0;
        }
        accRatio = accRatio + (double)ratio;
        numRatio++;
        of << (int)msgSizePvm << " " << ratePvm << " " << rateCPPvm 
           << " " << ratio << endl;
        writeOk = !(of.bad());
      }
    }
  }
  of << "# average ratio (CPPvm/Pvm): " 
       << accRatio / (double)numRatio
       << endl;
  

}


int
main(int argc, char *argv[])
{
  // convert command line parameters
  extractParameter(argc, argv);

  // merge file data
  mergeMeasurementResults();
  
  return 0;
}

