// transfer external (non-CPPvm) classes

#include "cppvm.h"
// external class
#include "point.h"
// CPPvm wrapper class
#include "cppvmpoint.h"
// mulitple inheritance class
#include "cppvmderivedpoint.h"

#ifndef CPPvmDoNotUseStdNamespace
// stl namespace
using namespace std;
#endif

void printVector(vector<point>& data)
{
  vector<point>::iterator i;
  for (i=data.begin(); i != data.end() ; i++) {
    cout << *i << " ";
  }
}

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  cppvmPoint p;
  cppvmDerivedPoint dp;
  vector<point> vectP;

  // spawn child process, redirect output
  cppvmCatchout catchOut(CPPvmStdout);
  cppvmSpawnConnection child("external2");
  
  /////////////////////////
  // distributed object  //
  /////////////////////////

  // register the objects 
  // as a distributed objects
  p.cppvmDistObj("point");
  dp.cppvmDistObj("derivedPoint");

  int i;
  for(i=0;i<500;i++){
    // update the local values with the values
    // in the message box
    p.cppvmReadMboxInfo();
    dp.cppvmReadMboxInfo();
    cout << '\r' 
         << p << " and: "
         << dp;
    cout.flush(); 
  }
  cout << endl;
  
  /////////////////////////
  // send/receive object //
  /////////////////////////

  // receive descriptor	  
  // (blocking receive 
  //  from child process)
  cppvmReceiveStream recStrm(child, 
                             CPPvmRBchild);

  // receive data from child process
  recStrm >> p >> dp >> vectP;
  
  cout << "received objects: " 
       << p << " and: "
       << dp << " and: ";
  printVector(vectP);
  cout << endl;

  // halt the virtual machine
  pvmConn.halt();

  return 0;
}


