// example how to write your own send/receivable classes
// class 'group' implementation

// include specification file
#include "group.h"

// static variables
CPPvmMethodsDefinitionVariable(groupClass, CPPvmMsgTagClassGroup)

// default constructor
groupClass::groupClass()
    : cppvmObject()
{
  groupId   = 0;
  groupName = "NoGroup";
}

// copy constructor
groupClass::groupClass (const groupClass& obj)
    : cppvmObject(obj)
{
  groupName  = obj.groupName;
  groupId    = obj.groupId;
}

// destructor
groupClass::~groupClass()
{
}

// operator=
groupClass&
groupClass::operator=(const groupClass& obj)
{
  // same object?
  if(cppvmIsSame(obj)){
    return *this;
  }
  // operator= of base class
  cppvmObject::operator=(obj);

  groupName  = obj.groupName;
  groupId    = obj.groupId;
  
  return *this;
}

// operator==
bool
groupClass::operator== (const groupClass& obj)
{
  return (cppvmObject::operator==(obj) &&
	  groupId    == obj.groupId  &&
	  groupName  == obj.groupName);
}

// operator==
bool
groupClass::operator== (const groupClass& obj) const
{
  return (cppvmObject::operator==(obj) &&
	  groupId    == obj.groupId  &&
	  groupName  == obj.groupName);
}


// cppvmTransfer
// used by send/receive/distributed/mailbox messages of this class
void
groupClass::cppvmTransfer()
{
  // standard 'cppvmString' class
  groupName.cppvmTransfer();

  // transfer standard types (int, float, double, ...)
  // with the functions 'cppvmTransfer<Type>'
  cppvmTransferInt(&groupId);

}

// cppvmPrintOn 
// (called when using the << stream operator "cout << obj;")
void
groupClass::cppvmPrintOn(ostream& strm)
{
  strm << "group: " 
       << " name: " << groupName
       << " id: " << groupId ;
}

// cppvmScanFrom
// (called when using the >> stream operator "cin >> obj;")
void
groupClass::cppvmScanFrom(istream& strm)
{
  strm >> groupName;
  strm >> groupId;
}

void 
groupClass::setName(cppvmString groupName_){
  groupName = groupName_;
}

void
groupClass::setId(int groupId_){
  groupId = groupId_;
}

