// transfer external (non-CPPvm) classes

#include "cppvm.h"
// external class
#include "hiddenpoint.h"
// CPPvm wrapper classes
#include "cppvmhiddenpoint.h"
#include "cppvmanotherhiddenpoint.h"

#ifndef CPPvmDoNotUseStdNamespace
// stl namespace
using namespace std;
#endif

void printVector(vector<hiddenPoint>& data)
{
  vector<hiddenPoint>::iterator i;
  for (i=data.begin(); i != data.end() ; i++) {
    cout << *i << " ";
  }
}

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  cppvmHiddenPoint p;
  cppvmAnotherHiddenPoint ap;
  vector<hiddenPoint> vectP;

  // spawn child process, redirect output
  cppvmCatchout catchOut(CPPvmStdout);
  cppvmSpawnConnection child("hiddenexternal2");
  
  /////////////////////////
  // distributed object  //
  /////////////////////////

  // register the objects 
  // as a distributed objects
  p.cppvmDistObj("p");
  ap.cppvmDistObj("ap");

  int i;
  for(i=0;i<500;i++){
    // update the local value with the value
    // in the message box
    p.cppvmReadMboxInfo();
    ap.cppvmReadMboxInfo();
    cout << '\r' 
         << p << " and: "
         << ap;
    cout.flush(); 
  }
  cout << endl;
  
  /////////////////////////
  // send/receive object //
  /////////////////////////

  // receive descriptor	  
  // (blocking receive 
  //  from child process)
  cppvmReceiveStream recStrm(child, 
                             CPPvmRBchild);

  // receive data from child process
  recStrm >> p >> ap >> vectP;
  
  cout << "received objects: " 
       << p << " and: "
       << ap << " and: ";
  printVector(vectP);
  cout << endl;

  // halt the virtual machine
  pvmConn.halt();

  return 0;
}


