// cppvm 'mcast'
// IMPORTANT: Multicasted (array of tids) and broadcasted
//            (to group members) messages are NOT send to
//            the caller even if the caller is in the
//            array of tids resp. member of the group

#include "cppvm.h"

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;
  
  // double array object
  cppvmArrayDouble douArrObj(4);
  // string object
  cppvmString stringObj("Hi child!");

  // direct all couts to stdout
  cppvmCatchout catchOut(CPPvmStdout);
  // spawn n children processes 'mcast2'
  cppvmSpawnMultiConnection children("mcast2", 5);

  // send descriptor	  
  cppvmSendStream sendStrm;

  ///////////////////////
  // multicast example //
  ///////////////////////
  unsigned int i;

  // tag all children except the last one
  for(i=0; i< children.numElems()-1; i++){
    children.tagElem(i);
  }

  // blocking send to all tagged child processes
  sendStrm.set(children, CPPvmSBtaggedChildren);
  
  sendStrm << douArrObj;

  // nonblocking send to the last (untagged) child
  // process in the collection
  sendStrm.set(children[children.numElems()-1], CPPvmSNBchild);
  sendStrm << douArrObj;

  // blocking send to all children in collection
  sendStrm.set(children, CPPvmSBallChildren);
  sendStrm << douArrObj;

  /////////////////////////////
  // group broadcast example //
  /////////////////////////////
    
  // add this process to a specified group
  pvmConn.joinGroup("my group name");
 
  // nonblocking send to all group members
  // groups can be dynamically changed (join, leave)
  // so a blocking send (wait for a handshake of the current
  // group members) can be dangerous
  sendStrm.set("my group name", CPPvmSNBgroup);
  sendStrm << stringObj;

  // receive descriptor	  
  cppvmReceiveStream recStrm;
  // blocking receive from all tagged child processes
  for(i=0; i< children.numElems(); i++){
    recStrm.set(children[i], CPPvmRBchild);
    recStrm >> stringObj;
    cout << "received from child number "
         << i << ": " << stringObj << endl;
  }
    
  cout << "The End." << endl;

  // halt the virtual machine
  pvmConn.halt();

  return 0;
}
