// cppvm 'send/rec/distributed of a own class'

#include "cppvm.h"
// include own classes 'userClass' and 'userKeyClass'
#include "user.h"
#include "userkey.h"

// return user id of an object
int
userId(cppvmObject* obj)
{
  int msgTag;
  msgTag = obj->cppvmMsgTag();
  if(msgTag != CPPvmMsgTagClassUser &&
     msgTag != CPPvmMsgTagClassUserkey){
    // not of class 'userClass' or 'userKeyClass' 
    // return dummy value
    return 0;
  }
  else{
    // read the id
    return ((userClass*)obj)->id;
  }
}

// compare function (descending) for two members of a collection
bool 
compareDescending(cppvmObject* obj1, cppvmObject* obj2){
  int val1, val2;

  val1 = userId(obj1);
  val2 = userId(obj2);

  if(val1<val2){
    return true;
  }
  else{
    return false;
  }
}

// compare function (ascending) for two members of a collection
bool 
compareAscending(cppvmObject* obj1, cppvmObject* obj2){
  return !(compareDescending(obj1, obj2));
}

int
main()
{
  int i;
  // objects of own classes
  userClass user;
  userKeyClass userKey;

  // connect to pvm
  cppvmConnection pvmConn;

  // direct all couts to stdout
  cppvmCatchout catchOut(CPPvmStdout);
  // spawn child process 'myclass2'
  cppvmSpawnConnection child("myclass2");
  
  /////////////////////////
  // send/receive object //
  /////////////////////////

  // receive descriptor	  
  // (blocking receive from child process)
  cppvmReceiveStream recStrm(child, CPPvmRBchild);

  // receive messages from child process
  recStrm >> user >> userKey;
  
  cout << user << endl;
  cout << userKey << endl;

  /////////////////////////
  // distributed objects //
  /////////////////////////

  // register the objects as a distributed objects
  user.cppvmDistObj("userObject");
  userKey.cppvmDistObj("userKeyObject");

  for(i=0;i<500;i++){
    // update the local value with the value
    // in the message box
    user.cppvmReadMboxInfo();
    userKey.cppvmReadMboxInfo();
    // show only one object
    cout << '\r' << user;
    cout.flush(); 
  }
  cout << endl;

  ////////////////////////////////
  // sort collection            //
  ////////////////////////////////

  // collection for objects of 
  // 'cppvmObject' and derived classes
  cppvmObjectCollection myColl;

  // computer developers (well, not a complete collection)
  myColl.add(new userClass("zuse", 6, "computer", 9));
  myColl.add(new userClass("schickard", 1, "computer", 9));
  myColl.add(new userClass("hollerith", 5, "computer", 9));
  myColl.add(new userClass("leibniz", 3, "computer", 9));
  myColl.add(new userClass("pascal", 2, "computer", 9));
  myColl.add(new userClass("babbage", 4, "computer", 9));
  myColl.add(new userClass("aiken", 7, "computer", 9));

  cout << "Collection before sort descending" << endl
       << myColl << endl;

  // a compare function which can compare all members
  // of the collection can be used to sort the collection
  // the compare function must be of type
  // 'bool myCompare(cppvmObject* obj1, cppvmObject* obj2)'
  myColl.sort(compareDescending);
  cout << "Collection after sort descending" << endl
       << myColl << endl;

  // sort again
  myColl.sort(compareAscending);
  cout << "Collection after sort ascending" << endl
       << myColl << endl;

  ////////////////////////////////
  // nested collection receive  //
  ////////////////////////////////
  
  // enable the collection to receive 'userClass' objects
  // the function is defined in 'user.h', implemented in 'user.cpp'
  myColl.addRecFct(userClassRecColl);

  // enable the collection to receive 'userKeyClass' objects
  // the function is defined in 'userkey.h', implemented in 'userkey.cpp'
  myColl.addRecFct(userKeyClassRecColl);

  // receive message from child process
  recStrm >> myColl;
  
  cout << "received collection: " << myColl << endl;  

  /////////////////////////
  // cin object          //
  /////////////////////////
  userKeyClass userKey2;

  cout << "enter new userKey: name(string) <RET> id(int) groupName(string) <RET> groupId(int) keyLen(int) keys('keyLen' floats) <RET>:"
       << endl;
  cin >> userKey2;
  cout << userKey2 << endl;

  cout << endl << "The End" << endl;

  // halt the virtual machine
  pvmConn.halt();

  return 0;
}
