// cppvm 'notify'
#include "cppvm.h"
// sleep
#include "examplesdefines.h"

int
main()
{
  unsigned int i;
  // connect to pvm
  cppvmConnection pvmConn;

  // message passing streams
  // for notification messages
  cppvmReceiveStream recStrmB(CPPvmRBany);
  cppvmReceiveStream recStrmNB(CPPvmRNBany);
  cppvmReceiveStream recStrmT(CPPvmRTany, 0, 500000);

  ///////////////////////////////////
  // notification for exited tasks //
  ///////////////////////////////////

  // spawn child processes 'notify2'
  cppvmSpawnMultiConnection children("notify2",3);

  // notification for task exits
  // message tag: 111
  cppvmNotificationTaskExit tasksExitNotify(111, children);

  // wait for notification messages
  for(i=0;i<children.numElems();i++){
    while((recStrmT >> tasksExitNotify).received()==false){
      cout << "notify: receive timout..." << endl;
    }
    cout << "Task with id: " << tasksExitNotify.getExitTid()
         << " ended." << endl;
  }

  ///////////////////////////////////
  // notification for killed tasks //
  ///////////////////////////////////

  // spawn child process 'notify2'
  cppvmSpawnConnection child("notify2");

  // notification for killed task
  // message tag: 222
  cppvmNotificationTaskExit taskExitNotify(222, child);

  child.killChild();

  // wait for notification messages
  recStrmB >> taskExitNotify;
  cout << "Task with id: " << taskExitNotify.getExitTid()
       << " killed." << endl;

  
  ///////////////////////////////////
  // notifications for added hosts //
  ///////////////////////////////////
  
  // notify for new hosts
  // message tag: 333
  cppvmNotificationHostAdd hostAddNotify(333, -1);
  
  // replace the next line by a host
  // of your choice
  char *hostName = "lin1";
  pvmConn.addHost(hostName);
  CPPVM_SLEEP(3);

  recStrmNB >> hostAddNotify;
  if(recStrmNB.received()){
    cppvmArrayInt hostDaemonTids;
    hostDaemonTids = hostAddNotify.getAddedHostDaemonTids();
    cout << "Hosts added:";
    for(i=0;i<hostDaemonTids.numElems();i++){
      cout << " name: "
           << pvmConn.getHostName(hostDaemonTids[i])
           << " daemon tid: "
           << hostDaemonTids[i];
    }
    cout << "." << endl;
  }

  ////////////////////////////////////
  // notifications for delted hosts //
  ////////////////////////////////////
  
  // notify for deleted host
  // message tag: 444
  cppvmNotificationHostDelete hostDeleteNotify(444, hostName);
  
  pvmConn.delHost(hostName);
  CPPVM_SLEEP(3);
  
  recStrmNB >> hostDeleteNotify;
  if(recStrmNB.received()){
    cout << "Host with PVM daemon tid: "
         << hostDeleteNotify.getDeletedHostDaemonTid()
         << " was deleted."
         << endl;
  }

  // halt the virtual machine
  pvmConn.halt();

  return 0;
}
