#include "pvm3.h"
#include "examplesdefines.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define PVMSPEED_MAX_STRING_LENGTH   1024
#define PVMSPEED_MSG_TAG_PARAM       1
#define PVMSPEED_MSG_TAG_DATA        2
#define PVMSPEED_MSG_TAG_HANDSHAKE   3
#define PVMSPEED_MAILBOX_OBJECT_NAME "m"

// global variables
int myTid;
int childTid;
int recBufid;

int minMsgSize, maxMsgSize, msgSizeStep, initSend;
int numMsg, msgMode, route, msgEncoding;
double numMsgFactor;
char hostSpeed2[PVMSPEED_MAX_STRING_LENGTH], outFile[PVMSPEED_MAX_STRING_LENGTH];
int handshake;

int currentMsgSize;
char *msg;

// timer variables
#ifdef REALTIME
long startSec;
long endSec;
long startUSec;
long endUSec;
#else
double startTimeSec;
double endTimeSec;
#endif 

FILE *of;

void
pvmPackCheck(int packResult)
{
  if(packResult<0){
    printf("pvmPackCheck: error while pack/unpack message!\n");
  }
}


void
printUsageAndExit()
{
  printf("Usage: pvmspeed <MinMsgSize> <MaxMsgSize> <MsgSizeStep> <NumMsg>\n");
  printf("                <MsgMode> <MsgEncoding> <Route> <HostSpeed2> <OutFile>}n");
  printf("\n       <MinMsgSize>: beginning messag size (bytes)");
  printf("\n       <MaxMsgSize>: ending message size (bytes)");
  printf("\n       <MsgSizeStep>: message size step (bytes)");
  printf("\n       <NumMsg>: number of messages for each measurement");
  printf("\n       <MsgMode>: 0: explicit message passing using non-blocking send");
  printf("\n                  1: explicit message passing using blocking send");
  printf("\n                  2: message mailbox using messages for synchronization");
  printf("\n       <MsgEncoding>: 0: PvmDefault: XDR encoding");
  printf("\n                      1: PvmDataInPlace: data left in place");
  printf("\n                      2: PvmDataRaw: no encoding;\n");
  printf("                                     only possible on homogeneous hosts!");
  printf("\n       <Route>:   0: route messages via the pvm daemon");
  printf("\n                  1: route messages directly between pvmspeed and pvmspeed2 (faster)");
  printf("\n       <HostSpeed2>: host of the child process pvmspeed2");
  printf("\n       <OutFile>: protocol file name\n");

  pvm_halt();
}

int
convertStringToInt(char * str)
{
  int val=0;

  sscanf(str, "%d", &val);
  return val;
}

void
extractParameter(int argc, char *argv[])
{
  int parseError=0;

  if(argc!=10){
    printUsageAndExit();
  }
  minMsgSize  = convertStringToInt(argv[1]);
  maxMsgSize  = convertStringToInt(argv[2]);
  msgSizeStep = convertStringToInt(argv[3]);
  numMsg      = convertStringToInt(argv[4]);
  msgMode     = convertStringToInt(argv[5]);
  msgEncoding = convertStringToInt(argv[6]);
  if((int)msgEncoding==1){
    initSend = PvmDataInPlace;
  }
  else if((int)msgEncoding==2){
    initSend = PvmDataRaw;
  }
  else{
    initSend = PvmDataDefault;
  }
  route = convertStringToInt(argv[7]);
  strncpy(hostSpeed2, argv[8], PVMSPEED_MAX_STRING_LENGTH);
  strncpy(outFile, argv[9], PVMSPEED_MAX_STRING_LENGTH);
  
  // check some values
  if(minMsgSize<=0||maxMsgSize<=0||
     msgSizeStep<=0||(int)numMsg<=0){
    parseError=1;
  }
  if((int)msgMode<0||(int)msgMode>2){
    parseError=1;
  }
  if((int)msgEncoding<0||(int)msgEncoding>2){
    parseError=1;
  }
  if((int)route<0||(int)route>1){
    parseError=1;
  }
  if(minMsgSize>maxMsgSize){
    parseError=1;
  }

  // exit if necessary
  if(parseError==1){
    printUsageAndExit();
  }
}

void
initMeasurement()
{
  // hand over parameters to child
  // send route init and msgEncoding
  pvm_initsend(PvmDataDefault);
  pvmPackCheck(pvm_pkint(&route, 1, 1));
  pvmPackCheck(pvm_pkint(&msgEncoding, 1, 1));
  pvmPackCheck(pvm_pkint(&maxMsgSize, 1, 1));
  pvm_send(childTid, PVMSPEED_MSG_TAG_PARAM);

  // allocate data
  msg = (char*) malloc(sizeof(char) * maxMsgSize);

  if(msgMode==2){
    // only number of messages
    numMsgFactor = 1;
  }
  else{
    // number of messages * 2
    numMsgFactor = 2;
  }
}

void
dataTransfer()
{

  if(msgMode==0){
    // send non-blocking, receive blocking
    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkbyte(msg, currentMsgSize, 1));
    pvm_send(childTid, PVMSPEED_MSG_TAG_DATA);

    recBufid = pvm_recv(childTid, PVMSPEED_MSG_TAG_DATA);
    pvmPackCheck(pvm_upkbyte(msg, currentMsgSize, 1));
  }
  else if(msgMode==1){
    // send blocking, receive blocking
    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkbyte(msg, currentMsgSize, 1));
    pvm_send(childTid, PVMSPEED_MSG_TAG_DATA);

    // simulate blocking send with handshake
    recBufid = pvm_recv(childTid, PVMSPEED_MSG_TAG_HANDSHAKE);
    pvmPackCheck(pvm_upkint(&handshake, 1, 1));

    recBufid = pvm_recv(childTid, PVMSPEED_MSG_TAG_DATA);
    pvmPackCheck(pvm_upkbyte(msg, currentMsgSize, 1));

    // simulate blocking send with handshake
    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkint(&handshake, 1, 1));
    pvm_send(childTid, PVMSPEED_MSG_TAG_HANDSHAKE);
  }
  else{
    // signal speed2 to put message in message mailbox
    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkint(&handshake, 1, 1));
    pvm_send(childTid, PVMSPEED_MSG_TAG_HANDSHAKE);

    // wait for speed2 finished
    recBufid = pvm_recv(childTid, PVMSPEED_MSG_TAG_HANDSHAKE);
    pvmPackCheck(pvm_upkint(&handshake, 1, 1));

    // read message from data base
    recBufid = pvm_recvinfo(PVMSPEED_MAILBOX_OBJECT_NAME, 0, PvmMboxDefault);
    pvmPackCheck(pvm_upkbyte(msg, currentMsgSize, 1));
  }
}

void
startTimer()
{
#ifdef REALTIME
  CPPVM_GET_REAL_TIME(startSec, startUSec);
#else
  CPPVM_GET_PROCESSOR_TIME(startTimeSec);
#endif
}

double 
stopTimer()
{
#ifdef REALTIME
  double result;
  CPPVM_GET_REAL_TIME(endSec, endUSec);
  CPPVM_DIFF_TIME(startSec, startUSec, endSec, endUSec, result);
  return result;
#else
  CPPVM_GET_PROCESSOR_TIME(endTimeSec);
  return endTimeSec - startTimeSec;
#endif
}

void
measurement()
{
  int cycles;
  int i, j;
  double time, rate;
  int loopNum = numMsg;

  // calculate cycle number
  cycles = (int)((float)(maxMsgSize - minMsgSize) / (float)(msgSizeStep));
  if((cycles*msgSizeStep)+minMsgSize!=maxMsgSize){
    cycles = cycles +2;
  }
  else{
    cycles++;
  }
  

  // hand over parameters to child
  // send route init and msgEncoding
  pvm_initsend(PvmDataDefault);
  pvmPackCheck(pvm_pkint(&msgMode, 1, 1));
  pvmPackCheck(pvm_pkint(&numMsg, 1, 1));
  pvm_send(childTid, PVMSPEED_MSG_TAG_PARAM);

  // open out file
  of = fopen(outFile, "w");
  if(of==NULL){
    printf("can not open outfile\n");
    printUsageAndExit();
  }
  // print header into file
  fprintf(of, "# CPPvm transfer rate measurement\n");
  fprintf(of, "# called with: \n");
  fprintf(of, "# cppvmspeed <MinMsgSize> <MaxMsgSize> <MsgSizeStep>");
  fprintf(of, " <NumMsg> <MsgMode>\n");
  fprintf(of, "#       <MsgEncoding> <Route> <HostSpeed2> <OutFile>\n");
  fprintf(of, "# cppvmspeed %d %d %d %d %d %d %d %s %s\n",
          minMsgSize, maxMsgSize, msgSizeStep, numMsg, msgMode,
          msgEncoding, route, hostSpeed2, outFile);
  fprintf(of, "# format:\n");
  fprintf(of, "# msgSize[byte] transferRate[MByte]\n");

  currentMsgSize = minMsgSize;
  for(i=0;i<cycles;i++){
    // send current message size to speed2
    pvm_initsend(PvmDataDefault);
    pvmPackCheck(pvm_pkint(&currentMsgSize, 1, 1));
    pvm_send(childTid, PVMSPEED_MSG_TAG_PARAM);

    // wait for notifiy of speed2
    // simulate blocking send with handshake
    recBufid = pvm_recv(childTid, PVMSPEED_MSG_TAG_HANDSHAKE);
    pvmPackCheck(pvm_upkint(&handshake, 1, 1));

    // begin time measurement
    
    // get start time
    startTimer();
    // transfer data
    for(j=0;j<loopNum;j++){
       dataTransfer();
    }
    // get end time
    time = stopTimer();

    if(time<=0.011){
      rate = 0.0;
      printf("!!! The processor time given by the function 'clock()' is probably to\n");
      printf("!!! unprecise for the measurement! Use a higher number of messages\n");
      printf("!!! (e.g. 100 times larger than current 'NumMsg') for better results.\n");
    }
    else{
      rate = (((double)((int)currentMsgSize)/ (double)(1024*1024)) * 
              ((double)((int)numMsg)) * numMsgFactor)
        / time;
    }

    printf("message size (byte): %d", currentMsgSize);
    printf(" time (s): %f", time);
    printf(" rate (MByte/s): %f\n", rate);

    // write data in file
    fprintf(of, "%d %f\n", currentMsgSize, rate);
    // calculate new msgSize
    currentMsgSize = currentMsgSize + msgSizeStep;
    if(currentMsgSize>maxMsgSize){
      currentMsgSize = maxMsgSize;
    }
  }
  fclose(of);
}

int
main(int argc, char *argv[])
{
  int info;
  int oldval;
  int numTasks;
  char *argsArray[1];

  // dummy argument
  argsArray[0] = NULL;
  
  // suppress PVM error messages
  // (e.g. "libpvm [t40002]: pvm_start_pvmd(): Duplicate host") 
  oldval = pvm_setopt(PvmAutoErr, 0);
  
  info = pvm_start_pvmd(0, argsArray, 1);

  // allow PVM error messages again
  pvm_setopt(PvmAutoErr, oldval);

  // connect to pvm
  myTid = pvm_mytid();

  // convert command line parameters
  extractParameter(argc, argv);

  if((int)route==1){
    // direct connection between two tasks
    pvm_setopt(PvmRoute, PvmRouteDirect);
  }

  // If the local host is added PVM
  // can have some problems if the
  // function "pvm_addhosts" is called
  // to fast after "pvm_start_pvmd".
  // Wait a second to avoid that.
  CPPVM_SLEEP(1);

  // add host
  argsArray[0] = hostSpeed2;
  pvm_addhosts(argsArray, 1, &info);
  
  // start speed2
  numTasks = pvm_spawn("pvmspeed2", (char**)0, PvmTaskHost, 
                       hostSpeed2, 1, &childTid);

  if(numTasks!=1){
    printf("pvmspeed: could not start pvmspeed2!\n");
    // halt the virtual machine
    pvm_halt();
    return 0;
  }
  
  // init evaluation
  initMeasurement();

  // evaluate transfer rate
  measurement();
  
  // halt the virtual machine
  pvm_halt();

  return 0;
}


