#include "pvm3.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define PVMSPEED_MAX_STRING_LENGTH   1024
#define PVMSPEED_MSG_TAG_PARAM       1
#define PVMSPEED_MSG_TAG_DATA        2
#define PVMSPEED_MSG_TAG_HANDSHAKE   3
#define PVMSPEED_MAILBOX_OBJECT_NAME "m"

// global variables
// global variables
int myTid;
int parentTid;
int recBufid;

int minMsgSize, maxMsgSize, msgSizeStep, initSend;
int numMsg, msgMode, route, msgEncoding;

char *msg;
int currentMsgSize;
int handshake;

void
pvmPackCheck(int packResult)
{
  if(packResult<0){
    printf("pvmPackCheck: error while pack/unpack message!\n");
  }
}

void
initMeasurement()
{
  // receive route and msgEncoding
  recBufid = pvm_recv(parentTid, PVMSPEED_MSG_TAG_PARAM);
  pvmPackCheck(pvm_upkint(&route, 1, 1));
  pvmPackCheck(pvm_upkint(&msgEncoding, 1, 1));
  pvmPackCheck(pvm_upkint(&maxMsgSize, 1, 1));

  // evaluate route
  if(route==1){
    // direct connection between two tasks
    pvm_setopt(PvmRoute, PvmRouteDirect);
  }

  // evaluate msgEncoding
  if(msgEncoding==1){
    initSend = PvmDataInPlace;
  }
  else if(msgEncoding==2){
    initSend = PvmDataRaw;
  }
  else{
    initSend = PvmDataDefault;
  }

  // allocate data
  msg = (char*) malloc(sizeof(char) * maxMsgSize);
  
  // receive some parameters
  recBufid = pvm_recv(parentTid, PVMSPEED_MSG_TAG_PARAM);
  pvmPackCheck(pvm_upkint(&msgMode, 1, 1));
  pvmPackCheck(pvm_upkint(&numMsg, 1, 1));

}

void
dataTransfer()
{
  // evaluate msgMode
  if(msgMode==0){
    // receive blocking, send non-blocking
    recBufid = pvm_recv(parentTid, PVMSPEED_MSG_TAG_DATA);
    pvmPackCheck(pvm_upkbyte(msg, currentMsgSize, 1));

    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkbyte(msg, currentMsgSize, 1));
    pvm_send(parentTid, PVMSPEED_MSG_TAG_DATA);
  }
  else if(msgMode==1){
    // receive blocking, send non-blocking
    recBufid = pvm_recv(parentTid, PVMSPEED_MSG_TAG_DATA);
    pvmPackCheck(pvm_upkbyte(msg, currentMsgSize, 1));

    // simulate blocking send with handshake
    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkint(&handshake, 1, 1));
    pvm_send(parentTid, PVMSPEED_MSG_TAG_HANDSHAKE);

    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkbyte(msg, currentMsgSize, 1));
    pvm_send(parentTid, PVMSPEED_MSG_TAG_DATA);

    // simulate blocking send with handshake
    recBufid = pvm_recv(parentTid, PVMSPEED_MSG_TAG_HANDSHAKE);
    pvmPackCheck(pvm_upkint(&handshake, 1, 1));
  }
  else{
    // wait for pvmspeed to write data
    recBufid = pvm_recv(parentTid, PVMSPEED_MSG_TAG_HANDSHAKE);
    pvmPackCheck(pvm_upkint(&handshake, 1, 1));

    // write message to data base
    pvm_initsend(PvmDataDefault);
    pvmPackCheck(pvm_pkbyte(msg, currentMsgSize, 1));
    pvm_putinfo(PVMSPEED_MAILBOX_OBJECT_NAME, pvm_getsbuf(), PvmMboxDefault);

    // signal speed to get message
    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkint(&handshake, 1, 1));
    pvm_send(parentTid, PVMSPEED_MSG_TAG_HANDSHAKE);
  }
}

void
measurement()
{
  int i;
  int loopNum = numMsg;
  
  while(1){
    // receive current message size 
    recBufid = pvm_recv(parentTid, PVMSPEED_MSG_TAG_PARAM);
    pvmPackCheck(pvm_upkint(&currentMsgSize, 1, 1));
    
    // send notification
    pvm_initsend(initSend);
    pvmPackCheck(pvm_pkint(&handshake, 1, 1));
    pvm_send(parentTid, PVMSPEED_MSG_TAG_HANDSHAKE);

    // begin transfer
    for(i=0;i<loopNum;i++){
      dataTransfer();
    }
  }
}

int
main()
{
  // connect to pvm
  myTid = pvm_mytid();
  parentTid = pvm_parent();

  initMeasurement();
  
  measurement();
  
  return 0;
}

