// cppvm 'sendrec'
#include "cppvm.h"

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  // send/receive objects
  cppvmInt intObj(1);
  cppvmDouble doubleObj(2.2);
  int counter=0;

  // redirect all output of 'sendrec2'
  // to standard out
  cppvmCatchout catchOut(CPPvmStdout);
  // spawn child process 'sendrec2'
  cppvmSpawnConnection child("sendrec2");

  // message passing descriptors
  cppvmReceiveStream recStrm;
  cppvmSendStream sendStrm;

  // initialize receive descriptor 
  // (nonblocking receive from child process)
  recStrm.set(child, CPPvmRNBchild);
  
  bool msgReceived = false;
  // wait until message arrives...
  while(msgReceived==false){
    recStrm >> intObj;
    msgReceived = recStrm.received();
    counter++;
  }
  cout << "sendrec: received message; counter: "
       <<  counter << endl;

  // initialize send descriptor 
  // (nonblocking send to child process)
  sendStrm.set(child, CPPvmSNBchild);
  
  // send nonblocking
  sendStrm << intObj;

  // initialize receive descriptor 
  // (timeout every 0.2 sec.
  // receive from child process)
  recStrm.set(child, CPPvmRTchild, 0, 200000);
  
  while((recStrm >> intObj).received()==false){
    cout << "sendrec: receive timout..." << endl;
  }

  // blocking send to child process
  sendStrm.set(child, CPPvmSBchild);
  
  sendStrm << intObj;

  // the next message will be forwarded
  // back to this process by the receiver
  // A message can only be forwarded
  // if it was sent with a nonblocking
  // send!

  // after the forward the object must have
  // these values again
  intObj = 1;
  doubleObj = 1.1;
  // nonblocking send to child process
  sendStrm.set(child, CPPvmSNBchild);
  
  sendStrm << intObj << doubleObj;
  intObj = 2;
  doubleObj = 2.2;
  cout << "sendrec: forward after send: " 
    << intObj << " " << doubleObj << endl;

  // blocking receive from child process
  recStrm.set(child, CPPvmRBchild);
  
  recStrm >> intObj >> doubleObj;
  cout << "sendrec: forward after receive: " 
    << intObj << " " << doubleObj << endl;

  // receive the objects in a different order
  // then they were sent
  recStrm >> intObj;

  // test if a double type message has
  // arrived
  int bufid, bytes, tid;
  
  if(recStrm.probe(cppvmDouble::cppvmClassMsgTag(),
                   bufid, bytes, tid)){
    cout << "Message of type 'cppvmDouble' has arrived:"
         << " Buffer id: " << bufid
         << " Bytes: " << bytes
         << " Source task id: " << tid
         << endl;
  }
  
  recStrm >> doubleObj;
  cout << "sendrec: different order receive1: " << intObj << endl;
  cout << "sendrec: different order receive2: " << doubleObj << endl;
  
  cout << "sendrec: end" << endl;

  // halt the virtual machine
  pvmConn.halt();

  return 0;
}






