#include "cppvm.h"
// CPPVM_SLEEP define
#include "examplesdefines.h"

int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  // send/receive object
  cppvmInt intObj(1);
  cppvmDouble doubleObj(3.3);
  int parentTid;

  // message passing descriptors
  cppvmSendStream sendStrm;
  cppvmReceiveStream recStrm, recStrm2;
  cppvmForwardStream forwStrm;

  // wait 1 second
  CPPVM_SLEEP(1);

  cout << "sendrec2: blocking send" << endl;
  // get task id of the parent process
  parentTid = pvmConn.getParentTid();
  // the normal pvm task id's can also
  // be used to send/receive messages
  sendStrm.set(parentTid, CPPvmSBtid);
  
  // send object to parent process
  sendStrm << intObj;

  cout << "sendrec2: blocking receive" << endl;
  // blocking receive from any pvm process
  recStrm.set(CPPvmRBany);

  // receive from any pvm process
  recStrm >> intObj;
  // blocking send back to the process
  // described in 'recStrm'
  // (from which the last message was
  // received with 'recStrm')
  sendStrm.set(recStrm, CPPvmSBback);
  
  CPPVM_SLEEP(1);

  cout << "sendrec2: blocking send" << endl;
  // send back
  sendStrm << intObj;

  // blocking receive from the process
  // described in 'recStrm'
  // (from which the last message was
  // received with 'recStrm')
  recStrm2.set(recStrm, CPPvmRBsame);
  
  cout << "sendrec2: blocking receive" << endl;
  // blocking receive object
  recStrm2 >> intObj;

  // an object can only be forwarded if it 
  // * was sent to this process
  //   with a nonblocking send
  // * is sent to another process 
  //   with nonblocking send
  sendStrm.set(pvmConn, CPPvmSNBparent);
  recStrm.set(pvmConn, CPPvmRBparent);
  forwStrm.set(recStrm, sendStrm);
  
  // forward a integer and a double object
  forwStrm | cppvmInt::cppvmClassMsgTag()
    | cppvmDouble::cppvmClassMsgTag();
  cout << "sendrec2: forward result: " 
    << forwStrm.forwarded() << endl;

  doubleObj = 4.4;
  intObj = 9;
  
  // send nonblocking
  sendStrm.set(pvmConn, CPPvmSNBparent);

  // send in a different order than the
  // objects are received
  sendStrm << doubleObj << intObj;
  
  cout << "sendrec2: end" << endl;

  return 0;
}
