// status of task/hosts
#include "cppvm.h"
// CPPVM_SLEEP define
#include "examplesdefines.h"

void 
printHosts()
{
  cppvmConnection pvmConn;
  // of course, you could 
  // also use this:
  //pvmConn.printHostInfo();

  int i;
  int nhost;
  int narch;
  struct pvmhostinfo *hostp;

  pvmConn.hostInfo(&nhost, &narch, &hostp);

  cout << "Number of host in the virtual machine: " 
       << nhost << endl;
  for(i=0;i<nhost;i++){
    cout << "Host name: " << hostp[i].hi_name
         << " Architecture: " << hostp[i].hi_arch
         << " Speed: " << hostp[i].hi_speed
         << " Host daemon tid: " << hostp[i].hi_tid 
         << endl;
  }
}

void 
printTasks()
{
  cppvmConnection pvmConn;
  // of course, you could 
  // also use this:
  // pvmConn.printTaskInfo();

  int i;
  int ntask;
  struct pvmtaskinfo *taskp;
  
  pvmConn.taskInfo(0, &ntask, &taskp);

  cout << "Number of task running on virtual machine: " 
       << ntask << endl;
  for(i=0;i<ntask;i++){
    cout << "Task name: " << taskp[i].ti_a_out
         << " Status: " << taskp[i].ti_flag
         << " Task id: " << taskp[i].ti_tid
         << " Task process id: " << taskp[i].ti_pid
         << " Parent tid: ";
    if(taskp[i].ti_ptid == PvmNoParent){
      cout << " No parent";
    }
    else{
      cout << taskp[i].ti_ptid;
    }
    cout << " This tasks daemon tid: " 
         << taskp[i].ti_host << endl;
  }
}


int
main()
{
  // connect to pvm
  cppvmConnection pvmConn;

  // spawn child 'status2' on 'sun1'
  //cppvmSpawnConnection child("status2", "", 
  //  PvmTaskHost, "sun1");

  // let pvm choose the host
  cppvmSpawnConnection child("status2", "");

  // print host info
  printHosts();
  // print task info
  printTasks();

  // is the host 'sun1' alive?
  // (is the pvm daemon on 'sun1' running?)
  if(pvmConn.getHostStat("sun1")==PvmOk){
    cout << "host 'sun1' is in the virtual machine" 
         << endl ;
  }
  else{
    cout << "host 'sun1' is NOT in the virtual machine" 
         << endl;
  }
  
  // is the child process still alive?
  if(child.getChildStat()==PvmOk){
    cout << "child 'status2' is alive" 
         << endl ;
  }
  else{
    cout << "child 'status2' is NOT alive" 
         << endl;
  }
  
  // well, game over...
  // send SIGTERM signal to child
  child.killChild();

  // wait until killChild()
  // is done
  CPPVM_SLEEP(1);
  
  // is the child process still alive?
  // you can also use the "getTaskStat"
  // methods of cppvmConnection to get
  // the status of a process 
  if(pvmConn.getTaskStat("status2")==PvmOk){
    cout << "child 'status2' is alive" 
         << endl ;
  }
  else{
    cout << "child 'status2' is NOT alive" 
         << endl;
  }

  // halt the virtual machine
  child.halt();

  return 0;
}

