// example how to write your own send/receivable classes
// class 'user' implementation

// include specification file
#include "user.h"

// static cppvm variables
CPPvmMethodsDefinitionVariable(userClass, CPPvmMsgTagClassUser)

// default constructor
userClass::userClass()
    : cppvmObject()
{
  name = "NoName";
  id   = 0;
  group.setName("NoGroup");
  group.setId(0);
}

// constructor
userClass::userClass(cppvmString name_, int id_,
		     cppvmString groupName, int groupId)
    : cppvmObject()
{
  name = name_;
  id   = id_;
  group.setName(groupName);
  group.setId(groupId);
}

// copy constructor
userClass::userClass (const userClass& obj)
    : cppvmObject(obj)
{
  name  = obj.name;
  id    = obj.id;
  group = obj.group;
}

// destructor
userClass::~userClass()
{
}

// operator=
userClass&
userClass::operator=(const userClass& obj)
{
  // same object?
  if(cppvmIsSame(obj)){
    return *this;
  }
  // operator= of base class
  cppvmObject::operator=(obj);

  name  = obj.name;
  id    = obj.id;
  group = obj.group;
  
  return *this;
}

// operator==
bool
userClass::operator== (const userClass& obj)
{
  return (name  == obj.name  &&
	  id    == obj.id    &&
	  group == obj.group);
}

// operator==
bool
userClass::operator== (const userClass& obj) const
{
  return (name  == obj.name  &&
	  id    == obj.id    &&
	  group == obj.group);
}


// cppvmTransfer
// used by send/receive/distributed/mailbox messages of this class
void
userClass::cppvmTransfer()
{
  // transfer 'cppvmObject' and derived objects
  // standard 'cppvmString' class
  name.cppvmTransfer();
  // your own class 'group'
  group.cppvmTransfer();

  // transfer standard types (int, float, double, ...)
  // with the functions 'cppvmTransfer<Type>'
  cppvmTransferInt(&id);

}

// cppvmPrintOn 
// (called when using the << stream operator "cout << obj;")
void
userClass::cppvmPrintOn(ostream& strm)
{
  strm << "user: " 
       << " name: " << name
       << " id: " << id 
       << " " << group;
}

// cppvmScanFrom
// (called when using the >> stream operator "cin >> obj;")
void
userClass::cppvmScanFrom(istream& strm)
{
  strm >> name;
  strm >> id;
  strm >> group;
}

// receive function for collections
// normal c-function, not a method
cppvmObject*
userClassRecColl(int classId)
{
  // is the next element to be received 
  // in the collection of this class?
  if(classId == userClass::cppvmClassMsgTag()){
    // yes, create new object
    userClass *objPtr;
    objPtr = new userClass();
    // receive object
    objPtr->cppvmTransfer();
    // return object to the collection
    return objPtr;
  }
  else{
    // NULL indicates that the object was not
    // of the class
    return NULL;
  }
}
