// example how to write your own send/receivable classes
// class 'userkey' implementation

// include specification file
#include "userkey.h"

// static variables
CPPvmMethodsDefinitionVariable(userKeyClass, CPPvmMsgTagClassUserkey)

// default constructor
userKeyClass::userKeyClass()
    : userClass()
{
  keyLen = 0;
  key = new float [0];
}

// constructor
userKeyClass::userKeyClass(cppvmString name_, int id_,
			   cppvmString groupName, int groupId,
			   unsigned int keyLen_, float *key_)
    : userClass(name_, id_, groupName, groupId)
{
  unsigned int i;

  // alloc array and copy values
  keyLen = (int) keyLen_;
  key = new float [keyLen];
  for(i=0; i<keyLen_; i++){
    key[i] = key_[i];
  }
}

// copy constructor
userKeyClass::userKeyClass (const userKeyClass& obj)
    : userClass(obj)
{
  int i;

  // copy all values
  keyLen = obj.keyLen;
  key = new float [keyLen];
  for(i=0; i<keyLen; i++){
    key[i] = obj.key[i];
  }
}

// destructor
userKeyClass::~userKeyClass()
{
  delete [] key;
}

// operator=
userKeyClass&
userKeyClass::operator=(const userKeyClass& obj)
{
  int i;

  // same object?
  if(cppvmIsSame(obj)){
    return *this;
  }
  // operator= of base class
  userClass::operator=(obj);

  delete [] key;

  // copy all values
  keyLen = obj.keyLen;
  key = new float [keyLen];
  for(i=0; i<keyLen; i++){
    key[i] = obj.key[i];
  }

  return *this;
}

// operator==
bool
userKeyClass::operator== (const userKeyClass& obj)
{
  int i;

  // compare base class values
  if(!(userClass::operator==(obj))){
    return false;
  }
  // compare keyLen
  if(obj.keyLen != keyLen){
    return false;
  }
  // compare array values
  for(i=0; i<keyLen; i++){
    if(key[i] != obj.key[i]){
      return false;
    }
  }
   return true;
}

// operator==
bool
userKeyClass::operator== (const userKeyClass& obj) const
{
  int i;

  // compare base class values
  if(!(userClass::operator==(obj))){
    return false;
  }
  // compare keyLen
  if(obj.keyLen != keyLen){
    return false;
  }
  // compare array values
  for(i=0; i<keyLen; i++){
    if(key[i] != obj.key[i]){
      return false;
    }
  }
  return true;
}


// cppvmTransfer
// used by send/receive/distributed/mailbox messages of this class
void
userKeyClass::cppvmTransfer()
{
  // transfer base class variables
  userClass::cppvmTransfer();
  
  // transfer standard types (int, float, double, ...)
  // with the functions 'cppvmTransfer<Type>'
  cppvmTransferInt(&keyLen);

  // if you have to differ between send and receive transfer,
  // use the functions 'cppvmIsTransferSend()' or 'cppvmIsTransferReceive'
  if(cppvmIsTransferReceive()==true){
    // delete old array, allocate new array with received new size
    // 'keyLen' and receive array values
    delete [] key;
    key = new float [keyLen];
  }
  // transfer array
  cppvmTransferFloat(key, keyLen);

}

// cppvmPrintOn 
// (called when using the << stream operator "cout << obj;")
void
userKeyClass::cppvmPrintOn(ostream& strm)
{
  int i;

  // call printOn of base class
  userClass::cppvmPrintOn(strm);
  strm << " key: ";
  for(i=0; i<keyLen; i++){
    strm << key[i] << " ";
  }
}

// cppvmScanFrom
// (called when using the >> stream operator "cin >> obj;")
void
userKeyClass::cppvmScanFrom(istream& strm)
{
  int i;

  // call scanFrom of base class
  userClass::cppvmScanFrom(strm);

  strm >> keyLen;

  // resize array to new length
  delete [] key;
  key = new float [keyLen];
  
  for(i=0;i<keyLen;i++){
    strm >> key[i];
  }
}

// receive function for collections
// normal c-function, not a method
cppvmObject*
userKeyClassRecColl(int classId)
{
  // is the next element to be received 
  // in the collection of this class?
  if(classId == userKeyClass::cppvmClassMsgTag()){
    // yes, create new object
    userKeyClass *objPtr;
    objPtr = new userKeyClass();
    // receive object
    objPtr->cppvmTransfer();
    // return object to the collection
    return objPtr;
  }
  else{
    // NULL indicates that the object was not
    // of the class
    return NULL;
  }
}
