/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_arbo.h 
@SE  CREATED     : 29.9.1999
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm bool array class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC PII 366
@SI  COMPILER        : gcc version 2.91.66
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_ARBO_INCLUDE
#define _PVM_ARBO_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_arda.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_ARBO_MODULE "pvm_arbo.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmArrayBool
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : bool array class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmArrayBool : public cppvmArrayData{
public:

  cppvmArrayBool();
  cppvmArrayBool(unsigned int len_, bool cleanArray=true);
  cppvmArrayBool(const cppvmArrayBool&);
  virtual ~cppvmArrayBool();

  cppvmArrayBool& operator=(const cppvmArrayBool&);
  bool operator==(const cppvmArrayBool&);
  bool operator==(const cppvmArrayBool&) const;
  bool operator!=(const cppvmArrayBool& obj) const {
    return ! (operator==(obj));
  }
  
  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  virtual void cppvmTransfer();
  
  CPPvmBuildInMsgTagMethodsDeclaration(CPPvmMsgTagClassArrayBool)

  virtual void cleanElem(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARBO_MODULE,
                                                   "cppvmArrayBool",
                                                   "cleanElem",
                                                   "index >= array length"));
    }
#endif
    array[index]=false;
  }
  virtual void cleanRange(unsigned int fromIndex, unsigned int toIndex);

  bool& operator[] (unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARBO_MODULE,
                                                   "cppvmArrayBool",
                                                   "operator[unsigned int]",
                                                   "index >= array length"));
    }
#endif
    return array[index];
  }
  bool* getPtr() {return array;}
  bool* getPtr() const {return array;}

  bool read(unsigned int index);
  void write(unsigned int index, bool value);

  void cycleShiftLeft();
  void cycleShiftRight();
  
  virtual double getElemDouble(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARBO_MODULE,
                                                   "cppvmArrayBool",
                                                   "getElemDouble",
                                                   "index >= array length"));
    }
#endif
    return (double)(array[index]);
  }
  
protected:
  virtual void addElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void subElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void multElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void divElem(int pos, cppvmArrayData& obj, int objPos);
  // assume that true==1 and false==0
  virtual void incElemPrae(int index){
    if(array[index]==false){
      array[index]=true;
    }
  }
  virtual void incElemPost(int index){
    if(array[index]==false){
      array[index]=true;
    }
  }
  virtual void decElemPrae(int index){
    if(array[index]==true){
      array[index]=false;
    }
  }
  virtual void decElemPost(int index){
    if(array[index]==true){
      array[index]=false;
    }
  }
  virtual void copyElems(int num, int localIndex, 
                         cppvmArrayData& extObj, int extIndex);
  
  
protected:
  bool*             array;  // char array
  unsigned char*    packedArray;
  
  virtual bool arrayIsNULL(){return (array==(bool*)NULL);}
  virtual void newArray(unsigned int newSize);
  virtual void newArrayCopy(unsigned int newSize, unsigned int numCopyElements);

  bool emptyArray[CPPvmArrayCleanFieldSize];
  bool initEmptyArray;
private:

  void packData();
  void unpackData(bool* to, int number);
  
  bool getPackedElem(int& byteIndex, int& bitIndex);
  void setPackedElem(int& byteIndex, int& bitIndex, bool value);

  int  getPackedArraySize(int newSize)
  {
    int result = newSize / (int)8;
    if( (newSize % (int)8) > 0){
      result++;
    }
    return result;
  }
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



