/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_arch.h 
@SE  CREATED     : 4.2.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm char array class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_ARCH_INCLUDE
#define _PVM_ARCH_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_arda.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_ARCH_MODULE "pvm_arch.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmArrayChar
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : char array class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmArrayChar : public cppvmArrayData{
public:

  cppvmArrayChar();
  cppvmArrayChar(unsigned int len_, bool cleanArray=true);
  cppvmArrayChar(const cppvmArrayChar&);
  virtual ~cppvmArrayChar();

  cppvmArrayChar& operator=(const cppvmArrayChar&);
  bool operator==(const cppvmArrayChar&);
  bool operator==(const cppvmArrayChar&) const;
  bool operator!=(const cppvmArrayChar& obj) const {
    return ! (operator==(obj));
  }
  
  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  virtual void cppvmTransfer();
  
  CPPvmBuildInMsgTagMethodsDeclaration(CPPvmMsgTagClassArrayChar)

  virtual void cleanElem(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARCH_MODULE,
                                                   "cppvmArrayChar",
                                                   "cleanElem",
                                                   "index >= array length"));
    }
#endif
    array[index]=0;
  }
  virtual void cleanRange(unsigned int fromIndex, unsigned int toIndex);
  
  char& operator[] (unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARCH_MODULE,
                                                   "cppvmArrayChar",
                                                   "operator[unsigned int]",
                                                   "index >= array length"));
    }
#endif
    return array[index];
  }
  char* getPtr() {return array;}
  char* getPtr() const {return array;}

  char read(unsigned int index);
  void write(unsigned int index, char value);

  void cycleShiftLeft();
  void cycleShiftRight();
  
  virtual double getElemDouble(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARCH_MODULE,
                                                   "cppvmArrayChar",
                                                   "getElemDouble",
                                                   "index >= array length"));
    }
#endif
    return (double)(array[index]);
  }
protected:
  virtual void addElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void subElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void multElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void divElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void incElemPrae(int index){++(array[index]);}
  virtual void incElemPost(int index){(array[index])++;}
  virtual void decElemPrae(int index){--(array[index]);}
  virtual void decElemPost(int index){(array[index])--;}

  virtual void copyElems(int num, int localIndex, 
                         cppvmArrayData& extObj, int extIndex);
  
  
protected:
  char*    array;  // char array

  virtual bool arrayIsNULL(){return (array==(char*)NULL);}
  virtual void newArray(unsigned int newSize);
  virtual void newArrayCopy(unsigned int newSize, unsigned int numCopyElements);

  char emptyArray[CPPvmArrayCleanFieldSize];
  bool initEmptyArray;
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



