/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_arda.h 
@SE  CREATED     : 3.2.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm array data base class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_ARDA_INCLUDE
#define _PVM_ARDA_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_type.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_ARDA_MODULE "pvm_arda.h"

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmArrayCleanFieldSize
@CD  DESCRIPTION   : size for the memcopy field when cleaning arrays
***
******************************************************************************/
#define CPPvmArrayCleanFieldSize 128

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmArrayData
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : array data base class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmArrayData : public cppvmType{
public:

  cppvmArrayData(unsigned int len_);
  cppvmArrayData(const cppvmArrayData&);
  virtual ~cppvmArrayData();

  cppvmArrayData& operator=(const cppvmArrayData&);
  bool operator==(const cppvmArrayData&);
  bool operator==(const cppvmArrayData&) const;
  bool operator!=(const cppvmArrayData& obj) const {
    return ! (operator==(obj));
  }

  void operator++();
  void operator++(int);
  void operator--();
  void operator--(int);

  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  virtual void cppvmTransfer();
  void setCPPvmTransferMode(int mode);
  int getCPPvmTransferMode(){return cppvmTransferMode;}
    
  CPPvmBuildInMsgTagMethodsDeclaration(CPPvmMsgTagClassArrayData)

  unsigned int numElems() const { return (unsigned int) len; }

  void clean() 
  { 
    if(len>0){ cleanRange(0, len-1); }
  }
  
  void append(cppvmArrayData&);
  void resize(unsigned int newLen);
  void resizeClean(unsigned int newSize, bool cleanOnlyNewElements=false);

  virtual void cleanElem(unsigned int){}
  virtual void cleanRange(unsigned int /*fromIndex*/, unsigned int /*toIndex*/){}

  virtual double getElemDouble(unsigned int){return 0.0;}

  cppvmArrayData& add(cppvmArrayData& obj);
  cppvmArrayData& subtract(cppvmArrayData& obj);
  cppvmArrayData& multiply(cppvmArrayData& obj);
  cppvmArrayData& divide(cppvmArrayData& obj);

  void setFitSizeToLength(bool value);
  bool getFitSizeToLength(){return fitSizeToLength;}

  unsigned int getSize(){return size;}
  unsigned int getLength(){return len;}

protected:
  void checkCleanRange(unsigned int fromIndex, unsigned int toIndex);
  void resizeWithoutFill(unsigned int newSize);
  void resizeWithoutFillAndCopy(unsigned int newSize);
  
  virtual bool arrayIsNULL(){return false;}
  virtual void newArray(unsigned int /*newSize*/){}
  virtual void newArrayCopy(unsigned int /*newSize*/, unsigned int /*numCopyElements*/){}

protected:
  virtual void addElem(int, cppvmArrayData&, int){}
  virtual void subElem(int, cppvmArrayData&, int){}
  virtual void multElem(int, cppvmArrayData&, int){}
  virtual void divElem(int, cppvmArrayData&, int){}
  virtual void incElemPrae(int){}
  virtual void incElemPost(int){}
  virtual void decElemPrae(int){}
  virtual void decElemPost(int){}

  virtual void copyElems(int /*num*/, int /*localIndex*/, 
                         cppvmArrayData& /*extObj*/, int /*extIndex*/){}
  
protected:
  int  len;  // array length
  int  cppvmTransferMode;
  int  size; // total array size
  bool fitSizeToLength;
  
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



