/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_ardo.h 
@SE  CREATED     : 4.2.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm double array class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_ARDO_INCLUDE
#define _PVM_ARDO_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_arda.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_ARDO_MODULE "pvm_ardo.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmArrayDouble
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : double array class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmArrayDouble : public cppvmArrayData{
public:

  cppvmArrayDouble();
  cppvmArrayDouble(unsigned int len_, bool cleanArray=true);
  cppvmArrayDouble(const cppvmArrayDouble&);
  virtual ~cppvmArrayDouble();

  cppvmArrayDouble& operator=(const cppvmArrayDouble&);
  bool operator==(const cppvmArrayDouble&);
  bool operator==(const cppvmArrayDouble&) const;
  bool operator!=(const cppvmArrayDouble& obj) const {
    return ! (operator==(obj));
  }

  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  virtual void cppvmTransfer();
  
  CPPvmBuildInMsgTagMethodsDeclaration(CPPvmMsgTagClassArrayDouble)

  virtual void cleanElem(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARDO_MODULE,
                                                   "cppvmArrayDouble",
                                                   "cleanElem",
                                                   "index >= array length"));
    }
#endif
    array[index]=0.0;
  }
  virtual void cleanRange(unsigned int fromIndex, unsigned int toIndex);

  double& operator[] (unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARDO_MODULE,
                                                   "cppvmArrayDouble",
                                                   "operator[unsigned int]",
                                                   "index >= array length"));
    }
#endif
    return array[index];
  }
  double* getPtr() {return array;}
  double* getPtr() const {return array;}

  double read(unsigned int index);
  void write(unsigned int index, double value);
  
  void cycleShiftLeft();
  void cycleShiftRight();
  
  virtual double getElemDouble(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARDO_MODULE,
                                                   "cppvmArrayDouble",
                                                   "getElemDouble",
                                                   "index >= array length"));
    }
#endif
    return (double)(array[index]);
  }
protected:
  virtual void addElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void subElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void multElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void divElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void incElemPrae(int index){++(array[index]);}
  virtual void incElemPost(int index){(array[index])++;}
  virtual void decElemPrae(int index){--(array[index]);}
  virtual void decElemPost(int index){(array[index])--;}

  virtual void copyElems(int num, int localIndex, 
                         cppvmArrayData& extObj, int extIndex);
  
  
protected:
  double*    array;  // double array

  virtual bool arrayIsNULL(){return (array==(double*)NULL);}
  virtual void newArray(unsigned int newSize);
  virtual void newArrayCopy(unsigned int newSize, unsigned int numCopyElements);

  double emptyArray[CPPvmArrayCleanFieldSize];
  bool initEmptyArray;
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



