/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_arfl.h 
@SE  CREATED     : 4.2.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm float array class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_ARFL_INCLUDE
#define _PVM_ARFL_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_arda.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_ARFL_MODULE "pvm_arfl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmArrayFloat
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : float array class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmArrayFloat : public cppvmArrayData{
public:

  cppvmArrayFloat();
  cppvmArrayFloat(unsigned int len_, bool cleanArray=true);
  cppvmArrayFloat(const cppvmArrayFloat&);
  virtual ~cppvmArrayFloat();

  cppvmArrayFloat& operator=(const cppvmArrayFloat&);
  bool operator==(const cppvmArrayFloat&);
  bool operator==(const cppvmArrayFloat&) const;
  bool operator!=(const cppvmArrayFloat& obj) const {
    return ! (operator==(obj));
  }
  
  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  virtual void cppvmTransfer();
  
  CPPvmBuildInMsgTagMethodsDeclaration(CPPvmMsgTagClassArrayFloat)

  virtual void cleanElem(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARFL_MODULE,
                                                   "cppvmArrayFloat",
                                                   "cleanElem",
                                                   "index >= array length"));
    }
#endif
    array[index]=0.0;
  }
  virtual void cleanRange(unsigned int fromIndex, unsigned int toIndex);

  float& operator[] (unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARFL_MODULE,
                                                   "cppvmArrayFloat",
                                                   "operator[unsigned int]",
                                                   "index >= array length"));
    }
#endif
    return array[index];
  }
  float* getPtr() {return array;}
  float* getPtr() const {return array;}

  float read(unsigned int index);
  void write(unsigned int index, float value);

  void cycleShiftLeft();
  void cycleShiftRight();
  
  virtual double getElemDouble(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARFL_MODULE,
                                                   "cppvmArrayFloat",
                                                   "getElemDouble",
                                                   "index >= array length"));
    }
#endif
    return (double)(array[index]);
  }

protected:
  virtual void addElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void subElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void multElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void divElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void incElemPrae(int index){++(array[index]);}
  virtual void incElemPost(int index){(array[index])++;}
  virtual void decElemPrae(int index){--(array[index]);}
  virtual void decElemPost(int index){(array[index])--;}

  virtual void copyElems(int num, int localIndex, 
                         cppvmArrayData& extObj, int extIndex);
  

  
protected:
  float*    array;  // float array

  virtual bool arrayIsNULL(){return (array==(float*)NULL);}
  virtual void newArray(unsigned int newSize);
  virtual void newArrayCopy(unsigned int newSize, unsigned int numCopyElements);

  float emptyArray[CPPvmArrayCleanFieldSize];
  bool initEmptyArray;
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



