/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_arin.h 
@SE  CREATED     : 4.2.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm integer array class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_ARIN_INCLUDE
#define _PVM_ARIN_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_arda.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_ARIN_MODULE "pvm_arin.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmArrayInt
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : integer array class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmArrayInt : public cppvmArrayData{
public:

  cppvmArrayInt();
  cppvmArrayInt(unsigned int len_, bool cleanArray=true);
  cppvmArrayInt(const cppvmArrayInt&);
  virtual ~cppvmArrayInt();

  cppvmArrayInt& operator=(const cppvmArrayInt&);
  bool operator==(const cppvmArrayInt&);
  bool operator==(const cppvmArrayInt&) const;
  bool operator!=(const cppvmArrayInt& obj) const {
    return ! (operator==(obj));
  }
 
  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  virtual void cppvmTransfer();
  
  CPPvmBuildInMsgTagMethodsDeclaration(CPPvmMsgTagClassArrayInt)

  virtual void cleanElem(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARIN_MODULE,
                                                   "cppvmArrayInt",
                                                   "cleanElem",
                                                   "index >= array length"));
    }
#endif
    array[index]=0;
  }
  virtual void cleanRange(unsigned int fromIndex, unsigned int toIndex);

  int& operator[] (unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARIN_MODULE,
                                                   "cppvmArrayInt",
                                                   "operator[unsigned int]",
                                                   "index >= array length"));
    }
#endif
    return array[index];
  }
  int* getPtr() {return array;}
  int* getPtr() const {return array;}

  int read(unsigned int index);
  void write(unsigned int index, int value);

  void cycleShiftLeft();
  void cycleShiftRight();
  
  virtual double getElemDouble(unsigned int index){
#ifndef CPPvmFastAccess
    if(((int)index)>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_ARIN_MODULE,
                                                   "cppvmArrayInt",
                                                   "getElemDouble",
                                                   "index >= array length"));
    }
#endif
    return (double)(array[index]);
  }
  
protected:
  virtual void addElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void subElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void multElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void divElem(int pos, cppvmArrayData& obj, int objPos);
  virtual void incElemPrae(int index){++(array[index]);}
  virtual void incElemPost(int index){(array[index])++;}
  virtual void decElemPrae(int index){--(array[index]);}
  virtual void decElemPost(int index){(array[index])--;}

  virtual void copyElems(int num, int localIndex, 
                         cppvmArrayData& extObj, int extIndex);
  
protected:
  int*    array;  // integer array

  virtual bool arrayIsNULL(){return (array==(int*)NULL);}
  virtual void newArray(unsigned int newSize);
  virtual void newArrayCopy(unsigned int newSize, unsigned int numCopyElements);

  int emptyArray[CPPvmArrayCleanFieldSize];
  bool initEmptyArray;
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



