/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_barr.h 
@SE  CREATED     : 18.09.2001
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : barrier class for process synchronization
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_BARR_INCLUDE
#define _PVM_BARR_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_def.h"
#include "pvm_exra.h"
#include "pvm_exre.h"
#include "pvm_exfo.h"
#include "pvm_expa.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_BARR_HEADER "pvm_barr.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
// class declarations to avoid cyclic includes
class cppvmString;

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmBarrier
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : notificatin class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmBarrier 
{
public:
  cppvmBarrier();
  cppvmBarrier(char *groupName, unsigned int count);
  virtual ~cppvmBarrier();
  cppvmBarrier(const cppvmBarrier&);
  cppvmBarrier& operator=(const cppvmBarrier& obj);
  bool operator!=(const cppvmBarrier& obj) const {
    return ! (operator==(obj));
  }
  bool operator==(const cppvmBarrier& obj) {
    return (barrierGroupName==obj.barrierGroupName &&
            barrierCount==obj.barrierCount);
  }
  bool operator==(const cppvmBarrier& obj) const {
    return (barrierGroupName==obj.barrierGroupName &&
            barrierCount==obj.barrierCount);
  }

  // test of same identity
  bool cppvmIsSame(const cppvmBarrier& obj) const{
    return cppvmObjId == obj.cppvmObjId;
  }

  virtual void cppvmPrintOn(ostream& strm= cout);
  virtual void cppvmScanFrom(istream& /*strm= cin*/) {};

  int getCPPvmObjId() { return cppvmObjId; }

public:
  void barrier();
  void set(char *groupName, unsigned int count, bool leaveFormerGroup=false);
  void setGroup(char *groupName, bool leaveFormerGroup=false);
  void setCount(unsigned int count);
  void leaveGroup();

protected:
  void joinGroup(char *group);

  void barrierTry();
  void joinGroupTry(char *group);
  void leaveGroupTry();
  
  cppvmString barrierGroupName;
  int barrierCount;
  
private:
  static int cppvmMaxId; // highest actual ID
  int cppvmObjId;        // specific object ID
  static int cppvmTotalObjNum; // total number of object for this class
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// inline functions 
inline ostream&
operator<<(ostream& strm, cppvmBarrier& obj){
  obj.cppvmPrintOn(strm);
  return strm;
}

inline istream&
operator>>(istream& strm, cppvmBarrier& obj){
  obj.cppvmScanFrom(strm);
  return strm;
}

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
