/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_catc.h 
@SE  CREATED     : 2.7.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : catchout for spawned processes
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : catchout causes the calling task (the parent) to catch
@SR                output from tasks spawned after creating a cppvmCatchout
@SR                object. Characters printed on stdout or stderr in children
@SR                tasks are collected by the pvmds and sent in control
@SR                messages to the parent task which tags each line and
@SR                appends it to the specified file.
@SR                Output from grandchildren (spawned by children) tasks is
@SR                also collected, provided the children don't reset
@SR                'PvmOutputTid'.
@SR                Each line of output has one of the following forms:
@SR                [txxxxx] BEGIN
@SR                [txxxxx] (text from child task)
@SR                [txxxxx] END
@SR                The output from each task includes one BEGIN line and one
@SR                END line, with whatever the task prints in between.
@SR                If pvm_exit is called while output collection is in effect,
@SR                it will block until all tasks sending it output have
@SR                exited, in order to print all their output. To avoid this,
@SR                output collection can be turned off by using 'CPPvmStop'
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_CATC_INCLUDE
#define _PVM_CATC_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_proh.h"
#include "pvm_cincl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED CONSTANTS                        *******
**********                                                           **********
******************************************************************************/
#define PVM_CATC_HEADER "pvm_catc.h"
/*@@@**************************************************************************
**********                                                           **********
*******                        EXPORTED TYPES                           *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmCatchoutMode
@TD  DESCRIPTION : describes how pvm should catch child tasks outputs
***
******************************************************************************/
typedef enum {
  CPPvmStdout     = 0,
  CPPvmStderr     = 1,
  CPPvmFile       = 2,
  CPPvmStop       = 3
} cppvmCatchoutMode;

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmCatchout
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm message buffer
@KC  CONSTRAINTS   : 
******************************************************************************/

class cppvmCatchout : public cppvmProcessHandling {

public:
  cppvmCatchout(cppvmCatchoutMode catchoutFlag = CPPvmStdout,
                char *catchoutFileName = NULL);
  
  virtual ~cppvmCatchout();

  bool operator==(const cppvmCatchout&);
  bool operator==(const cppvmCatchout&) const;
  bool operator!=(const cppvmCatchout&);

  virtual void cppvmPrintOn(ostream& = cout);
  virtual void cppvmScanFrom(istream& = cin) {};

  void set(cppvmCatchoutMode catchoutFlag = CPPvmStdout,
           char *catchoutFileName = NULL);

  void reset();
  void unset();

protected:
  cppvmCatchoutMode flag;
  cppvmString *fileName;
protected:
  static FILE *catchoutFileDesc;
  static bool isFile;
  static int numObjs;
  
private:
  // do not allow the operator=
  // do not copy a opened file descriptor
  cppvmCatchout& operator=(const cppvmCatchout& obj);
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
