/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_cincl.h 
@SE  CREATED     : 08.07.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : C include files; standard and namespace versions
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_CINCL_INCLUDE
#define _PVM_CINCL_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

/* The order of the include files is important!
   1) In 'pvm3.h' the function 'pvm_catchout' is only
      declared when 'EOF' is defined. 'EOF' is defined
      in 'stdio.h' resp. 'cstdio'. 
   2) the PVM functions 'pvm_vpackf' and 'pvm_vunpackf'
      are not part of 'pvm3.h'. They have to be declared
      as 'extern "C"'. Since these functions use
      'va_list' as parameter type they have to be
      declared after including 'stdarg.h' resp. 
      'cstdarg'. 
   3) On Windows systems 'pvm3.h' includes 'pvmwin.h'. 
      In 'pvmwin.h' the standard C defines 'va_start' 
      and 'va_end' are undefined! This is why these 
      defines must be redefined again according 
      to the Windows compiler version
      (see 'pvm_tran.cpp' and 'pvm_str.cpp').

   To summarize: The include order must be:

   #include <stdio.h> (resp. <cstdio>)
   #include <stdarg.h> (resp. <cstdarg>)
   extern "C" 
   {
     #include "pvm3.h"
     int pvm_vpackf(char *fmt, va_list ap);
     int pvm_vunpackf(char *fmt, va_list ap);
   }
   here define 'va_start' and 'va_list' again!
*/

extern "C" 
{
  // common includes
  #include <sys/types.h>
  // multithreading unix (C functions)
#ifdef CPPvmThread
#ifndef WIN32
  #include<pthread.h>
#endif
#endif
}

// multithreading windows (MFC classes)
#ifdef CPPvmThread
#ifdef WIN32
  #include <afxmt.h>
#endif
#endif

// includes with and without STD namespace
#ifdef CPPvmUseStandardCIncludeFiles
#include <ctype.h>
#include <iomanip.h>
#include <iostream.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#else

#ifdef CPPvmUseStandardCIncludeFileCtype
#include <ctype.h>
#else
#include <cctype>
#endif

#ifdef CPPvmUseStandardCIncludeFileIomanip
#include <iomanip.h>
#else
#include <iomanip>
#endif

#ifdef CPPvmUseStandardCIncludeFileIostream
#include <iostream.h>
#else
#include <iostream>
#ifndef CPPvmDoNotUseStdNamespace
using std::ostream;
using std::istream;
using std::cout;
using std::cin;
using std::cerr;
using std::endl;
#endif // CPPvmDoNotUseStdNamespace
#endif

#ifdef CPPvmUseStandardCIncludeFileLimits
#include <limits.h>
#else
#include <climits>
#endif

#ifdef CPPvmUseStandardCIncludeFileStdarg
#include <cstdarg>
#else
#include <stdarg.h>
#endif

#ifdef CPPvmUseStandardCIncludeFileStdio
#include <cstdio>
#else
#include <stdio.h>
#endif

#ifdef CPPvmUseStandardCIncludeFileStdlib
#include <cstdlib>
#else
#include <stdlib.h>
#endif

#ifdef CPPvmUseStandardCIncludeFileString
#include <cstring>
#else
#include <string.h>
#endif

#ifdef CPPvmUseStandardCIncludeFileTime
#include <ctime>
#else
#include <time.h>
#endif

#endif // CPPvmUseStandardCIncludes 


extern "C" 
{
  // include PVM
  #include "pvm3.h"
  #include "pvmtev.h"

  // special pvm pack/unpack functions
  // with 'va_list' argument type.
  // they are not declared in "pvm3.h" 
  int pvm_vpackf(char *fmt, va_list ap);
  int pvm_vunpackf(char *fmt, va_list ap);

  // pvm_freezgroup prototype
  // exists only since PVM 
  // version 3.4.4 in "pvm3.h";
  // declare prototype for
  // backward compatibility
  int pvm_freezegroup(char *group, int size);

  // In PVM version 3.4.x with x < 2
  // the define 'PvmIPLoopback'
  // is not set; define it for
  // backward compatibility
#ifndef PvmIPLoopback
#define	PvmIPLoopback	-36	/* Master Host's IP is Loopback */
#endif
}



#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
