/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_coll.h 
@SE  CREATED     : 16.12.1997
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm object ordered collection class
@SU  USAGE       : To use this cppvmObject collection with classes derived from
@SU                cppvmObject use the castdown command
@SU                  (PVMExample&)pvmObjCollection[i]
@SU                to access specific subclass methods
@SC  CONSTRAINTS : only for cppvmObject or derived classes
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_COLL_INCLUDE
#define _PVM_COLL_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_cincl.h"
#include "pvm_arbo.h"
#include "pvm_arch.h"
#include "pvm_arda.h"
#include "pvm_ardo.h"
#include "pvm_arfl.h"
#include "pvm_arin.h"
#include "pvm_arlo.h"
#include "pvm_bool.h"
#include "pvm_char.h"
#include "pvm_doub.h"
#include "pvm_floa.h"
#include "pvm_int.h"
#include "pvm_long.h"
#include "pvm_obj.h"
#include "pvm_str.h"
#include "pvm_task.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_COLL_MODULE "pvm_coll.h"

/*@@@**************************************************************************
**********                                                           **********
*******                              TYPES                              *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : pvmCollCompFctPtr
@TD  DESCRIPTION : pointer to a compare function for *ALL* 'cppvmObject's
@TD                in the collection (the compare function MUST be usable for all 
@TD                objects in the collection) 
***
******************************************************************************/
typedef bool (*pvmCollCompFctPtr)(cppvmObject* obj1, cppvmObject* obj2);

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : pvmCollRecFctPtr
@TD  DESCRIPTION : pointer to a function which can receive cppvmObjects
@TD                the function must:
@TD                * if the 'objClassId' fits the object(s) which can be
@TD                  received by the function:
@TD                  - create a new object with 'new'  operator
@TD                  - receive the object
@TD                  - give the pointer to this object back ('recObj')
@TD                * if the 'objClassId' does not fit the object(s) which can 
@TD                  be received by the function:
@TD                  - return NULL pointer
***
******************************************************************************/
typedef cppvmObject* (*pvmCollRecFctPtr)(int objClassId);

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmObjectCollection
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm object collection
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmObjectCollection : public cppvmObject{
public:

  cppvmObjectCollection();
  virtual ~cppvmObjectCollection();

  bool operator==(const cppvmObjectCollection&);
  bool operator==(const cppvmObjectCollection&) const;
  bool operator!=(const cppvmObjectCollection& obj) const {
    return ! (operator==(obj));
  }

  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  virtual void cppvmTransfer();
  int getCPPvmTransferMode(){
    return cppvmTransferMode;
  }
  void setCPPvmTransferMode(int cppvmTransferMode_);

  CPPvmBuildInMsgTagMethodsDeclaration(CPPvmMsgTagClassObjcoll)

public:
  void add(cppvmObject* objPtr);
  void remove(cppvmObject& obj);
  void remove(cppvmObject* objPtr);
  void remove(unsigned int i);
  void clean();

  void append(cppvmObjectCollection* objColl);
  void resize(unsigned int);
  unsigned int numElems() const { return num; }

  cppvmObject& operator[] (unsigned int i);
  cppvmObject* getObjPtr(unsigned int i);
  cppvmObject* getConstObjPtr(unsigned int i) const;

  // message box method
  cppvmObject& read (unsigned int i);

  void swap(unsigned int index1, unsigned int index2);
  void sort(pvmCollCompFctPtr cmpPtr);

public:
  // only useful if the collection contains cppvmTask objects
  cppvmTask& getTaskWithName(cppvmString taskName);

public:
  void addRecFct(pvmCollRecFctPtr fctPtr);
  void removeRecFct(pvmCollRecFctPtr fctPtr);

private:
  void cppvmTransferSend();
  void cppvmTransferRec();
  bool cppvmTransferRecBase(int classId);
  bool cppvmTransferRecFcts(int classId);

  int cppvmTransferMode;
  
  void shiftLeft(unsigned int indexMin,
                 unsigned int indexMax);

private:
  cppvmObject** elems;  // dynamic pointer array
  unsigned int size;  // array size
  unsigned int num;   // number of entries (elems[0..num-1])

  int transferNumObj;    // only for send objects
  int classIdMsgTag;     // only for send objects
  
private:
  pvmCollRecFctPtr *recFcts; // dynamic receive function pointer array
  unsigned int recFctSize;   // array size
  unsigned int recFctNum;    // number of entries (recFcts[0..recFctNum-1])

private:
  void resizeRecFct(unsigned int newSize);
  void shiftLeftRecFct(unsigned int indexMin,
                       unsigned int indexMax);

  // do not allow the operator=
  // a deep copy must be available for this operator
  cppvmObjectCollection& operator=(const cppvmObjectCollection& obj);
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



