/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_conn.h 
@SE  CREATED     : 5.12.1997
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : parallel process base class (local contact to pvm)
@SD                This class should be used for slaves
@SD                or for masters to set pvm options
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : IMPORTANT: Multicasted (array of tids) and broadcasted
@SR                           (to group members) messages are NOT send to
@SR                           the caller even if the caller is in the
@SR                           array of tids resp. member of the group
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_CONN_INCLUDE
#define _PVM_CONN_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_proh.h"

/*@@@**************************************************************************
**********                                                           **********
*******                     EXPORTED CONSTANTS                          *******
**********                                                           **********
******************************************************************************/

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmMsgEncoding
@CD  DESCRIPTION   : message encoding specification
***
******************************************************************************/
#define CPPvmMsgEncoding -99

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmAllPvmDataLocal
@CD  DESCRIPTION   : use local message encoding specifications of each
@CD                  send stream (local setting)
***
******************************************************************************/
#define CPPvmAllPvmDataLocal 0

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmAllPvmDataDefault
@CD  DESCRIPTION   : message encoding type 'PvmDataDefault' for all send
@CD                  operations (global setting)
***
******************************************************************************/
#define CPPvmAllPvmDataDefault 1

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmAllPvmDataInPlace
@CD  DESCRIPTION   : message encoding type 'PvmDataInPlace' for all send
@CD                  operations (global setting)
***
******************************************************************************/
#define CPPvmAllPvmDataInPlace 2

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmAllPvmDataRaw
@CD  DESCRIPTION   : message encoding type 'PvmDataRaw' for all send
@CD                  operations (global setting)
***
******************************************************************************/
#define CPPvmAllPvmDataRaw 3

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmConnection
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : Pvm class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmConnection : public cppvmProcessHandling {
 
public:
  cppvmConnection();
  cppvmConnection(int argc, char **argv, int block=1);
  cppvmConnection(const cppvmConnection&);
  virtual ~cppvmConnection();

  cppvmConnection& operator=(const cppvmConnection&);
  bool operator==(const cppvmConnection&);
  bool operator==(const cppvmConnection&) const;
  bool operator!=(const cppvmConnection& obj) const {
    return ! (operator==(obj));
  }

  virtual void cppvmPrintOn(ostream& = cout);
  virtual void cppvmScanFrom(istream& = cin) {}

public:
  void halt();
  void exit();

public:
  cppvmTask getMyTask();
  cppvmTask getParentTask();
  int getContextId(){return contextId;}

public:
  // informations on the message box
  void cppvmGetMboxInfo(char *pattern,
		      int *nclasses,
		      struct pvmmboxinfo **classes);
  
public:
  void setDebugLevel(int debugLevel_);
  int getDebugLevel();
  
  
public:
  // group methods (for broadcast messages to a group)

  int joinGroup(char *group);
  void leaveGroup(char *group);
  void freezeGroup(char *group, int size = -1);

protected:

  int joinGroupTry(char *group);
  void leaveGroupTry(char *group);
  void freezeGroupTry(char *group, int size = -1);

public:
  
  int getInst(char *group, int tid) {return pvm_getinst(group, tid);}
  int getMyInst(char *group) {return pvm_getinst(group, myTid);}
  int getTid(char *group, int inum) {return pvm_gettid(group, inum);}
  int getGroupSize(char *group) {return pvm_gsize(group);}

  void barrier(char *group, int count); 
  void gather(void *result, void *data, int count, int datatype,	
              int msgtag, char *group, int rootginst);		
  void scatter(void *result, void *data, int count, int datatype,	
              int msgtag, char *group, int rootginst);		
  void reduce(void (*func)(int*, void*, void*, int*, int*), 
              void *data, int count, int datatype,
              int msgtag, char *group, int rootginst);
  
public:

  int getTid() { return myTid; }
  int getParentTid() { return parentTid; }

  void exportEnvVar(char *environmentVar){
    pvm_export(environmentVar);
  }
  void unexportEnvVar(char *environmentVar){
    pvm_unexport(environmentVar);
  }
  
  unsigned long int getEnrolledUSec() {
    return enrolledMicroseconds;
  }

  double getEnrolledSecs() {
    return (((double) enrolledMicroseconds) / ((double) CLOCKS_PER_SEC));
  }

  void setOpt(int what, int val);
  
  void getMsgBufInfo(int bufid, int *bytes,
                     int *msgtag, int *tid);

  int  getHostDaemonTid(char* taskOrHostName);
  int  getHostDaemonTid(cppvmTask& task);
  int  getHostDaemonTid(int tid);
  int  getHostDaemonTid() { return getHostDaemonTid(myTid); }
  char *getHostName(char* taskname){return getHostName(getTidOfTask(taskname));}
  char *getHostName(cppvmTask& task);
  char *getHostName(int tid);
  char *getHostName(){return getHostName(myTid);}
  char *getHostArchitecture(){return getHostArchitecture(myTid);};
  char *getHostArchitecture(char* taskname){
    return getHostArchitecture(getTidOfTask(taskname));
  }
  char *getHostArchitecture(cppvmTask& task);
  char *getHostArchitecture(int tid);
  int getHostSpeed(int tid);
  int getHostSpeed(cppvmTask& task);
  int getHostSpeed(char* taskname){return getHostSpeed(getTidOfTask(taskname));}
  int getHostSpeed(){return getHostSpeed(myTid);};
  
  int getHostStat(char *host);

  int getTaskStat(char* taskname){
    return getTaskStat(getTidOfTask(taskname));
  }
  int getTaskStat(cppvmTask& task);
  int getTaskStat(int tid);

  cppvmArrayInt getAllTaskIds(char* hostname);
  cppvmArrayInt getAllTaskIds(int pvmDaemonTaskId);

  void hostInfo(int *nhost,
                int *narch,
                struct pvmhostinfo **hostp);
  void printHostInfo();
  void printOneHostInfo(struct pvmhostinfo& h);

  void taskInfo(int where,
                int *ntask,
                struct pvmtaskinfo **taskp);
  void printTaskInfo();
  void printOneTaskInfo(struct pvmtaskinfo& t);

  int getTidOfTask(char *taskName);
  
  bool msgIsObjOfClass(int classMsgTag);
  bool msgIsObjOfClass(cppvmObject& obj);

  void getTMask(int who, Pvmtmask mask);
  void setTMask(int who, Pvmtmask mask);
  
public:
  // these methods are only for advanced pvm programming
  // they are not necessary for normal send/receive communication
  // See pvm manual for further information on the pvm functions
  // you also have to understand the cppvmConnection,
  // cppvmSpawnConnection classes and the
  // init/exit Send/Receive methods in cppvmObject
  int archcode(char *arch) {return pvm_archcode(arch);}
  void addHost(char *host);
  void addHostTry(char *host);
  void delHost(char *host);
  int freebuf(int bufid) {return pvm_freebuf(bufid);}
  int getfds(int **fds) {return pvm_getfds(fds);}
  int getopt(int what) {return pvm_getopt(what);}
  int getrbuf() {return pvm_getrbuf();}
  int getsbuf() {return pvm_getsbuf();}
  int hostsync(int host, struct timeval *clk, struct timeval *delta){
    return pvm_hostsync(host, clk, delta);
  }

  int siblings(int **tids){
    return pvm_siblings(tids);
  }
  
  int addMhf(int src, int tag, int ctx,
             int (*mhf)(int)){
    // return mhid
    return pvm_addmhf(src, tag, ctx, mhf);
  }
  int delMhf(int mhid){
    return pvm_delmhf(mhid);
  }

  // include messages in other messages as data elements
  //stillmiss
  void pkMesg(int bufid);
  //stillmiss
  void pkMesgBody(int bufid);
  //stillmiss
  int  upkMesg();
  
  // stillmiss
  //int deleteDB(char *name, int index) {return pvm_delete(name, index);}
  //int insertDB(char *name, int index, int data){
  //  return pvm_insert(name, index, data);
  //}
  //int lookupDB(char *name, int index, int *data) {
  //  return pvm_lookup(name, index, data);
  //}

  int mkbuf(int encoding=PvmDataDefault);
  int notify(int what, int msgtag, int cnt, int *tids){
    return pvm_notify(what, msgtag, cnt, tids);
  }
  int perror(char *msg) {return pvm_perror(msg);}
  int probe(int tid, int msgtag);
  void* recvf(int(*newvf)(int, int, int));
  int regHoster() {return pvm_reg_hoster();}
  int regRm(struct pvmhostinfo **hip) {return pvm_reg_rm(hip);}
  int regTasker() {return pvm_reg_tasker();}
  int sendsig(int tid, int signum);
  // stillmiss
  //int getmwid(int bufid) {return pvm_getmwid(bufid);}
  //int setmwid(int bufid, int waitid) {return pvm_setmwid(bufid, waitid);}
  
  int setrbuf(int bufid) {return pvm_setrbuf(bufid);}
  int setsbuf(int bufid) {return pvm_setsbuf(bufid);}
  int tidtohost(int tid) {return pvm_tidtohost(tid);} 

public:
  static int getGlobalMsgEncoding(){return globalMsgEncoding;}
  
protected:
  int myTid;
  int parentTid;
  unsigned long int enrolledMicroseconds;

  int  getMsgIsObjOfClass();

  void setOwnContext(){
    oldContextId = pvm_setcontext(contextId);
  }
  void resetContext(){
    pvm_setcontext(oldContextId);
  }
  void initContext(){
    contextId = pvm_getcontext();
  }
  int contextId;
  int oldContextId;
  
 
private:

  void initPvmConnection(int argc, char **argv, int block);

  static int totalNumObjs;
  static int globalMsgEncoding;
  
  // do not allow send/receive/forward for this class
  void cppvmSend(cppvmSendStream& sendStrm);
  bool cppvmReceive(cppvmReceiveStream& recStrm);
  bool cppvmForward(cppvmReceiveStream& recStrm, cppvmSendStream& sendStrm);

  int debugLevel;
  int debugLevelLock;
  static bool debugLevelEntryChecked;
  void sendDebugToMsgBox();
  bool recDebugFromMsgBox();
  bool setDebugLevelLock();

  void setDebugLevelTry(int debugLevel_);
  int getDebugLevelTry();
  // informations on the message box
  void cppvmGetMboxInfoTry(char *pattern,
                           int *nclasses,
                           struct pvmmboxinfo **classes);

};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

