/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_ctil.h 
@SE  CREATED     : 20.4.1999
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : inline operators for explicit message passing
@SD                of standard C types
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_CTIL_INCLUDE
#define _PVM_CTIL_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_cincl.h"
#include "pvm_bool.h"
#include "pvm_char.h"
#include "pvm_doub.h"
#include "pvm_floa.h"
#include "pvm_int.h"
#include "pvm_long.h"
#include "pvm_str.h"
#include "pvm_send.h"
#include "pvm_rec.h"
#include "pvm_forw.h"
#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

#include "pvm_stl.h"
#include "pvm_stlbitset.h"
#include "pvm_stlcomplex.h"
#include "pvm_stlstring.h"
#include "pvm_stldeque.h"
#include "pvm_stllist.h"
#include "pvm_stlslist.h"
#include "pvm_stlset.h"
#include "pvm_stlstack.h"
#include "pvm_stlmap.h"
#include "pvm_stlpair.h"
#include "pvm_stlpriorityqueue.h"
#include "pvm_stlqueue.h"
#include "pvm_stlvalarray.h"
#include "pvm_stlvector.h"

#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

///////////////////////
// receive operators //
///////////////////////

// receive standard C++ Types
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, bool& var){
  cppvmBool obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy value only if a message was received
    var = (bool)obj;
  }
  return recStrm;
}

inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, char& var){
  cppvmChar obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy value only if a message was received
    var = (char)obj;
  }
  return recStrm;
}

inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, double& var){
  cppvmDouble obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy value only if a message was received
    var = (double)obj;
  }
  return recStrm;
}

inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, float& var){
  cppvmFloat obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy value only if a message was received
    var = (float)obj;
  }
  return recStrm;
}

inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, int& var){
  cppvmInt obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy value only if a message was received
    var = (int)obj;
  }
  return recStrm;
}

inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, long& var){
  cppvmLong obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy value only if a message was received
    var = (long)obj;
  }
  return recStrm;
}

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

#ifdef CPPvmUseStlBitset
template <unsigned int T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, bitset<T> & var){
  cppvmStlBitset<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    int i;
    for(i=0;i<(int)var.size(); i++){
      var[i] = obj[i];
    }
  }
  return recStrm;
}
#endif // CPPvmUseStlBitset

inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, string& var){
  cppvmStlString obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy value only if a message was received
    var = (string)obj;
  }
  return recStrm;
}

template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, complex<T> & var){
  cppvmStlComplex<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var = obj;
  }
  return recStrm;
}

template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, deque<T> & var){
  cppvmStlDeque<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var.resize(obj.size());
    typename deque<T>::iterator i;
    typename deque<T>::iterator j;
    for(i=obj.begin(), j=var.begin(); i!= obj.end(); i++, j++){
      *j = *i;
    }
  }
  return recStrm;
}

template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, list<T> & var){
  cppvmStlList<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var.resize(obj.size());
    typename list<T>::iterator i;
    typename list<T>::iterator j;
    for(i=obj.begin(), j=var.begin(); i!= obj.end(); i++, j++){
      *j = *i;
    }
  }
  return recStrm;
}

#ifdef CPPvmUseStlSlist
template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, slist<T> & var){
  cppvmStlSlist<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var.resize(obj.size());
    typename slist<T>::iterator i;
    typename slist<T>::iterator j;
    for(i=obj.begin(), j=var.begin(); i!= obj.end(); i++, j++){
      *j = *i;
    }
  }
  return recStrm;
}
#endif // CPPvmUseStlSlist


template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, set<T> & var){
  cppvmStlSet<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var.clear();
    typename set<T>::iterator i;
    for(i=obj.begin(); i!= obj.end(); i++){
      var.insert(*i);
    }
  }
  return recStrm;
}

template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, multiset<T> & var){
  cppvmStlMultiset<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var.clear();
    typename multiset<T>::iterator i;
    for(i=obj.begin(); i!= obj.end(); i++){
      var.insert(*i);
    }
  }
  return recStrm;
}

template <class Key, class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, map<Key, T> & var){
  cppvmStlMap<Key, T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    typename map<Key, T>::iterator i;
    var.clear();
    for(i=obj.begin(); i!= obj.end(); i++){
      typename map<Key, T>::value_type elem(i->first, i->second);
      var.insert(elem);
    }
  }
  return recStrm;
}

template <class T1, class T2>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, pair<T1, T2> & var){
  cppvmStlPair<T1, T2> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var.first = obj.first;
    var.second = obj.second;
  }
  return recStrm;
}

template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, queue<T> & var){
  cppvmStlQueue<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    // remove all elements in queue
    while(var.size()>0){
      var.pop();
    }
    while(obj.size()>0){
      var.push(obj.front());
      obj.pop();
    }
  }
  return recStrm;
}

template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, priority_queue<T> & var){
  cppvmStlPriorityqueue<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    // remove all elements in priorityqueue
    while(var.size()>0){
      var.pop();
    }
    while(obj.size()>0){
      var.push(obj.top());
      obj.pop();
    }
  }
  return recStrm;
}

template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, stack<T> & var){
  cppvmStlStack<T> obj;
  vector<T> varVector;
  int i;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    // remove all elements in stack
    while(var.size()>0){
      var.pop();
    }
    // copy values
    varVector.resize(obj.size());
    for(i=0;i< (int)varVector.size(); i++){
      varVector[i] = obj.top();
      obj.pop();
    }
    for(i=(int)(varVector.size()-1); i>=0; i--){
      var.push(varVector[i]);
    }
  }
  return recStrm;
}

template <class Key, class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, multimap<Key, T> & var){
  cppvmStlMultimap<Key, T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    typename multimap<Key, T>::iterator i;
    var.clear();
    for(i=obj.begin(); i!= obj.end(); i++){
      typename multimap<Key, T>::value_type elem(i->first, i->second);
      var.insert(elem);
    }
  }
  return recStrm;
}

template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, vector<T> & var){
  cppvmStlVector<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var.resize(obj.size());
    typename vector<T>::iterator i;
    typename vector<T>::iterator j;
    for(i=obj.begin(), j=var.begin(); i!= obj.end(); i++, j++){
      *j = *i;
    }
  }
  return recStrm;
}

#ifdef CPPvmUseStlValarray
template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, valarray<T> & var){
  cppvmStlValarray<T> obj;
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  if(recStrm.received()==true){
    // copy values only if a message was received
    var.resize(obj.size());
    int i;
    for(i=0; i<(int)obj.size();i++){
      var[i] = obj[i];
    }
  }
  return recStrm;
}
#endif // CPPvmUseStlValarray


#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

///////////////////////
// send operators    //
///////////////////////

// send standard C++ Types
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, bool var){
  cppvmBool obj(var);
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, char var){
  cppvmChar obj(var);
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, double var){
  cppvmDouble obj(var);
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, float var){
  cppvmFloat obj(var);
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, int var){
  cppvmInt obj(var);
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, long var){
  cppvmLong obj(var);
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, string var){
  cppvmStlString obj;
  obj.assign(var);
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

#ifdef CPPvmUseStlBitset
template <unsigned int T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, bitset<T> & var){
  cppvmStlBitset<T> obj;
  int i;
  for(i=0; i<(int)var.size();i++){
    obj[i] = var[i];
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}
#endif // CPPvmUseStlBitset

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, complex<T> & var){
  cppvmStlComplex<T> obj;
  obj = var;
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, deque<T> & var){
  cppvmStlDeque<T> obj;
  obj.resize(var.size());
  typename deque<T>::iterator i;
  typename deque<T>::iterator j;
  for(i=obj.begin(), j=var.begin(); j!= var.end(); i++, j++){
    *i = *j;
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, list<T> & var){
  cppvmStlList<T> obj;
  obj.resize(var.size());
  typename list<T>::iterator i;
  typename list<T>::iterator j;
  for(i=obj.begin(), j=var.begin(); j!= var.end(); i++, j++){
    *i = *j;
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

#ifdef CPPvmUseStlSlist
template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, slist<T> & var){
  cppvmStlSlist<T> obj;
  obj.resize(var.size());
  typename slist<T>::iterator i;
  typename slist<T>::iterator j;
  for(i=obj.begin(), j=var.begin(); j!= var.end(); i++, j++){
    *i = *j;
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}
#endif // CPPvmUseStlSlist

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, set<T> & var){
  cppvmStlSet<T> obj;
  typename set<T>::iterator i;
  for(i=var.begin(); i!= var.end(); i++){
    obj.insert(*i);
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, multiset<T> & var){
  cppvmStlMultiset<T> obj;
  typename multiset<T>::iterator i;
  for(i=var.begin(); i!= var.end(); i++){
    obj.insert(*i);
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

template <class Key, class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, map<Key, T> & var){
  cppvmStlMap<Key, T> obj;
  typename map<Key, T>::iterator i;
  for(i=var.begin(); i!= var.end(); i++){
    typename map<Key, T>::value_type elem(i->first, i->second);
    obj.insert(elem);
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

template <class T1, class T2>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, pair<T1, T2> & var){
  cppvmStlPair<T1, T2> obj;
  obj.first = var.first;
  obj.second = var.second;
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

template <class Key, class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, multimap<Key, T> & var){
  cppvmStlMultimap<Key, T> obj;
  typename multimap<Key, T>::iterator i;
  for(i=var.begin(); i!= var.end(); i++){
    typename multimap<Key, T>::value_type elem(i->first, i->second);
    obj.insert(elem);
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, queue<T> & var){
  cppvmStlQueue<T> obj;
  // copy values
  while(var.size()>0){
    obj.push(var.front());
    var.pop();
  }
  obj.cppvmSend(sendStrm);
  // copy value back again
  while(obj.size()>0){
    var.push(obj.front());
    obj.pop();
  }
  return sendStrm;
}

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, priority_queue<T> & var){
  cppvmStlPriorityqueue<T> obj;
  // copy values
  while(var.size()>0){
    obj.push(var.top());
    var.pop();
  }
  obj.cppvmSend(sendStrm);
  // copy value back again
  while(obj.size()>0){
    var.push(obj.top());
    obj.pop();
  }
  return sendStrm;
}

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, stack<T> & var){
  cppvmStlStack<T> obj;
  vector<T> varVector;
  int i;
  // copy values
  varVector.resize(var.size());
  for(i=0;i< (int)varVector.size(); i++){
    varVector[i] = var.top();
    var.pop();
  }
  // copy values
  for(i=(int)(varVector.size()-1); i>=0; i--){
    obj.push(varVector[i]);
    var.push(varVector[i]);
  }
  obj.cppvmSend(sendStrm);

  return sendStrm;
}

#ifdef CPPvmUseStlValarray
template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, valarray<T> & var){
  cppvmStlValarray<T> obj;
  obj.resize(var.size());
  int i;
  for(i=0; i<(int)obj.size();i++){
    obj[i] = var[i];
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}
#endif // CPPvmUseStlValarray

template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, vector<T> & var){
  cppvmStlVector<T> obj;
  obj.resize(var.size());
  typename vector<T>::iterator i;
  typename vector<T>::iterator j;
  for(i=obj.begin(), j=var.begin(); j!= var.end(); i++, j++){
    *i = *j;
  }
  obj.cppvmSend(sendStrm);
  return sendStrm;
}


#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
