/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_exce.h 
@SE  CREATED     : 5.12.1997
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm classes for exception handling
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : the messages can be suppressed by setting 'CPPvmErrorMsg'
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_EXCE_INCLUDE
#define _PVM_EXCE_INCLUDE

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_def.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                        EXPORTED TYPES                           *******
**********                                                           **********
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

extern void
pvmPrintDefaultException(void);


/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED CONSTANTS                        *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmTryAll
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : global exception handling
@MD                 all exceptions will be handled
@MU  USAGE        : CPPvmTryAll(expression)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : example: CPPvmTryAll(O.print())
@MR                 if the method 'print()' of the object 'O' throws an
@MR                 exception, it will be handled
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : expression  expression to check for exceptions
@MP  INOUT   : -
@MP  OUT     : - 
@MP  RETURNS : - 
******************************************************************************/
#ifdef CPPvmDoNotUseExceptions
#define CPPvmTryAll(expression) (expression);
#else
#define CPPvmTryAll(expression)                           \
  CPPvmInternalTry {                                      \
    (expression);                                         \
  }                                                       \
  catch (cppvmException& exception) {                     \
    cerr << exception;                                    \
  }                                                       \
  catch (...) {                                           \
    pvmPrintDefaultException();                           \
  }
#endif

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmTry
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : exception handling for cppvm
@MD                 all exception classes in this module are handled with this
@MD                 macro. Other exceptions are not handled.
@MU  USAGE        : CPPvmTry(expression)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : example: CPPvmTry(O.print())
@MR                 if the method 'print()' of the object 'O' throws one of the
@MR                 exceptions in this module, it will be handled.
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : expression  expression to check for exceptions
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#ifdef CPPvmDoNotUseExceptions
#define CPPvmTry(expression) (expression);
#else
#define CPPvmTry(expression)                              \
try {                                                     \
    (expression);                                         \
  }                                                       \
  catch (cppvmException& exception) {                     \
    cerr << exception;                                    \
  }
#endif

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmTryAndErr
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : global exception handling
@MD                 all exceptions will be handled
@MD                 if an exception occurs, the 'error_expression'
@MD                 is evaluated.
@MD            
@MU  USAGE        : CPPvmTryAndErr(check_expression, error_expression)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : example:
@MR                 CPPvmTryAndErr(O.print(),print("Exception occurred."))
@MR                 if the method 'print()' of the object 'O' throws an
@MR                 exception, it will be handled.
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : check_expression  expression to check for exceptions
@MP            error_expression  expression to evaluate if a exception occurs
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#ifdef CPPvmDoNotUseExceptions
#define CPPvmTryAndErr(check_expression, error_expression) (check_expression);
#else
#define CPPvmTryAndErr(check_expression, error_expression)   \
try {                                                        \
    (check_expression);                                      \
  }                                                          \
  catch (...) {                                              \
    (error_expression);                                      \
  }
#endif

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmTryAndSpecificErr
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : global exception handling
@MD                 all exceptions will be handled
@MD                 if the exception is a cppvm exception the
@MD                 'cppvm_error_expression' is evaluated.
@MD                 if the exception is not one of this module, the
@MD                 'error_expression' is evaluated.
@MD            
@MU  USAGE        : CPPvmTryAndSpecificErr(check_expression,
@MU                   cppvm_error_expression, error_expression)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : example:
@MR                 CPPvmTryAndSpecificErr(O.print(),
@MR                                print("Cppvm Exception."),
@MR                                print("Other Exception.")) 
@MR                 if the method 'print()' of the object 'O' throws an
@MR                 exception, it will be handled.
@MR                 if the exception is not one of this module, the
@MR                 'print()' expression is evaluated.
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : check_expression       expression to check for exceptions
@MP            cppvm_error_expression expression to evaluate for cppvm
@MP                                   exceptions
@MP            error_expression  expression to evaluate for other exceptions
@MP                              than the exceptions in this module
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#ifdef CPPvmDoNotUseExceptions
#define CPPvmTryAndSpecificErr(check_expression, error_expression) (check_expression);
#else
#define CPPvmTryAndSpecificErr(check_expression, cppvm_error_expression, error_expression)   \
try {                                                     \
    (check_expression);                                   \
  }                                                       \
  catch (cppvmException& exception) {                     \
    (cppvm_error_expression);                             \
  }                                                       \
  catch (...) {                                           \
    (error_expression);                                   \
  }
#endif

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmWarn
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : prints a warning message to cerr
@MU  USAGE        : CPPvmWarn(module_name, class_name, method_name, msg)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : 
@MP  INOUT   : -
@MP  OUT     : - 
@MP  RETURNS : - 
******************************************************************************/
#define CPPvmWarn(module_name, class_name, method_name, msg)\
  cerr <<                                                 \
    "WARNING:" << endl <<                                 \
    "Module:          " << module_name << endl <<         \
    "Class:           " << class_name  << endl <<         \
    "Method/Function: " << method_name << endl <<         \
    "Message:           " << msg     << endl 


/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED VARIABLES                        *******
**********                                                           **********
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmException
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : base class for all exceptions
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmException {
public:
  
  cppvmException();
  cppvmException(const char *mo, const char *cl, 
                 const char *me, const char *er);
  cppvmException(const cppvmException&);
  virtual ~cppvmException();

  cppvmException& operator=(const cppvmException&);
  bool operator==(const cppvmException&);
  bool operator==(const cppvmException&) const;
  bool operator!=(const cppvmException& obj) const {
    return ! (operator==(obj));
  }

  virtual void printOn(ostream& = cout);

public:
  char *getModuleName(){return module_name;}
  char *getClassName(){return class_name;}
  char *getMethodName(){return method_name;}
  char *getErrMsg(){return err_msg;}

protected:
  char *module_name;
  char *class_name;
  char *method_name;
  char *err_msg;
};

inline ostream&
operator<<(ostream& strm, cppvmException& obj){
  obj.printOn(strm);
  return strm;
}

/*@@@**************************************************************************
**********                                                           **********
*******                 MACROS FOR INTERNAL USE ONLY                    *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmInternalTry
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : try call
@MD                 if CPPvmDoNotUseExeptions is defined, just go on
@MU  USAGE        : CPPvmInternalTry
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter: 
@M#  
@MP  IN      : expression: try expression
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#ifdef CPPvmDoNotUseExceptions
#define CPPvmInternalTry if(true)
#else
#define CPPvmInternalTry try
#endif

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmInternalCatch/CPPvmInternalCatchDots
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : catch call
@MD                 if CPPvmDoNotUseExeptions is defined, do not enter the next
@MD                 block
@MU  USAGE        : CPPvmInternalCatch(expression)/CPPvmInternalCatchDots
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter: 
@M#  
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#ifdef CPPvmDoNotUseExceptions
#define CPPvmInternalCatchDots if(false)
#define CPPvmInternalCatch(expression) if(false)
#else
#define CPPvmInternalCatchDots catch(...)
#define CPPvmInternalCatch(expression) catch(expression)
#endif

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmInternalThrow
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : throw a CPPvm exception; 
@MD                 if CPPvmDoNotUseExeptions is defined, print
@MD                 a standard message and exit
@MU  USAGE        : CPPvmInternalThrow(exception)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : exit, if CPPvmDoNotUseExeptions is defined
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter: 
@M#  
@MP  IN      : exception: cppvmException object
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#ifdef CPPvmDoNotUseExceptions
#define CPPvmInternalThrow(exception)                        \
{                                                            \
  cppvmException *cppvmNewExcep= new exception;              \
  cppvmNewExcep->printOn(cerr); cerr << endl;                \
  cerr << "CPPvm exception handling switched off " << endl   \
       << "(compiler flag -DCPPvmDoNotUseExceptions). "      \
       << "exiting..." << endl;                              \
  exit(1);                                                   \
}
#else
#define CPPvmInternalThrow(exception)                        \
  {throw exception;}
#endif

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmInternalThrowAgain
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : throw a exception again; 
@MD                 if CPPvmDoNotUseExeptions is defined, print
@MD                 a standard message and exit
@MU  USAGE        : CPPvmInternalThrowAgain
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : exit, if CPPvmDoNotUseExeptions is defined
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter: 
@M#  
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#ifdef CPPvmDoNotUseExceptions
#define CPPvmInternalThrowAgain                              \
{                                                            \
  cerr << "throw exception again. "                          \
       << "exception handling switched off " << endl         \
       << "(compiler flag -DCPPvmDoNotUseExceptions). "      \
       << "exit..." << endl;                                 \
  exit(1);                                                   \
}
#else
#define CPPvmInternalThrowAgain throw
#endif


#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

