/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_forw.h 
@SE  CREATED     : 8.4.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : description of a pvm forward command
@SU  USAGE       : 
@SC  CONSTRAINTS : Attention! The send description 'cppvmSendStream' MUST NOT
@SC                           be a blocking send! 
@SC                           The received message MUST NOT be sent with
@SC                           a blocking send!
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_FORW_INCLUDE
#define _PVM_FORW_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_strm.h"
#include "pvm_send.h"
#include "pvm_rec.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_FORW_HEADER "pvm_forw.h"


/*@@@**************************************************************************
**********                                                           **********
*******                        EXPORTED TYPES                           *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED CONSTANTS                        *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                            MACROS                               *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmForwardStream
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm forward command class
@KC  CONSTRAINTS   : 
******************************************************************************/

class cppvmForwardStream : public cppvmStream
{
public:
  cppvmForwardStream();
  
public:
  cppvmForwardStream(cppvmReceiveStream& recStrm_, cppvmSendStream& sendStrm_);
  virtual ~cppvmForwardStream();
  cppvmForwardStream(const cppvmForwardStream&);

  cppvmForwardStream& operator=(const cppvmForwardStream&);
  bool operator==(const cppvmForwardStream&);
  bool operator==(const cppvmForwardStream&) const;
  bool operator!=(const cppvmForwardStream&);

  virtual void cppvmPrintOn(ostream& = cout);
  virtual void cppvmScanFrom(istream& = cin) {};

public:
  void set(cppvmReceiveStream& recStrm_, cppvmSendStream& sendStrm_);
  bool forward(int msgTag);

  cppvmReceiveStream& getRecStrm();
  cppvmSendStream& getSendStrm();

  bool forwarded(){return lastForwResult;}

protected:
  bool forwardTry(int msgTag);

  cppvmReceiveStream recStrm;
  cppvmSendStream sendStrm;
  
  bool initialized;

  // result of the last forward
  bool lastForwResult;

private:
  // has this object initialized pvm receiving?
  static bool pvmInitializedRec;
  // old context receive
  static int pvmInitRecContext;
  // old context send
  static int pvmInitSendContext;
  // old message send buffer id
  static int pvmInitSendMsgBuf;
};

inline cppvmForwardStream&
operator|(cppvmForwardStream& forwardStrm, int msgTag){
  forwardStrm.forward(msgTag);
  return forwardStrm;
}

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
