/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_inmu.h 
@SE  CREATED     : 16.10.2001
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : internal mutex class; the implementation depends on:
@SD                1) compiler switch 'CPPvmThread'
@SD                2) operating system (UNIX or Windows)
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_INMU_INCLUDE
#define _PVM_INMU_INCLUDE

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_cincl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                              TYPES                              *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmInternalMutex
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : mutex class only for internal use
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmInternalMutex {
public:
  
  cppvmInternalMutex();
#ifdef CPPvmThread
#ifndef WIN32
  cppvmInternalMutex(pthread_mutex_t *mutexPtr);
#endif
#endif
  virtual ~cppvmInternalMutex();


public:
  void lock();
  void unlock();

private:
  // do not allow copy constructor
  cppvmInternalMutex(const cppvmInternalMutex&){}
  // do not allow assignments
  cppvmInternalMutex& operator=(const cppvmInternalMutex&){ return *this; }

  // multi-threading
#ifdef CPPvmThread
#ifdef WIN32
  CSingleLock* winLock;
  CCriticalSection winMutex;
#else
  pthread_mutex_t *unixMutexPtr;
#endif
#endif
};

#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



