/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_noti.h 
@SE  CREATED     : 24.07.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : notification class 
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_NOTI_INCLUDE
#define _PVM_NOTI_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_def.h"
#include "pvm_exra.h"
#include "pvm_exre.h"
#include "pvm_exfo.h"
#include "pvm_expa.h"
#include "pvm_arin.h"
#include "pvm_rec.h"
#include "pvm_spaw.h"
#include "pvm_spmu.h"
#include "pvm_task.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_NOTI_HEADER "pvm_noti.h"

/*@@@**************************************************************************
**********                                                           **********
*******                        EXPORTED TYPES                           *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                            MACROS                               *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmNotificationThrowIfNotInit
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : throws an exception if the object was not
@MD                 initialized (made with the default constructor)
@MU  USAGE        : only for internal use!
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmNotificationThrowIfNotInit                                        \
if(!initialized){                                                              \
  CPPvmInternalThrow (cppvmExceptionForbiddenCall(PVM_NOTI_HEADER,             \
                  "cppvmNotification",                                         \
                  "Macro: CPPvmNotificationThrowIfNotInit",                    \
                  "object not initialized, it is not allowed to use it"));     \
}

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmNotification
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : notification class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmNotification 
{
public:
  cppvmNotification();
  cppvmNotification(int event, int msgTag, int number, int* tids);
  virtual ~cppvmNotification();
  cppvmNotification(const cppvmNotification&);
  cppvmNotification& operator=(const cppvmNotification& obj);
  bool operator!=(const cppvmNotification& obj) const {
    return ! (operator==(obj));
  }
  bool operator==(const cppvmNotification& obj); 
  bool operator==(const cppvmNotification& obj) const;

  // test of same identity
  bool cppvmIsSame(const cppvmNotification& obj) const{
    return cppvmObjId == obj.cppvmObjId;
  }

  virtual void cppvmPrintOn(ostream& strm= cout);
  virtual void cppvmScanFrom(istream& /*strm= cin*/) {};

  void set(int event, int msgTag);
  void set(int event, int msgTag, int number, int* tids);

  int getCPPvmObjId() { return cppvmObjId; }

  bool cppvmReceive(cppvmReceiveStream& recStrm);

  cppvmArrayInt getReceivedTids();
  
  int getEvent();
  int getMsgTag();
  cppvmArrayInt getTids();
  int getNumber();
  

protected:
  bool initialized;
  
  // init parameters
  int notificationEvent;
  int notificationMsgTag;
  cppvmArrayInt notificationTids;
  int notificationNumber;

  // received results
  cppvmArrayInt receivedTids;
  // host task ids as names
  cppvmArrayChar addedHostNames;
  // this paramter is only used for 
  // receiving data
  int numberResults;

  void unpackMsg();
  void unpackInt(int *ptr, int nitem, int stride);
  
  int getHostDaemonTid(char* taskOrHostName);
  int getTidOfTask(char *taskName);
  void hostInfo(int *nhost, int *narch, struct pvmhostinfo **hostp);
  void taskInfo(int where, int *ntask, struct pvmtaskinfo **taskp);

private:
  static int cppvmMaxId; // highest actual ID
  int cppvmObjId;        // specific object ID
  static int cppvmTotalObjNum; // total number of object for this class

};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// inline functions 
inline ostream&
operator<<(ostream& strm, cppvmNotification& obj){
  obj.cppvmPrintOn(strm);
  return strm;
}

inline istream&
operator>>(istream& strm, cppvmNotification& obj){
  obj.cppvmScanFrom(strm);
  return strm;
}

// receive cppvmNotification objects
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmNotification& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
