/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_obj.h 
@SE  CREATED     : 5.12.1997
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm object base class, defines all standard methods for the
@SD                objects of the framework
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_OBJ_INCLUDE
#define _PVM_OBJ_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_cincl.h"
#include "pvm_def.h"
#include "pvm_exfo.h"
#include "pvm_expa.h"
#include "pvm_exra.h"
#include "pvm_exre.h"
#include "pvm_send.h"
#include "pvm_rec.h"
#include "pvm_forw.h"
#include "pvm_mem.h"
#include "pvm_tran.h"
#include "pvm_globvar.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_OBJ_HEADER "pvm_obj.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmTransferModeClean
@CD  DESCRIPTION   : clean own data, then receive new ones
***
******************************************************************************/
#define CPPvmTransferModeClean  0

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmTransferModeAppend
@CD  DESCRIPTION   : append received data to the own ones
***
******************************************************************************/
#define CPPvmTransferModeAppend 1

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmMethodsDeclaration
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : generate the message tag methods of a class
@MU  USAGE        : CPPvmMethodsDeclaration(msgtag)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : the methods are inserted 'public'
@MR                 IMPORTANT: the next methods after this macro will
@MR                 be 'public', too. So use 'protected' or 'private'
@MR                 before writing the next methods if you do not want
@MR                 them public.
@MR                 Only for use in a class description 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : msgtag    message tag for this class
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmMethodsDeclaration(msgtag)                          \
public:                                                          \
  virtual int cppvmMsgTag(){                                     \
    checkMsgTagRange(msgtag);                                    \
    if(cppvmUseObjectMsgTag()==true){                            \
      return cppvmObjectMsgTag();                                \
    }                                                            \
    else{                                                        \
      return msgtag;                                             \
    }                                                            \
  }                                                              \
  static int cppvmClassMsgTag(){                                 \
    checkMsgTagRange(msgtag);                                    \
    return msgtag;                                               \
  }                                                              \
  virtual void setCPPvmObjectMsgTag(int newMsgTag){              \
    checkMsgTagRange(newMsgTag);                                 \
    cppvmProtectedSetObjectMsgTag(newMsgTag);                    \
  }                                                              

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmMethodsDeclarationVariable
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : generate the message tag methods of a class
@MU  USAGE        : CPPvmMethodsDeclarationVariable
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : the methods are inserted 'public'
@MR                 IMPORTANT: the next methods after this macro will
@MR                 be 'public', too. So use 'protected' or 'private'
@MR                 before writing the next methods if you do not want
@MR                 them public.
@MR                 Only for use in a class description 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      :
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmMethodsDeclarationVariable                          \
private:                                                         \
  static int cppvmMsgtag;                                        \
public:                                                          \
  virtual int cppvmMsgTag(){                                     \
    checkMsgTagRange(cppvmMsgtag);                               \
    if(cppvmUseObjectMsgTag()==true){                            \
      return cppvmObjectMsgTag();                                \
    }                                                            \
    else{                                                        \
      return cppvmMsgtag;                                        \
    }                                                            \
  }                                                              \
  static int cppvmClassMsgTag(){                                 \
    checkMsgTagRange(cppvmMsgtag);                               \
    return cppvmMsgtag;                                          \
  }                                                              \
  static void setCPPvmClassMsgTag(int tag){                      \
    checkMsgTagRange(cppvmMsgtag);                               \
    cppvmMsgtag = tag;                                           \
  }                                                              \
  virtual void setCPPvmObjectMsgTag(int newMsgTag){              \
    checkMsgTagRange(newMsgTag);                                 \
    cppvmProtectedSetObjectMsgTag(newMsgTag);                    \
  }                                                              


/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmMethodsDefinitionVariable
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : generate the message tag methods of a class
@MU  USAGE        : CPPvmMethodsDefinitionVariable(classname, msgtag)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : the methods are inserted 'public'
@MR                 IMPORTANT: the next methods after this macro will
@MR                 be 'public', too. So use 'protected' or 'private'
@MR                 before writing the next methods if you do not want
@MR                 them public.
@MR                 Only for use in a class description 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : msgtag    message tag for this class
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmMethodsDefinitionVariable(classname, msgtag) \
int classname::cppvmMsgtag = msgtag;

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmBuildInMsgTagMethodsStlDeclaration 
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : generate the message tag methods of a STL class
@MU  USAGE        : CPPvmBuildInMsgTagMethodsStlDeclaration(msgtag)          
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : the methods are inserted 'public'
@MR                 IMPORTANT: the next methods after this macro will
@MR                 be 'public', too. So use 'protected' or 'private'
@MR                 before writing the next methods if you do not want
@MR                 them public.
@MR                 Only for use in a class description 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : msgtag    message tag for this class
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmBuildInMsgTagMethodsStlDeclaration(msgtag)          \
public:                                                          \
  virtual int cppvmMsgTag(){                                     \
    if(cppvmUseObjectMsgTag()==true){                            \
      return cppvmObjectMsgTag();                                \
    }                                                            \
    else{                                                        \
      return msgtag;                                             \
    }                                                            \
  }                                                              \
  static int cppvmClassMsgTag(){                                 \
    return msgtag;                                               \
  }                                                              \
  virtual void setCPPvmObjectMsgTag(int newMsgTag){              \
    if(newMsgTag!=msgtag){                                       \
      checkMsgTagRange(newMsgTag);                               \
    }                                                            \
    cppvmProtectedSetObjectMsgTag(newMsgTag);                    \
  }                                                              

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmBuildInMsgTagMethodsDeclaration
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : generate the message tag methods of a class
@MU  USAGE        : CPPvmBuildInMsgTagMethodsDeclaration(msgtag)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : only for internal use!
@MR                 the methods are inserted 'public'
@MR                 IMPORTANT: the next methods after this macro will
@MR                 be 'public', too. So use 'protected' or 'private'
@MR                 before writing the next methods if you do not want
@MR                 them public.
@MR                 Only for use in a class description 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : msgtag    message tag for this class
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmBuildInMsgTagMethodsDeclaration(msgtag)             \
private:                                                         \
  static int cppvmMsgtag;                                        \
public:                                                          \
  virtual int cppvmMsgTag(){                                     \
    if(cppvmUseObjectMsgTag()==true){                            \
      return cppvmObjectMsgTag();                                \
    }                                                            \
    else{                                                        \
      return cppvmMsgtag;                                        \
    }                                                            \
  }                                                              \
  static int cppvmClassMsgTag(){return cppvmMsgtag;}             \
  static void setCPPvmClassMsgTag(int newMsgTag){                \
    if(newMsgTag!=msgtag){                                       \
      checkMsgTagRange(newMsgTag);                               \
    }                                                            \
   cppvmMsgtag = newMsgTag;                                      \
  }                                                              \
  virtual void setCPPvmObjectMsgTag(int newMsgTag){              \
    if(newMsgTag!=msgtag){                                       \
      checkMsgTagRange(newMsgTag);                               \
    }                                                            \
    cppvmProtectedSetObjectMsgTag(newMsgTag);                    \
  }                                                              


/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmBuildInMsgTagMethodsDefinition
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : generate the message tag methods of a class
@MU  USAGE        : CPPvmBuildInMsgTagMethodsDefinition(classname, msgtag)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : only for internal use!
@MR                 the methods are inserted 'public'
@MR                 IMPORTANT: the next methods after this macro will
@MR                 be 'public', too. So use 'protected' or 'private'
@MR                 before writing the next methods if you do not want
@MR                 them public.
@MR                 Only for use in a class description 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : msgtag    message tag for this class
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmBuildInMsgTagMethodsDefinition(classname, msgtag) \
int classname::cppvmMsgtag = msgtag;


/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmObject
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm framework base class, defines all standard methods 
@KD                  for pvm objects
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmObject 
{
public:
  cppvmObject(int objectMsgTag=CPPvmNoObjectMsgTag);
  virtual ~cppvmObject();
  cppvmObject(const cppvmObject&);
  cppvmObject& operator=(const cppvmObject& obj);
  bool operator!=(const cppvmObject& obj) const {
    return ! (operator==(obj));
  }
  bool operator==(const cppvmObject& obj) {return cppvmIsSame(obj);}
  bool operator==(const cppvmObject& obj) const {return cppvmIsSame(obj);}

  // test of same identity
  bool cppvmIsSame(const cppvmObject& obj) const{
    return cppvmObjId == obj.cppvmObjId;
  }

  virtual void cppvmPrintOn(ostream& = cout) {};
  virtual void cppvmScanFrom(istream& = cin) {};

  int getCPPvmObjId() { return cppvmObjId; }

  void cppvmSend(cppvmSendStream& sendStrm);
  bool cppvmReceive(cppvmReceiveStream& recStrm);

  virtual int cppvmMsgTag()=0;
  // this static methods MUST be defined by every subclass!
  static int cppvmClassMsgTag(){
    return CPPvmMsgTagNoMessagePassing;
  }
  virtual void cppvmTransfer()=0;

  // static message tag check function
  static void checkMsgTagRange(int msgTag_)
  {
    if(msgTag_>=CPPvmMsgTagMin &&
       msgTag_<=CPPvmMsgTagMax){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_OBJ_HEADER,
                                                   "cppvmObject",
                                                   "checkMsgTagRange",
                                                   "this message tag is reserved for CPPvm classes! Range for CPPvm classes: [CPPvmMsgTagMin, CPPvmMsgTagMax] (see 'pvm_def.h')"));
    }
  }
  
public:
  

  static bool cppvmClassIsTransferSend(){
    return cppvmTransferSend;
  }
  static bool cppvmClassIsTransferReceive(){
    return cppvmTransferRec;
  }

public:
  void cppvmDistObj(char *keyName,
                    bool persistent=true,
                    int sendEncoding=PvmDataDefault,
                    bool useOwnSendBuffer=false);
  void cppvmSetMbox(char *keyName,
                    int writeFlags=PvmMboxDefault + 
                               PvmMboxOverWritable,
                    int readFlags=PvmMboxDefault,
                    int delFlags=0,
                    int sendEncoding=PvmDataDefault,
                    bool useOwnSendBuffer=false);
  void cppvmSetMboxRead(char *keyName,
                        int readFlags=PvmMboxDefault);
  void cppvmSetMboxWrite(char *keyName,
                         int writeFlags=PvmMboxDefault + PvmMboxPersistent +
		                      PvmMboxOverWritable,
                         int sendEncoding=PvmDataDefault,
                         bool useOwnSendBuffer=false);
  void cppvmSetMboxDel(char *keyName,
                       int delFlags=0);
  
  void cppvmWriteMboxInfo();
  void cppvmReadMboxInfo(int keyIndex = 0);
  void cppvmDelMboxInfo(int keyIndex = 0);

  void cppvmSetMboxKey(char *keyName,
		      int keyIndex);
  void cppvmSetMboxKeyName(char *keyName);
  void cppvmSetMboxKeyIndex(int keyIndex);
  void cppvmSetMboxSend(int sendEncoding,
		      bool useOwnSendBuffer);

  int cppvmGetMboxKeyIndex(){return mboxKeyIndex;}
  char *cppvmGetMboxKeyName(){return mboxKeyName;}

  void cppvmGetMboxInfo(struct pvmmboxinfo *classInfo);
  void cppvmGetMboxInfo(char *pattern,
                      int *nclasses,
                      struct pvmmboxinfo **classes);
  virtual void setCPPvmObjectMsgTag(int newMsgTag){
    cppvmProtectedSetObjectMsgTag(newMsgTag);
  }
  void cppvmDeactivateObjectMsgTag(){cppvmUseMsgTagObject=false;}
  
protected:
  void cppvmProtectedSetObjectMsgTag(int newMsgTag){
    cppvmUseMsgTagObject=true;
    cppvmMsgTagObject = newMsgTag;
  }
  int  cppvmObjectMsgTag() {return cppvmMsgTagObject;}
  bool cppvmUseObjectMsgTag() {return cppvmUseMsgTagObject;}

private:

  void cppvmWriteMboxInfoBase();
  int cppvmReadMboxInfoBase();

private:
  // try methods, use mutex lock before
  void cppvmDistObjTry(char *keyName,
                       bool persistent,
                       int sendEncoding,
                       bool useOwnSendBuffer);

  void cppvmWriteMboxInfoTry();
  void cppvmReadMboxInfoTry(int keyIndex);
  void cppvmDelMboxInfoTry(int keyIndex);

  void cppvmSendTry(cppvmSendStream& sendStrm);
  bool cppvmReceiveTry(cppvmReceiveStream& recStrm);
  void cppvmGetMboxInfoTry(char *pattern,
                           int *nclasses,
                           struct pvmmboxinfo **classes);

private:
  void cppvmResizeHandshakeArray(unsigned int newLen);
  void cppvmResetHandshakeArray(){pvmSendHandshakeTidsNum = 0;}
  void cppvmAddToHandshakeArray(int tid);

private:
  void cppvmSetHandshakeTidArrayGroupBroadcast(cppvmSendStream& send);
  void cppvmSetHandshakeTidArrayOneOrMulticast(cppvmSendStream& send);

private:
  static int cppvmMaxId;       // highest actual ID
  int cppvmObjId;              // specific object ID
  static int cppvmTotalObjNum; // total number of objects

  // indicates if the transfer command is a send of a receive
  static bool cppvmTransferRec ;
  static bool cppvmTransferSend;
  
private:
  // has this object initialized pvm sending?
  static bool pvmInitializedSend;
  // old context send
  static int pvmInitSendContext;
  // old message send buffer id
  static int pvmInitSendMsgBuf;
  // new created message send buffer id
  static int pvmNewSendMsgBuf;
  // message tag of the object which has initialized pvm sending
  static int pvmObjTagInitSend;
  // id of the object which has initialized pvm sending
  static int pvmObjIdInitSend;
  // is the sending blocking or nonblocking
  static int pvmSendHandshake;
  // handshake tids (receive handshakes from these tids)
  static int* pvmSendHandshakeTids;
  // current number of handshake tids
  static unsigned int pvmSendHandshakeTidsNum;
  // total size of the handshake tids array
  static unsigned int pvmSendHandshakeTidsSize;
  
  // has this object initialized pvm receiving?
  static bool pvmInitializedRec;
  // old context receive
  static int pvmInitRecContext;
  // message tag of the object which has initialized pvm receiving
  static int pvmObjTagInitRec;
  // id of the object which has initialized pvm receiving
  static int pvmObjIdInitRec;
  // was the received message delivered with blocking or nonblocking send
  static int pvmRecHandshake;

  void exitSend(cppvmSendStream& send,
                int msgTag);
  void initSend(cppvmSendStream& send,
                int msgTag);
  bool initRec(cppvmReceiveStream& rec,
               int msgTag);
  void exitRec(cppvmReceiveStream& rec,
               int msgTag);

private:
  int cppvmMsgTagObject;
  bool cppvmUseMsgTagObject;

  // was the mbox (or distributed object) initialized?
  // for reading
  bool mboxInitRead;
  // for writing
  bool mboxInitWrite;
  // for deleting?
  bool mboxInitDel;
  // use a own (new) send buffer for message box?
  bool mboxOwnSendBuf;
  // database key name
  char *mboxKeyName;
  // database key index
  int mboxKeyIndex;
  // flags for write
  int mboxFlagsWrite;
  // encoding for initSend
  int mboxSendEncoding;
  // flags for read
  int mboxFlagsRead;
  // flags for delete
  int mboxFlagsDel;

};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// inline functions 
inline ostream&
operator<<(ostream& strm, cppvmObject& obj){
  obj.cppvmPrintOn(strm);
  return strm;
}

inline istream&
operator>>(istream& strm, cppvmObject& obj){
  obj.cppvmScanFrom(strm);
  return strm;
}

// send cppvmObject classes
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmObject& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmObject classes
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmObject& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
