/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_proh.h 
@SE  CREATED     : 26.6.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm process handling base class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_PROH_INCLUDE
#define _PVM_PROH_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_def.h"
#include "pvm_send.h"
#include "pvm_rec.h"
#include "pvm_forw.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_PROH_HEADER "pvm_proh.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

// class declarations to avoid cyclic includes
class cppvmArrayInt;
class cppvmObject;
class cppvmObjectCollection;
class cppvmString;
class cppvmTask;

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmProcessHandling
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm system base class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmProcessHandling 
{
public:
  cppvmProcessHandling();
  virtual ~cppvmProcessHandling();
  cppvmProcessHandling(const cppvmProcessHandling&);
  cppvmProcessHandling& operator=(const cppvmProcessHandling& obj);
  bool operator!=(const cppvmProcessHandling& obj) const {
    return ! (operator==(obj));
  }
  bool operator==(const cppvmProcessHandling& obj) {return cppvmIsSame(obj);}
  bool operator==(const cppvmProcessHandling& obj) const {return cppvmIsSame(obj);}

  // test of same identity
  bool cppvmIsSame(const cppvmProcessHandling& obj) const{
    return cppvmObjId == obj.cppvmObjId;
  }

  virtual void cppvmPrintOn(ostream& = cout) {};
  virtual void cppvmScanFrom(istream& = cin) {};

  int getCPPvmObjId() { return cppvmObjId; }

protected:
  bool startPvmDaemon(int argc, char **argv, int block);

private:
  static int cppvmMaxId; // highest actual ID
  int cppvmObjId;        // specific object ID
  static int cppvmTotalObjNum; // total number of object for this class
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// inline functions 
inline ostream&
operator<<(ostream& strm, cppvmProcessHandling& obj){
  obj.cppvmPrintOn(strm);
  return strm;
}

inline istream&
operator>>(istream& strm, cppvmProcessHandling& obj){
  obj.cppvmScanFrom(strm);
  return strm;
}

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
