/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_rec.h 
@SE  CREATED     : 5.12.1997
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : description of a pvm receive command
@SD                Just use the default constructor for this class!
@SD                Just use the get/set methods if you REALLY know
@SD                what you are doing!
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : Warning: receive from a non existant TIDs is not possible.
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_REC_INCLUDE
#define _PVM_REC_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_strm.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_REC_HEADER "pvm_rec.h"


/*@@@**************************************************************************
**********                                                           **********
*******                        EXPORTED TYPES                           *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmRecMode
@TD  DESCRIPTION : describes how pvm should receive a message
***
******************************************************************************/
typedef enum{
  CPPvmBlockingRec     = 0,
  CPPvmNonblockingRec  = 1,
  CPPvmTimeoutRec      = 2
} cppvmRecMode;


/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmRecTask
@TD  DESCRIPTION : receive options for cppvmTask variables
***
******************************************************************************/
typedef enum{
  CPPvmRBtask  = 0,
  CPPvmRNBtask = 1,
  CPPvmRTtask  = 2
} cppvmRecTask;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmRecConnection
@TD  DESCRIPTION : receive options for cppvmConnection variables
***
******************************************************************************/
typedef enum{
  CPPvmRBparent  = 0,
  CPPvmRNBparent = 1,
  CPPvmRTparent  = 2
} cppvmRecConnection;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmRecSpawnConnection
@TD  DESCRIPTION : receive options for cppvmSpawnConnection variables
***
******************************************************************************/
typedef enum{
  CPPvmRBchild  = 0,
  CPPvmRNBchild = 1,
  CPPvmRTchild  = 2
} cppvmRecSpawnConnection;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmRecTid
@TD  DESCRIPTION : receive options for task id receives
***
******************************************************************************/
typedef enum{
  CPPvmRBtid  = 0,
  CPPvmRNBtid = 1,
  CPPvmRTtid  = 2
} cppvmRecTid;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmRecAny
@TD  DESCRIPTION : receive options for receive from any pvm process
***
******************************************************************************/
typedef enum{
  CPPvmRBany  = 0,
  CPPvmRNBany = 1,
  CPPvmRTany  = 2
} cppvmRecAny;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmRecSame
@TD  DESCRIPTION : receive options for receive from the same process
@TD                as the last receive
***
******************************************************************************/
typedef enum{
  CPPvmRBsame  = 0,
  CPPvmRNBsame = 1,
  CPPvmRTsame  = 2
} cppvmRecSame;

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED CONSTANTS                        *******
**********                                                           **********
******************************************************************************/

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmRecNoLastReceive
@CD  DESCRIPTION   : value to express no last receives in 'lastRecFromTid'
***
******************************************************************************/
#define CPPvmRecNoLastReceive -2

/*@@@**************************************************************************
**********                                                           **********
*******                            MACROS                               *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmRecThrowIfNotInit
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : throws an exception if the object was not
@MD                 initialized (made with the default constructor)
@MU  USAGE        : only for internal use!
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmRecThrowIfNotInit                                                 \
if(!initialized){                                                              \
  CPPvmInternalThrow (cppvmExceptionForbiddenCall(PVM_REC_HEADER,              \
                  "cppvmReceiveStream",                                        \
                  "Macro: CPPvmRecThrowIfNotInit",                             \
                  "object not initialized, it is not allowed to use it"));     \
}

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmReceiveStream
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm receive command class
@KC  CONSTRAINTS   : 
******************************************************************************/

class cppvmReceiveStream : public cppvmStream
{
protected:
  int context;
  
  bool initialized;

  // tid of the process to receive from (-1 = receive from any)
  int recFromTid;
  // tid of the process, from which the last message was received.
  // e.g. if you want to receive from any process ('recFromTid' = -1)
  // 'lastRecFromTid' contains the real tid of the 'any' process
  int lastRecFromTid;

  struct timeval timeout;
  cppvmRecMode recMode;

  // result of the last receive 
  // this variable is set externaly by the 
  // operator>>(cppvmReceiveStream& recStrm, cppvmObject& obj)
  // in file 'pvm_obj.h'
  bool lastRecResult;
  
public:
  cppvmReceiveStream();
  
public:
  // constructor method for 'cppvmTask' objects
  cppvmReceiveStream(cppvmTask& task,
                     cppvmRecTask mode,
                     int seconds=0, int useconds=0);
  // constructor method for 'cppvmConnection' objects
  cppvmReceiveStream(cppvmConnection& conn,
                     cppvmRecConnection mode,
                     int seconds=0, int useconds=0);
  // constructor method for 'cppvmSpawnConnection' objects
  cppvmReceiveStream(cppvmSpawnConnection& sconn,
                     cppvmRecSpawnConnection mode,
                     int seconds=0, int useconds=0);
  // constructor method for processes specified by 'taskId'
  // with current context
  cppvmReceiveStream(int taskId,
                     cppvmRecTid mode,
                     int seconds=0, int useconds=0);
  // constructor method for processes specified by 'taskId' with given context
  cppvmReceiveStream(int taskId, int context_,
                     cppvmRecTid mode,
                     int seconds=0, int useconds=0);
  // constructor method for receive from any processes with current context
  cppvmReceiveStream(cppvmRecAny mode,
                     int seconds=0, int useconds=0);
  // constructor method for receive from any processes with given context
  cppvmReceiveStream(int context_,
                     cppvmRecAny mode,
                     int seconds=0, int useconds=0);
  // constructor method for receive from the same processes, from which
  // a receive was done by 'oldStrm'
  // with current context
  cppvmReceiveStream(cppvmReceiveStream& oldStrm,
                     cppvmRecSame mode,
                     int seconds=0, int useconds=0);
  // constructor method for receive from the same processes, from which
  // a receive was done by 'oldStrm'
  // with given context
  cppvmReceiveStream(cppvmReceiveStream& oldStrm,
                     int context_,
                     cppvmRecSame mode,
                     int seconds=0, int useconds=0);
  
public:
  virtual ~cppvmReceiveStream();
  cppvmReceiveStream(const cppvmReceiveStream&);

  cppvmReceiveStream& operator=(const cppvmReceiveStream&);
  bool operator==(const cppvmReceiveStream&);
  bool operator==(const cppvmReceiveStream&) const;
  bool operator!=(const cppvmReceiveStream&);

  virtual void cppvmPrintOn(ostream& = cout);
  virtual void cppvmScanFrom(istream& = cin) {};

public:
  // set method for 'cppvmTask' objects
  void set(cppvmTask& task,
           cppvmRecTask mode,
           int seconds=0, int useconds=0);
  // set method for 'cppvmConnection' objects
  void set(cppvmConnection& conn,
           cppvmRecConnection mode,
           int seconds=0, int useconds=0);
  // set method for 'cppvmSpawnConnection' objects
  void set(cppvmSpawnConnection& sconn,
           cppvmRecSpawnConnection mode,
           int seconds=0, int useconds=0);
  // set method for processes specified by 'taskId' with current context
  void set(int taskId,
           cppvmRecTid mode,
           int seconds=0, int useconds=0);
  // set method for processes specified by 'taskId' with given context
  void set(int taskId, int context_,
           cppvmRecTid mode,
           int seconds=0, int useconds=0);
  // set method for receive from any processes with current context
  void set(cppvmRecAny mode,
           int seconds=0, int useconds=0);
  // set method for receive from any processes with given context
  void set(int context_,
           cppvmRecAny mode,
           int seconds=0, int useconds=0);
  // set method for receive from the same processes, from which
  // a receive was done by 'oldStrm'
  // with current context
  void set(cppvmReceiveStream& oldStrm,
           cppvmRecSame mode,
           int seconds=0, int useconds=0);
  // set method for receive from the same processes, from which
  // a receive was done by 'oldStrm'
  // with given context
  void set(cppvmReceiveStream& oldStrm,
           int context_,
           cppvmRecSame mode,
           int seconds=0, int useconds=0);
  
public:
  bool received(){return lastRecResult;}

  void setContext(int context_);
  int getContext(){return context;}

  bool probe(int msgTag);
  bool probe(int msgTag, int& bufid, int& bytes, int& tid);
  bool probeAny();
  bool probeAny(int& msgTag);
  bool probeAny(int& msgTag, int& bufid, int& bytes, int& tid);
  
  void setLastRecResult(bool lastRecResult_){ lastRecResult = lastRecResult_;}
  
  void setRecFromTid(int newTid);
  void setLastRecFromTid(int newTid);
  void setRecMode(cppvmRecMode newMode);
  void setTimeout(struct timeval& newTimeout);
  void setTimeout(int seconds, int useconds);
  
  int            getRecFromTid() {
    CPPvmRecThrowIfNotInit
      return recFromTid;
  }
  int            getLastRecFromTid() {
    CPPvmRecThrowIfNotInit
    return lastRecFromTid;
  }
  cppvmRecMode   getRecMode()    {
    CPPvmRecThrowIfNotInit
    return recMode;
  }
  struct timeval* getTimeoutDirect()    {
    CPPvmRecThrowIfNotInit
    return (&timeout);
  }

protected:
  void resetObject();

  bool probeTry(int msgTag);
  bool probeTry(int msgTag, int& bufid, int& bytes, int& tid);
  bool probeAnyTry(int& msgTag);
  bool probeAnyTry(int& msgTag, int& bufid, int& bytes, int& tid);

};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
