/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_sead.h 
@SE  CREATED     : 18.2.1999
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : semaphore administration class for semaphore server
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_SEAD_INCLUDE
#define _PVM_SEAD_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_def.h"
#include "pvm_exra.h"
#include "pvm_exre.h"
#include "pvm_exfo.h"
#include "pvm_expa.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_SEAD_HEADER "pvm_sead.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

// class declarations to avoid cyclic includes
class cppvmObjectCollection;
class cppvmString;

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmSemaphoreAdministrator
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm system base class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmSemaphoreAdministrator 
{
public:
  cppvmSemaphoreAdministrator();
  virtual ~cppvmSemaphoreAdministrator();
  bool operator!=(const cppvmSemaphoreAdministrator& obj) const {
    return ! (operator==(obj));
  }
  bool operator==(const cppvmSemaphoreAdministrator& obj) {return cppvmIsSame(obj);}
  bool operator==(const cppvmSemaphoreAdministrator& obj) const {return cppvmIsSame(obj);}

  // test of same identity
  bool cppvmIsSame(const cppvmSemaphoreAdministrator& obj) const{
    return cppvmObjId == obj.cppvmObjId;
  }

  virtual void cppvmPrintOn(ostream& strm= cout);
  virtual void cppvmScanFrom(istream& /* strm=cin */) {};

  int getCPPvmObjId() { return cppvmObjId; }
  
public:
  bool enter(cppvmString& semaName, int tid);
  bool enterNB(cppvmString& semaName);
  int leave(cppvmString& semaName);
  
  bool setNumber (cppvmString& semaName, int newNumber);
  int getNumber (cppvmString& semaName);
  void resetNumber(cppvmString& semaName, int newNumber);

protected:
  void addNewSemaphoreToList(cppvmString& semaName);
  int getIndexOfSemaphoreInList(cppvmString& semaName);
  
  cppvmObjectCollection complexSemaphoreList;

private:
  static int cppvmMaxId; // highest actual ID
  int cppvmObjId;        // specific object ID
  static int cppvmTotalObjNum; // total number of object for this class

private:
  // do not allow copy operation
  cppvmSemaphoreAdministrator(const cppvmSemaphoreAdministrator&);
  cppvmSemaphoreAdministrator& operator=(const cppvmSemaphoreAdministrator& obj);
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// inline functions 
inline ostream&
operator<<(ostream& strm, cppvmSemaphoreAdministrator& obj){
  obj.cppvmPrintOn(strm);
  return strm;
}

inline istream&
operator>>(istream& strm, cppvmSemaphoreAdministrator& obj){
  obj.cppvmScanFrom(strm);
  return strm;
}

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
