/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_sema.h 
@SE  CREATED     : 18.2.1999
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : semaphore class for process synchronization
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_SEMA_INCLUDE
#define _PVM_SEMA_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_def.h"
#include "pvm_exra.h"
#include "pvm_exre.h"
#include "pvm_exfo.h"
#include "pvm_expa.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_SEMA_HEADER "pvm_sema.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

// class declarations to avoid cyclic includes
class cppvmString;
class cppvmInt;
class cppvmReceiveStream;
class cppvmSendStream;
class cppvmConnection;

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmSemaphore
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : notificatin class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmSemaphore 
{
public:
  cppvmSemaphore();
  cppvmSemaphore(char *semaName, char *hostName=NULL);
  virtual ~cppvmSemaphore();
  cppvmSemaphore(const cppvmSemaphore&);
  cppvmSemaphore& operator=(const cppvmSemaphore& obj);
  bool operator!=(const cppvmSemaphore& obj) const {
    return ! (operator==(obj));
  }
  bool operator==(const cppvmSemaphore& obj) {
    return semaKeyName==obj.semaKeyName;}
  bool operator==(const cppvmSemaphore& obj) const {
    return semaKeyName==obj.semaKeyName;}

  // test of same identity
  bool cppvmIsSame(const cppvmSemaphore& obj) const{
    return cppvmObjId == obj.cppvmObjId;
  }

  virtual void cppvmPrintOn(ostream& strm= cout);
  virtual void cppvmScanFrom(istream& /*strm= cin*/) {};

  int getCPPvmObjId() { return cppvmObjId; }

public:
  void setName(char *semaName);

public:
  void enter();
  bool enterNB();
  void leave();
  
  void setNumber(unsigned int number);
  bool tryToSetNumber(unsigned int number);
  unsigned int getNumber();
  void resetNumber(unsigned int number);

public:
  void killServer();
  void startServer(char *hostName);

protected:
  cppvmString semaKeyName;

protected:
  void connectSemaphoreServer(char *hostName);
  void connectSemaphoreServerTry(char *hostName);
  bool checkSemaphoreServerRunning();
  void startSemaphoreServer(char *hostName);
  void getSemaphoreServerTid();
  void taskInfo(int where,
                int *ntask,
                struct pvmtaskinfo **taskp);

  int mboxEntry;

protected:
  void enterTry();
  bool enterNBTry();
  void leaveTry();
  bool tryToSetNumberTry(unsigned int number);

  unsigned int getNumberTry();
  void setNumberTry(unsigned int number);

  void resetNumberTry(unsigned int number);
  
protected:
  // static variables
  static int serverTid;
  static bool semaphoreServerConnected;
  static cppvmReceiveStream receiveBStream;
  static cppvmSendStream sendNBStream;
  

protected:
  cppvmInt operation;
  cppvmInt operationResult;
  cppvmInt numSemaSet;
  cppvmInt numSemaGet;
  bool serverTransfer();
  
private:
  static int cppvmMaxId; // highest actual ID
  int cppvmObjId;        // specific object ID
  static int cppvmTotalObjNum; // total number of object for this class
public:
  // only for 'cppvmMutexCollection'
  bool internalTryToSetNumber(unsigned int number);

};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// inline functions 
inline ostream&
operator<<(ostream& strm, cppvmSemaphore& obj){
  obj.cppvmPrintOn(strm);
  return strm;
}

inline istream&
operator>>(istream& strm, cppvmSemaphore& obj){
  obj.cppvmScanFrom(strm);
  return strm;
}

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
