/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_send.h 
@SE  CREATED     : 5.12.1997
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : description of a pvm send command
@SD                Just use the default constructor for this class!
@SD                Just use the get/set methods if you REALLY know
@SD                what you are doing!
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : Warning: sending to non existant TIDs is not possible.
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_SEND_INCLUDE
#define _PVM_SEND_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_strm.h"

/*@@@**************************************************************************
**********                                                           **********
*******                        EXPORTED TYPES                           *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendMode
@TD  DESCRIPTION : describes how pvm should send a message
***
******************************************************************************/
typedef enum {
  CPPvmBlockingSend    = 0,
  CPPvmNonblockingSend = 1
} cppvmSendMode;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendConnectionType
@TD  DESCRIPTION : describes the which send model is used
***
******************************************************************************/
typedef enum {
  CPPvmOneToOne  = 0,
  CPPvmMulticast   = 1,
  CPPvmGroupBroadcast = 2
} cppvmSendConnectionType;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendGroup
@TD  DESCRIPTION : send to a group of pvm processes
***
******************************************************************************/
typedef enum {
  CPPvmSBgroup  = 0,
  CPPvmSNBgroup = 1
} cppvmSendGroup;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendTid
@TD  DESCRIPTION : send to a process specified by its task id
***
******************************************************************************/
typedef enum {
  CPPvmSBtid  = 0,
  CPPvmSNBtid = 1
} cppvmSendTid;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendTids
@TD  DESCRIPTION : send to processes specified by their task ids
***
******************************************************************************/
typedef enum {
  CPPvmSBtids  = 0,
  CPPvmSNBtids = 1
} cppvmSendTids;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendBack
@TD  DESCRIPTION : send to a process from which a message was received
***
******************************************************************************/
typedef enum {
  CPPvmSBback  = 0,
  CPPvmSNBback = 1
} cppvmSendBack;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendSpawnMultiConnection
@TD  DESCRIPTION : send to several children
***
******************************************************************************/
typedef enum {
  CPPvmSBtaggedChildren  = 0,
  CPPvmSNBtaggedChildren = 1,
  CPPvmSBallChildren  = 3,
  CPPvmSNBallChildren = 4
} cppvmSendSpawnMultiConnection;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendConnection
@TD  DESCRIPTION : send to the parent process
***
******************************************************************************/
typedef enum {
  CPPvmSBparent  = 0,
  CPPvmSNBparent = 1
} cppvmSendConnection;


/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendSpawnConnection
@TD  DESCRIPTION : send to a child process
***
******************************************************************************/
typedef enum {
  CPPvmSBchild  = 0,
  CPPvmSNBchild = 1
} cppvmSendSpawnConnection;

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : cppvmSendTask
@TD  DESCRIPTION : send to a process specified by cppvmTask
***
******************************************************************************/
typedef enum {
  CPPvmSBtask  = 0,
  CPPvmSNBtask = 1
} cppvmSendTask;

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED CONSTANTS                        *******
**********                                                           **********
******************************************************************************/
#define PVM_SEND_HEADER "pvm_send.h"


/*@@@**************************************************************************
**********                                                           **********
*******                            MACROS                               *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmSendThrowIfNotInit
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : throws an exception if the object was not
@MD                 initialized (made with the default constructor)
@MU  USAGE        : only for internal use!
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmSendThrowIfNotInit                                                \
if(!initialized){                                                              \
  CPPvmInternalThrow (cppvmExceptionForbiddenCall(PVM_SEND_HEADER,             \
                       "cppvmSendStream",                                      \
                       "Macro: CPPvmSendThrowIfNotInit",                       \
                       "object not initialized, it is not allowed to use it"));\
}


/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmSendStream
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : description of a pvm send command
@KC  CONSTRAINTS   : 
******************************************************************************/

class cppvmSendStream : public cppvmStream
{
protected:
  bool initialized;
  
  // own send message buffer?
  bool msgBufSend;
  // encoding of own send buffer
  int sendBufEncoding;
  
  cppvmSendMode sendMode;
  cppvmSendConnectionType sendConnection;
  
  int           numTid;
  int           *sendToTids;
  int           *context;

  int           initSend;

  char *group;
  bool groupAllocated;

public:
  cppvmSendStream();

public:
  // set for cppvmTask objects
  cppvmSendStream(cppvmTask& task,
                  cppvmSendTask mode,
                  int initSend_ = PvmDataDefault);
  // set for cppvmConnection objects
  cppvmSendStream(cppvmConnection& conn,
                  cppvmSendConnection mode,
                  int initSend_ = PvmDataDefault);
  // set for cppvmSpawnConnection objects
  cppvmSendStream(cppvmSpawnConnection& sconn,
                  cppvmSendSpawnConnection mode,
                  int initSend_ = PvmDataDefault);
  // set for cppvmSpawnMultiConnection objects
  cppvmSendStream(cppvmSpawnMultiConnection& msconn,
                  cppvmSendSpawnMultiConnection mode,
                  int initSend_ = PvmDataDefault);
  // set for group multicasting with current context
  cppvmSendStream(char *group_,
                  cppvmSendGroup mode,
                  int initSend_ = PvmDataDefault);
  // set for group multicasting with given context
  cppvmSendStream(char *group_, int context_,
                  cppvmSendGroup mode,
                  int initSend_ = PvmDataDefault);
  // set for sending to a specified process with current context
  cppvmSendStream(int taskId,
                  cppvmSendTid mode,
                  int initSend_ = PvmDataDefault);
  // set for sending to a specified process with given context
  cppvmSendStream(int taskId, int context_,
                  cppvmSendTid mode,
                  int initSend_ = PvmDataDefault);
  // set for sending to specified processes with current context
  cppvmSendStream(int* taskIds,                             
                  unsigned int numTasks,
                  cppvmSendTids mode,
                  int initSend_ = PvmDataDefault);
  // set for sending to specified processes with given context
  cppvmSendStream(int* taskIds, 
                  unsigned int numTasks,
                  int context_,
                  cppvmSendTids mode,
                  int initSend_ = PvmDataDefault);
  // set for sending to specified processes with given context array
  cppvmSendStream(int* taskIds, 
                  unsigned int numTasks,
                  int* contextArr,
                  cppvmSendTids mode,
                  int initSend_ = PvmDataDefault);
  // set for sending back to a process from which a message was
  // received with 'recStrm'; send with current context
  cppvmSendStream(cppvmReceiveStream& recStrm,
                  cppvmSendBack mode,
                  int initSend_ = PvmDataDefault);
  // set for sending back to a process from which a message was
  // received with 'recStrm'; send with given context
  cppvmSendStream(cppvmReceiveStream& recStrm,
                  int context_,
                  cppvmSendBack mode,
                  int initSend_ = PvmDataDefault);
public:
  
  virtual ~cppvmSendStream();
  cppvmSendStream(const cppvmSendStream&);

  cppvmSendStream& operator=(const cppvmSendStream&);
  bool operator==(const cppvmSendStream&);
  bool operator==(const cppvmSendStream&) const;
  bool operator!=(const cppvmSendStream&);

  virtual void cppvmPrintOn(ostream& = cout);
  virtual void cppvmScanFrom(istream& = cin) {};

public:
  // set for cppvmTask objects
  void set(cppvmTask& task,
           cppvmSendTask mode,
           int initSend_ = PvmDataDefault);
  // set for cppvmConnection objects
  void set(cppvmConnection& conn,
           cppvmSendConnection mode,
           int initSend_ = PvmDataDefault);
  // set for cppvmSpawnConnection objects
  void set(cppvmSpawnConnection& sconn,
           cppvmSendSpawnConnection mode,
           int initSend_ = PvmDataDefault);
  // set for cppvmSpawnMultiConnection objects
  void set(cppvmSpawnMultiConnection& msconn,
           cppvmSendSpawnMultiConnection mode,
           int initSend_ = PvmDataDefault);
  // set for group multicasting with current context
  void set(char *group_,
           cppvmSendGroup mode,
           int initSend_ = PvmDataDefault);
  // set for group multicasting with given context
  void set(char *group_, int context_,
           cppvmSendGroup mode,
           int initSend_ = PvmDataDefault);
  // set for sending to a specified process with current context
  void set(int taskId,
           cppvmSendTid mode,
           int initSend_ = PvmDataDefault);
  // set for sending to a specified process with given context
  void set(int taskId, int context_,
           cppvmSendTid mode,
           int initSend_ = PvmDataDefault);
  // set for sending to specified processes with current context
  void set(int* taskIds,                             
           unsigned int numTasks,
           cppvmSendTids mode,
           int initSend_ = PvmDataDefault);
  // set for sending to specified processes with given context
  void set(int* taskIds, 
           unsigned int numTasks,
           int context_,
           cppvmSendTids mode,
           int initSend_ = PvmDataDefault);
  // set for sending to specified processes with given context array
  void set(int* taskIds, 
           unsigned int numTasks,
           int* contextArr,
           cppvmSendTids mode,
           int initSend_ = PvmDataDefault);
  // set for sending back to a process from which a message was
  // received with 'recStrm'; send with current context
  void set(cppvmReceiveStream& recStrm,
           cppvmSendBack mode,
           int initSend_ = PvmDataDefault);
  // set for sending back to a process from which a message was
  // received with 'recStrm'; send with given context
  void set(cppvmReceiveStream& recStrm,
           int context_,
           cppvmSendBack mode,
           int initSend_ = PvmDataDefault);
  
public:
  void setMsgBuf(bool ownMsgBuffer,
		 int encoding=PvmDataDefault);
  
  void setContext(int context_);
  int getContext(int indexContext){
    if ( indexContext>= numTid){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_SEND_HEADER,
                                                   "cppvmSendStream",
                                                   "getContext(int indexContext)",
                                                   "Index is not in the array limits"));
    }
    return context[indexContext];
  }
  
  void addTask(int tid, int context_);
  void remTask(int tid);

  void setTid(int tid);
  void setTid(int tid, int indexTid);
  
  void setGroup(char *group_);
  void setSendMode(cppvmSendMode newMode);
  void setSendConnection(cppvmSendConnectionType newConnection);
  void setInitSend(int initSend_);

  int           getNumTid() {
    CPPvmSendThrowIfNotInit
    return numTid;
  }
  int*          getTidsDirect() {
    CPPvmSendThrowIfNotInit
    return sendToTids;
  }
  int           getTid() {
    CPPvmSendThrowIfNotInit
    return sendToTids[0];
  }
  int getTid(int indexTid){
    if ( indexTid>= numTid){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_SEND_HEADER,
                                                   "cppvmSendStream",
                                                   "getTid(int indexTid)",
                                                   "Index is not in the array limits"));
    }
    return sendToTids[indexTid];
  }
  int& operator[] (int indexTid){
    if ( indexTid>= numTid){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_SEND_HEADER,
                                                   "cppvmSendStream",
                                                   "operator[]",
                                                   "Index is not in the array limits"));
    }
    return sendToTids[indexTid];
  }
  void          getTids(int *tids, int *num);
  cppvmSendMode getSendMode()    {
    CPPvmSendThrowIfNotInit
    return sendMode;
  }
  cppvmSendConnectionType getSendConnection()    {
    CPPvmSendThrowIfNotInit
    return sendConnection;
  }
  int getInitSend();
  char* getGroup();

  bool getMsgBuf(){
    CPPvmSendThrowIfNotInit
    return msgBufSend;
  };
  int getMsgBufEncoding(){
    CPPvmSendThrowIfNotInit
    return sendBufEncoding;
  };

protected:
  void resetObject();

};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
