/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_spaw.h 
@SE  CREATED     : 5.12.1997
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm spawn class
@SD                this class should be used for parents
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : IMPORTANT: Multicasted (array of tids) and broadcasted
@SR                           (to group members) messages are NOT send to
@SR                           the caller even if the caller is in the
@SR                           array of tids resp. member of the group
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_SPAW_INCLUDE
#define _PVM_SPAW_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_conn.h"
#include "pvm_coll.h"

/*@@@**************************************************************************
**********                                                           **********
*******                        EXPORTED TYPES                           *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : pvmContextMode
@TD  DESCRIPTION : describes how pvm should set the context while spawning
***
******************************************************************************/
typedef enum {
  CPPvmCurrentContext = 0,
  CPPvmNewContext     = 1,
  CPPvmGivenContext   = 2
} cppvmContextMode;

/*@@@**************************************************************************
**********                                                           **********
*******                            MACROS                               *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : PVM_THROW_IF_CHILD_KILLED
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : throws an exception if the child process was killed
@MD                 by the use of 'killChild()'
@MU  USAGE        : only for internal use!
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define PVM_THROW_IF_CHILD_KILLED                                                  \
if(childKilled){                                                                   \
  CPPvmInternalThrow (cppvmExceptionForbiddenCall(PVM_REC_HEADER,                  \
                       "cppvmSpawnConnection",                                     \
                       "Macro: PVM_THROW_IF_CHILD_KILLED",                         \
                       "spawned process was killed, it is not allowed to use it"));\
}


/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmSpawnConnection
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm spawn class
@KD                  use this class for pvm masters
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmSpawnConnection : public cppvmConnection{

public:
  // spawn process
  cppvmSpawnConnection(char *task,
                       char *args = "",
                       int optionFlag = PvmTaskDefault,
                       char *where = "",
                       cppvmContextMode contextFlag = CPPvmCurrentContext,
                       int   contextId_=0);
  cppvmSpawnConnection(char *task,
                       char **argv,
                       int optionFlag = PvmTaskDefault,
                       char *where = "",
                       cppvmContextMode contextFlag = CPPvmCurrentContext,
                       int   contextId_=0);
  // copy constructor
  cppvmSpawnConnection(const cppvmSpawnConnection&);
  virtual ~cppvmSpawnConnection();

  cppvmSpawnConnection& operator=(const cppvmSpawnConnection&);
  bool operator==(const cppvmSpawnConnection&);
  bool operator==(const cppvmSpawnConnection&) const;
  bool operator!=(const cppvmSpawnConnection& obj) const {
    return ! (operator==(obj));
  }
  
  virtual void cppvmPrintOn(ostream& = cout);
  virtual void cppvmScanFrom(istream& = cin) {}

public:
  int getChildParentTid() { return myTid; }
  cppvmTask getChildTask();
  cppvmString getChildName(){return childName;}
  unsigned int getNumChildParams(){
    return childParams.numElems();
  }
  cppvmString& getChildParams(unsigned int index=0){
    return (cppvmString&)(childParams[index]);
  }

  void killChild();

  int  getChildHostDaemonTid() { 
    PVM_THROW_IF_CHILD_KILLED
    return childHostDaemonTid;
  }

  cppvmString getChildHostName(){
    PVM_THROW_IF_CHILD_KILLED
    return childHostName;
  }
  cppvmString getChildHostArchitecture(){
    PVM_THROW_IF_CHILD_KILLED
    return childHostArchitecture;
  }
  int getChildHostSpeed(){
    PVM_THROW_IF_CHILD_KILLED
    return childHostSpeed;
  }

  int getChildStat(){
    return getTaskStat(childTid[0]);
  }

  int getChildTid() {
    PVM_THROW_IF_CHILD_KILLED
    return childTid[0];
  }
  unsigned long int getEnrolledChildUSec() {
    PVM_THROW_IF_CHILD_KILLED
    return enrolledChild;
  }
  double getEnrolledChildSecs() {
    return (((double) enrolledChild) / ((double) CLOCKS_PER_SEC));
  }

protected:
  int childTid[1];
  unsigned long int enrolledChild;
  bool childKilled;
  cppvmString childName;
  cppvmObjectCollection childParams;
  int childHostDaemonTid;
  cppvmString childHostName;
  cppvmString childHostArchitecture;
  int childHostSpeed;
  

private:
  void spawnIt(char *task,
               char **argv,
               int optionFlag,
               char *where,
               cppvmContextMode contextFlag,
               int   contextId_);
  void setContextMode(cppvmContextMode contextFlag,
                      int   contextId_);
  void setHost(int optionFlag, char *where);

  void setHostEnvironment();
  
  bool spawnError(int infoErr, int tidErr, char errMsg[]);

  bool flagMemberOfSet(int flagSet, int flag);
  
  // do not allow send/receive/forward for this class
  void cppvmSend(cppvmSendStream& sendStrm);
  bool cppvmReceive(cppvmReceiveStream& recStrm);
  bool cppvmForward(cppvmReceiveStream& recStrm, cppvmSendStream& sendStrm);

};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

