/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_spmu.h 
@SE  CREATED     : 4.2.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm spawn multi class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_SPMU_INCLUDE
#define _PVM_SPMU_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_proh.h"
#include "pvm_syco.h"
#include "pvm_spaw.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_SPMU_MODULE "pvm_spmu.h"

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmSpawnMultiConnectionTagCleanFieldSize
@CD  DESCRIPTION   : size for the memcopy field when cleaning tags
***
******************************************************************************/
#define CPPvmSpawnMultiConnectionTagCleanFieldSize 32

/*@@@**************************************************************************
**********                                                           **********
*******                        EXPORTED TYPES                           *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : pvmMultiContextMode
@TD  DESCRIPTION : describes how pvm should set the context while spawning
***
******************************************************************************/
typedef enum {
  CPPvmMultiOneNewContext  = 0,
  CPPvmMultiAllNewContext  = 1
} cppvmMultiContextMode;

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmSpawnMultiConnection
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : spawn multi class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmSpawnMultiConnection : public cppvmProcessHandling{
public:

  // spawn multi processes
  cppvmSpawnMultiConnection(char *task,
                            unsigned int numTasks = 1,
                            char *args = "",
                            int optionFlag = PvmTaskDefault,
                            char *where = "",
                            cppvmContextMode contextFlag = CPPvmCurrentContext,
                            int   contextId_=0,
                            cppvmMultiContextMode multiContextFlag =
                            CPPvmMultiAllNewContext);

  // spawn multi processes
  cppvmSpawnMultiConnection(char *task,
                            unsigned int numTasks,
                            char **argv,
                            int optionFlag = PvmTaskDefault,
                            char *where = "",
                            cppvmContextMode contextFlag = CPPvmCurrentContext,
                            int   contextId_=0,
                            cppvmMultiContextMode multiContextFlag =
                            CPPvmMultiAllNewContext);

  cppvmSpawnMultiConnection();
  cppvmSpawnMultiConnection(const cppvmSpawnMultiConnection&){};
  virtual ~cppvmSpawnMultiConnection();

  bool operator==(cppvmSpawnMultiConnection&);
  bool operator!=(cppvmSpawnMultiConnection& obj){
    return ! (operator==(obj));
  }
  
  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  void add(cppvmSpawnConnection* objPtr);
  void remove(cppvmSpawnConnection& obj);
  void remove(cppvmSpawnConnection* objPtr);
  void remove(unsigned int index);
  void clean();

  unsigned int numElems() { return spawnCollection.numElems(); }

  cppvmSpawnConnection& operator[] (unsigned int index){
    if(index>spawnCollection.numElems()){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_SPMU_MODULE,
                                                   "cppvmSpawnMultiConnection",
                                                   "operator[unsigned int]",
                                                   "bad index"));
    }
    return ((cppvmSpawnConnection&)spawnCollection[index]);
  }

  void append(cppvmSpawnMultiConnection& obj);

  void tagElem(unsigned int index);
  void untagElem(unsigned int index);

  void tagAllElems();
  void untagAllElems();

  bool isElemTagged(unsigned int index);

  void killTaggedChildren();
  void killAllChildren();
  
  void halt();
public:
  // cppvmTask collection
  cppvmObjectCollection *getTaggedChildTasks();
  // cppvmTask collection
  cppvmObjectCollection *getAllChildTasks();
  
protected:
  // collection of cppvmSpawnConnection objects
  cppvmProcessHandlingCollection spawnCollection;
  // indicates, weather an cppvmSpawnConnection object is tagged or not
  // has the same length as spawnCollection.numElems()
  char*      tags;
  
private:
  // spawn multi processes
  void spawnIt(char *task,
               unsigned int numTasks,
               char **argv,
               int optionFlag,
               char *where,
               cppvmContextMode contextFlag,
               int   contextId_,
               cppvmMultiContextMode multiContextFlag);
  void initDaemon();
  void initTag(unsigned int numElem);
  void resizeTag(unsigned int numElem,
                 unsigned int oldElems);
  void shiftTagsLeft(unsigned int indexMin,
                     unsigned int indexMax);
  
  // do not allow assignments
  cppvmSpawnMultiConnection& operator=(const cppvmSpawnMultiConnection&){
    // dummy return
    return *this;
  }
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



