/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_stl.h 
@SE  CREATED     : 31.1.1999
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : CPPvm stl transfer functions
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STL_INCLUDE
#define _PVM_STL_INCLUDE

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl
/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

// cppvm includes
#include "pvm_cincl.h"
#include "pvm_tran.h"
#include "pvm_int.h"
#include "pvm_nssdata.h"

// stl includes
#include <vector>
#include <deque>
#include <list>
#include <set>
#include <map>
#include <utility>
#include <string>
#include <complex>
#ifdef CPPvmUseStlBitset
#include <bitset>
#endif // CPPvmUseStlBitset
#ifdef CPPvmUseStlSlist
#ifdef __GNUC__
// gcc version < 3.1.0 ?
#if __GNUC__ < 3 || \
    (__GNUC__ == 3 && __GNUC_MINOR__ < 1)
#include <slist>
#else
#include <ext/slist>
#endif // GCC_VERSION
#else
#include <slist>
#endif // __GNUC__
#endif // CPPvmUseStlSlist
#include <queue>
#include <stack>
#ifdef CPPvmUseStlValarray
#include <valarray>
#endif // CPPvmUseStlValarray

#ifndef CPPvmDoNotUseStdNamespace

// use stl classes
using std::vector;
using std::deque;
using std::list;
using std::set;
using std::multiset;
using std::map;
using std::multimap;
using std::pair;
using std::string;
using std::complex;
#ifdef CPPvmUseStlBitset
using std::bitset;
#endif // CPPvmUseStlBitset
#ifdef CPPvmUseStlSlist
#ifdef __GNUC__
// gcc version < 3.1.0 ?
#if __GNUC__ < 3 || \
    (__GNUC__ == 3 && __GNUC_MINOR__ < 1)
using std::slist;
#else
using __gnu_cxx::slist;
#endif // GCC_VERSION
#else
using std::slist;
#endif // __GNUC__
#endif // CPPvmUseStlSlist
using std::queue;
using std::priority_queue;
using std::stack;
#ifdef CPPvmUseStlValarray
using std::valarray;
#endif // CPPvmUseStlValarray

#endif // CPPvmDoNotUseStdNamespace

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                            FUNCTIONS                            *******
**********                                                           **********
******************************************************************************/


// stl transfer functions
void cppvmTransferStlString (string& data, int stride);
void cppvmTransferStlString (string * data, int nitem, int stride);

template <class T>
inline
void cppvmTransferStlVector (vector<T> & data, int stride ) {
  typename vector<T>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      for (i=data.begin(); i != data.end() ; i++) {
        cppvmTransferTemplate( *i, stride);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    if((int)(tmpSize)>0){
      for (i=data.begin(); i != data.end() ; i++) {
        cppvmTransferTemplate( *i, stride);
      }
    }
  }
}
// specialized transfer functions for vector
template <>
void cppvmTransferStlVector (vector<bool> & data, int stride);

template <class T>
inline
void cppvmTransferStlVector (vector<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class T>
inline
void cppvmTransferStlDeque (deque<T> & data, int stride ) {
  typename deque<T>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      for (i=data.begin(); i != data.end() ; i++) {
        cppvmTransferTemplate( *i, stride);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    if((int)(tmpSize)>0){
      for (i=data.begin(); i != data.end() ; i++) {
        cppvmTransferTemplate( *i, stride);
      }
    }
  }
}

// specialized transfer functions for deque
template <>
void cppvmTransferStlDeque (deque<bool> & data, int stride);

template <class T>
inline
void cppvmTransferStlDeque (deque<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class T>
inline
void cppvmTransferStlList (list<T> & data, int stride ) {
  typename list<T>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      for (i=data.begin(); i != data.end() ; i++) {
        cppvmTransferTemplate( *i, stride);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    for (i=data.begin(); i != data.end() ; i++) {
      cppvmTransferTemplate( *i, stride);
    }
  }
}

// specialized transfer functions for list
template <>
void cppvmTransferStlList (list<bool> & data, int stride);


template <class T>
inline
void cppvmTransferStlList (list<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class T>
inline
void cppvmTransferStlSet (set<T> & data, int stride ) {
  typename set<T>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if(data.size()>0){
      cppvmNewSendStackData<T> array(data.size());
      int j=0;
      for (i=data.begin(); i != data.end() ; i++) {
        array.data[j]=(*i);
        cppvmTransferTemplate( array.data[j], stride);
        j++;
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.clear();
    T elem;
    int j;
    for (j=0; j<(int)(tmpSize); j++) {
      cppvmTransferTemplate( elem, stride);
      data.insert(elem);
    }
  }
}

template <class T>
inline
void cppvmTransferStlSet (set<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class T>
inline
void cppvmTransferStlMultiset (multiset<T> & data, int stride ) {
  typename multiset<T>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if(data.size()>0){
      cppvmNewSendStackData<T> array(data.size());
      int j=0;
      for (i=data.begin(); i != data.end() ; i++) {
        array.data[j]=(*i);
        cppvmTransferTemplate( array.data[j], stride);
        j++;
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.clear();
    T elem;
    int j;
    for (j=0; j<(int)(tmpSize); j++) {
      cppvmTransferTemplate( elem, stride);
      data.insert(elem);
    }
  }
}

// specialized transfer functions for multiset
template <>
void cppvmTransferStlMultiset (multiset<bool> & data, int stride);

template <class T>
inline
void cppvmTransferStlMultiset (multiset<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class Key, class T>
inline
void cppvmTransferStlMap (map<Key, T> & data, int stride ) {
  typename map<Key, T>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if(data.size()>0){
      cppvmNewSendStackData<Key> posArray(data.size());
      cppvmNewSendStackData<T> varArray(data.size());
      int j=0;
      for (i=data.begin(); i != data.end() ; i++) {
        posArray.data[j]=i->first;
        varArray.data[j]=i->second;
        cppvmTransferTemplate( posArray.data[j], stride);
        cppvmTransferTemplate( varArray.data[j], stride);
        j++;
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.clear();
    Key pos;
    T var;
    int j;
    for (j=0; j<(int)(tmpSize); j++) {
      cppvmTransferTemplate( pos, stride);
      cppvmTransferTemplate( var, stride);
      typename map<Key, T>::value_type elem(pos, var);
      data.insert(elem);
    }
  }
}

// Visual C++ can not handle partly instantiated templates
#ifdef CPPVM_NODEADCODE
template <class Key>
inline
void cppvmTransferStlMap (map<Key, bool> & data, int stride ) {
  typename map<Key, bool>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if(data.size()>0){
      cppvmNewSendStackData<Key> posArray(data.size());
      cppvmNewSendStackData<cppvmArrayBool> varArray(1);
      varArray.data[0].resize(data.size());
      int j=0;
      for (i=data.begin(); i != data.end() ; i++) {
        posArray.data[j]=i->first;
        (varArray.data[0])[j]=i->second;
        cppvmTransferTemplate( posArray.data[j], stride);
        j++;
      }
      varArray.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.clear();
    if((int)(tmpSize)>0){
      cppvmNewSendStackData<Key> pos((int)(tmpSize));
      cppvmArrayBool var;
      var.resize((int)(tmpSize));
      int j;
      for (j=0; j<(int)(tmpSize); j++) {
        cppvmTransferTemplate( pos.data[j], stride);
      }
      var.cppvmTransfer();
      for (j=0; j<(int)(tmpSize); j++) {
          typename map<Key, bool>::value_type elem(pos.data[j], var[j]);
        data.insert(elem);
      }
    }
  }
}
#endif // CPPVM_NODEADCODE



template <class Key, class T>
inline
void cppvmTransferStlMap (map<Key, T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class Key, class T>
inline
void cppvmTransferStlMultimap (multimap<Key, T> & data, int stride ) {
  typename multimap<Key, T>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if(data.size()>0){
      cppvmNewSendStackData<Key> posArray(data.size());
      cppvmNewSendStackData<T> varArray(data.size());
      int j=0;
      for (i=data.begin(); i != data.end() ; i++) {
        posArray.data[j]=i->first;
        varArray.data[j]=i->second;
        cppvmTransferTemplate( posArray.data[j], stride);
        cppvmTransferTemplate( varArray.data[j], stride);
        j++;
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.clear();
    Key pos;
    T var;
    int j;
    for (j=0; j<(int)(tmpSize); j++) {
      cppvmTransferTemplate( pos, stride);
      cppvmTransferTemplate( var, stride);
        typename multimap<Key, T>::value_type elem(pos, var);
      data.insert(elem);
    }
  }
}

// Visual C++ can not handle partly instantiated templates
#ifdef CPPVM_NODEADCODE
template <class Key>
inline
void cppvmTransferStlMultimap (multimap<Key, bool> & data, int stride ) {
  typename multimap<Key, bool>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if(data.size()>0){
      cppvmNewSendStackData<Key> posArray(data.size());
      cppvmNewSendStackData<cppvmArrayBool> varArray(1);
      varArray.data[0].resize(data.size());
      int j=0;
      for (i=data.begin(); i != data.end() ; i++) {
        posArray.data[j]=i->first;
        (varArray.data[0])[j]=i->second;
        cppvmTransferTemplate( posArray.data[j], stride);
        j++;
      }
      varArray.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.clear();
    if((int)(tmpSize)>0){
      cppvmNewSendStackData<Key> pos((int)(tmpSize));
      cppvmArrayBool var;
      var.resize((int)(tmpSize));
      int j;
      for (j=0; j<(int)(tmpSize); j++) {
        cppvmTransferTemplate( pos.data[j], stride);
      }
      var.cppvmTransfer();
      for (j=0; j<(int)(tmpSize); j++) {
          typename multimap<Key, bool>::value_type elem(pos.data[j], var[j]);
        data.insert(elem);
      }
    }
  }
}
#endif // CPPVM_NODEADCODE


template <class Key, class T>
inline
void cppvmTransferStlMultimap (multimap<Key, T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class T1, class T2>
inline
void cppvmTransferStlPair (pair<T1, T2> & data, int stride ) {
  cppvmTransferTemplate(data.first, stride);
  cppvmTransferTemplate(data.second, stride);
}

template <class T1, class T2>
inline
void cppvmTransferStlPair (pair<T1, T2> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class T>
inline
void cppvmTransferStlComplex (complex<T> & data, int stride ) {
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<T> array(2);
    array.data[0] = data.real();
    array.data[1] = data.imag();
    cppvmTransferTemplate( array.data[0], stride);
    cppvmTransferTemplate( array.data[1], stride);
  }
  else{
    T real;
    T imag;
    cppvmTransferTemplate( real, stride);
    cppvmTransferTemplate( imag, stride);
    complex<T> val(real, imag);
    data = val;
  }
}

template <class T>
inline
void cppvmTransferStlComplex (complex<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

#ifdef CPPvmUseStlBitset
template <unsigned int T>
inline
void cppvmTransferStlBitset (bitset<T> & data, int stride ) {
  int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(T);
    tmpSize.data[0].cppvmTransfer();
    if((int)(T)>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(T));
      for (i=0; i<((int)(T)); i++){
        (tmpData.data[0])[i] = data[i];
      }
      tmpData.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    if(((int)(T))!=(int)((tmpSize))){
      CPPvmInternalThrow (cppvmExceptionForbiddenCall("pvm_st.h",
                                                      "-",
                                                      "Function: cppvmTransferStlBitset",
                      "the sended bitset has a different size than the receive bitset. Since it is not possible to resize bitset, this is not allowed"));
    }
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      for (i=0; i<((int)(T)); i++){
        data[i] = tmpData[i];
      }
    }
  }
}

template <unsigned int T>
inline
void cppvmTransferStlBitset (bitset<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}
#endif // CPPvmUseStlBitset

#ifdef CPPvmUseStlSlist
template <class T>
inline
void cppvmTransferStlSlist (slist<T> & data, int stride ) {
  typename slist<T>::iterator i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      for (i=data.begin(); i != data.end() ; i++) {
        cppvmTransferTemplate( *i, stride);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    for (i=data.begin(); i != data.end() ; i++) {
      cppvmTransferTemplate( *i, stride);
    }
  }
}
// specialized transfer functions for slist
template <>
void cppvmTransferStlSlist (slist<bool> & data, int stride);

template <class T>
inline
void cppvmTransferStlSlist (slist<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}
#endif // CPPvmUseStlSlist

template <class T>
inline
void cppvmTransferStlQueue (queue<T> & data, int stride ) {
  int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<T> varArray(data.size());
      i=0;
      while(data.size()>0){
        varArray.data[i] = data.front();
        cppvmTransferTemplate( varArray.data[i], stride);
        data.pop();
        i++;
      }
      // copy data back
      for(i=0; i< (int)(tmpSize.data[0]); i++){
        data.push(varArray.data[i]);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    T value;
    tmpSize.cppvmTransfer();
    // delete all data
    while(data.size()>0){
      data.pop();
    }
    if((int)(tmpSize)>0){
      for(i=0; i<(int)(tmpSize); i++){
        cppvmTransferTemplate( value, stride);
        data.push(value);
      }
    }
  }
}

// specialized transfer functions for queue
template <>
void cppvmTransferStlQueue (queue<bool> & data, int stride);

template <class T>
inline
void cppvmTransferStlQueue (queue<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class T>
inline
void cppvmTransferStlPriorityqueue (priority_queue<T> & data, int stride ) {
  int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<T> varArray(data.size());
      i=0;
      while(data.size()>0){
        varArray.data[i] = data.top();
        cppvmTransferTemplate( varArray.data[i], stride);
        data.pop();
        i++;
      }
      // copy data back
      for(i=0; i< (int)(tmpSize.data[0]); i++){
        data.push(varArray.data[i]);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    T value;
    tmpSize.cppvmTransfer();
    // delete all data
    while(data.size()>0){
      data.pop();
    }
    if((int)(tmpSize)>0){
      for(i=0; i<(int)(tmpSize); i++){
        cppvmTransferTemplate( value, stride);
        data.push(value);
      }
    }
  }
}

#ifdef CPPVM_NODEADCODE
// The next specialized template function causes an internal error in Visual C++
template <>
void cppvmTransferStlPriorityqueue (priority_queue<bool> & data, int stride);
#endif // CPPVM_NODEADCODE

template <class T>
inline
void cppvmTransferStlPriorityqueue (priority_queue<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

template <class T>
inline
void cppvmTransferStlStack (stack<T> & data, int stride ) {
  int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<T> varArray(data.size());
      i=0;
      while(data.size()>0){
        varArray.data[i] = data.top();
        data.pop();
        i++;
      }
      // send all data
      // in different order
      for(i=(int)(tmpSize.data[0])-1; i>=0 ; i--){
        cppvmTransferTemplate( varArray.data[i], stride);
      }
      
      // copy data back
      for(i=(int)(tmpSize.data[0])-1; i>=0 ; i--){
        data.push(varArray.data[i]);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    T value;
    tmpSize.cppvmTransfer();
    // delete all data
    while(data.size()>0){
      data.pop();
    }
    if((int)(tmpSize)>0){
      // receive all data
      // in different order
      for(i=0; i<(int)(tmpSize); i++){
        cppvmTransferTemplate( value, stride);
        data.push(value);
      }
    }
  }
}

// specialized transfer functions for stack
template <>
void cppvmTransferStlStack (stack<bool> & data, int stride);

template <class T>
inline
void cppvmTransferStlStack (stack<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}

#ifdef CPPvmUseStlValarray
template <class T>
inline
void cppvmTransferStlValarray (valarray<T> & data, int stride ) {
  int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      for(i=0; i< (int)(data.size()); i++){
        cppvmTransferTemplate( data[i], stride);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    if((int)(tmpSize)>0){
      for(i=0; i< (int)(data.size()); i++){
        cppvmTransferTemplate( data[i], stride);
      }
    }
  }
}
// specialized transfer functions for valarray
template <>
void cppvmTransferStlValarray (valarray<bool> & data, int stride);

template <class T>
inline
void cppvmTransferStlValarray (valarray<T> * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferTemplate( data[i], stride);
  }
}
#endif // CPPvmUseStlValarray


// cppvmTransferTemplate functions for the stl classes
// Currently only 'string' can be supported
// due to Visual C++ constraints

template <>
inline
void cppvmTransferTemplate(string& data, int stride )
{
  cppvmTransferStlString(data, stride);
}

template <>
inline
void cppvmTransferTemplate(string * data, int nitem, int stride )
{
  cppvmTransferStlString(data, nitem, stride);
}


#ifdef CPPVM_NODEADCODE
// The next functions are currently not supported by Visual C++
// This prevents nested stl calls like:
// vector< deque<int> > vd;

template <class T>
inline
void cppvmTransferTemplate(vector<T> & data, int stride ) {
  cppvmTransferStlVector(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(vector<T> * data, int nitem, int stride ) {
  cppvmTransferStlVector(data, nitem, stride);
}

template <class T>
inline
void cppvmTransferTemplate(deque<T> & data, int stride ) {
  cppvmTransferStlDeque (data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(deque<T> * data, int nitem, int stride ) {
  cppvmTransferStlDeque (data, nitem, stride);
}

template <class T>
inline
void cppvmTransferTemplate(list<T> & data, int stride ) {
  cppvmTransferStlList(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(list<T> * data, int nitem, int stride ) {
  cppvmTransferStlList(data, nitem, stride);
}

template <class T>
inline
void cppvmTransferTemplate(set<T> & data, int stride ) {
  cppvmTransferStlSet(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(set<T> * data, int nitem, int stride ) {
  cppvmTransferStlSet(data, nitem, stride);
}

template <class T>
inline
void cppvmTransferTemplate(multiset<T> & data, int stride ) {
  cppvmTransferStlMultiset(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(multiset<T> * data, int nitem, int stride ) {
  cppvmTransferStlMultiset(data, nitem, stride);
}

template <class Key, class T>
inline
void cppvmTransferTemplate(map<Key, T> & data, int stride ) {
  cppvmTransferStlMap(data, stride);
}

template <class Key, class T>
inline
void cppvmTransferTemplate(map<Key, T> * data, int nitem, int stride ) {
  cppvmTransferStlMap(data, nitem, stride);
}

template <class Key, class T>
inline
void cppvmTransferTemplate(multimap<Key, T> & data, int stride ) {
  cppvmTransferStlMultimap(data, stride);
}

template <class Key, class T>
inline
void cppvmTransferTemplate(multimap<Key, T> * data, int nitem, int stride ) {
  cppvmTransferStlMultimap(data, nitem, stride);
}

template <class T1, class T2>
inline
void cppvmTransferTemplate(pair<T1, T2> & data, int stride ) {
  cppvmTransferStlPair(data, stride);
}

template <class T1, class T2>
inline
void cppvmTransferTemplate(pair<T1, T2> * data, int nitem, int stride ) {
  cppvmTransferStlPair(data, nitem, stride);
}

template <class T>
inline
void cppvmTransferTemplate(complex<T> & data, int stride ) {
  cppvmTransferStlComplex(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(complex<T> * data, int nitem, int stride ) {
  cppvmTransferStlComplex(data, nitem, stride);
}

#ifdef CPPvmUseStlBitset
template <class T>
inline
void cppvmTransferTemplate(bitset<T> & data, int stride ) {
  cppvmTransferStlBitset(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(bitset<T> * data, int nitem, int stride ) {
  cppvmTransferStlBitset(data, nitem, stride);
}
#endif // CPPvmUseStlBitset

#ifdef CPPvmUseStlSlist
template <class T>
inline
void cppvmTransferTemplate(slist<T> & data, int stride ) {
  cppvmTransferStlSlist(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(slist<T> * data, int nitem, int stride ) {
  cppvmTransferStlSlist(data, nitem, stride);
}
#endif // CPPvmUseStlSlist

template <class T>
inline
void cppvmTransferTemplate(queue<T> & data, int stride ) {
  cppvmTransferStlQueue(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(queue<T> * data, int nitem, int stride ) {
  cppvmTransferStlQueue(data, nitem, stride);
}

template <class T>
inline
void cppvmTransferTemplate(priority_queue<T> & data, int stride ) {
  cppvmTransferStlPriorityqueue(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(priority_queue<T> * data, int nitem, int stride ) {
  cppvmTransferStlPriorityqueue(data, nitem, stride);
}

template <class T>
inline
void cppvmTransferTemplate(stack<T> & data, int stride ) {
  cppvmTransferStlStack(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(stack<T> * data, int nitem, int stride ) {
  cppvmTransferStlStack(data, nitem, stride);
}

#ifdef CPPvmUseStlValarray
template <class T>
inline
void cppvmTransferTemplate(valarray<T> & data, int stride ) {
  cppvmTransferStlValarray(data, stride);
}

template <class T>
inline
void cppvmTransferTemplate(valarray<T> * data, int nitem, int stride ) {
  cppvmTransferStlValarray(data, nitem, stride);
}
#endif // CPPvmUseStlValarray



#endif // CPPVM_NODEADCODE


#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates


#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



