/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_stlbitset.h 
@SE  CREATED     : 4.7.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : cppvm stl bitset class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STLBITSET_INCLUDE
#define _PVM_STLBITSET_INCLUDE

#ifdef CPPvmUseStlBitset

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_obj.h"
#include "pvm_stl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

template <unsigned int T>
class cppvmStlBitset: public cppvmObject, public bitset<T> {
  
public:
  // common constructors
  cppvmStlBitset(): cppvmObject(), bitset<T>() {}
  cppvmStlBitset(unsigned long value): cppvmObject(), bitset<T>(value) {}
  explicit cppvmStlBitset(const string& str): cppvmObject(), bitset<T>(str) {}
  cppvmStlBitset(const string& str, typename string::size_type strIdx): 
      cppvmObject(), bitset<T>(str, strIdx) {}
  cppvmStlBitset(const string& str, typename string::size_type strIdx,
                 typename string::size_type strAnz): 
      cppvmObject(), bitset<T>(str, strIdx, strAnz) {}
  // destructor
  virtual ~cppvmStlBitset() {}
  
  // cppvmStlBitset assignment
  bitset<T>& operator=(const cppvmStlBitset<T>& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return bitset<T>::operator=((const bitset<T>&) obj);
  }
  // bitset assignment
  bitset<T>& operator=(const bitset<T>& obj){
    return bitset<T>::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlBitset<T>& obj) const {
    return !(operator==((const bitset<T>&) *this, (const bitset<T>&) obj));
  }
  bool operator==(const cppvmStlBitset& obj) {
    return operator==((const bitset<T>&) *this, (const bitset<T>&) obj);
  }
  bool operator==(const cppvmStlBitset& obj) const {
    return operator==((const bitset<T>&) *this, (const bitset<T>&) obj);
  }

  virtual void cppvmTransfer() {
    cppvmTransferStlBitset ( (bitset<T>&) *this, 1);
  }
  
  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlBitset)
        
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

#ifndef CPPvmDoNotUseStdNamespace
// inline functions 
template <unsigned int T>
inline std::ostream&
operator<<(std::ostream& strm, cppvmStlBitset<T>& obj){
  return std::operator<<(strm, (bitset<T>&)obj);
}

template <unsigned int T>
inline std::istream&
operator>>(std::istream& strm, cppvmStlBitset<T>& obj){
  return std::operator>>(strm, (bitset<T>&)obj);
}
#else
// inline functions 
template <unsigned int T>
inline ostream&
operator<<(ostream& strm, cppvmStlBitset<T>& obj){
  return operator<<(strm, (bitset<T>&)obj);
}

template <unsigned int T>
inline istream&
operator>>(istream& strm, cppvmStlBitset<T>& obj){
  return operator>>(strm, (bitset<T>&)obj);
}
#endif

// send cppvmStlBitset classes
template <unsigned int T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlBitset<T>& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlBitset classes
template <unsigned int T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlBitset<T>& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates


#endif // CPPvmUseStlBitset

#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



