/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_stlcomplex.h 
@SE  CREATED     : 3.2.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : cppvm stl complex class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STLCOMPLEX_INCLUDE
#define _PVM_STLCOMPLEX_INCLUDE

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_obj.h"
#include "pvm_stl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

template <class T>
class cppvmStlComplex: public cppvmObject, public complex<T> {
  
public:
  // common constructors
  cppvmStlComplex(): cppvmObject(), complex<T>() {}
  cppvmStlComplex(const T& real): cppvmObject(), complex<T>(real) {}
  cppvmStlComplex(const T& real, const T& imag): 
      cppvmObject(), complex<T>(real, imag) {}
  // constructors for cppvmStlComplex
  cppvmStlComplex(const cppvmStlComplex<T>& obj) : 
      cppvmObject(), complex<T>((const complex<T>&)obj){}
  // constructors for complex
  cppvmStlComplex(const complex<T>& obj) : 
      cppvmObject(), complex<T>(obj) {}

  // destructor
  virtual ~cppvmStlComplex() {}
  
  // cppvmStlComplex assignment
  complex<T>& operator=(const cppvmStlComplex<T>& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return complex<T>::operator=((const complex<T>&) obj);
  }
  // complex assignment
  complex<T>& operator=(const complex<T>& obj){
    return complex<T>::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlComplex<T>& obj) const {
    return !(operator==((const complex<T>&) *this, (const complex<T>&) obj));
  }
  bool operator==(const cppvmStlComplex& obj) {
    return operator==((const complex<T>&) *this, (const complex<T>&) obj);
  }
  bool operator==(const cppvmStlComplex& obj) const {
    return operator==((const complex<T>&) *this, (const complex<T>&) obj);
  }

  virtual void cppvmTransfer() {
    cppvmTransferStlComplex ( (complex<T>&) *this, 1);
  }
  
  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlComplex)
        
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

#ifndef CPPvmDoNotUseStdNamespace
// inline functions 
template <class T>
inline std::ostream&
operator<<(std::ostream& strm, cppvmStlComplex<T>& obj){
  return std::operator<<(strm, (complex<T>&)obj);
}

template <class T>
inline std::istream&
operator>>(std::istream& strm, cppvmStlComplex<T>& obj){
  return std::operator>>(strm, (complex<T>&)obj);
}
#else
// inline functions 
template <class T>
inline ostream&
operator<<(ostream& strm, cppvmStlComplex<T>& obj){
  return operator<<(strm, (complex<T>&)obj);
}

template <class T>
inline istream&
operator>>(istream& strm, cppvmStlComplex<T>& obj){
  return operator>>(strm, (complex<T>&)obj);
}
#endif


// send cppvmStlComplex classes
template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlComplex<T>& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlComplex classes
template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlComplex<T>& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



