/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_stlmap.h 
@SE  CREATED     : 31.1.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : cppvm stl map class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STLMAP_INCLUDE
#define _PVM_STLMAP_INCLUDE

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_obj.h"
#include "pvm_stl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

template <class Key, class T>
class cppvmStlMap: public cppvmObject, public map<Key, T> {
  
public:
  // common constructors
  cppvmStlMap(): cppvmObject(), map<Key, T>() {}
  cppvmStlMap(typename map<Key, T>::const_iterator first, 
              typename map<Key, T>::const_iterator last): 
      cppvmObject(), map<Key, T>(first, last) {}
  // constructors for cppvmStlMap
  cppvmStlMap(const cppvmStlMap<Key, T>& obj) : 
      cppvmObject(), map<Key, T>((const map<Key, T>&)obj){}
  // constructors for map
  cppvmStlMap(const map<Key, T>& obj) : 
      cppvmObject(), map<Key, T>(obj) {}
  // destructor
  virtual ~cppvmStlMap() {}
  
  // cppvmStlMap assignment
  map<Key, T>& operator=(const cppvmStlMap<Key, T>& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return map<Key, T>::operator=((const map<Key, T>&) obj);
  }
  // map assignment
  map<Key, T>& operator=(const map<Key, T>& obj){
    return map<Key, T>::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlMap<Key, T>& obj) const {
    return !(operator==((const map<Key, T>&) *this, (const map<Key, T>&) obj));
  }
  bool operator==(const cppvmStlMap& obj) {
    return operator==((const map<Key, T>&) *this, (const map<Key, T>&) obj);
  }
  bool operator==(const cppvmStlMap& obj) const {
    return operator==((const map<Key, T>&) *this, (const map<Key, T>&) obj);
  }

  virtual void cppvmTransfer() {
    cppvmTransferStlMap ( (map<Key, T>&) *this, 1);
  }
  
  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlMap)
        
};

template <class Key, class T>
class cppvmStlMultimap: public cppvmObject, public multimap<Key, T> {
  
public:
  // common constructors
  cppvmStlMultimap(): cppvmObject(), multimap<Key, T>() {}
  cppvmStlMultimap(typename multimap<Key, T>::const_iterator first, 
                   typename multimap<Key, T>::const_iterator last): 
      cppvmObject(), multimap<Key, T>(first, last) {}
  // constructors for cppvmStlMultimap
  cppvmStlMultimap(const cppvmStlMultimap<Key, T>& obj) : 
      cppvmObject(), multimap<Key, T>((const multimap<Key, T>&)obj){}
  // constructors for multimap
  cppvmStlMultimap(const multimap<Key, T>& obj) : 
      cppvmObject(), multimap<Key, T>(obj) {}
  // destructor
  virtual ~cppvmStlMultimap() {}
  
  // cppvmStlMultimap assignment
  multimap<Key, T>& operator=(const cppvmStlMultimap<Key, T>& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return multimap<Key, T>::operator=((const multimap<Key, T>&) obj);
  }
  // multimap assignment
  multimap<Key, T>& operator=(const multimap<Key, T>& obj){
    return multimap<Key, T>::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlMultimap<Key, T>& obj) const {
    return !(operator==((const multimap<Key, T>&) *this, (const multimap<Key, T>&) obj));
  }
  bool operator==(const cppvmStlMultimap& obj) {
    return operator==((const multimap<Key, T>&) *this, (const multimap<Key, T>&) obj);
  }
  bool operator==(const cppvmStlMultimap& obj) const {
    return operator==((const multimap<Key, T>&) *this, (const multimap<Key, T>&) obj);
  }

  virtual void cppvmTransfer() {
    cppvmTransferStlMultimap ( (multimap<Key, T>&) *this, 1);
  }
  
  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlMultimap)
        
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// send cppvmStlMap classes
template <class Key, class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlMap<Key, T>& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlMap classes
template <class Key, class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlMap<Key, T>& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

// send cppvmStlMultimap classes
template <class Key, class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlMultimap<Key, T>& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlMultimap classes
template <class Key, class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlMultimap<Key, T>& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



