/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_stlset.h 
@SE  CREATED     : 31.1.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : cppvm stl set class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STLSET_INCLUDE
#define _PVM_STLSET_INCLUDE

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_obj.h"
#include "pvm_stl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

template <class T>
class cppvmStlSet: public cppvmObject, public set<T> {
  
public:
  // common constructors
  cppvmStlSet(): cppvmObject(), set<T>() {}
  cppvmStlSet(typename set<T>::const_iterator first, 
              typename set<T>::const_iterator last): 
      cppvmObject(), set<T>(first, last) {}
  // constructors for cppvmStlSet
  cppvmStlSet(const cppvmStlSet<T>& obj) : 
      cppvmObject(), set<T>((const set<T>&)obj){}
  // constructors for set
  cppvmStlSet(const set<T>& obj) : 
      cppvmObject(), set<T>(obj) {}
  // destructor
  virtual ~cppvmStlSet() {}
  
  // cppvmStlSet assignment
  set<T>& operator=(const cppvmStlSet<T>& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return set<T>::operator=((const set<T>&) obj);
  }
  // set assignment
  set<T>& operator=(const set<T>& obj){
    return set<T>::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlSet<T>& obj) const {
    return !(operator==((const set<T>&) *this, (const set<T>&) obj));
  }
  bool operator==(const cppvmStlSet& obj) {
    return operator==((const set<T>&) *this, (const set<T>&) obj);
  }
  bool operator==(const cppvmStlSet& obj) const {
    return operator==((const set<T>&) *this, (const set<T>&) obj);
  }

  virtual void cppvmTransfer() {
    cppvmTransferStlSet ( (set<T>&) *this, 1);
  }
  
  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlSet)
        
};

template <class T>
class cppvmStlMultiset: public cppvmObject, public multiset<T> {
  
public:
  // common constructors
  cppvmStlMultiset(): cppvmObject(), multiset<T>() {}
  cppvmStlMultiset(typename multiset<T>::const_iterator first, 
                   typename multiset<T>::const_iterator last): 
      cppvmObject(), multiset<T>(first, last) {}
  // constructors for cppvmStlMultiset
  cppvmStlMultiset(const cppvmStlMultiset<T>& obj) : 
      cppvmObject(), multiset<T>((const multiset<T>&)obj){}
  // constructors for multiset
  cppvmStlMultiset(const multiset<T>& obj) : 
      cppvmObject(), multiset<T>(obj) {}
  // destructor
  virtual ~cppvmStlMultiset() {}
  
  // cppvmStlMultiset assignment
  multiset<T>& operator=(const cppvmStlMultiset<T>& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return multiset<T>::operator=((const multiset<T>&) obj);
  }
  // multiset assignment
  multiset<T>& operator=(const multiset<T>& obj){
    return multiset<T>::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlMultiset<T>& obj) const {
    return !(operator==((const multiset<T>&) *this, (const multiset<T>&) obj));
  }
  bool operator==(const cppvmStlMultiset& obj) {
    return operator==((const multiset<T>&) *this, (const multiset<T>&) obj);
  }
  bool operator==(const cppvmStlMultiset& obj) const {
    return operator==((const multiset<T>&) *this, (const multiset<T>&) obj);
  }

  virtual void cppvmTransfer() {
    cppvmTransferStlMultiset ( (multiset<T>&) *this, 1);
  }
  
  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlMultiset)
        
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// send cppvmStlSet classes
template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlSet<T>& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlSet classes
template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlSet<T>& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

// send cppvmStlMultiset classes
template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlMultiset<T>& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlMultiset classes
template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlMultiset<T>& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



