/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_stlstring.h 
@SE  CREATED     : 5.2.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : cppvm stl string class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STLSTRING_INCLUDE
#define _PVM_STLSTRING_INCLUDE

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_obj.h"
#include "pvm_stl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

class cppvmStlString: public cppvmObject, public string {
  
public:
  // common constructors
  cppvmStlString(): cppvmObject(), string() {}
  cppvmStlString(const char *zf): cppvmObject(), string(zf) {}
#ifdef CPPVM_NODEADCODE   
  cppvmStlString(typename string::size_type anz, char c): 
      cppvmObject(), string(anz, c) {}
#endif // CPPVM_NODEADCODE   
  // constructors for cppvmStlString class
  cppvmStlString(const cppvmStlString& str): 
      cppvmObject((const cppvmObject&)str), string((const string&)str) {}
// Visual C++ can not handle 'size_type' 
#ifdef CPPVM_NODEADCODE   
  cppvmStlString(const cppvmStlString& str, typename string::size_type idx): 
      cppvmObject((const cppvmObject&)str), 
      string((const string&)str, idx) {}
  cppvmStlString(const cppvmStlString& str, typename string::size_type idx, 
                 typename string::size_type anz): 
      cppvmObject((const cppvmObject&)str), string((const string&)str, idx, anz) {}
#endif // CPPVM_NODEADCODE   
  // constructors for string class
  cppvmStlString(const string& str): cppvmObject(), string(str) {}
#ifdef CPPVM_NODEADCODE   
  cppvmStlString(const string& str, typename string::size_type idx): 
      cppvmObject(), string(str, idx) {}
  cppvmStlString(const string& str, typename string::size_type idx, 
                 typename string::size_type anz): cppvmObject(), 
    string(str, idx, anz) {}
#endif // CPPVM_NODEADCODE   

  // destructor
  virtual ~cppvmStlString() {}

  // common assignment
  string& operator=(const char* zf){
    return string::operator=(zf);
  }
  string& operator=(char c){
    return string::operator=(c);
  }
  // cppvmStlString assignment
  string& operator=(const cppvmStlString& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return string::operator=((const string&) obj);
  }
  // string assignment
  string& operator=(const string& obj){
    return string::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlString& obj) const {
    if (compare((const string&)obj)==0){
      return false;
    }
    else{
      return true;
    }
  }
  bool operator==(const cppvmStlString& obj) {
    if (compare((const string&)obj)==0){
      return true;
    }
    else{
      return false;
    }
  }
  bool operator==(const cppvmStlString& obj) const {
    if (compare((const string&)obj)==0){
      return true;
    }
    else{
      return false;
    }
  }
  
  virtual void cppvmTransfer() {
    cppvmTransferStlString ( (string&) *this, 1);
  }

  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlString)
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/


#ifndef CPPvmDoNotUseStdNamespace
// inline functions 
inline std::ostream&
operator<<(std::ostream& strm, cppvmStlString& obj){
  return std::operator<<(strm, (string&)obj);
}

inline std::istream&
operator>>(std::istream& strm, cppvmStlString& obj){
  return std::operator>>(strm, (string&)obj);
}
#else
// inline functions 
inline ostream&
operator<<(ostream& strm, cppvmStlString& obj){
  return operator<<(strm, (string&)obj);
}

inline istream&
operator>>(istream& strm, cppvmStlString& obj){
  return operator>>(strm, (string&)obj);
}
#endif

// send cppvmStlString classes
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlString& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlString classes
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlString& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



