/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_stlvalarray.h 
@SE  CREATED     : 31.1.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : cppvm stl valarray class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STLVALARRAY_INCLUDE
#define _PVM_STLVALARRAY_INCLUDE

#ifdef CPPvmUseStlValarray

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_obj.h"
#include "pvm_stl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

template <class T>
class cppvmStlValarray: public cppvmObject, public valarray<T> {
  
public:
  // common constructors
  cppvmStlValarray(): cppvmObject(), valarray<T>() {}
// gcc can not handle 'size_t' 
#ifdef CPPVM_NODEADCODE   
  explicit cppvmStlValarray(typename valarray<T>::size_t n): 
      cppvmObject(), valarray<T>(n) {}
  cppvmStlValarray(const T& value, typename valarray<T>::size_t n): 
      cppvmObject(), valarray<T>(value, n) {}
  cppvmStlValarray(const T* array, typename valarray<T>::size_t n): 
      cppvmObject(), valarray<T>(array, n) {}
#endif // CPPVM_NODEADCODE   
  // constructors for cppvmStlValarray
  cppvmStlValarray(const cppvmStlValarray<T>& obj) : 
      cppvmObject(), valarray<T>((const valarray<T>&)obj){}
  // constructors for valarray
  cppvmStlValarray(const valarray<T>& obj) : 
      cppvmObject(), valarray<T>(obj) {}
  // destructor
  virtual ~cppvmStlValarray() {}
  
  // cppvmStlValarray assignment
  valarray<T>& operator=(const cppvmStlValarray<T>& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return valarray<T>::operator=((const valarray<T>&) obj);
  }
  // valarray assignment
  valarray<T>& operator=(const valarray<T>& obj){
    return valarray<T>::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlValarray<T>& obj) const {
    return !(operator==((const valarray<T>&) *this, (const valarray<T>&) obj));
  }
  bool operator==(const cppvmStlValarray& obj) {
    return operator==((const valarray<T>&) *this, (const valarray<T>&) obj);
  }
  bool operator==(const cppvmStlValarray& obj) const {
    return operator==((const valarray<T>&) *this, (const valarray<T>&) obj);
  }

  virtual void cppvmTransfer() {
    cppvmTransferStlValarray ( (valarray<T>&) (*this), 1);
  }
  
  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlValarray)
        
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// send cppvmStlValarray classes
template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlValarray<T>& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlValarray classes
template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlValarray<T>& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

#endif // CPPvmUseStlValarray

#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



