/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_stlvector.h 
@SE  CREATED     : 31.1.2000
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : cppvm stl vector class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STLVECTOR_INCLUDE
#define _PVM_STLVECTOR_INCLUDE

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_obj.h"
#include "pvm_stl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

template <class T>
class cppvmStlVector: public cppvmObject, public vector<T> {
  
public:
  // common constructors
  cppvmStlVector(): cppvmObject(), vector<T>() {}
  cppvmStlVector(typename vector<T>::size_type n): cppvmObject(), vector<T>(n) {}
  cppvmStlVector(typename vector<T>::size_type n, const T& value): 
      cppvmObject(), vector<T>(n, value) {}
  cppvmStlVector(typename vector<T>::const_iterator first, 
                 typename vector<T>::const_iterator last): 
      cppvmObject(), vector<T>(first, last) {}
  // constructors for cppvmStlVector
  cppvmStlVector(const cppvmStlVector<T>& obj) : 
      cppvmObject(), vector<T>((const vector<T>&)obj){}
  // constructors for vector
  cppvmStlVector(const vector<T>& obj) : 
      cppvmObject(), vector<T>(obj) {}
  // destructor
  virtual ~cppvmStlVector() {}
  
  // cppvmStlVector assignment
  vector<T>& operator=(const cppvmStlVector<T>& obj){
    cppvmObject::operator=((const cppvmObject&) obj);
    return vector<T>::operator=((const vector<T>&) obj);
  }
  // vector assignment
  vector<T>& operator=(const vector<T>& obj){
    return vector<T>::operator=(obj);
  }

  // compare operators
  bool operator!=(const cppvmStlVector<T>& obj) const {
    return !(operator==((const vector<T>&) *this, (const vector<T>&) obj));
  }
  bool operator==(const cppvmStlVector& obj) {
    return operator==((const vector<T>&) *this, (const vector<T>&) obj);
  }
  bool operator==(const cppvmStlVector& obj) const {
    return operator==((const vector<T>&) *this, (const vector<T>&) obj);
  }

  virtual void cppvmTransfer() {
    cppvmTransferStlVector ( (vector<T>&) (*this), 1);
  }
  
  CPPvmBuildInMsgTagMethodsStlDeclaration(CPPvmMsgTagClassStlVector)
        
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// send cppvmStlVector classes
template <class T>
inline cppvmSendStream&
operator<<(cppvmSendStream& sendStrm, cppvmStlVector<T>& obj){
  obj.cppvmSend(sendStrm);
  return sendStrm;
}

// receive cppvmStlVector classes
template <class T>
inline cppvmReceiveStream&
operator>>(cppvmReceiveStream& recStrm, cppvmStlVector<T>& obj){
  // set the receive result in 'recStrm'
  recStrm.setLastRecResult(obj.cppvmReceive(recStrm));
  return recStrm;
}

#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

#endif 
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



