/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_str.h 
@SE  CREATED     : 5.12.1997
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm string class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STR_INCLUDE
#define _PVM_STR_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_sida.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_STR_MODULE "pvm_str.h"

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmStringCleanFieldSize
@CD  DESCRIPTION   : size for the memcopy field when cleaning strings
***
******************************************************************************/
#define CPPvmStringCleanFieldSize 128

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmString
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : string class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmString : public cppvmSingleData{
public:

  cppvmString();
  cppvmString(unsigned int len_, bool cleanString=true);
  cppvmString(char* s);
  cppvmString(const char* s);
  cppvmString(const cppvmString&);
  virtual ~cppvmString();

  void write(char *value_);
  char *read();

  virtual double readValDouble();
  virtual void   writeValDouble(double value_);
  
  cppvmString& operator=(const cppvmString&);
  bool operator==(const cppvmString&);
  bool operator==(const cppvmString&) const;
  bool operator!=(const cppvmString& obj) const {
    return ! (operator==(obj));
  }
  
  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

  virtual void cppvmTransfer();
  void setCPPvmTransferMode(int mode);
  int getCPPvmTransferMode(){return cppvmTransferMode;}
  

  CPPvmBuildInMsgTagMethodsDeclaration(CPPvmMsgTagClassString)

  void clean();
  void append(cppvmString& obj);
  void append(char c);
  void append(char *str);

  void printFormat(const char *format, ...);
  // 
  // void scanFormat(const char *format, ...);
  
  char* getPtr() {return cstring;} 
  char* getPtr() const {return cstring;} 

  int countChars(){return (int)len;}
  int countWords();
  char& operator[] (unsigned int index){
    if(index>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_STR_MODULE,
                                                   "cppvmString",
                                                   "operator[unsigned int]",
                                                   "wrong index"));
    }
    return cstring[index];
  }
  bool copyWord(unsigned int index, cppvmString& target);

  void cycleShiftLeft();
  void cycleShiftRight();

  void resize(unsigned int newSize);
  void resizeClean(unsigned int newLen, bool cleanOnlyNewElements=false);

  void cleanElem(unsigned int index)
  {
#ifndef CPPvmFastAccess
    if(index>=len){
      CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_STR_MODULE,
                                                   "cppvmString",
                                                   "cleanElem",
                                                   "index >= string length"));
    }
#endif
    cstring[index]=(char)0;
  }
  void cleanRange(unsigned int fromIndex, unsigned int toIndex);

  void setFitSizeToLength(bool value);
  bool getFitSizeToLength(){return fitSizeToLength;}

  unsigned int getSize(){return size;}
  unsigned int getLength(){return len;}
  
  
protected:
  char*    cstring;  // dynamic c string
  unsigned int len;  // number of chars (without '\0')
  unsigned int size; // total size of cstring
  int  cppvmTransferMode; // transfer mode
  bool fitSizeToLength;

private:
  void resizeWithoutFill(unsigned int newSize);
  void resizeWithoutFillAndCopy(unsigned int newSize);
  bool arrayIsNULL(){return (cstring==(char*)NULL);}
  void newArray(unsigned int newSize);
  void newArrayCopy(unsigned int newSize, unsigned int numCopyElements);
  
  int setOnWord(unsigned int i);
  int setOnNextWordBegin(unsigned int index);

  int strLength; // used for send/recieve only

  char emptyString[CPPvmStringCleanFieldSize];
  bool initEmptyString;
};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



