/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_strm.h 
@SE  CREATED     : 26.6.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm stream base class
@SU  USAGE       : 
@SC  CONSTRAINTS : 
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_STRM_INCLUDE
#define _PVM_STRM_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_def.h"
#include "pvm_exra.h"
#include "pvm_exre.h"
#include "pvm_exfo.h"
#include "pvm_expa.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define PVM_STRM_HEADER "pvm_strm.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/

// class declarations to avoid cyclic includes
class cppvmObject;
class cppvmConnection;
class cppvmTask;
class cppvmSpawnConnection;
class cppvmSpawnMultiConnection;
class cppvmReceiveStream;

/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmStream
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm system base class
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmStream 
{
public:
  cppvmStream();
  virtual ~cppvmStream();
  cppvmStream(const cppvmStream&);
  cppvmStream& operator=(const cppvmStream& obj);
  bool operator!=(const cppvmStream& obj) const {
    return ! (operator==(obj));
  }
  bool operator==(const cppvmStream& obj) {return cppvmIsSame(obj);}
  bool operator==(const cppvmStream& obj) const {return cppvmIsSame(obj);}

  // test of same identity
  bool cppvmIsSame(const cppvmStream& obj) const{
    return cppvmObjId == obj.cppvmObjId;
  }

  virtual void cppvmPrintOn(ostream& = cout) {};
  virtual void cppvmScanFrom(istream& = cin) {};

  int getCPPvmObjId() { return cppvmObjId; }
  
private:
  static int cppvmMaxId; // highest actual ID
  int cppvmObjId;        // specific object ID
  static int cppvmTotalObjNum; // total number of object for this class
};

/*@@@**************************************************************************
**********                                                           **********
*******                       EXPORTED FUNCTIONS                        *******
**********                                                           **********
******************************************************************************/

// inline functions 
inline ostream&
operator<<(ostream& strm, cppvmStream& obj){
  obj.cppvmPrintOn(strm);
  return strm;
}

inline istream&
operator>>(istream& strm, cppvmStream& obj){
  obj.cppvmScanFrom(strm);
  return strm;
}

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
