/*@@@**************************************************************************
**********                                                           **********
*******                      SPECIFICATION FILE                         *******
**********                                                           **********
*******************************************************************************
@SF  FILE        : pvm_syco.h 
@SE  CREATED     : 25.6.1998
@SA  AUTHOR      : goerzig
@SJ  PROJECT     : cppvm
@S#  
@SD  DESCRIPTION : pvm system ordered collection class
@SU  USAGE       : To use this cppvmProcessHandling collection with classes derived from
@SU                cppvmProcessHandling use the castdown command
@SU                  (PVMExample&)pvmObjCollection[i]
@SU                to access specific subclass methods
@SC  CONSTRAINTS : only for cppvmProcessHandling or derived classes
@S#  
@SR  NOTES       : 
@S#  --------------------------------------------------------------------------
@S#  Technical Information:
@S#  
@SI  LANGUAGE        : C++
@SI  HARDWARE        : PC 386
@SI  COMPILER        : gcc version 2.7.2.3
@SI  OPERATINGSYSTEM : Linux
@S#  --------------------------------------------------------------------------
@S#  Version Manager Informations:
@S#  
@SV  FILENAME    : $Source: $
@SV  AUTHOR      : $Author: $
@SV  REVISION    : $Revision: $
@SV  DATE        : $Date: $
@SV  STATE       : $State: $
@SV  LOCKER      : $Locker: $
@S#  
@SV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef _PVM_SYCO_INCLUDE
#define _PVM_SYCO_INCLUDE


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

#include "pvm_cincl.h"
#include "pvm_proh.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define PVM_SYCO_MODULE "pvm_syco.h"

/*@@@**************************************************************************
**********                                                           **********
*******                              TYPES                              *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   TYPE   ************************************
***
@TN  TYPE NAME   : pvmSysCollCompFctPtr
@TD  DESCRIPTION : pointer to a compare function for *ALL* 'cppvmProcessHandling's
@TD                in the collection (the compare function MUST be usable for all 
@TD                objects in the collection) 
***
******************************************************************************/
typedef bool (*pvmSysCollCompFctPtr)(cppvmProcessHandling* obj1, cppvmProcessHandling* obj2);

/*@@@**************************************************************************
**********                                                           **********
*******                             CLASSES                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   CLASS   ***********************************
***
@KN  CLASS NAME    : cppvmProcessHandlingCollection
@KB  BASE CLASS    :
@KA  AUTHOR        : goerzig
@K#  
@KD  DESCRIPTION   : pvm object collection
@KC  CONSTRAINTS   : 
******************************************************************************/
class cppvmProcessHandlingCollection : public cppvmProcessHandling{
public:

  cppvmProcessHandlingCollection();
  virtual ~cppvmProcessHandlingCollection();

  bool operator==(const cppvmProcessHandlingCollection&);
  bool operator==(const cppvmProcessHandlingCollection&) const;
  bool operator!=(const cppvmProcessHandlingCollection& obj) const {
    return ! (operator==(obj));
  }

  virtual void cppvmScanFrom(istream& = cin);
  virtual void cppvmPrintOn(ostream& = cout);

public:
  void add(cppvmProcessHandling* objPtr);
  void remove(cppvmProcessHandling& obj);
  void remove(cppvmProcessHandling* objPtr);
  void remove(unsigned int i);
  void clean();

  void append(cppvmProcessHandlingCollection* objColl);
  void resize(unsigned int);
  unsigned int numElems() const { return num; }

  cppvmProcessHandling& operator[] (unsigned int i);
  cppvmProcessHandling* getObjPtr(unsigned int i);

  void swap(unsigned int index1, unsigned int index2);
  void sort(pvmSysCollCompFctPtr cmpPtr);

private:
  void shiftLeft(unsigned int indexMin,
                 unsigned int indexMax);

  // do not allow the operator=
  // a deep copy must be available for this operator
  cppvmProcessHandlingCollection& operator=(const cppvmProcessHandlingCollection& obj);

private:
  cppvmProcessHandling** elems;  // dynamic pointer array
  unsigned int size;  // array size
  unsigned int num;   // number of entries (elems[0..num-1])

};

#endif
/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



