// cppvm semaphore server

// this is the counterpart to the method 
// serverTransfer()
// in the class cppvmSemaphore (pvm_sema.h/.cpp)

#include "cppvm.h"
#include "pvm_cose.h"
#include "pvm_sead.h"
#include "pvm_sepr.h"

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : recNum
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : receive semaphore set counter number value
@FU  USAGE         : int recNum(cppvmReceiveStream& recStrm)
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : recStrm  receive number from this stream
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : int      number value 
******************************************************************************/
int 
recNum(cppvmReceiveStream& recStrm)
{
  static cppvmInt setNum;
  
  recStrm >> setNum;
  return (int)setNum;
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : sendNum
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : send semaphore number to a specified task
@FU  USAGE         : void sendNum(int num, int tid)
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : num  semaphore counter number
@FP            tid  send to task with this id
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void 
sendNum(int num, int tid)
{
  static cppvmSendStream sendStrm;
  static cppvmInt resultNum;

  resultNum = (cppvmInt) num;
  sendStrm.set(tid, PvmBaseContext, CPPvmSNBtid);
  sendStrm << resultNum;
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : sendResult
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : send result code to a specified task
@FU  USAGE         : void sendResult(bool result, int tid)
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : bool result
@FP            tid  send to task with this id
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void 
sendResult(bool result, int tid)
{
  static cppvmSendStream sendStrm;
  static cppvmInt resultCode;
  
  if(result==true){
    resultCode = (int)CPPvmFromSemaphoreServerOK;
  }
  else{
    resultCode = (int)CPPvmFromSemaphoreServerFailed;
  }
  sendStrm.set(tid, PvmBaseContext, CPPvmSNBtid);
  sendStrm << resultCode;
}


/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : enter
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : call enter operation for a semaphore
@FU  USAGE         : void enter(cppvmSemaphoreAdministrator& semaAdmin,
@FU                             cppvmString& semaName, int recTid);
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : semaAdmin semaphore administrator
@FP            semaName  semaphore name
@FP            recTid    receive from this task id
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void 
enter(cppvmSemaphoreAdministrator& semaAdmin,
      cppvmString& semaName, int recTid)
{
  bool result;

  result = semaAdmin.enter(semaName, recTid);
  
  if(result==true){
    // process can continue
    sendResult(true, recTid);
  }
  // otherwise let the process sleep
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : enterNB
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : call non-blocking enter operation for a semaphore
@FU  USAGE         : void enterNB(cppvmSemaphoreAdministrator& semaAdmin,
@FU                               cppvmString& semaName, int recTid);
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : semaAdmin semaphore administrator
@FP            semaName  semaphore name
@FP            recTid    receive from this task id
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void 
enterNB(cppvmSemaphoreAdministrator& semaAdmin,
        cppvmString& semaName, int recTid)
{
  bool result;
  
  result = semaAdmin.enterNB(semaName);
  sendResult(result, recTid);
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : leave
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : call leave operation for a semaphore
@FU  USAGE         : void leave(cppvmSemaphoreAdministrator& semaAdmin,
@FU                             cppvmString& semaName, int recTid);
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : semaAdmin semaphore administrator
@FP            semaName  semaphore name
@FP            recTid    receive from this task id
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void 
leave(cppvmSemaphoreAdministrator& semaAdmin,
      cppvmString& semaName, int recTid)
{
  int result;
  
  result = semaAdmin.leave(semaName);
  if(result>=0){
    // a sleeping process must be awaiked
    sendResult(true, result);
    // let the process continue
    sendResult(true, recTid);
  }
  else{
    // no sleeping process, let the process continue
    sendResult(true, recTid);
  }
}
  
/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : setNumber
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : set the counter number of a semaphore
@FU  USAGE         : void setNumber(cppvmReceiveStream recStrm,
@FP                                 cppvmSemaphoreAdministrator& semaAdmin,
@FU                                 cppvmString& semaName, int recTid);
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : recStrm   receive stream to get number from
@FP            semaAdmin semaphore administrator
@FP            semaName  semaphore name
@FP            recTid    receive from this task id
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void 
setNumber(cppvmReceiveStream& recStrm,
          cppvmSemaphoreAdministrator& semaAdmin,
          cppvmString& semaName, int recTid)
{
  static int newNum;
  static bool result;
  
  newNum = recNum(recStrm);
  // set new number
  result = semaAdmin.setNumber(semaName, newNum);

  sendResult(result, recTid);
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : getNumber
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : get the counter number of a semaphore and send it to
@FD                  the calling process
@FU  USAGE         : void getNumber(cppvmSemaphoreAdministrator& semaAdmin,
@FU                                 cppvmString& semaName, int recTid);
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : semaAdmin semaphore administrator
@FP            semaName  semaphore name
@FP            recTid    receive from this task id
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void 
getNumber(cppvmSemaphoreAdministrator& semaAdmin,
          cppvmString& semaName, int recTid)
{
  static int num;
  
  // get number
  num = semaAdmin.getNumber(semaName);

  sendNum(num, recTid);
  sendResult(true, recTid);
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : resetNumber
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : reset the counter number of a semaphore
@FU  USAGE         : void resetNumber(cppvmReceiveStream recStrm,
@FP                                   cppvmSemaphoreAdministrator& semaAdmin,
@FU                                   cppvmString& semaName, int recTid);
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : recStrm   receive stream to get number from
@FP            semaAdmin semaphore administrator
@FP            semaName  semaphore name
@FP            recTid    receive from this task id
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void 
resetNumber(cppvmReceiveStream& recStrm,
            cppvmSemaphoreAdministrator& semaAdmin,
            cppvmString& semaName, int recTid)
{
  static int newNum;
  
  newNum = recNum(recStrm);
  // reset
  semaAdmin.resetNumber(semaName, newNum);

  sendResult(true, recTid);
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : main
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : semaphore server; counterpart to transferSend() of
@FD                  the class cppvmSemaphore (pvm_sema.cpp)
@FU  USAGE         : int main()
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : 
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
int
main()
{
  bool endless=true;
  
  // connect to pvm 
  cppvmConnection pvmConn;

  // variables
  cppvmString semaName;
  cppvmReceiveStream firstRecStrm;
  cppvmReceiveStream nextRecStrm;
  cppvmSemaphoreAdministrator semaAdmin;
  cppvmInt operation;
  int lastRecFromTid;

  // set receive stream
  firstRecStrm.set(PvmBaseContext, CPPvmRBany);
  
  while(endless==true){
    // wait for the next server access
    // read operation and semaphore name
    firstRecStrm >> operation;
    lastRecFromTid = firstRecStrm.getLastRecFromTid();
    // ensure that the next values are send 
    // from the same proces
    nextRecStrm.set(lastRecFromTid, PvmBaseContext, CPPvmRBtid);
    // read semaphore name
    nextRecStrm >> semaName;
    
    if((int)operation==CPPvmToSemaphoreServerEnter){
      enter(semaAdmin, semaName, lastRecFromTid);
    }
    else if((int)operation==CPPvmToSemaphoreServerEnterNB){
      enterNB(semaAdmin, semaName, lastRecFromTid);
    }
    else if((int)operation==CPPvmToSemaphoreServerLeave){
      leave(semaAdmin, semaName, lastRecFromTid);
    }
    else if((int)operation==CPPvmToSemaphoreServerSetNumber){
      setNumber(nextRecStrm, semaAdmin, semaName, lastRecFromTid);
    }
    else if((int)operation==CPPvmToSemaphoreServerResetNumber){
      resetNumber(nextRecStrm, semaAdmin, semaName, lastRecFromTid);
    }
    else if((int)operation==CPPvmToSemaphoreServerGetNumber){
      getNumber(semaAdmin, semaName, lastRecFromTid);
    }
    else{
      cerr << "cppvms.cpp: server operation from client unknown. "
           << "Possibly a protocol error." << endl;
    }
  }
  return 0;
  
}



