/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_arbo.cpp 
@IE  CREATED     : 29.9.1999
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm bool array class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC PII 366
@II  COMPILER        : gcc version 2.91.66
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_arbo.h"


/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_arbo.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                           VARIABLES                             *******
**********                                                           **********
******************************************************************************/
CPPvmBuildInMsgTagMethodsDefinition(cppvmArrayBool,
                                    CPPvmMsgTagClassArrayBool)


/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmArrayBool
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : cppvmArrayBool
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmArrayBool::cppvmArrayBool()
    : cppvmArrayData(0), array(NULL), packedArray(NULL), initEmptyArray(false)
{
  // base class constructor sets 'len=0'
  // resize works only because 'array == NULL'

  resize(0);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmArrayBool
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmArrayBool
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : len_       length of the array
@MP            cleanArray clean array after allocation
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmArrayBool::cppvmArrayBool(unsigned int len_, bool cleanArray)
    : cppvmArrayData(len_), array(NULL), packedArray(NULL), initEmptyArray(false)
{
  // base class constructor sets 'len=len_'
  // resize works only because 'array == NULL'

  if(cleanArray==true){
    resizeClean(len_);
  }
  else{
    resize(len_);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmArrayBool
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmArrayBool::cppvmArrayBool (const cppvmArrayBool& obj)
    : cppvmArrayData(obj), array(NULL), packedArray(NULL), initEmptyArray(false)
{
  // base class copy constructor sets 'len=obj.len'
  // resize works only because 'array == NULL'

  resizeWithoutFillAndCopy((unsigned int)obj.len);
  
  if(len>0){
    // copy elements
    memcpy((void*)array, (void*)obj.array, sizeof(bool) * len);
  }

}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmArrayBool
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmArrayBool
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmArrayBool::~cppvmArrayBool()
{
  if(array)
    delete [] array;
  if(packedArray)
    delete [] packedArray;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to assign
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmArrayBool&
cppvmArrayBool::operator= (const cppvmArrayBool& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }

  // adapt array length
  // 'cppvmArrayData::operator=' sets
  // len to obj.len, so resize has
  // to be called before
  resizeWithoutFillAndCopy((unsigned int)obj.len);

  cppvmArrayData::operator=(obj);
  
  
  if(obj.len>0){
    // copy elements
    memcpy((void*)array, (void*)obj.array, sizeof(bool) * obj.len);
  }
    
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmArrayBool::operator== (const cppvmArrayBool& obj)
{
  int i;

  // same length ?
  if((cppvmArrayData::operator==(obj))==false){
    return false;
  }
  // same values ?
  for(i=0;i<len;i++){
    if(array[i]!=obj.array[i]){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmArrayBool::operator== (const cppvmArrayBool& obj) const
{
  int i;

  // same length ?
  if((cppvmArrayData::operator==(obj))==false){
    return false;
  }
  // same values ?
  for(i=0;i<len;i++){
    if(array[i]!=obj.array[i]){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmTransfer
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : transfer object from/to buffer
@MU  USAGE         : cppvmTransfer()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::cppvmTransfer()
{
  // transfer base class variables
  cppvmArrayData::cppvmTransfer();

  if(cppvmIsTransferReceive()==true){
    
    // receive array
    // remind old length
    int oldLen=len;
    // new array length
    int newLen;
    // transfer array length
    cppvmTransferInt(&newLen);

    if(cppvmTransferMode==CPPvmTransferModeAppend){
      // received array length plus old length
      resizeWithoutFill(oldLen+newLen);
      // receive the incomming array
      cppvmTransferByte((char*)packedArray, getPackedArraySize(newLen));
      unpackData(&array[oldLen], newLen);
    }
    else{
      resizeWithoutFillAndCopy(newLen);
      // receive the incomming array
      cppvmTransferByte((char*)packedArray, getPackedArraySize(len));
      unpackData(array, len);
    }
  }
  else{
    // send branch
    
    // send array
    // transfer array length
    cppvmTransferInt(&len);

    // pack and send array
    packData();
    cppvmTransferByte((char*)packedArray, getPackedArraySize(len));
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::cppvmPrintOn(ostream& strm)
{
  int i;
  
  cppvmArrayData::cppvmPrintOn(strm);
  
  strm << "( ";
  for(i=0;i<len;i++){
    strm << array[i];
    strm << " ";
  }
  strm << ")";
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmScanFrom
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : scan method for << operator
@MU  USAGE         : cppvmScanFrom()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to scan object from
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::cppvmScanFrom(istream& strm)
{
  char dummy;
  int i;
  int scanValue;

  cppvmArrayData::cppvmScanFrom(strm);

  // resize array to new length
  resizeWithoutFillAndCopy((unsigned int)len);
  
  strm >> dummy;
  for(i=0;i<len;i++){
    strm >> scanValue;
    if(scanValue==0){
      array[i]=false;
    }
    else{
      array[i]=true;
    }
  }
  strm >> dummy;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cleanRange
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : clean a range in the array
@MU  USAGE         : cleanRange(unsigned int fromIndex, unsigned int toIndex)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the class variables 'len' and 'size' must have
@MR                  correct values, 'array' must not be NULL
@M#  --------------------------------------------------------------------------
@MP  IN      : fromIndex    index of first element to clean
@MP            toIndex      index of last element to clean
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::cleanRange(unsigned int fromIndex,
                           unsigned int toIndex)
{
  checkCleanRange(fromIndex, toIndex);

  unsigned int i;
  unsigned int numMemCopy;
  unsigned int restSingleCopyIndex;
  
  // clean memory to copy
  if(initEmptyArray==false){
    for(i=0; i< CPPvmArrayCleanFieldSize; i++){
      emptyArray[i]=false;
    }
    initEmptyArray=true;
  }
  
  numMemCopy = ((toIndex-fromIndex)+1) / CPPvmArrayCleanFieldSize;
  restSingleCopyIndex = fromIndex + (numMemCopy * CPPvmArrayCleanFieldSize);
  
  unsigned int index = fromIndex;
  // use memcpy
  for(i=0; i< numMemCopy; i++){
    memcpy((void*)(&array[index]), (void*)(&emptyArray[0]), 
           sizeof(bool) *  CPPvmArrayCleanFieldSize);
    index = index +  CPPvmArrayCleanFieldSize;
  }
  
  // copy the rest as single values
  for(i=restSingleCopyIndex; i<=toIndex; i++){
    array[i]=false;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : newArray
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : allocate new array
@MU  USAGE         : void newArray(unsignd int newSize)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : newSize  new array size
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::newArray(unsigned int newSize)
{
  if(array!=NULL){
    // the former array is not needed any more
    delete [] array;
  }
  if(packedArray!=NULL){
    // delete packed array
    delete [] packedArray;
  }
  // allocate memory
  array = new bool [newSize];  
  packedArray = new unsigned char [getPackedArraySize(newSize)];
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : newArrayCopy
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : allocate new array and copy elements from old array
@MU  USAGE         : void newArrayCopy(unsigned int newSize, 
@MU                                    unsigned int numCopyElements)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : newSize         new array size
@MP            numCopyElements number of element to copy from the old into the
@MP                            new array
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::newArrayCopy(unsigned int newSize, unsigned int numCopyElements)
{
  if(array==NULL){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmArrayBool",
                                                 "newArrayCopy",
                                                 "array to copy from is NULL!"));
  }
  if(packedArray!=NULL){
    // delete packed array
    delete [] packedArray;
  }
  // new packed array
  packedArray = new unsigned char [getPackedArraySize(newSize)];
  
  // new array is needed
  bool *tmpArray = array;

  // allocate memory  
  array = new bool [newSize];  

  if(numCopyElements>0){
    // copy memory
    memcpy((void*)array, (void*)tmpArray, sizeof(bool) * numCopyElements);
  }

  // the former array is not needed any more
  delete [] tmpArray;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : addElem
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add a value from another array to this value
@MU  USAGE         : void addElem(int pos, cppvmArrayData& obj, int objPos);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pos     index of the element of this array (the value is
@MP                    added to the element with this index)
@MP            obj     array object (from which an value should be added)
@MP            objPos  index for the value to add in the given array object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::addElem(int pos, cppvmArrayData& obj, int objPos)
{
  if(obj.cppvmMsgTag()==CPPvmMsgTagClassArrayBool){
    // array object of this class
    array[pos] = array[pos] || ((cppvmArrayBool&)(obj))[objPos];
  }
  else{
    // add value via double conversion
    bool addVal;
    if(obj.getElemDouble(objPos)<=0.0){
      addVal=false;
    }
    else{
      addVal=true;
    }
    array[pos] = array[pos] || addVal; 
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : subElem
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : subtract a value from another array to this value
@MU  USAGE         : void subElem(int pos, cppvmArrayData& obj, int objPos);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pos     index of the element of this array (the value is
@MP                    subtracted to the element with this index)
@MP            obj     array object (from which an value should be subtracted)
@MP            objPos  index for the value to subtract in the given
@MP                    array object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::subElem(int pos, cppvmArrayData& obj, int objPos)
{
  if(obj.cppvmMsgTag()==CPPvmMsgTagClassArrayBool){
    // array object of this class
    if(array[pos]==true && ((cppvmArrayBool&)(obj))[objPos]==true){
      array[pos] = false;
    }
  }
  else{
    // subtract value via double conversion
    if(array[pos]== true && obj.getElemDouble(objPos)>0.0){
      array[pos] = false;
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : multElem
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : multiply a value from another array to this value
@MU  USAGE         : void multElem(int pos, cppvmArrayData& obj, int objPos);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pos     index of the element of this array (the value is
@MP                    multiplied to the element with this index)
@MP            obj     array object (from which an value should be multiplied)
@MP            objPos  index for the value to multiply in the given
@MP                    array object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::multElem(int pos, cppvmArrayData& obj, int objPos)
{
  if(obj.cppvmMsgTag()==CPPvmMsgTagClassArrayBool){
    // array object of this class
    array[pos] = array[pos] && ((cppvmArrayBool&)(obj))[objPos];
  }
  else{
    // multiply value via double conversion
    bool mulValue;
    if(obj.getElemDouble(objPos)<=0.0){
      mulValue=false;
    }
    else{
      mulValue=true;
    }
    array[pos] = array[pos] && mulValue;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : divElem
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : divide a value from another array to this value
@MU  USAGE         : void divElem(int pos, cppvmArrayData& obj, int objPos);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pos     index of the element of this array (the value is
@MP                    divided from the element with this index)
@MP            obj     array object (from which an value should be devided)
@MP            objPos  index for the value to divide in the given
@MP                    array object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::divElem(int pos, cppvmArrayData& obj, int objPos)
{
  if(obj.cppvmMsgTag()==CPPvmMsgTagClassArrayBool){
    // array object of this class
    array[pos] = array[pos] && ((cppvmArrayBool&)(obj))[objPos];
  }
  else{
    // devide value via double conversion
    bool divValue;
    if(obj.getElemDouble(objPos)<=0.0){
      divValue=false;
    }
    else{
      divValue=true;
    }
    array[pos] = array[pos] && divValue;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : copyElems
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy values from an external objects into own array
@MU  USAGE         : void copyElems(int num, int localIndex,
@MU                                 cppvmArrayData& extObj, int extIndex);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : Copy 'num' elements: 
@MR                  array[localIndex] = obj.array[extIndex]; etc.
@M#  --------------------------------------------------------------------------
@MP  IN      : num        number of elements to copy
@MP            localIndex start index for local object
@MP            extObj     external object to copy data from
@MP            extIndex   start index for external object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::copyElems(int num, int localIndex, 
                          cppvmArrayData& extObj, int extIndex)
{
  if(extObj.cppvmMsgTag()==CPPvmMsgTagClassArrayBool){
    // array object of this class
    // can be a reference to own object
    // -> use 'memmove' instead of memcpy
    if(num>0){
      memmove((void*) &array[localIndex], 
              (void*) &((cppvmArrayBool&)extObj).array[extIndex], 
              sizeof(bool) * (num));
    }
  }
  else{
    // set value via double conversion
    int i;
    bool value;
    for(i=0; i< num; i++){
      if(extObj.getElemDouble(extIndex+i)<=0.0){
        value=false;
      }
      else{
        value=true;
      }
      array[localIndex+i] = value;
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cycleShiftLeft
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : cycle shift left
@MD                  (1 2 3) -> (2 3 1)
@MU  USAGE         : void cycleShiftLeft();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::cycleShiftLeft()
{
  bool tmp;

  if(len <= 1){
    return;
  }
  
  tmp = array[0];
  memmove((void*) &array[0], (void*) &array[1], sizeof(bool) * (len-1));
  array[len-1]=tmp;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cycleShiftRight
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : cycle shift right
@MD                  (1 2 3)-> (3 1 2)
@MU  USAGE         : void cycleShiftRight();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::cycleShiftRight()
{
  bool tmp;

  if(len <= 1){
    return;
  }
  
  tmp = array[len-1];
  memmove((void*) &array[1], (void*) &array[0], sizeof(bool) * (len-1));
  array[0] = tmp;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : read
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : return current value
@MU  USAGE         : bool read(unsigned int index);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : index element index in array
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
bool
cppvmArrayBool::read(unsigned int index)
{
  // get object from the mailbox (if the mailbox was initialized)
  cppvmReadMboxInfo();

#ifndef CPPvmFastAccess
  if(index>(unsigned int)len){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmArrayBool",
                                                 "read",
                                                 "index > array length"));
  }
#endif

  return array[index];
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : write
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set array value
@MU  USAGE         : void write(unsigned int index, bool value)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : index index of element
@MP            value new value for element
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::write(unsigned int index,
                      bool value)
{
#ifndef CPPvmFastAccess
  if(index>(unsigned int)len){
    CPPvmInternalThrow(cppvmExceptionOutOfRange(THIS_MODULE,
                                                "cppvmArrayBool",
                                                "write",
                                                "index > array length"));
  }
#endif
  array[index] = value;

  // send it to the mailbox (if the mailbox was initialized)
  cppvmWriteMboxInfo();
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : packData
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : pack all data from 'array' to 'packedArray'
@MU  USAGE         : void packData()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayBool::packData()
{
  int i;
  int byteIndex = 0;
  int bitIndex  = 0;

  // clean packedArray
  for(i=0; i<getPackedArraySize(len); i++){
    packedArray[i]=(unsigned char)0;
  }
  
  // pack new data
  // array is already cleared 
  // set only true values
  for(i=0; i<len; i++){
    if(array[i]==true){
      setPackedElem(byteIndex, bitIndex, true);
    }
    bitIndex++;
    if(bitIndex==8){
      bitIndex = 0;
      byteIndex++;
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : unpackData
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unpack a array of data
@MU  USAGE         : void unpackData(bool* to, int number)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the 'to' array length must be long enough for the data
@M#  --------------------------------------------------------------------------
@MP  IN      : number number of data to unpack
@MP  INOUT   : to     unpack data into this array
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void 
cppvmArrayBool::unpackData(bool* to, int number)
{
  int i;
  int byteIndex = 0;
  int bitIndex  = 0;
  
  for(i=0; i<number; i++)
  {
    to[i] = getPackedElem(byteIndex, bitIndex);
    bitIndex++;
    if(bitIndex==8){
      bitIndex = 0;
      byteIndex++;
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setPackedElem
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set a packed element
@MU  USAGE         : void setPackedElem(int &byteIndex, int& bitIndex, 
@MU                                     bool value);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the 'packedArray' length must be long enough for the data
@M#  --------------------------------------------------------------------------
@MP  IN      : byteIndex byte index
@MP            bitIndex  index of the bit value in the byte
@MP            value     set to this value
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void 
cppvmArrayBool::setPackedElem(int& byteIndex, int& bitIndex, bool value)
{
  
  static const unsigned char addBit[] = {
    0x01,
    0x02,
    0x04,
    0x08,
    0x10,
    0x20,
    0x40,
    0x80
  };
  
  
  static const unsigned char delBit[] = {
    0xFE,
    0xFD,
    0xFB,
    0xF7,
    0xEF,
    0xDF,
    0xBF,
    0x7F
  };

  
  if(value==true){
    // add bit
    packedArray[byteIndex] = (unsigned char)(packedArray[byteIndex] | addBit[bitIndex]);
    //      cout << "pvm_arbo: add bit: " << (int)packedArray[byteIndex] 
    //           << " byteIndex: " << byteIndex 
    //           << " bitIndex: " << bitIndex 
    //           << " addBit:" << (int)addBit[bitIndex] 
    //           << " or: " << (int)(packedArray[byteIndex] | addBit[bitIndex])
    //           << endl;
  }
  else{
    // delete bit
    packedArray[byteIndex] = (unsigned char)(packedArray[byteIndex] & delBit[bitIndex]);
    //      cout << "pvm_arbo: del bit: " << (int)packedArray[byteIndex] 
    //           << " byteIndex: " << byteIndex 
    //           << " bitIndex: " << bitIndex 
    //           << " addBit:" << (int)delBit[bitIndex] 
    //           << " or: " << (int)(packedArray[byteIndex] & delBit[bitIndex])
    //           << endl;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getPackedElem
@MK  CLASS NAME    : cppvmArrayBool
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get a packed element
@MU  USAGE         : bool getPackedElem(int& byteIndex, int& bitIndex);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the 'packedArray' length must be long enough for the data
@M#  --------------------------------------------------------------------------
@MP  IN      : byteIndex byte index
@MP            bitIndex  index of the bit value in the byte
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool   element value
******************************************************************************/
bool
cppvmArrayBool::getPackedElem(int& byteIndex, int& bitIndex)
{
  
  static const unsigned char checkBit[] = {
    0x01,
    0x02,
    0x04,
    0x08,
    0x10,
    0x20,
    0x40,
    0x80
  };
  
  unsigned char result;

  result = (unsigned char)(packedArray[byteIndex] & checkBit[bitIndex]);
  //    cout << "getPackedElem: byteIndex: " << byteIndex
  //         << " bitIndex: " << bitIndex 
  //         << " packedArray[byteIndex]: " << (int)packedArray[byteIndex]
  //         << " checkBit[bitIndex]: " << (int)checkBit[bitIndex]
  //         << " and: " << (int)(packedArray[byteIndex] & checkBit[bitIndex])
  //         << endl;

  if((int)result==0){
    return false;
  }
  else{
    return true;
  }
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/









