/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_arfl.cpp 
@IE  CREATED     : 3.2.1998
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm float array class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_arfl.h"


/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_arfl.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                           VARIABLES                             *******
**********                                                           **********
******************************************************************************/
CPPvmBuildInMsgTagMethodsDefinition(cppvmArrayFloat,
                                      CPPvmMsgTagClassArrayFloat)

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmArrayFloat
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : cppvmArrayFloat
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmArrayFloat::cppvmArrayFloat()
    : cppvmArrayData(0), array(NULL), initEmptyArray(false)
{
  // base class constructor sets 'len=0'
  // resize works only because 'array == NULL'

  resize(0);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmArrayFloat
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmArrayFloat
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : len_  length of the array
@MP            cleanArray clean array after allocation
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmArrayFloat::cppvmArrayFloat(unsigned int len_, bool cleanArray)
    : cppvmArrayData(len_), array(NULL), initEmptyArray(false)
{
  // base class constructor sets 'len=len_'
  // resize works only because 'array == NULL'

  if(cleanArray==true){
    resizeClean(len_);
  }
  else{
    resize(len_);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmArrayFloat
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmArrayFloat::cppvmArrayFloat (const cppvmArrayFloat& obj)
    : cppvmArrayData(obj), array(NULL), initEmptyArray(false)
{
  // base class copy constructor sets 'len=obj.len'
  // resize works only because 'array == NULL'

  resizeWithoutFillAndCopy((unsigned int)obj.len);

  if(len>0){
    // copy elements
    memcpy((void*)array, (void*)obj.array, sizeof(float) * len);
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmArrayFloat
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmArrayFloat
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmArrayFloat::~cppvmArrayFloat()
{
  delete [] array;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to assign
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmArrayFloat&
cppvmArrayFloat::operator= (const cppvmArrayFloat& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }
  
  // adapt array length
  // 'cppvmArrayData::operator=' sets
  // len to obj.len, so resize has
  // to be called before
  resizeWithoutFillAndCopy((unsigned int)obj.len);

  cppvmArrayData::operator=(obj);

  if(obj.len>0){
    // copy elements
    memcpy((void*)array, (void*)obj.array, sizeof(float) * obj.len);
  }
    
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmArrayFloat::operator== (const cppvmArrayFloat& obj)
{
  int result;

  // same length ?
  if((cppvmArrayData::operator==(obj))==false){
    return false;
  }
  // same values ?
  if(len>0){
    result = memcmp((void*)&array[0], (void*) &obj.array[0], sizeof(float) * len);
    if (result !=0){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmArrayFloat::operator== (const cppvmArrayFloat& obj) const
{
  int result;

  // same length ?
  if((cppvmArrayData::operator==(obj))==false){
    return false;
  }
  // same values ?
  if(len>0){
    result = memcmp((void*)&array[0], (void*) &obj.array[0], sizeof(float) * len);
    if (result !=0){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmTransfer
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : transfer object from/to buffer
@MU  USAGE         : cppvmTransfer()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::cppvmTransfer()
{
  // transfer base class variables
  cppvmArrayData::cppvmTransfer();

  if(cppvmIsTransferReceive()==true){
    // receive array
    // remind old length
    int oldLen=len;
    // new array length
    int newLen;
    // transfer array length
    cppvmTransferInt(&newLen);

    if(cppvmTransferMode==CPPvmTransferModeAppend){
      // received array length plus old length
      resizeWithoutFill(oldLen+newLen);
      // add the incomming array
      cppvmTransferFloat(&(array[oldLen]), newLen);
    }
    else{
      resizeWithoutFillAndCopy(newLen);
      cppvmTransferFloat(array, len);
    }
  }
  else{
    // send array
    // transfer array length
    cppvmTransferInt(&len);
    cppvmTransferFloat(array, len);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::cppvmPrintOn(ostream& strm)
{
  int i;
  
  cppvmArrayData::cppvmPrintOn(strm);
  
  strm << "( ";
  for(i=0;i<len;i++){
    strm << array[i];
    strm << " ";
  }
  strm << ")";
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmScanFrom
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : scan method for << operator
@MU  USAGE         : cppvmScanFrom()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to scan object from
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::cppvmScanFrom(istream& strm)
{
  char dummy;
  int i;

  cppvmArrayData::cppvmScanFrom(strm);

  // resize array to new length
  resizeWithoutFillAndCopy((unsigned int)len);
  
  strm >> dummy;
  for(i=0;i<len;i++){
    strm >> array[i];
  }
  strm >> dummy;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cleanRange
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : clean a range in the array
@MU  USAGE         : cleanRange(unsigned int fromIndex, unsigned int toIndex)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the class variables 'len' and 'size' must have
@MR                  correct values, 'array' must not be NULL
@M#  --------------------------------------------------------------------------
@MP  IN      : fromIndex    index of first element to clean
@MP            toIndex      index of last element to clean
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::cleanRange(unsigned int fromIndex,
                            unsigned int toIndex)
{
  checkCleanRange(fromIndex, toIndex);

  unsigned int i;
  unsigned int numMemCopy;
  unsigned int restSingleCopyIndex;
  
  // clean memory to copy
  if(initEmptyArray==false){
    for(i=0; i< CPPvmArrayCleanFieldSize; i++){
      emptyArray[i]=0.0;
    }
      initEmptyArray=true;
  }
    
  numMemCopy = ((toIndex-fromIndex)+1) / CPPvmArrayCleanFieldSize;
  restSingleCopyIndex = fromIndex + (numMemCopy * CPPvmArrayCleanFieldSize);
  
  unsigned int index = fromIndex;
    // use memcpy
  for(i=0; i< numMemCopy; i++){
    memcpy((void*)(&array[index]), (void*)(&emptyArray[0]), 
           sizeof(float) *  CPPvmArrayCleanFieldSize);
      index = index +  CPPvmArrayCleanFieldSize;
  }
  
  // copy the rest as single values
  for(i=restSingleCopyIndex; i<=toIndex; i++){
    array[i]=0.0;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : newArray
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : allocate new array
@MU  USAGE         : void newArray(unsignd int newSize)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : newSize  new array size
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::newArray(unsigned int newSize)
{
  if(array!=NULL){
    // the former array is not needed any more
    delete [] array;
  }
  // allocate memory
  array = new float [newSize];  // allocate memory
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : newArrayCopy
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : allocate new array and copy elements from old array
@MU  USAGE         : void newArrayCopy(unsigned int newSize, 
@MU                                    unsigned int numCopyElements)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : newSize         new array size
@MP            numCopyElements number of element to copy from the old into the
@MP                            new array
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::newArrayCopy(unsigned int newSize, unsigned int numCopyElements)
{
  if(array==NULL){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmArrayFloat",
                                                 "newArrayCopy",
                                                 "array to copy from is NULL!"));
  }
  
  // new array is needed
  float *tmpArray = array;

  // allocate memory  
  array = new float [newSize];  

  if(numCopyElements>0){
    // copy memory
    memcpy((void*)array, (void*)tmpArray, sizeof(float) * numCopyElements);
  }

  // the former array is not needed any more
  delete [] tmpArray;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : addElem
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add a value from another array to this value
@MU  USAGE         : void addElem(int pos, cppvmArrayData& obj, int objPos);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pos     index of the element of this array (the value is
@MP                    added to the element with this index)
@MP            obj     array object (from which an value should be added)
@MP            objPos  index for the value to add in the given array object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::addElem(int pos, cppvmArrayData& obj, int objPos)
{
  if(obj.cppvmMsgTag()==CPPvmMsgTagClassArrayFloat){
    // array object of this class
    array[pos] = array[pos] + ((cppvmArrayFloat&)(obj))[objPos];
  }
  else{
    // add value via double conversion
    array[pos] = array[pos] + (float) (obj.getElemDouble(objPos));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : subElem
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : subtract a value from another array to this value
@MU  USAGE         : void subElem(int pos, cppvmArrayData& obj, int objPos);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pos     index of the element of this array (the value is
@MP                    subtracted to the element with this index)
@MP            obj     array object (from which an value should be subtracted)
@MP            objPos  index for the value to subtract in the given
@MP                    array object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::subElem(int pos, cppvmArrayData& obj, int objPos)
{
  if(obj.cppvmMsgTag()==CPPvmMsgTagClassArrayFloat){
    // array object of this class
    array[pos] = array[pos] - ((cppvmArrayFloat&)(obj))[objPos];
  }
  else{
    // subtract value via double conversion
    array[pos] = array[pos] - (float) (obj.getElemDouble(objPos));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : multElem
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : multiply a value from another array to this value
@MU  USAGE         : void multElem(int pos, cppvmArrayData& obj, int objPos);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pos     index of the element of this array (the value is
@MP                    multiplied to the element with this index)
@MP            obj     array object (from which an value should be multiplied)
@MP            objPos  index for the value to multiply in the given
@MP                    array object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::multElem(int pos, cppvmArrayData& obj, int objPos)
{
  if(obj.cppvmMsgTag()==CPPvmMsgTagClassArrayFloat){
    // array object of this class
    array[pos] = array[pos] * ((cppvmArrayFloat&)(obj))[objPos];
  }
  else{
    // multiply value via double conversion
    array[pos] = array[pos] * (float) (obj.getElemDouble(objPos));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : divElem
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : divide a value from another array to this value
@MU  USAGE         : void divElem(int pos, cppvmArrayData& obj, int objPos);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pos     index of the element of this array (the value is
@MP                    divided from the element with this index)
@MP            obj     array object (from which an value should be devided)
@MP            objPos  index for the value to divide in the given
@MP                    array object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::divElem(int pos, cppvmArrayData& obj, int objPos)
{
  if(obj.cppvmMsgTag()==CPPvmMsgTagClassArrayFloat){
    // array object of this class
    array[pos] = array[pos] / ((cppvmArrayFloat&)(obj))[objPos];
  }
  else{
    // devide value via double conversion
    array[pos] = array[pos] / (float) (obj.getElemDouble(objPos));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : copyElems
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy values from an external objects into own array
@MU  USAGE         : void copyElems(int num, int localIndex,
@MU                                 cppvmArrayData& extObj, int extIndex);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : Copy 'num' elements: 
@MR                  array[localIndex] = obj.array[extIndex]; etc.
@M#  --------------------------------------------------------------------------
@MP  IN      : num        number of elements to copy
@MP            localIndex start index for local object
@MP            extObj     external object to copy data from
@MP            extIndex   start index for external object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::copyElems(int num, int localIndex, 
                         cppvmArrayData& extObj, int extIndex)
{
  if(extObj.cppvmMsgTag()==CPPvmMsgTagClassArrayFloat){
    // array object of this class
    // can be a reference to own object
    // -> use 'memmove' instead of memcpy
    if(num>0){
      memmove((void*) &array[localIndex], 
              (void*) &((cppvmArrayFloat&)extObj).array[extIndex], 
              sizeof(float) * (num));
    }
  }
  else{
    int i;
    for(i=0; i<num; i++){
      // set value via double conversion
      array[localIndex+i] = (float) (extObj.getElemDouble(extIndex+i));
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cycleShiftLeft
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : cycle shift left
@MD                  (1 2 3) -> (2 3 1)
@MU  USAGE         : void cycleShiftLeft();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::cycleShiftLeft()
{
  float tmp;

  if(len <= 1){
    return;
  }
  
  tmp = array[0];
  memmove((void*) &array[0], (void*) &array[1], sizeof(float) * (len-1));
  array[len-1]=tmp;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cycleShiftRight
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : cycle shift right
@MD                  (1 2 3)-> (3 1 2)
@MU  USAGE         : void cycleShiftRight();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::cycleShiftRight()
{
  float tmp;

  if(len <= 1){
    return;
  }
  
  tmp = array[len-1];
  memmove((void*) &array[1], (void*) &array[0], sizeof(float) * (len-1));
  array[0] = tmp;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : read
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : return current value
@MU  USAGE         : float read(unsigned int index);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : index element index in array
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
float
cppvmArrayFloat::read(unsigned int index)
{
  // get object from the mailbox (if the mailbox was initialized)
  cppvmReadMboxInfo();

#ifndef CPPvmFastAccess
  if(index>(unsigned int)len){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmArrayFloat",
                                                 "read",
                                                 "index > array length"));
  }
#endif

  return array[index];
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : write
@MK  CLASS NAME    : cppvmArrayFloat
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set array value
@MU  USAGE         : void write(unsigned int index, int value)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : index index of element
@MP            value new value for element
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmArrayFloat::write(unsigned int index,
                     float value)
{
#ifndef CPPvmFastAccess
  if(index>(unsigned int)len){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmArrayFloat",
                                                 "write",
                                                 "index > array length"));
  }
#endif
  array[index] = value;

  // send it to the mailbox (if the mailbox was initialized)
  cppvmWriteMboxInfo();
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

