/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_barr.cpp 
@IE  CREATED     : 18.09.2001
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : barrier class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// cppvm classes
#include "pvm_obj.h"
#include "pvm_str.h"
// class header
#include "pvm_barr.h" 

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_barr.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmMaxId
@VD  DESCRIPTION   : unique id number for all framework objects
@VD                  highest actual ID
@VT  TYPE          : int
***
******************************************************************************/
int cppvmBarrier::cppvmMaxId = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmTotalObjNum
@VD  DESCRIPTION   : total number of cppvmObjects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmBarrier::cppvmTotalObjNum = 0; 

/*@@@**************************************************************************
**********                                                           **********
*******                            MACROS                               *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmBarrierThrowIfNotInGroup
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : throws an exception if the group name of the object was not
@MD                 initialized 
@MU  USAGE        : only for internal use!
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmBarrierThrowIfNotInGroup                                                 \
if(barrierGroupName.getLength()==0){                                                  \
  CPPvmInternalThrow (cppvmExceptionForbiddenCall("pvm_barr.cpp",                     \
                  "cppvmBarrier",                                                     \
                  "Macro: CPPvmBarrierThrowIfNotInGroup",                             \
                  "Group name is not set. Probably you called the 'leaveGroup' method. The group name must be set with method 'setGroup' before calling 'barrier' again.")); \
}

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmBarrier
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : cppvmBarrier
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmBarrier::cppvmBarrier()
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmBarrier
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmBarrier
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : groupName     barrier group name
@MP            count         number of group members that must call the
@MP                          method 'barrier' until they are released
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmBarrier::cppvmBarrier(char *groupName, unsigned int count)
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  // set barrier group and count
  setGroup(groupName, false);
  setCount(count);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmBarrier
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmBarrier::cppvmBarrier (const cppvmBarrier& obj)
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  if(barrierGroupName.getLength()==0){
    // the barrier is currently in no group
    // -> do not try to leave group
    setGroup(obj.barrierGroupName.getPtr(), false);
  }
  else{
    setGroup(obj.barrierGroupName.getPtr(), true);
  }
  
  setCount(obj.barrierCount);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmBarrier
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmBarrier()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      :
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
cppvmBarrier::~cppvmBarrier()
{
  // dec total number of objects
  cppvmTotalObjNum--;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  info for configuration object attributes 
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmBarrier&
cppvmBarrier::operator= (const cppvmBarrier& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }

  if(barrierGroupName.getLength()==0){
    // the barrier is currently in no group
    // -> do not try to leave group
    setGroup(obj.barrierGroupName.getPtr(), false);
  }
  else{
    setGroup(obj.barrierGroupName.getPtr(), true);
  }
  setCount(obj.barrierCount);
  
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : barrier
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : block the process until 'count' processes in the group
@MD                  have called it
@MU  USAGE         : void barrier();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmBarrier::barrier()
{
  CPPvmBarrierThrowIfNotInGroup  

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    barrierTry();
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : barrierTry
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : block the process until 'count' processes in the group
@MD                  have called it
@MU  USAGE         : void barrierTry();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmBarrier::barrierTry()
{
  int info;

  info =  pvm_barrier(barrierGroupName.getPtr(), barrierCount);

  if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "barrierTry",
                      "pvm-daemon was not started or has crashed."));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "barrierTry",
                      "count < 1"));
  }
  else if(info == PvmNoGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "barrierTry",
                      "giving a non-existent group name"));
  }
  else if(info == PvmNotInGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "barrierTry",
                      "calling process is not in specified group"));
  }
}

  
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print on stream method
@MU  USAGE         : void cppvmPrintOn(ostream& strm);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm out stream
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmBarrier::cppvmPrintOn(ostream& strm)
{
  strm << "barrier group name: " << barrierGroupName
       << " count: " << barrierCount
       << endl;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the barrier group name and count
@MU  USAGE         : void set(char *groupName, unsigned int count, 
@MU                           bool leaveFormerGroup);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : groupName        barrier group name
@MP            count            number of group members that must call the
@MP                             method 'barrier' until they are released
@MP            leaveFormerGroup leave the former barrier group
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmBarrier::set(char *groupName, unsigned int count, bool leaveFormerGroup)
{
  setGroup(groupName, leaveFormerGroup);
  setCount(count);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setCount
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the barrier count
@MU  USAGE         : void setCount(unsigned int count);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN        count         number of group members that must call the
@MP                          method 'barrier' until they are released
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmBarrier::setCount(unsigned int count)
{
  barrierCount = (int) count;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setGroup
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the barrier group name
@MU  USAGE         : void setGroup(char *groupName, bool leaveFormerGroup);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : groupName        barrier group name
@MP            leaveFormerGroup leave the former barrier group
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmBarrier::setGroup(char *groupName, bool leaveFormerGroup)
{
  if(groupName==NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmBarrier",
                                                   "setGroup",
                                                   "groupName is a NULL pointer."));
  }
  
  // leave old group
  if(leaveFormerGroup==true){
    leaveGroup();
  }

  // assign new group name
  barrierGroupName = groupName;

  // join new group
  joinGroup(groupName);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : joinGroup
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : enrolls the calling process in a named group
@MU  USAGE         : int joinGroup(char* groupName);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : groupName  group name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  instance number; the number starts at 0 and count up
@MP                 < 0 indicates an error
******************************************************************************/
void
cppvmBarrier::joinGroup(char *groupName)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    joinGroupTry(groupName);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : joinGroupTry
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : enrolls the calling process in a named group
@MU  USAGE         : int joinGroupTry(char* groupName);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : groupName  group name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  instance number; the number starts at 0 and count up
@MP                 < 0 indicates an error
******************************************************************************/
void
cppvmBarrier::joinGroupTry(char *groupName)
{
  int info;

  info =  pvm_joingroup(groupName);

  if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "joinGroup",
                      "giving a NULL group name"));
  }
  else if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "joinGroup",
                      "pvmd was not started or has crashed"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : leaveGroup
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unenrolls the calling process from a named group
@MU  USAGE         : void leaveGroup();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : groupName  group name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmBarrier::leaveGroup()
{
  CPPvmBarrierThrowIfNotInGroup  

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    leaveGroupTry();
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : leaveGroupTry
@MK  CLASS NAME    : cppvmBarrier
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unenrolls the calling process from a named group
@MU  USAGE         : void leaveGroupTry();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : groupName  group name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmBarrier::leaveGroupTry()
{
  int info;

  info =  pvm_lvgroup(barrierGroupName.getPtr());  

  // delete old group name
  // to indicate that the process is not
  // part of the old group any more
  barrierGroupName.clean();
  
  if(info == PvmNoGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "leaveGroup",
                      "giving a non-existent group name"));
  }
  else if(info == PvmNotInGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "leaveGroup",
                      "asking to leave a group you are not a member of"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "leaveGroup",
                      "giving a NULL group name"));
  }
  else if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmBarrier",
                      "leaveGroup",
                      "pvmd was not started or has crashed"));
  }
}


/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



