/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_catc.cpp 
@IE  CREATED     : 2.7.1998
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : catchout for spawned processes
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// includes for predeclared classes to avoid cyclic includes
#include "pvm_str.h"
// include own class
#include "pvm_catc.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_catc.cpp"


/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : catchoutFileDesc
@VD  DESCRIPTION   : file pointer; the output is redirected to this file
@VT  TYPE          : FILE *
***
******************************************************************************/
FILE * cppvmCatchout::catchoutFileDesc = NULL; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : isFile
@VD  DESCRIPTION   : indicates if 'catchoutFileDesc' is a file (must be closed)
@VD                  or other outputs (stdout, stderr)
@VT  TYPE          : bool
***
******************************************************************************/
bool cppvmCatchout::isFile = false; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : numObjs
@VD  DESCRIPTION   : total number of 'cppvmCatchout' objects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmCatchout::numObjs = 0; 

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmCatchout
@MK  CLASS NAME    : cppvmCatchout
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmCatchout(cppvmCatchoutMode catchoutFlag,
@MU                              char *catchoutFileName);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : catchoutFlag specifies, where to catch the output of the child:
@MP               * 'CPPvmStdout'    : write the output on stdout
@MP               * 'CPPvmStderr'    : write the output on stderr
@MP               * 'CPPvmFile'      : write output on file specified by
@MP                                    'catchoutFileName'
@MP            catchoutFileName specifies a catchout file name;  only used if 
@MP               'catchoutFlag' is set to 'CPPvmFile'
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmCatchout::cppvmCatchout(cppvmCatchoutMode catchoutFlag,
                             char *catchoutFileName)
  : cppvmProcessHandling()
{
  // a new object is born...
  numObjs = numObjs + 1;
  if(numObjs>=2){
    // there are allready objects activ
    // unset their output redirection
    unset();
  }
  
  fileName = new cppvmString();
  
  // set new catchout 
  set(catchoutFlag, catchoutFileName);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmCatchout
@MK  CLASS NAME    : cppvmCatchout
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmCatchout
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmCatchout::~cppvmCatchout()
{
  numObjs = numObjs - 1;

  if(numObjs == 0){
    // this was the last cppvmCatchout object
    unset();
  }
  delete fileName;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmCatchout
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmCatchout object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmCatchout::operator== (const cppvmCatchout& obj) 
{
  return((*fileName)==(*(obj.fileName)) &&
         flag==obj.flag);
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmCatchout
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmCatchout object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmCatchout::operator== (const cppvmCatchout& obj) const
{
  return((*fileName)==(*(obj.fileName)) &&
         flag==obj.flag);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator!=
@MK  CLASS NAME    : cppvmCatchout
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are not equal
@MU  USAGE         : !=
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : i  cppvmCatchout value
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmCatchout::operator!= (const cppvmCatchout& obj) 
{
  return ! (operator==(obj));
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : printOn
@MK  CLASS NAME    : cppvmCatchout
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : printOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmCatchout::cppvmPrintOn(ostream& strm)
{
  strm << "Catchout cout/cerr from child processes to: ";
  if(flag==CPPvmStop){
    strm << "stoped catchout";
  }
  else if(flag==CPPvmStdout){
    strm << "stdout";
  }
  else if(flag==CPPvmStderr){
    strm << "stderr";
  }
  else if(flag==CPPvmFile){
    strm << "file " << (*fileName);
  }
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the child catchout
@MU  USAGE         : void set(cppvmCatchoutMode catchoutFlag,
@MU                                       char *catchoutFileName); 
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : catchoutFlag specifies, where to catch the output of the child:
@MP               * 'CPPvmStdout'    : write the output on stdout
@MP               * 'CPPvmStderr'    : write the output on stderr
@MP               * 'CPPvmFile'      : write output on file specified by
@MP                                    'catchoutFileName'
@MP            catchoutFileName spwcifies a coutout file name;  only used if 
@MP               'catchoutFlag' is set to 'CPPvmFile'
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmCatchout::set(cppvmCatchoutMode catchoutFlag,
                   char *catchoutFileName)
{
  int result=0;
  flag=catchoutFlag;
  // no file output
  isFile=false;
  
  
  if(catchoutFlag==CPPvmStdout){
    // send all cout and cerr of processes spawned after this function
    // to stderr
    catchoutFileDesc = stdout;
    (*fileName) = "";
    result = pvm_catchout(catchoutFileDesc);
  }
  else if(catchoutFlag==CPPvmStderr){
    // send all cout and cerr of processes spawned after this function
    // to stdout
    catchoutFileDesc = stderr;
    (*fileName) = "";
    result = pvm_catchout(catchoutFileDesc);
  }
  else if(catchoutFlag==CPPvmStop){
    // stop catchout
    catchoutFileDesc = NULL;
    (*fileName) = "";
    result = pvm_catchout(catchoutFileDesc);
  }
  else if(catchoutFlag==CPPvmFile){
    if(catchoutFileName == NULL){
      CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                     "cppvmCatchout",
                                     "set",
                                     "'CPPvmFile' set and no log file name given."));
    }

    isFile = true;
    // remind file name
    (*fileName) = catchoutFileName;
    // open log file
    catchoutFileDesc =  fopen(catchoutFileName, "w");
    if(catchoutFileDesc == NULL){
      // no file name given
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmCatchout",
                                                  "set",
                                                  "cannot open catchout file"));
    }

    // log all cout and cerr of processes spawned after this function
    // in this new file
    result = pvm_catchout(catchoutFileDesc);
  }
  else{
    // unknown 'catchoutFlag'
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmCatchout",
                                                "set",
                                                "unknown catchout flag"));
  }
  

  if(result<0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                    "cppvmCatchout",
                                    "set",
                                    "cannot redirect output."));
  }
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : reset
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the child catchout to the former given values
@MU  USAGE         : void reset();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmCatchout::reset()
{
  // unset output redirection
  unset();
  // set catchout again
  set(flag, fileName->getPtr());
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : unset
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unset the current catchout redirection
@MU  USAGE         : void unset();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmCatchout::unset()
{
  if(catchoutFileDesc != NULL){
    // write all buffered data 
    fflush(catchoutFileDesc);
    if(isFile == true){
      // is cout and cerr caught in a file? If yes, delete file handle
      // (close file).
      fclose(catchoutFileDesc);
    }
  }
  isFile = false;
  catchoutFileDesc = NULL;
}


/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

