/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_coll.cpp 
@IE  CREATED     : 16.12.1997
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm object collection
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_coll.h"


/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_coll.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                           VARIABLES                             *******
**********                                                           **********
******************************************************************************/
CPPvmBuildInMsgTagMethodsDefinition(cppvmObjectCollection,
                                      CPPvmMsgTagClassObjcoll)

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmObjectCollection
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmObjectCollection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmObjectCollection::cppvmObjectCollection()
    : cppvmObject()
{
  // init object collection
  size = 0;
  elems = new cppvmObject*[size];
  num = 0;
  // init receive function array
  recFctSize = 0;
  recFcts = new pvmCollRecFctPtr[recFctSize];
  recFctNum = 0;
  // transfer mode is clean old collection
  cppvmTransferMode=CPPvmTransferModeClean;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmObjectCollection
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmObjectCollection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmObjectCollection::~cppvmObjectCollection()
{
  clean();
  delete [] elems;
  delete [] recFcts;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmObjectCollection::operator== (const cppvmObjectCollection& obj)
{
  unsigned int i;
  
  if(this == &obj){
    return true;
  }
  if(num !=  obj.num){
    return false;
  }
  for (i=0; i<num; i++){
    if (elems[i] != obj.elems[i]){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmObjectCollection::operator== (const cppvmObjectCollection& obj) const
{
  unsigned int i;
  
  if(this == &obj){
    return true;
  }
  if(num !=  obj.num){
    return false;
  }
  for (i=0; i<num; i++){
    if (elems[i] != obj.elems[i]){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmTransfer
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : transfer object from/to buffer
@MU  USAGE         : cppvmTransfer()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::cppvmTransfer()
{
  if(cppvmIsTransferSend()==true){
    cppvmTransferSend();
  }
  if(cppvmIsTransferReceive()==true){
    cppvmTransferRec();
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmTransferSend
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : transfer object to buffer
@MU  USAGE         : cppvmTransferSend()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::cppvmTransferSend()
{
  unsigned int i;

  // send total number of objects
  transferNumObj = (int) num;
  cppvmTransferInt(&transferNumObj);
  // send objects
  for(i=0; i<num; i++){
    // send unique! message tag for the class type of the object
    // (unique means that the message tag of every class differs to other
    //  classes)
    classIdMsgTag = elems[i]->cppvmMsgTag();
    cppvmTransferInt(&classIdMsgTag);
    // finally send the object itself
    elems[i]->cppvmTransfer();
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmTransferRec
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : transfer object from buffer
@MU  USAGE         : cppvmTransferRec()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::cppvmTransferRec()
{
  int i;
  int sendNum;
  int classId;
  char errMsg[300];
  
  // get the number of total sent objects
  cppvmTransferInt(&sendNum);
    
  if(cppvmTransferMode==CPPvmTransferModeClean) {
    // receive with deleting old collection

    clean();  
  }
  
  for(i=0; i<sendNum; i++){
    // get the class id (message tag) of the next object to receive
    cppvmTransferInt(&classId);

    if(cppvmTransferRecBase(classId)!=true){
      // no object of the base classes were found
      if(cppvmTransferRecFcts(classId)!=true){
        // no object which can received by the users were found
        sprintf(errMsg, "The sent object with hash class id (msg tag) %i in a cppvmObjectCollection collection cannot be received. You have to add a special receive function for this object.", classId);
      CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,                   
                                                     "cppvmObjectCollection",
                                                     "cppvmTransferRec",
                                                     errMsg));  
      }
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmTransferRecBase
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : receive the base objects from the buffer
@MU  USAGE         : bool cppvmTransferRecBase(int classId);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : classId   message tag (class id) of the next object
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : bool  true: object was received
@MP                  false: object was not received; not an object of the
@MP                         base objects
******************************************************************************/
bool
cppvmObjectCollection::cppvmTransferRecBase(int classId)
{
  if(classId == cppvmObjectCollection::cppvmClassMsgTag()){
    cppvmObjectCollection *objPtr;
    unsigned int i;
    
    objPtr = new cppvmObjectCollection;
    // inherit all external receive functions of this collection
    // to this new collection object
    for(i=0;i<recFctNum;i++){
      objPtr->addRecFct(recFcts[i]);
    }
    // receive the collection
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmString::cppvmClassMsgTag()){
    cppvmString *objPtr;
    objPtr = new cppvmString;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmBool::cppvmClassMsgTag()){
    cppvmBool *objPtr;
    objPtr = new cppvmBool;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmInt::cppvmClassMsgTag()){
    cppvmInt *objPtr;
    objPtr = new cppvmInt;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmFloat::cppvmClassMsgTag()){
    cppvmFloat *objPtr;
    objPtr = new cppvmFloat;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmDouble::cppvmClassMsgTag()){
    cppvmDouble *objPtr;
    objPtr = new cppvmDouble;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmLong::cppvmClassMsgTag()){
    cppvmLong *objPtr;
    objPtr = new cppvmLong;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmChar::cppvmClassMsgTag()){
    cppvmChar *objPtr;
    objPtr = new cppvmChar;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmArrayInt::cppvmClassMsgTag()){
    cppvmArrayInt *objPtr;
    objPtr = new cppvmArrayInt;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmArrayBool::cppvmClassMsgTag()){
    cppvmArrayBool *objPtr;
    objPtr = new cppvmArrayBool;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmArrayLong::cppvmClassMsgTag()){
    cppvmArrayLong *objPtr;
    objPtr = new cppvmArrayLong;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmArrayFloat::cppvmClassMsgTag()){
    cppvmArrayFloat *objPtr;
    objPtr = new cppvmArrayFloat;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmArrayDouble::cppvmClassMsgTag()){
    cppvmArrayDouble *objPtr;
    objPtr = new cppvmArrayDouble;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmArrayChar::cppvmClassMsgTag()){
    cppvmArrayChar *objPtr;
    objPtr = new cppvmArrayChar;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else if(classId == cppvmTask::cppvmClassMsgTag()){
    cppvmTask *objPtr;
    objPtr = new cppvmTask;
    objPtr->cppvmTransfer();
    add((cppvmObject*)(objPtr));
    return true;
  }
  else{
    return false;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmTransferRecFcts
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : receive user defined objects with user defined receive
@MD                  functions
@MU  USAGE         : bool cppvmTransferRecFcts(int classId);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : classId   message tag (class id) of the next object
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : bool  true: object was received
@MP                  false: object was not received; not an object of the
@MP                         base objects
******************************************************************************/
bool
cppvmObjectCollection::cppvmTransferRecFcts(int classId)
{
  unsigned int i;
  cppvmObject *tmp;

  for(i=0; i<recFctNum; i++){
    // call external receive function
    tmp = (recFcts[i])(classId);
    if(tmp != NULL){
      add((cppvmObject*)(tmp));
      return true;
    }
  }
  return false;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::cppvmPrintOn(ostream& strm)
{
  unsigned int i;

  strm << "( ";
  for(i=0; i<num; i++){
    strm << *elems[i];
    strm << " ";
  }
  strm << ")";
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmScanFrom
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : scan method for << operator
@MU  USAGE         : cppvmScanFrom()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to scan object from
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::cppvmScanFrom(istream& /*strm*/)
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : remove
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove element (cppvmObject pointer) from collection
@MU  USAGE         : remove(obj)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   cppvmObject to remove
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::remove(cppvmObject& obj)
{
  unsigned int i;
  
  for(i=0; i<num; i++){
    if(elems[i]->cppvmIsSame(obj)){
      num = num - 1;
      delete elems[i];
      shiftLeft((unsigned int) i, (unsigned int) num);
      return;
    }
  }
  CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                 "cppvmObjectCollection",
                                                 "remove(cppvmObject&)",
                                                 "Object not in collection"));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : add
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add pointer to an element to collection
@MU  USAGE         : add(cppvmObject *objPtr)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   pointer to cppvmObject to add
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::add(cppvmObject *objPtr)
{
  // dont add same collection to collection (endless recursion)
  if(cppvmIsSame(*objPtr)){
    return;
  }
  unsigned int newSize;
  
  if (num >= size){
    if(size==0){
      newSize = 10;
    }
    else{
      newSize = size + 10;
    }
    resize(newSize);
  }
  elems[num] = objPtr;

  num = num + 1;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : remove
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove element (cppvmObject pointer) from collection
@MU  USAGE         : remove(cppvmObject *objPtr)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   pointer to cppvmObject to remove
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::remove(cppvmObject *objPtr)
{
  unsigned int i;
  
  for(i=0; i<num; i++){
    if(elems[i]->cppvmIsSame(*objPtr)){
      num = num - 1;
      delete elems[i];
      shiftLeft((unsigned int) i, (unsigned int) num);
      return;
    }
  }
  CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                 "cppvmObjectCollection",
                                                 "remove(T&)",
                                                 "Object not in collection"));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : remove
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove element (cppvmObject pointer) from collection
@MU  USAGE         : remove(unsigned int)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : i  index of object to remove
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::remove(unsigned int i)
{
  if(i >=num){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmObjectCollection",
                                                 "remove(unsigned int)",
                                                 "Bad index"));
  }
  num = num - 1;
  delete elems[i];
  shiftLeft((unsigned int) i, (unsigned int) num);

  return;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : clean
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove all elements (cppvmObject pointer) from collection
@MD                  and remove references
@MU  USAGE         : clean()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::clean()
{
  if(size==0){
    // is clean
    return;
  }
  
  unsigned int i;

  for(i=0; i<num; i++){
    delete elems[i];
  }

  delete [] elems;
  
  size = 0;
  elems = new cppvmObject*[size];
  num = 0;

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmObjectCollection
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : append another collection to this collection; the appended collection
@MD                  is deleted
@MU  USAGE         : void append(cppvmObjectCollection *objColl);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : objColl   collection to append
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmObjectCollection::append(cppvmObjectCollection *objColl)
{
  unsigned int i;

  // copy all object pointer
  for (i=0; i<(objColl->num); i++){
    add(objColl->elems[i]);
  }
  // set num elem of appended collection to NULL
  // DO NOT delete objects, otherwise they are gone forever
  objColl->num=0;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resize
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : rezise collection length
@MD                  if the new size is smaller than the current size, 
@MD                  nothing is done
@MU  USAGE         : resize(unsigned int)
@MU                  this can be done explicit after creating a new collection
@MU                  or implicit by adding an object to an empty collection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newSize   new size 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::resize(unsigned int newSize)
{
  cppvmObject** tmp;

  if(newSize < num){
    return;
  }
  // create new array
  size = newSize;
  tmp = new cppvmObject* [size];
  
  if(num>0){
    // copy all object pointer
    memcpy((void*) tmp, (void*) elems, sizeof(cppvmObject*) * num);
  }

  // delete old array
  delete [] elems;
  elems = tmp;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : swap
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : swap two members of the collection
@MU  USAGE         : swap(int index1, int index2);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : index1   index of the first cppvmObject*
@MP            index2   index of the second cppvmObject*
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::swap(unsigned int index1, unsigned int index2)
{
  cppvmObject *tmp;

  if (index1>= num || index2 >= num){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_COLL_MODULE,
                                            "cppvmObjectCollection",
                                            "swap",
                                            "index1 or index2 > collection length"));
  }

  tmp = elems[index1];
  elems[index1] = elems[index2];
  elems[index2] = tmp;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : sort
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : sorts the members of the collection by the given compare
@MD                  function with the direct insert method
@MU  USAGE         : sort(pvmCollCompFctPtr cmpPtr);
@MC  CONSTRAINTS   : the compare function MUST be usable for all cppvmObjects
@MC                  in the collection
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : cmpPtr   compare function for objects
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::sort(pvmCollCompFctPtr cmpPtr)
{
  cppvmObject *tmp;
  unsigned int i,j;
  
  // sort here
  for(i=1; i<num; i++){
    tmp = elems[i];
    j=i;
    while((j>0) && cmpPtr(elems[j-1],tmp)){
      elems[j] = elems[j-1];
      j--;
    }
    elems[j] = tmp;
  }

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : shiftLeft
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : shift all object pointers to the left to keep 
@MD                  the collection ordered
@MU  USAGE         : shiftLeft(unsigned int indexMin,
@MU                            unsigned int indexMax)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : example: collection(a,b,c,d,e)
@MR                  collection.shiftLeft(2,4) -> collection(a,b,d,e,?)
@MR                  
@MR                  
@M#  --------------------------------------------------------------------------
@MP  IN      : indexMin  the collection is shifted left up to this element
@MP            indexMax  all objects up to this index are shifted
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::shiftLeft(unsigned int indexMin,
                      unsigned int indexMax)
{
  if(indexMin > indexMax){
    return;
  }
  
  unsigned int numCopy = indexMax-indexMin;
  if(numCopy>0){
    // copy all object pointer
    memmove((void*) &elems[indexMin], (void*) &elems[indexMin+1], 
           sizeof(cppvmObject*) * numCopy);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : addRecFct
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add pointer to a function which receives an element
@MD                  and gives the pointer to the received element back.
@MD                  the element is added to the elem collection.
@MU  USAGE         : addRecFct(pvmCollRecFctPtr fctPtr);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : fctPtr  pointer to cppvmObject receive function 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::addRecFct(pvmCollRecFctPtr fctPtr)
{
  // check if the function was already added
  unsigned int i;

  for(i=0;i<recFctNum;i++){
    if(recFcts[i]==fctPtr){
      // do not add the same function again
      return;
    }
  }
  
  unsigned int newSize;
  
  if (recFctNum >= recFctSize){
    if(recFctSize==0){
      newSize = 10;
    }
    else{
      newSize = recFctSize + 10;
    }
    resizeRecFct(newSize);
  }
  recFcts[recFctNum] = fctPtr;

  recFctNum = recFctNum + 1;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : removeRecFct
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove receive function from collection of receive functions
@MU  USAGE         : removeRecFct(pvmCollRecFctPtr fctPtr);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : fctPtr  pointer to receive function which should be removed
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::removeRecFct(pvmCollRecFctPtr fctPtr)
{
  unsigned int i;
  
  for(i=0; i<recFctNum; i++){
    if(recFcts[i]==fctPtr){
      recFctNum = recFctNum - 1;
      shiftLeftRecFct((unsigned int) i, (unsigned int) recFctNum);
      return;
    }
  }
  CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                     "cppvmObjectCollection",
                     "removeRecFct",
                     "function is not in the collection"));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resizeRecFct
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : rezise recFcts collection length
@MD                  if the new size is smaller than the current size, 
@MD                  nothing is done
@MU  USAGE         : resizeRecFct(unsigned int)
@MU                  this can be done explicit after creating a new collection
@MU                  or implicit by adding an object to an empty collection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newSize   new size 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::resizeRecFct(unsigned int newSize)
{
  pvmCollRecFctPtr *tmp;

  if(newSize < recFctNum){
    // new size is smaller than current size
    return;
  }
  // create new array
  recFctSize = newSize;
  tmp = new pvmCollRecFctPtr[recFctSize];

  // copy all function pointer
  if(recFctNum>0){
    // copy all object pointer
    memcpy((void*) tmp, (void*) recFcts, sizeof(pvmCollRecFctPtr*) * recFctNum);
  }

  // delete old array
  delete [] recFcts;
  recFcts = tmp;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : shiftLeftRecFct
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : shift all function pointers to the left to keep 
@MD                  the collection ordered
@MU  USAGE         : shiftLeftRecFct(unsigned int indexMin,
@MU                                  unsigned int indexMax)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : example: collection(a,b,c,d,e)
@MR                  collection.shiftLeftRecFct(2,4) -> collection(a,b,d,e,?)
@MR                  
@MR                  
@M#  --------------------------------------------------------------------------
@MP  IN      : indexMin  the collection is shifted left up to this element
@MP            indexMax  all function pointer up to this index are shifted
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::shiftLeftRecFct(unsigned int indexMin,
                            unsigned int indexMax)
{
  if(indexMin > indexMax){
    return;
  }
  
  unsigned int numCopy = indexMax-indexMin;
  if(numCopy>0){
    // copy all object pointer
    memmove((void*) &recFcts[indexMin], (void*) &recFcts[indexMin+1], 
           sizeof(pvmCollRecFctPtr*) * numCopy);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setCPPvmTransferMode
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : determins weather a received object should be placed
@MD                  instead (CLEAN) of the current collection or should be appended
@MD                  (APPEND) to the current collection
@MU  USAGE         : void setCPPvmTransferMode(int cppvmTransferMode_);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : cppvmTransferMode_ CPPvmTransferModeClean:
@MP                                remove current collection, receive new collection
@MP                             CPPvmTransferModeAppend:  
@MP                                append received collection to current collection
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmObjectCollection::setCPPvmTransferMode(int cppvmTransferMode_){
  if(cppvmTransferMode_!=CPPvmTransferModeClean &&
     cppvmTransferMode_!=CPPvmTransferModeAppend){
    CPPvmInternalThrow( cppvmExceptionBadParameter(PVM_COLL_MODULE,
                                                   "cppvmObjectCollection",
                                                   "setCPPvmTransferMode",
                                                   "unknown transfer mode"));
  }
  cppvmTransferMode = cppvmTransferMode_;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator[]
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : access to the objects in the collection
@MU  USAGE         : []
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : i  object index in collection
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmObject&
cppvmObjectCollection::operator[] (unsigned int i)
{
#ifndef CPPvmFastAccess
  if (i>= num){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_COLL_MODULE,
                                                 "cppvmObjectCollection",
                                                 "operator[]",
                                                 "index > collection length"));
  }
#endif
  return *(elems[i]);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getObjPtr
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : access to the pointer of objects in the collection
@MU  USAGE         :  getObjPtr(unsigned int i)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : i  object index in collection
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmObject*
cppvmObjectCollection::getObjPtr(unsigned int i)
{
#ifndef CPPvmFastAccess
  if (i>= num){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_COLL_MODULE,
                                                 "cppvmObjectCollection",
                                                 "getObjPtr",
                                                 "index > collection length"));
  }
#endif
  return elems[i];
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getConstObjPtr
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : access to the pointer of objects in the collection
@MU  USAGE         :  getConstObjPtr(unsigned int i)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : i  object index in collection
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmObject*
cppvmObjectCollection::getConstObjPtr(unsigned int i) const
{
#ifndef CPPvmFastAccess
  if (i>= num){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_COLL_MODULE,
                                                 "cppvmObjectCollection",
                                                 "getConstObjPtr",
                                                 "index > collection length"));
  }
#endif
  return elems[i];
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : read
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : read object from collection; update local collection with the
@MD                  message box (if this object was initialized in the
@MD                  message box)
@MU  USAGE         : cppvmObject& read (unsigned int i)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : i  object index in collection
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : cppvmObject&  element number i in collection
******************************************************************************/
cppvmObject&
cppvmObjectCollection::read (unsigned int i)
{
  
  // get object from the mailbox (if the mailbox was initialized)
  cppvmReadMboxInfo();
#ifndef CPPvmFastAccess
  if (i>= num){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(PVM_COLL_MODULE,
                                                 "cppvmObjectCollection",
                                                 "read",
                                                 "index > collection length"));
  }
#endif
  return *(elems[i]);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTaskWithName
@MK  CLASS NAME    : cppvmObjectCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the cppvmTask collection member with the given name
@MU  USAGE         : cppvmTask& getTaskWithName (cppvmString taskName)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : only useful if the collection contains cppvmTask objects
@M#  --------------------------------------------------------------------------
@MP  IN      : taskName  name of the cppvmTask to get
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : cppvmTask&  description of the task with the given name
******************************************************************************/
cppvmTask&
cppvmObjectCollection::getTaskWithName(cppvmString taskName)
{
  unsigned int tmpIndex=0;
  unsigned int foundIndex=0;
  bool found = false;

  while(found == false && tmpIndex<num){
    if(elems[tmpIndex]->cppvmMsgTag() ==CPPvmMsgTagClassTask){
      // element is of type cppvmTask
      if( ((cppvmTask*)(elems[tmpIndex]))->getName() == taskName ){
        found = true;
        foundIndex = tmpIndex;
      }
    }
    tmpIndex++;
  }
  if(found == true){
    return (cppvmTask&)(*(elems[foundIndex]));
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,                   
                     "cppvmObjectCollection",                    
                     "getTaskWithName",
                     "there is no cppvmTask with the given name in the collection"));  
  }
  // dummy return
  return (cppvmTask&)(*(elems[0]));

}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

