/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_conn.cpp 
@IE  CREATED     : 5.12.1997
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm base class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// includes for predeclared classes to avoid cyclic includes
#include "pvm_task.h"
#include "pvm_arin.h"
// include own class
#include "pvm_conn.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_conn.cpp"

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmHostInfoArrayLength
@CD  DESCRIPTION   : length of the array which contains information about
@CD                  the hosts
***
******************************************************************************/
#define CPPvmHostInfoArrayLength 10

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmMboxDebugLevelKey
@CD  DESCRIPTION   : key for the debugLevel functions
***
******************************************************************************/
#define CPPvmMboxDebugLevelKey "cppvmDebugLevel"

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmMboxDebugLevelKey
@CD  DESCRIPTION   : key for the debugLevel functions
***
******************************************************************************/
#define CPPvmMboxDebugLevelLockKey "cppvmDebugLevelLock"

/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : totalNumObjs
@VD  DESCRIPTION   : number of cppvmConnection objects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmConnection::totalNumObjs = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : globalMsgEncoding
@VD  DESCRIPTION   : pvm_initsend message encoding; used by "pvm_obj.cpp'
@VT  TYPE          : int
***
******************************************************************************/
int cppvmConnection::globalMsgEncoding = CPPvmAllPvmDataLocal; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : debugLevelEntryChecked
@VD  DESCRIPTION   : indicate if the check for the debugLevel message mailbox
@VD                  entry was done or not
@VT  TYPE          : bool
***
******************************************************************************/
bool cppvmConnection::debugLevelEntryChecked = false; 

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmConnection
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmConnection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmConnection::cppvmConnection()
    : cppvmProcessHandling()
{
  totalNumObjs = totalNumObjs + 1;

  //set local debug level
  debugLevel = 0;
  
  char *argsArray[1];
  // dummy arguments
  argsArray[0] = NULL;

  // start pvm if it is not already running
  initPvmConnection(0, argsArray, 1);

  // init context
  initContext();
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmConnection
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor, start the pvm daemon with the given
@MD                  arguments if not already running
@MU  USAGE         : cppvmConnection(int argc, char **argv)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : argc   number of arguments for the daemon
@MP            argv   arguments for the daemon
@MP            block  integer specifying whether to block until startup of
@MP                   all hosts complete or return immediately
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmConnection::cppvmConnection(int argc, char **argv, int block)
    : cppvmProcessHandling()
{
  totalNumObjs = totalNumObjs + 1;
  //set local debug level
  debugLevel = 0;

  if(totalNumObjs!=1) {
    // this is not the first object
    // i.e. the daemon is already started and can not
    // be started again 
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "initPvmConnection",
                      "cannot get my tid"));
  }

  // start pvm if it is not already running
  initPvmConnection(argc, argv, block);

  // init context
  initContext();
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmConnection
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmConnection::cppvmConnection (const cppvmConnection& obj)
    : cppvmProcessHandling(obj), enrolledMicroseconds(obj.enrolledMicroseconds)
{
  totalNumObjs = totalNumObjs + 1;
  //set local debug level
  debugLevel = 0;

  myTid = obj.myTid;
  parentTid = obj.parentTid;
  contextId = obj.contextId;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmConnection
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmConnection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmConnection::~cppvmConnection()
{
  // decrease the number of cppvmConnection objects
  totalNumObjs = totalNumObjs - 1;
  
  if(totalNumObjs == 0){
    // the process is leaving PVM automatically if the last
    // cppvmConnection object is deleted
    pvm_exit();
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to assign
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmConnection&
cppvmConnection::operator= (const cppvmConnection& obj)
{
  // no self-assignment
  if(cppvmIsSame(obj)){
    return *this;
  }
  cppvmProcessHandling::operator=(obj);
  
  myTid = obj.myTid;
  enrolledMicroseconds = obj.enrolledMicroseconds;
  parentTid = obj.parentTid;
  contextId = obj.contextId;
  
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmConnection::operator== (const cppvmConnection& obj)
{
  return (myTid     == obj.myTid  &&
          parentTid == obj.parentTid &&
          contextId == obj.contextId);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmConnection::operator== (const cppvmConnection& obj) const
{
  return (myTid     == obj.myTid  &&
          parentTid == obj.parentTid &&
          contextId == obj.contextId);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::cppvmPrintOn(ostream& strm)
{
  strm << "Process TID: " << myTid << endl;
  strm << "Parent TID: ";
  if(parentTid == PvmNoParent){
    strm << "No parent process." << endl;
  }
  else{
    strm << parentTid << endl;
  }
  strm << "Context id: " << contextId << endl;
  strm << "Host information: name: " << getHostName()
       << " architecture: " << getHostArchitecture()
       << " speed: " << getHostSpeed()
       << " daemon task id: " << getHostDaemonTid() << endl;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : addHost
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add a host to the virtual pvm machine
@MU  USAGE         : void addHost(char *host)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : - For every addHost you should use delHost
@MR                    when you don't need the host any more or at
@MR                    the end of the program. Otherwise a pvmd
@MR                    is still running on the added host when
@MR                    your program has finished
@MR                  - If a pvmd is already running on the host
@MR                    and the pvmd was not started from this
@MR                    program (or from a former run of this program)
@MR                    addHost DOES NOT WORK
@MR                    Solution: halt the already running pvmd
@MR                              e.g. > rlogin host
@MR                                   > cd $PVM_ROOT/lib/ARCH
@MR                                   > pvm
@MR                                   pvm> halt
@MR                    To avoid leaving such pvmd zombies, use
@MR                    delHost in your program (see above) 
@M#  --------------------------------------------------------------------------
@MP  IN      : host  name of the host to add to the virtual machine
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::addHost(char *host)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif
  
  CPPvmInternalTry{
    addHostTry(host);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : addHostTry
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add a host to the virtual pvm machine
@MU  USAGE         : void addHostTry(char *host)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : - For every addHost you should use delHost
@MR                    when you don't need the host any more or at
@MR                    the end of the program. Otherwise a pvmd
@MR                    is still running on the added host when
@MR                    your program has finished
@MR                  - If a pvmd is already running on the host
@MR                    and the pvmd was not started from this
@MR                    program (or from a former run of this program)
@MR                    addHost DOES NOT WORK
@MR                    Solution: halt the already running pvmd
@MR                              e.g. > rlogin host
@MR                                   > cd $PVM_ROOT/lib/ARCH
@MR                                   > pvm
@MR                                   pvm> halt
@MR                    To avoid leaving such pvmd zombies, use
@MR                    delHost in your program (see above) 
@M#  --------------------------------------------------------------------------
@MP  IN      : host  name of the host to add to the virtual machine
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::addHostTry(char *host)
{
  int info;
  int infos[1];
  char errMsg[80];

  if(host == NULL){
    CPPvmInternalThrow(cppvmExceptionBadParameter(THIS_MODULE,
                                                  "cppvmConnection",
                                                  "addHost",
                                                  "host is a NULL pointer."));
  }

  if(strcmp(getHostName(), host)==0){
    // There is a race condition in PVM:
    // if you try to add the local host
    // (pvm_addhosts) to fast after starting
    // the PVM (pvm_start_pvmd) there can
    // occur an exception.
    // To avoid that, do not add a host
    // when it is equal to the local host.
    return;
  }
  
  
  // preset return value
  infos[0] = PvmOk;
  info = pvm_addhosts(&host, 1, infos);

  if (info == PvmAlready){
    // host already been added
    return;
  }
  if (infos[0] == PvmDupHost){
    // host already in configuration
    return;
  }
  if (infos[0] == PvmIPLoopback){
    // this error can also occur on Windows
    // systems when the hosts is already
    // in the configuration
    return;
  }

  if (info == PvmBadParam){
    strcpy(errMsg,"invalide argument value");
  }
  else if (info == PvmSysErr){
    strcpy(errMsg,"local pvmd is not responding");
  }
  else if (infos[0] == PvmNoHost){
    strcpy(errMsg,"no such host");
  }
  else if (infos[0] == PvmBadParam){
    strcpy(errMsg,"bad hostname syntax");
  }
  else if (infos[0] == PvmCantStart){
    strcpy(errMsg,"cannot start pvmd on host");
  }
  else if (infos[0] == PvmBadVersion){
    strcpy(errMsg,"remote pvmd version does not match");
  }
  else if (infos[0] == PvmOutOfRes){
    strcpy(errMsg,"pvm has run out of system resources");
  }
  else{
    strcpy(errMsg,"can not add host");
  }
  
  if(info != 1){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmConnection",
                                                "addHost",
                                                errMsg));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : delHost
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : delete a host from the virtual pvm machine
@MU  USAGE         : void delHost(char *host)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : host  name of the host to delete from the virtual machine
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::delHost(char *host)
{
  int info;
  int infos[1];
  char errMsg[80];
  
  if(host == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmConnection",
                                                   "delHost",
                                                   "host is a NULL pointer."));
  }
  
  // preset return value
  infos[0] = PvmOk;
  info = pvm_delhosts(&host, 1, infos);

  if (info == PvmBadParam){
    strcpy(errMsg,"invalide argument value");
  }
  else if (info == PvmSysErr){
    strcpy(errMsg,"local pvmd is not responding");
  }
  else if (infos[0] == PvmNoHost){
    strcpy(errMsg,"no such host");
  }
  else if (infos[0] == PvmBadParam){
    strcpy(errMsg,"bad hostname syntax");
  }
  else if (infos[0] == PvmCantStart){
    strcpy(errMsg,"cannot start pvmd on host");
  }
  else if (infos[0] == PvmBadVersion){
    strcpy(errMsg,"remote pvmd version does not match");
  }
  else if (infos[0] == PvmOutOfRes){
    strcpy(errMsg,"pvm has run out of system resources");
  }
  else{
    strcpy(errMsg,"could not delete a host from the virtual machine.");
  }

  if(info != 1){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmConnection",
                                                "delHost",
                                                errMsg));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : hostInfo
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : configuration of the virtual machine
@MU  USAGE         : void hostInfo(int *nhost,int *narch,
@MU                                struct pvmhostinfo **hostp)
@MC  CONSTRAINTS   : the space for the struct pvmhostinfo is allocated by
@MC                  this method
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : nhost  number of hosts on the virtual machine
@MP            narch  number of different data formats being used
@MP            pvmhostinfo  pointer to an array of structures containing
@MP                         information about each host
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::hostInfo(int *nhost,
                          int *narch,
                          struct pvmhostinfo **hostp)
{
  int info;

  info = pvm_config(nhost, narch, hostp);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                   "cppvmConnection",
                   "hostInfo",
                   "could not get configuration informations of the virtual machine."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : taskInfo
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : information about tasks running on the virtual machine
@MU  USAGE         : void taskInfo(int where,int *ntask,
@MU                                struct pvmtaskinfo **taskp);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : the space for the struct pvmtaskinfo array is given
@MR                  by this method
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : where   information specifier:
@MP                    0:        all tasks on the virtual machine
@MP                    pvmd tid: all task on this host (tid of the pvm daemon, 
@MP                              use getHostDaemonTid to get this id)
@MP                    tid:      specific task
@MP            ntask   number of task (depending on 'where')
@MP            taskp   pointer to an array of structures which contain
@MP                    information about each task
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::taskInfo(int where,
                int *ntask,
                struct pvmtaskinfo **taskp)
{
  int info;

  if(ntask == NULL ||
     taskp == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmConnection",
                                                   "taskInfo",
                                                   "a parameter is a NULL pointer."));
  }
  
  info = pvm_tasks(where, ntask, taskp);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "taskInfo",
                      "could not get task informations of the virtual machine."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : printOneTaskInfo
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print information about one tasks
@MU  USAGE         : void printOneTaskInfo(struct pvmtaskinfo& t)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::printOneTaskInfo(struct pvmtaskinfo& t)
{
  cout << "Task name: " << t.ti_a_out
       << " Status: " << t.ti_flag
       << " Task id: " << t.ti_tid
       << " Task process id: " << t.ti_pid
       << " Parent tid: ";
    if(t.ti_ptid == PvmNoParent){
      cout << " No parent";
    }
    else{
      cout << t.ti_ptid;
    }
  cout << " This tasks daemon tid: " << t.ti_host << endl;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : printOneHostInfo
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print information about one host
@MU  USAGE         : void printOneHostInfo(struct pvmhostinfo& h)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::printOneHostInfo(struct pvmhostinfo& h)
{
  cout << "Host name: " << h.hi_name
       << " Architecture: " << h.hi_arch
       << " Speed: " << h.hi_speed
       << " Host daemon tid: " << h.hi_tid << endl;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : printTaskInfo
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print information about tasks running on the
@MD                  virtual machine to cout
@MU  USAGE         : void printTaskInfo()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::printTaskInfo()
{
  int i;
  int ntask;
  struct pvmtaskinfo *taskp;
  
  taskInfo(0, &ntask, &taskp);

  cout << "Number of task running on virtual machine: " << ntask << endl;
  for(i=0;i<ntask;i++){
    printOneTaskInfo(taskp[i]);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTidOfTask
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the task id of a task running in pvm specified
@MD                  by its name
@MU  USAGE         : int getTidOfTask(char *taskName)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskName  name of the pvm task
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : int > 0: task id of the process
@MP                <=0: task does not exists in pvm
******************************************************************************/
int
cppvmConnection::getTidOfTask(char *taskName)
{
  int i;
  int ntask;
  struct pvmtaskinfo *taskp;

  if(taskName == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmConnection",
                                                   "getTidOfTask",
                                                   "taskName is a NULL pointer"));
  }

  taskInfo(0, &ntask, &taskp);

  for(i=0;i<ntask;i++){
    if(strcmp(taskp[i].ti_a_out, taskName)==0){
      return taskp[i].ti_tid;
    }
  }
  
  return -1;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : printHostInfo
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print information about the configuration of the
@MD                  virtual machine to cout
@MU  USAGE         : void printHostInfo()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::printHostInfo()
{
  int i;
  int nhost;
  int narch;
  struct pvmhostinfo *hostp;

  //hostp = new struct pvmhostinfo[CPPvmHostInfoArrayLength];
  
  hostInfo(&nhost, &narch, &hostp);

  cout << "Number of host in the virtual machine: " << nhost << endl;
  for(i=0;i<nhost;i++){
    printOneHostInfo(hostp[i]);
  }

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostDaemonTid
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the host daemon task id for a task or host name
@MU  USAGE         : int  getHostDaemonTid(char* taskOrHostName);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskOrHostName  name of a task or a host
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int pvm daemon task id
******************************************************************************/
int
cppvmConnection::getHostDaemonTid(char* taskOrHostName)
{
  int tid = getTidOfTask(taskOrHostName);
  if(tid>0){
    int dtid = pvm_tidtohost(tid);
    if(dtid == PvmSysErr){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmConnection",
                                                  "getHostDaemonTid",
                                                  "pvmd is not responding."));
    }
    else if(dtid>0){
      return dtid;
    }
  }
  
  int i=0;
  int nhost;
  int narch;
  struct pvmhostinfo *hostp;
  bool found = false;
  int position=0;
  
  hostInfo(&nhost, &narch, &hostp);
  
  while(i<nhost && found==false){
    if(strcmp(hostp[i].hi_name,taskOrHostName)==0){
      found = true;
      position = i;
    }
    i++;
  }
  
  if(found==true){
    return (hostp[position].hi_tid);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                              "cppvmConnection",
                              "getHostDaemonTid",
                              "no pvm daemon tid found for the given task/host name."));
    return -1;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostDaemonTid
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the tid of the daemon of a task specified by a task
@MD                  object
@MU  USAGE         : int  getHostDaemonTid(cppvmTask& task);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  id of the daemon
******************************************************************************/
int
cppvmConnection::getHostDaemonTid(cppvmTask& task)
{
  return getHostDaemonTid(task.getTid());
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostDaemonTid
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the tid of the daemon of a task specified by its tid
@MU  USAGE         : int  getHostDaemonTid(int tid);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid     task id
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  id of the daemon
******************************************************************************/
int
cppvmConnection::getHostDaemonTid(int tid)
{
  int dtid = pvm_tidtohost(tid);
  if(dtid < 0){
    if(dtid == PvmBadParam){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                             "cppvmConnection",
                                             "getHostDaemonTid",
                                             "an invalid tid value was specified."));
    }
    else if(dtid == PvmSysErr){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmConnection",
                                                  "getHostDaemonTid",
                                                  "pvmd is not responding."));
    }
    else{
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmConnection",
                                                  "mkbuf",
                                                  "cannot get host daemon tid."));
    }
  }
  return dtid;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostName
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the host name of a task specified by a task object
@MU  USAGE         : char*  getHostName(cppvmTask& task);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : char* name of the host, where the task is running
******************************************************************************/
char*
cppvmConnection::getHostName(cppvmTask& task)
{
  return getHostName(task.getTid());
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostName
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the host name of a task specified by its tid
@MU  USAGE         : char*  getHostName(int tid);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid  task id
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : char* name of the host, where the task is running
******************************************************************************/
char*
cppvmConnection::getHostName(int tid)
{
  int dtid = getHostDaemonTid(tid);
  int i=0;
  int nhost;
  int narch;
  struct pvmhostinfo *hostp;
  bool found = false;
  int position=0;

  hostInfo(&nhost, &narch, &hostp);

  while(i<nhost && found==false){
    if(hostp[i].hi_tid==dtid){
      found = true;
      position = i;
    }
    i++;
  }
  
  if(found==true){
    return (hostp[position].hi_name);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                        "cppvmConnection",
                                        "getHostName",
                                        "no host name for the given tid found."));
    return NULL;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostArchitecture
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the host architecture of a task specified by a
@MD                  task object
@MU  USAGE         : char*  getHostArchitecture(cppvmTask& task);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : char* architecture of the host, where the task is running
******************************************************************************/
char*
cppvmConnection::getHostArchitecture(cppvmTask& task)
{
  return getHostArchitecture(task.getTid());
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostArchitecture
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the host architecture of a task specified by its tid
@MU  USAGE         : char*  getHostArchitecture(int tid);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid  task id
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : char* architecture of the host, where the task is running
******************************************************************************/
char*
cppvmConnection::getHostArchitecture(int tid)
{
  int dtid = getHostDaemonTid(tid);
  int i=0;
  int nhost;
  int narch;
  struct pvmhostinfo *hostp;
  bool found = false;
  int position=0;

  hostInfo(&nhost, &narch, &hostp);

  while(i<nhost && found==false){
    if(hostp[i].hi_tid==dtid){
      found = true;
      position = i;
    }
    i++;
  }
  
  if(found==true){
    return (hostp[position].hi_arch);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                   "cppvmConnection",
                                   "getHostArchitecture",
                                   "no host architecture for the given tid found."));
    return NULL;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostSpeed
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the host speed of a task specified by a task object
@MU  USAGE         : int  getHostSpeed(cppvmTask& task);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task object
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  speed of the host, where the task is running
******************************************************************************/
int
cppvmConnection::getHostSpeed(cppvmTask& task)
{
  return getHostSpeed(task.getTid());
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostSpeed
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the host speed of a task specified by its tid
@MU  USAGE         : int  getHostSpeed(int tid);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid  task id
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  speed of the host, where the task is running
******************************************************************************/
int
cppvmConnection::getHostSpeed(int tid)
{
  int dtid = getHostDaemonTid(tid);
  int i=0;
  int nhost;
  int narch;
  struct pvmhostinfo *hostp;
  bool found = false;
  int position=0;

  hostInfo(&nhost, &narch, &hostp);

  while(i<nhost && found==false){
    if(hostp[i].hi_tid==dtid){
      found = true;
      position = i;
    }
    i++;
  }
  
  if(found==true){
    return (hostp[position].hi_speed);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                        "cppvmConnection",
                                        "getHostSpeed",
                                        "no host speed for the given tid found."));
    return -1;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getAllTaskIds
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get all task ids of the processes running on the specified
@MD                  host
@MU  USAGE         : cppvmArrayInt getAllTaskIds(char* hostname);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : hostname  host name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : cppvmArrayInt array of task ids
******************************************************************************/
cppvmArrayInt
cppvmConnection::getAllTaskIds(char* hostname)
{
  int dtid = getHostDaemonTid(hostname);
  
  return getAllTaskIds(dtid);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getAllTaskIds
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get all task ids of the processes running on the specified
@MD                  host
@MU  USAGE         : cppvmArrayInt getAllTaskIds(int pvmDaemonTaskId)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : pvmDaemonTaskId  task id of the pvm daemon
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : cppvmArrayInt array of task ids
******************************************************************************/
cppvmArrayInt
cppvmConnection::getAllTaskIds(int pvmDaemonTaskId)
{
  cppvmArrayInt tids;

  int dtid = pvmDaemonTaskId;
  
  int i;
  int j=0;
  int ntask;
  struct pvmtaskinfo *taskp;
  
  taskInfo(0, &ntask, &taskp);

  for(i=0;i<ntask;i++){
    if(taskp[i].ti_host==dtid){
      tids.resize(tids.numElems()+1);
      tids[j]=taskp[i].ti_tid;
      j++;
    }
  }
  return tids;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setOpt
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set various libpvm option
@MD                  for detail see pvm manual function pvm_setopt
@MU  USAGE         : void setOpt(int what, int val)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : what   option what is being set
@MP                   additional for CPPvm:
@MP                   CPPvmMsgEncoding : set global or local message encoding
@MP            val    new value of the option
@MP                   additional for CPPvm:
@MP                   CPPvmMsgEncoding : 
@MP                     CPPvmAllPvmDataLocal: use local send stream settings
@MP                     CPPvmAllPvmDataDefault: use PvmDataDefault for all
@MP                                   send operations (ignore local settings)
@MP                     CPPvmAllPvmDataInPlace: use PvmDataInPlace for all
@MP                                   send operations (ignore local settings)
@MP                     CPPvmAllPvmDataRaw: use PvmDataRaw for all
@MP                                   send operations (ignore local settings)
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::setOpt(int what, int val)
{
  // check for CPPvm setting
  if(what==CPPvmMsgEncoding){
    if(val!=CPPvmAllPvmDataLocal   && val!=CPPvmAllPvmDataDefault &&
       val!=CPPvmAllPvmDataInPlace && val!=CPPvmAllPvmDataRaw){
      CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                "cppvmConnection",
                                "setOpt",
                                "unknown 'val' for 'what'=CPPvmMsgEncoding"));
    }
    globalMsgEncoding = val;
    return;
  }
  
  // PVM settings
  int info;
  info = pvm_setopt(what, val);
  
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmConnection",
                                                "setOpt",
                                                "could not set option"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : mkbuf
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : allocate a new message buffer
@MU  USAGE         : int mkbuf(int encoding);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : encoding   PvmDataDefault: XDR encoding
@MP                       PvmDataRaw:     no encoding
@MP                                       only for homogenous architectures
@MP                       PvmDataInPlace: data left in place during packing;
@MP                                       not allowd for shared memory arch.
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : bufid      new message buffer identifier
******************************************************************************/
int
cppvmConnection::mkbuf(int encoding)
{
  int bufid;

  bufid = pvm_mkbuf(encoding);
  if(bufid < 0){
    if(bufid == PvmBadParam){
	    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                         "cppvmConnection",
                         "mkbuf",
                         "giving an invalid encoding value"));
    }
    else if(bufid == PvmNoMem){	
	    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                         "cppvmConnection",
                         "mkbuf",
                         "not enough memory to create the buffer"));
    }
    else{
	    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                         "cppvmConnection",
                         "mkbuf",
                         "cannot create new buffer"));
    }
  }
  return bufid;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getMsgBufInfo
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get informations about the last message
@MU  USAGE         : void getMsgBufInfo(int bufid, int *bytes,
@MU                                     int *msgtag, int *tid);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : bufid  id of the message buffer to get informations from
@MP  INOUT   : 
@MP  OUT     : bytes  length of the message in bytes
@MP            msgtag message label
@MP            tid    source of the message
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::getMsgBufInfo(int bufid,
                     int *bytes,
                     int *msgtag,
                     int *tid)
{
  int info;
  
  if(bytes  == NULL  ||
     msgtag == NULL  ||
     tid    == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                         "cppvmConnection",
                         "getMsgBufInfo(int bufid,...",
                         "one of the parameter is a NULL pointer."));
  }
  
  info = pvm_bufinfo(bufid, bytes, msgtag, tid);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "getMsgBufInfo(int bufid,...",
                      "could not get buffer informations."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getMsgIsObjOfClass
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : gives the message tag of the current received
@MD                  message buffer; this message tag designes also
@MD                  the class of the incomming object
@MD                  if the return value is < 0; there is no current
@MD                  message
@MU  USAGE         : int  getMsgIsObjOfClass(int classMsgTag);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : int >0: message tag of the current message (==class description)
@MP                <0: no current message
******************************************************************************/
int
cppvmConnection::getMsgIsObjOfClass()
{
  int bufId;
  int bytes;
  int msgTag;
  int tid;

  bufId = probe(-1,-1);
  if(bufId>0){
    getMsgBufInfo(bufId, &bytes, &msgTag,&tid);
    return msgTag;
  }
  else{
    return -1;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : msgIsObjOfClass
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : checks, weather the current message has the given
@MD                  message tag; this message tag designes also
@MD                  the class of the incomming object so you can check
@MD                  which object is in the message buffer!
@MD                  if the message buffer is empty, false is returned
@MU  USAGE         : bool msgIsObjOfClass(int classMsgTag);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : bool true:  object of the given type is in message buffer
@MP                 false: there are other data in the message buffer
@MP                        or the message buffer is empty
******************************************************************************/
bool
cppvmConnection::msgIsObjOfClass(int classMsgTag)
{
  return (classMsgTag == getMsgIsObjOfClass());
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : msgIsObjOfClass
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : checks, weather the current message has the given
@MD                  message tag; this message tag designes also
@MD                  the class of the incomming object so you can check
@MD                  which object is in the message buffer!
@MD                  if the message buffer is empty, false is returned
@MU  USAGE         : bool msgIsObjOfClass(cppvmObject& obj);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : cppvmObject   object, for which should be determinated, wheather
@MP                        there is a message for it or not
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : bool true:  object of the given type is in message buffer
@MP                 false: there are other data in the message buffer
@MP                        or the message buffer is empty
******************************************************************************/
bool
cppvmConnection::msgIsObjOfClass(cppvmObject& obj)
{
  return (obj.cppvmMsgTag() == getMsgIsObjOfClass());
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getMyTask
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the description of the current task as cppvmTask
@MD                  object
@MU  USAGE         : getMyTask
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : cppvmTask task description
******************************************************************************/
cppvmTask
cppvmConnection::getMyTask()
{
  return cppvmTask(myTid, contextId, "");
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getParentTask
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the description of the parent task as cppvmTask
@MD                  object
@MU  USAGE         : getParentTask();
@MC  CONSTRAINTS   : if there is no parent task, an exception is thrown
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : cppvmTask task description
******************************************************************************/
cppvmTask
cppvmConnection::getParentTask()
{
  if(parentTid == PvmNoParent){
    CPPvmInternalThrow( cppvmExceptionForbiddenCall(THIS_MODULE,
                                "cppvmConnection",
                                "getParentTask",
                                "no parent task"));
  }
  return cppvmTask(parentTid, contextId, "");
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTaskStat
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get status of a specified PVM process
@MU  USAGE         : int getTaskStat(int tid);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid  task id of the process to get status from
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : int  PvmOk:       task is running
@MP                 PvmBadParam: bad parameter (most likely, an invalid
@MP                              tid value)
@MP                 PvmSysErr:   pvmd not responding
@MP                 PvmNoTask:   task not running
******************************************************************************/
int
cppvmConnection::getTaskStat(int tid)
{
  int result;
  int oldval;
  
  // suppress PVM error messages
  // (e.g. "libpvm [t40001]: pvm_pstat(): Bad parameter") 
  oldval = pvm_setopt(PvmAutoErr, 0);

  result=pvm_pstat(tid);

  // allow PVM error messages again
  pvm_setopt(PvmAutoErr, oldval);

  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTaskStat
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get status of a process
@MU  USAGE         : int getTaskStat(cppvmTask& task);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task description of the process to get status from
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : int  PvmOk:       task is running
@MP                 PvmBadParam: bad parameter (most likely, an invalid
@MP                              tid value)
@MP                 PvmSysErr:   pvmd not responding
@MP                 PvmNoTask:   task not running
******************************************************************************/
int
cppvmConnection::getTaskStat(cppvmTask& task)
{
  return getTaskStat(task.getTid());
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostStat
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get status of a specified host
@MU  USAGE         : int getHostStat(char *host);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : host  name of the host to get status from
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : int  PvmOk:       host is ok
@MP                 PvmNoHost:   host is not in virtual machine
@MP                 PvmHostFail: host is unreachable (and thus possibly failed)
******************************************************************************/
int
cppvmConnection::getHostStat(char *host)
{
  return pvm_mstat(host);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : initPvmConnection
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : initialize the class cppvmConnection
@MU  USAGE         : initPvmConnection(int argc, char **argv, int block);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : argc   number of arguments for the daemon
@MP            argv   arguments for the daemon
@MP            block  integer specifying whether to block until startup of
@MP                   all hosts complete or return immediately
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::initPvmConnection(int argc, char **argv, int block)
{
  // start pvm if it is not already running
  if(totalNumObjs==1) {
    bool result;
  
    // start daemon only for the first pvm object
    result=startPvmDaemon(argc, argv, block);

    // tried to start a daemon with arguments
    // but the daemon is already running
    if(argc>0 && result==false){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
     "cppvmConnection",
     "initPvmConnection",
     "tried to start the PVM daemon with arguments, but the daemon is already running"));
    }
  }

  myTid = pvm_mytid();
  if(myTid < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "initPvmConnection",
                      "cannot get my tid"));
  }

  enrolledMicroseconds = clock();
  if(enrolledMicroseconds == (unsigned long int) ((clock_t) -1)){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "initPvmConnection()",
                      "can get no value from function clock()"));
    

  }
  parentTid = pvm_parent();

  // enable direct route communication between PVM tasks
  // for faster (about two times faster!)
  // sometimes :-) stillmiss
  // pvm_setopt(PvmRoute, PvmRouteDirect);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : barrier
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : blocks the calling process until all processes in a
@MD                  group have called it
@MU  USAGE         : void barrier(char *group, int count);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group  group name; the group must exist and the calling
@MP                   process must be a member of the group
@MP            count  number of group members that must call barrier before
@MP                   they are released. Though not required, count is
@MP                   expected to be the total number of members of the
@MP                   specified group
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::barrier(char *group, int count)
{
  int info;

  info = pvm_barrier(group, count);

  if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "barrier",
                      "local pvmd is not responding"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "barrier",
                      "count is < 1"));
  }
  else if(info == PvmNoGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "barrier",
                      "a non-existent group name was given"));
  }
  else if(info == PvmNotInGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "barrier",
                      "calling process is not in specified group"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : gather
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : a specified member of the group receives messages
@MD                  from each member of the group and gathers these
@MD                  messages into a single array
@MU  USAGE         : void gather(void *result, void *data, int count,
@MU                              int datatype, int msgtag, char *group,
@MU                              int rootginst);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : result   on the root this is a pointer to the starting
@MP                     address of an array datatype of local values
@MP                     which are to be accumulated from the members of
@MP                     the group; if n is the number of members in the
@MP                     group, then this array of datatype should be of 
@MP                     length at least n * count
@MP            data     pointer to the starting address of an array of local
@MP                     datatype values to be send by each member of the
@MP                     group to the root
@MP            count    integer specifying the number of elements of datatype
@MP                     in the data array
@MP            datatype type of the entries in the data array
@MP            msgtag   message tag for the gather operation
@MP            group    group name; the group must exist and the calling
@MP                     process must be a member of the group
@MP            rootginst instance number of group member who performs the
@MP                      gather of the messages from the member of the group
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::gather(void *result, void *data, int count, int datatype,
             int msgtag, char *group, int rootginst)
{
  int info;
  info =  pvm_gather(result, data, count, datatype, msgtag, group, rootginst);

  if(info == PvmNoInst){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "gather",
                      "calling task is not in the group"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "gather",
                      "the datatype specified is not appropriate"));
  }
  else if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "gather",
                      "pvm system error"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : scatter
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : send to each member of a group a section of an
@MD                  array from a specified member of the group
@MU  USAGE         : void scatter(void *result, void *data, int count,
@MU                               int datatype, int msgtag, char *group,
@MU                               int rootginst);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : result   pointer to the starting address of an array datatype
@MP                     which will be overwritten by the message from the
@MP                     specified root member of the group
@MP            data     on the root this is a pointer to the starting address
@MP                     of an array datatype of local values which are to be
@MP                     distributed to the members of the group
@MP                     if n is the number of members in the
@MP                     group, then this array of datatype should be of 
@MP                     length at least n * count
@MP            count    integer specifying the number of elements of datatype
@MP                     to be sent to each member of the group from the root
@MP            datatype type of the entries in the data array
@MP            msgtag   message tag for the scatter operation
@MP            group    group name; the group must exist and the calling
@MP                     process must be a member of the group
@MP            rootginst instance number of group member who performs the
@MP                      scatter of its array to the members of the group
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::scatter(void *result, void *data, int count, int datatype,
               int msgtag, char *group, int rootginst)
{
  int info;

  info =  pvm_scatter(result, data, count, datatype, msgtag, group, rootginst);

  if(info == PvmNoInst){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "scatter",
                      "calling task is not in the group"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "scatter",
                      "the datatype specified is not appropriate"));
  }
  else if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "scatter",
                      "pvm system error"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : reduce
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : performs a reduction operation over members
@MD                  of the specified group
@MU  USAGE         : void reduce(void (*func)(), void *data, int count, 
@MU                              int datatype, int msgtag, char *group,
@MU                              int rootginst);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : func      function which defines the operation performed
@MP                      on the global data
@MP            count     integer specifying the number of elements of datatype
@MP                      in the data array
@MP            datatype  type of the entries in the data array
@MP            msgtag    message tag for the reduce operation
@MP            group     group name; the group must exist and the calling
@MP                      process must be a member of the group
@MP            rootginst instance number of group member who gets the result
@MP  INOUT   : data  pointer to the starting address of an array of local
@MP                  values. On return, the data array on the root will
@MP                  be overwritten with the result of the reduce operation.
@MP                  For the other members of the group the values after
@MP                  the operation are not defined
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::reduce(void (*func)(int*, void*, void*, int*, int*),
              void *data, int count, int datatype,
              int msgtag, char *group, int rootginst)
{
  int info;

  info =  pvm_reduce(func, data, count, datatype, msgtag, group, rootginst);

  if(info == PvmNoInst){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "reduce",
                      "calling task is not in the group"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "reduce",
                      "the datatype specified is not appropriate"));
  }
  else if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "reduce",
                      "pvm system error"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : joinGroup
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : enrolls the calling process in a named group
@MU  USAGE         : int joinGroup(char* group);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group  group name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  instance number; the number starts at 0 and count up
@MP                 < 0 indicates an error
******************************************************************************/
int
cppvmConnection::joinGroup(char *group)
{
  int result=-1;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    result = joinGroupTry(group);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
    // dummy return
    return result;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : joinGroupTry
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : enrolls the calling process in a named group
@MU  USAGE         : int joinGroupTry(char* group);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group  group name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  instance number; the number starts at 0 and count up
@MP                 < 0 indicates an error
******************************************************************************/
int
cppvmConnection::joinGroupTry(char *group)
{
  int info;

  info =  pvm_joingroup(group);

  if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmConnection",
                                                "joinGroup",
                                                "giving a NULL group name"));
  }
  else if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "joinGroup",
                      "pvmd was not started or has crashed"));
  }
  return info;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : leaveGroup
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unenrolls the calling process from a named group
@MU  USAGE         : void leaveGroup(char *group);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group  group name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::leaveGroup(char *group)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    leaveGroupTry(group);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : leaveGroupTry
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unenrolls the calling process from a named group
@MU  USAGE         : void leaveGroupTry(char *group);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group  group name
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::leaveGroupTry(char *group)
{
  int info;

  info =  pvm_lvgroup(group);

  if(info == PvmNoGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "leaveGroup",
                      "giving a non existent group name"));
  }
  else if(info == PvmNotInGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "leaveGroup",
                      "asking to leave a group you are not a member of"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "leaveGroup",
                      "giving a NULL group name"));
  }
  else if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "leaveGroup",
                      "pvmd was not started or has crashed"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : freezeGroup
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : freezes dynamical group membership and caches info locally
@MU  USAGE         : void freezeGroup(char* group, int size);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group  group name
@MP            size   size of the group when it is frozen
@MP                   -1 indicates that the current size of the group should
@MP                   be used
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::freezeGroup(char * group, int size)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif
  
  CPPvmInternalTry{
    freezeGroupTry(group, size);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : freezeGroupTry
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : freezes dynamical group membership and caches info locally
@MU  USAGE         : void freezeGroupTry(char* group, int size);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group  group name
@MP            size   size of the group when it is frozen
@MP                   -1 indicates that the current size of the group should
@MP                   be used
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::freezeGroupTry(char * group, int size)
{
  int info=0;

  info =  pvm_freezegroup(group, size);

  if(info == PvmDupGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "freezeGroup",
                      "trying to freeze a group that is already frozen"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "freezeGroup",
                      "giving a NULL group name"));
  }
  else if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "freezeGroup",
                      "pvmd was not started or has crashed"));
  }
  else if(info == PvmNotInGroup){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "freezeGroup",
                      "trying to freeze a group that you are not in"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : pkMesg
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : packs the specified message into the current send buffer
@MU  USAGE         : pkMesg(int bufid);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : bufid  message buffer to pack
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::pkMesg(int /*bufid*/)
{
  int info=0;

  // stillmiss
  //info =  pvm_pkmesg(bufid);

  if(info == PvmNoBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesg",
                      "no current send buffer"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesg",
                      "invalid value for bufid or bufid is the current send buffer"));
  }
  else if(info == PvmNoSuchBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesg",
                      "message buffer bufid does not exist"));
  }
  else if(info == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesg",
                      "libpvm is unable to allocate memory to pack data"));
  }
  else if(info == PvmNoData){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesg",
                      "end of message reached, perhaps pack and unpack calls were mismatched"));
  }
  else if(info == PvmMismatch){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesg",
                      "pvm_pkmegbody was called and the encoding types of bufid and the current send buffer do not match"));
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : pkMesgBody
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : similar to pkMesg() but packs only the body of the
@MD                  specified message omitting the header fields
@MU  USAGE         : pkMesgBody(int bufid);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : bufid  message buffer to pack
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::pkMesgBody(int /*bufid*/)
{
  int info=0;

  // stillmiss
  //info =  pvm_pkmesgbody(bufid);

  if(info == PvmNoBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesgBody",
                      "no current send buffer"));
  }
  else if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesgBody",
                      "invalid value for bufid or bufid is the current send buffer"));
  }
  else if(info == PvmNoSuchBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesgBody",
                      "message buffer bufid does not exist"));
  }
  else if(info == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesgBody",
                      "libpvm is unable to allocate memory to pack data"));
  }
  else if(info == PvmNoData){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesgBody",
                      "end of message reached, perhaps pack and unpack calls were mismatched"));
  }
  else if(info == PvmMismatch){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "pkMesgBody",
                      "pvm_pkmegbody was called and the encoding types of bufid and the current send buffer do not match"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : upkMesg
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unpacks a message from the current receive buffer
@MU  USAGE         : int upkMesg();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int new message buffer unpacked or result code
******************************************************************************/
int
cppvmConnection::upkMesg()
{
  int newbufid=0;
  
  //stillmiss
  //newbufid = pvm_upkmesg();

  if(newbufid == PvmNoBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "upkMesg",
                      "no current send buffer"));
  }
  else if(newbufid == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "upkMesg",
                      "invalid value for bufid or bufid is the current send buffer"));
  }
  else if(newbufid == PvmNoSuchBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "upkMesg",
                      "message buffer bufid does not exist"));
  }
  else if(newbufid == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "upkMesg",
                      "libpvm is unable to allocate memory to pack data"));
  }
  else if(newbufid == PvmNoData){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "upkMesg",
                      "end of message reached, perhaps pack and unpack calls were mismatched"));
  }
  else if(newbufid == PvmMismatch){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "upkMesg",
                      "pvm_pkmegbody was called and the encoding types of bufid and the current send buffer do not match"));
  }
  return newbufid;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTMask
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get trace mask of a task or its children
@MU  USAGE         : getTMask(int who, Pvmtmask mask);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : who  specifies which trace mask to get
@MP                 PvmTaskSelf  : the current task
@MP                 PvmTaskChild : children tasks to be spawned
@MP            mask trace mask which is filled in with the current trace 
@MP                 vector on calling pvm_gettmask or used to set the current
@MP                 trace vector on calling pvm_settmask
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::getTMask(int who, Pvmtmask mask)
{
  int info;
  
  info = pvm_gettmask(who, mask);

  if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "getTMask",
                      "invalid value for argument 'who'"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setTMask
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set trace mask of a task or its children
@MU  USAGE         : setTMask(int who, Pvmtmask mask);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : who  specifies which trace mask to set
@MP                 PvmTaskSelf  : the current task
@MP                 PvmTaskChild : children tasks to be spawned
@MP            mask trace mask which is filled in with the current trace 
@MP                 vector on calling pvm_gettmask or used to set the current
@MP                 trace vector on calling pvm_settmask
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::setTMask(int who, Pvmtmask mask)
{
  int info;
  
  info = pvm_settmask(who, mask);

  if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmConnection",
                      "setTMask",
                      "invalid value for argument 'who'"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : sendDebugToMsgBox
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : send the value in the variable 'debugLevel' to the
@MD                  message box
@MU  USAGE         : sendDebugToMsgBox();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::sendDebugToMsgBox()
{
  int info;
  int oldSBuf;
  int newSBuf;
  
  oldSBuf = pvm_getsbuf();
  newSBuf = pvm_mkbuf(PvmDataDefault);
  if(newSBuf < 0){
    if(newSBuf == PvmBadParam){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                  "cppvmConnection",
                                  "sendDebugToMsgBox",
                                  "giving an invalid encoding value"));
    }
    else if(newSBuf == PvmNoMem){	
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                  "cppvmConnection",
                                  "sendDebugToMsgBox",
                                  "not enough memory to create the buffer"));
    }
    else{
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                  "cppvmConnection",
                                  "sendDebugToMsgBox",
                                  "cannot create new buffer"));
    }
  }

  
  pvm_setsbuf(newSBuf);

  // pack debugLevel
  pvm_pkint(&debugLevel,1,1);

  info = pvm_putinfo(CPPvmMboxDebugLevelKey, 
                     pvm_getsbuf(),
	             PvmMboxDefault + PvmMboxPersistent + PvmMboxOverWritable);
  if(info == PvmBadParam ||
     info == PvmNoSuchBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "sendDebugToMsgBox",
                       "invalid or non existant send buffer. Maybe you have set a new message buffer with 'cppvmSetMboxMsgBufSend(buffer)' and deleted 'buffer' afterwards."));
  }
  else if(info == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "sendDebugToMsgBox",
                       "unable to allocate memory to pack data"));
  }
  else if(info == PvmExists){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "sendDebugToMsgBox",
                       "the requested key already exists"));
  }
  else if(info == PvmDenied){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "sendDebugToMsgBox",
                       "the key is locked by another task and cannot be replaced or deleted"));
  }
  else if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "sendDebugToMsgBox",
                       "error writing in database"));
  }

  // reset values
  pvm_setsbuf(oldSBuf);
  pvm_freebuf(newSBuf);

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : recDebugFromMsgBox
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : receive the value of the variable 'debugLevel' from the
@MD                  message box
@MU  USAGE         : recDebugFromMsgBox();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool =true:  read was possible
@MP                 =false: read was not possible (no such object in database)
******************************************************************************/
bool
cppvmConnection::recDebugFromMsgBox()
{

  int bufid;

  bufid = pvm_recvinfo(CPPvmMboxDebugLevelKey, 
                       0,
	               PvmMboxDefault);
  
  if(bufid == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "recDebugFromMsgBox",
                       "unable to allocate memory to pack data"));
    return false;
  }
  else if(bufid == PvmDenied){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "recDebugFromMsgBox",
                       "the key is locked by another task and cannot be replaced or deleted"));
    return false;
  }
  else if(bufid == PvmNotFound){
    // object does not exist in database!
    return false;
  }
  else if(bufid < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "recDebugFromMsgBox",
                       "error reading from database"));
    return false;
  }

  // receive was successful
  // unpack debugLevel
  pvm_upkint(&debugLevel, 1, 1);
  
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setDebugLevelLock
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : try to set the debugLevel lock in the message mailbox 
@MU  USAGE         : bool setDebugLevelLock()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : - 
@MP  RETURNS : bool true:  new debugLevelLock was set
@MP                 false: debugLevelLock already exists, no new lock set
******************************************************************************/
bool
cppvmConnection::setDebugLevelLock()
{

  int info;
  int oldSBuf;
  int newSBuf;
  
  oldSBuf = pvm_getsbuf();
  newSBuf = pvm_mkbuf(PvmDataDefault);
  if(newSBuf < 0){
    if(newSBuf == PvmBadParam){
     CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                 "cppvmConnection",
                                                 "setDebugLevelLock",
                                                 "giving an invalid encoding value"));
     return false;
    }
    else if(newSBuf == PvmNoMem){	
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                            "cppvmConnection",
                                            "setDebugLevelLock",
                                            "not enough memory to create the buffer"));
     return false;
    }
    else{
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmConnection",
                                                  "setDebugLevelLock",
                                                  "cannot create new buffer"));
     return false;
    }
  }

  
  pvm_setsbuf(newSBuf);

  // pack debugLevelLock
  pvm_pkint(&debugLevelLock,1,1);

  info = pvm_putinfo(CPPvmMboxDebugLevelLockKey,
                     pvm_getsbuf(),
	             PvmMboxDefault + PvmMboxPersistent);

  // reset values
  pvm_setsbuf(oldSBuf);
  pvm_freebuf(newSBuf);

  if(info == PvmExists || info == PvmDenied){
    // message mailbox entry exists
    return false;
  }
  else if(info == PvmBadParam ||
     info == PvmNoSuchBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "setDebugLevelLock",
                       "invalid or non existant send buffer. Maybe you have set a new message buffer with 'cppvmSetMboxMsgBufSend(buffer)' and deleted 'buffer' afterwards."));
    return false;
  }
  else if(info == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                            "cppvmConnection",   
                                            "setDebugLevelLock",
                                            "unable to allocate memory to pack data"));
    return false;
  }
  else if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmConnection",   
                                                "setDebugLevelLock",
                                                "error writing in database"));
    return false;
  }

  // new lock was set
  return true;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setDebugLevel
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set debug level in global message box
@MU  USAGE         : setDebugLevel();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : debugLevel_ new global debug level
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::setDebugLevel(int debugLevel_)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif
  
  CPPvmInternalTry{
    setDebugLevelTry(debugLevel_);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setDebugLevelTry
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set debug level in global message box
@MU  USAGE         : setDebugLevelTry();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : debugLevel_ new global debug level
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::setDebugLevelTry(int debugLevel_)
{
  debugLevel = debugLevel_;
  if(debugLevelEntryChecked==false){
    // set debugLevelLock, if not already set
    setDebugLevelLock();
    debugLevelEntryChecked=true;
  }
  // set new debugLevel in message mailbox
  sendDebugToMsgBox();
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getDebugLevel
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get global debug level from message box
@MU  USAGE         : int getDebugLevel();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  current debug level in message box
******************************************************************************/
int
cppvmConnection::getDebugLevel()
{
  int levelRes;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif
  
  CPPvmInternalTry{
    levelRes = getDebugLevelTry();
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
    return 0;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  
  return levelRes;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getDebugLevelTry
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get global debug level from message box
@MU  USAGE         : int getDebugLevelTry();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  current debug level in message box
******************************************************************************/
int
cppvmConnection::getDebugLevelTry()
{
  if(debugLevelEntryChecked==false){
    // set debugLevelLock, if not already set
    if(setDebugLevelLock()==true){
      // there was no debugLevel object in the database until now
      sendDebugToMsgBox();
      // well, now there is one
    }
    debugLevelEntryChecked=true;
  }

  bool result=false;
  while(result==false){
    // wait until the process who set
    // the debugLevelLock has set
    // the debugLevel value in the
    // message mailbox
    result = recDebugFromMsgBox();
  }
  return debugLevel;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmGetMboxInfo
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get information about database entries
@MU  USAGE         : cppvmGetMboxInfo(char *pattern,
@MU                                 int *nclasses,
@MU                                 struct pvmmboxinfo **classes);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 'nclasses' and 'classes' are allocated by this method
@MR                  and freed on the next call of this method or the
@MR                  call of the other method 'cppvmGetMboxInfo'
@MR                  struct pvmmboxinfo{
@MR                    char *mi_name;      // class name
@MR                    int   mi_nentries;  // # of entries for this class
@MR                    int  *mi_indices;   // mbox entry indices
@MR                    int  *mi_owners;    // mbox entry owner tids
@MR                    int  *mi_flags;     // mbox entry flags
@MR                  }  
@M#  --------------------------------------------------------------------------
@MP  IN      : pattern  GNU regular expression to match on names in mailbox
@MP                     database. Additionally, the singular '*' will match
@MP                     all names
@MP  INOUT   : 
@MP  OUT     : nclasses number of classes matching pattern
@MP            classes  array of pvmmboxinfo mailbox entries matching pattern
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::cppvmGetMboxInfo(char *pattern,
		          int *nclasses,
		          struct pvmmboxinfo **classes)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    cppvmGetMboxInfoTry(pattern, nclasses, classes);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmGetMboxInfoTry
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get information about database entries
@MU  USAGE         : cppvmGetMboxInfoTry(char *pattern,
@MU                                 int *nclasses,
@MU                                 struct pvmmboxinfo **classes);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 'nclasses' and 'classes' are allocated by this method
@MR                  and freed on the next call of this method or the
@MR                  call of the other method 'cppvmGetMboxInfo'
@MR                  struct pvmmboxinfo{
@MR                    char *mi_name;      // class name
@MR                    int   mi_nentries;  // # of entries for this class
@MR                    int  *mi_indices;   // mbox entry indices
@MR                    int  *mi_owners;    // mbox entry owner tids
@MR                    int  *mi_flags;     // mbox entry flags
@MR                  }  
@M#  --------------------------------------------------------------------------
@MP  IN      : pattern  GNU regular expression to match on names in mailbox
@MP                     database. Additionally, the singular '*' will match
@MP                     all names
@MP  INOUT   : 
@MP  OUT     : nclasses number of classes matching pattern
@MP            classes  array of pvmmboxinfo mailbox entries matching pattern
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::cppvmGetMboxInfoTry(char *pattern,
                                     int *nclasses,
                                     struct pvmmboxinfo **classes)
{
  int info;
  info = pvm_getmboxinfo(pattern, nclasses, classes);
  if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "cppvmGetMboxInfoTry",
                       "invalid argument"));
  }    
  else if(info == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmConnection",   
                       "cppvmGetMboxInfoTry",
                       "unable to allocate memory"));
  }    
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : halt
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : shut down the entire \pvm{} system
@MU  USAGE         : void halt()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::halt()
{
  int info;
  info = pvm_halt();
  
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmConnection",
                                                "halt",
                                                "local pvmd is not responding"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : exit
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : disconnect process from \pvm{}
@MU  USAGE         : void exit()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmConnection::exit()
{
  int info;
  info = pvm_exit();
  
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmConnection",
                                                "exit",
                                                "local pvmd is not responding"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : sendsig
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : sends a signal to another PVM process
@MU  USAGE         : int sendsig(int tid, int signum);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid    integer task identifier of PVM process to receive 
@MP                   the signal
@MP            signum integer signal number
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : int ==0: sendsig successful
@MP                <0 : error
******************************************************************************/
int
cppvmConnection::sendsig(int tid, int signum)
{
  int info;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  info = pvm_sendsig(tid, signum);

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return info;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probe
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message of a certain tag has arrived
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MU  USAGE         : bool probe(int msgtag)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag message tag (for example cppvmFloat::cppvmClassMsgTag())
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : bool  true: message with this message tag has arrived
@MP                  false: no message with this message tag has arrived
******************************************************************************/
int
cppvmConnection::probe(int tid, int msgTag)
{
  int info;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  info = pvm_probe(tid, msgTag);

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return info;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : recvfsendsig
@MK  CLASS NAME    : cppvmConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : redefines the comparison function used to accept messages
@MU  USAGE         : void * recvf(int(*newvf)(int, int, int));
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newvf  pointer to new comparison function
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : void * pointer to old comparison function
******************************************************************************/
void *
cppvmConnection::recvf(int(*newvf)(int, int, int))
{
  void *old;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  old = (void*)(pvm_recvf(newvf));

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return old;
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/


