/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_exce.cpp 
@IE  CREATED     : 5.12.1997
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : classes for exception handling
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_exce.h"

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : pvmPrintDefaultException
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : print a message for default exceptions
@FU  USAGE         : void pvmPrintDefaultException(void)
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : 
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void
pvmPrintDefaultException(void)
{
#ifdef CPPvmErrorMsg
  cerr << "EXCEPTION OCCURRED. This exception is not part of the exceptions in module 'pvm_exce.h'." << endl;
#endif
}

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmException
@MK  CLASS NAME    : cppvmException
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : cppvmException
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmException::cppvmException() 
{
  char *mo="NO_MODULE_NAME_GIVEN";
  char *cl="NO_CLASS_NAME_GIVEN";
  char *me="NO_METHOD_NAME_GIVEN";
  char *er="NO_ERROR_MESSAGE_GIVEN";
  
  module_name = new char [strlen(mo)+1];
  class_name  = new char [strlen(cl)+1];
  method_name = new char [strlen(me)+1];
  err_msg     = new char [strlen(er)+1];

  strcpy(module_name,mo);
  strcpy(class_name,cl);
  strcpy(method_name,me);
  strcpy(err_msg,er);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmException
@MK  CLASS NAME    : cppvmException
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmException
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : mo  name of the module in which the error occurred
@MP            cl  name of the class  ...
@MP            me  name of the method ...
@MP            er  error message
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmException::cppvmException(const char *mo, const char *cl, 
                               const char *me, const char *er)
{
  if (mo == NULL){
    mo = "NO_MODULE_NAME_GIVEN";
  }
  if (cl == NULL){
    cl = "NO_CLASS_NAME_GIVEN";
  }
  if (me == NULL){
    me = "NO_METHOD_NAME_GIVEN";
  }
  if (er == NULL){
    er = "NO_ERROR_MESSAGE_GIVEN";
  }
  
  module_name = new char [strlen(mo)+1];
  class_name  = new char [strlen(cl)+1];
  method_name = new char [strlen(me)+1];
  err_msg     = new char [strlen(er)+1];
  
  strcpy(module_name,mo);
  strcpy(class_name,cl);
  strcpy(method_name,me);
  strcpy(err_msg,er);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmException
@MK  CLASS NAME    : cppvmException
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor 
@MU  USAGE         : ~cppvmException
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : cppvmException
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmException::~cppvmException () 
{
  delete [] module_name;
  delete [] class_name;
  delete [] method_name;
  delete [] err_msg;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmException
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : lane  object which has to be assigned to the local object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmException&
cppvmException::operator= (const cppvmException& obj)
{
  if(this == &obj){
    return *this;
  }
  delete [] module_name;
  delete [] class_name;
  delete [] method_name;
  delete [] err_msg;

  module_name = new char [strlen(obj.module_name)+1];
  class_name  = new char [strlen(obj.class_name)+1];
  method_name = new char [strlen(obj.method_name)+1];
  err_msg     = new char [strlen(obj.err_msg)+1];
  
  strcpy(module_name,obj.module_name);
  strcpy(class_name,obj.class_name);
  strcpy(method_name,obj.method_name);
  strcpy(err_msg,obj.err_msg);

  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmException
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  obj which has to be assigned to the local object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmException::operator== (const cppvmException& obj)
{
  return (
      strcmp(module_name,obj.module_name) && 
      strcmp(class_name,obj.class_name)   &&
      strcmp(method_name,obj.method_name) &&
      strcmp(err_msg,obj.err_msg));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmException
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object which has to be assigned to the local object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmException::operator== (const cppvmException& obj) const
{
  return (
      strcmp(module_name,obj.module_name) && 
      strcmp(class_name,obj.class_name)   &&
      strcmp(method_name,obj.method_name) &&
      strcmp(err_msg,obj.err_msg));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmException
@MK  CLASS NAME    : cppvmException
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmException::cppvmException (const cppvmException& obj)
{
  module_name = new char [strlen(obj.module_name)+1];
  class_name  = new char [strlen(obj.class_name)+1];
  method_name = new char [strlen(obj.method_name)+1];
  err_msg     = new char [strlen(obj.err_msg)+1];
  
  strcpy(module_name,obj.module_name);
  strcpy(class_name,obj.class_name);
  strcpy(method_name,obj.method_name);
  strcpy(err_msg,obj.err_msg);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : printOn
@MK  CLASS NAME    : cppvmException
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : printOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmException::printOn(ostream& strm)
{
#ifdef CPPvmErrorMsg
  strm <<
    "EXCEPTION:" << endl <<
    "Module:          " << module_name << endl <<
    "Class:           " << class_name  << endl <<
    "Method/Function: " << method_name << endl <<
    "Error:           " << err_msg     << endl;
#endif
}


/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

