/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_forw.cpp 
@IE  CREATED     : 8.4.1998
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : description of a pvm forward command
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// includes for predeclared classes to avoid cyclic includes
#include "pvm_obj.h"
// include own class
#include "pvm_forw.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_forw.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitializedRec
@VD  DESCRIPTION   : indicates if pvm receiving was initialized
@VT  TYPE          : bool
***
******************************************************************************/
bool cppvmForwardStream::pvmInitializedRec=false; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitRecContext
@VD  DESCRIPTION   : current context id before receiving
@VT  TYPE          : int
***
******************************************************************************/
int cppvmForwardStream::pvmInitRecContext; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitSendMsgBuf
@VD  DESCRIPTION   : current message buffer id before sending
@VT  TYPE          : int
***
******************************************************************************/
int cppvmForwardStream::pvmInitSendMsgBuf; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitSendContext
@VD  DESCRIPTION   : current context id before sending
@VT  TYPE          : int
***
******************************************************************************/
int cppvmForwardStream::pvmInitSendContext; 

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmForwardStream
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : only use it if it is later on initialized by the
@MU                  method 'set'
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmForwardStream::cppvmForwardStream()
    : cppvmStream(),
      initialized(false), lastForwResult(false)
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmForwardStream
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmForwardStream
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : recStrm_  receive descriptor
@MP            sendStrm_ send descriptor
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmForwardStream::cppvmForwardStream(cppvmReceiveStream& recStrm_,
                                   cppvmSendStream& sendStrm_)
    : cppvmStream(),
      initialized(true), lastForwResult(false)
{
  recStrm = recStrm_;
  sendStrm = sendStrm_;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmForwardStream
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmForwardStream
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmForwardStream::~cppvmForwardStream()
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmForwardStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmForwardStream& 
cppvmForwardStream::operator= (const cppvmForwardStream& obj) 
{
  if(cppvmIsSame(obj)){
    return *this;
  }
  cppvmStream::operator=(obj);
  recStrm = obj.recStrm;
  sendStrm = obj.sendStrm;
  lastForwResult = obj.lastForwResult;
  
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmForwardStream
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmForwardStream::cppvmForwardStream (const cppvmForwardStream& obj)
    : cppvmStream(obj),
      initialized(obj.initialized), lastForwResult(obj.lastForwResult)
{
  recStrm = obj.recStrm;
  sendStrm = obj.sendStrm;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmForwardStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmForwardStream::operator== (const cppvmForwardStream& obj) 
{
  return ( initialized  == obj.initialized &&
           recStrm      == obj.recStrm     &&
           sendStrm     == obj.sendStrm);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmForwardStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmForwardStream::operator== (const cppvmForwardStream& obj) const
{
  return ( initialized  == obj.initialized &&
           recStrm      == obj.recStrm     &&
           sendStrm     == obj.sendStrm);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator!=
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are not equal
@MU  USAGE         : !=
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmForwardStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmForwardStream::operator!= (const cppvmForwardStream& obj) 
{
  return ! (operator==(obj));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the receive and send descriptors
@MU  USAGE         : void set(cppvmReceiveStream& recStrm_,
@MU                                  cppvmSendStream& sendStrm_);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : recStrm_  receive descriptor
@MP            sendStrm_ send descriptor
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmForwardStream::set(cppvmReceiveStream& recStrm_,
                             cppvmSendStream& sendStrm_)
{
  recStrm = recStrm_;
  sendStrm = sendStrm_;
  initialized = true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : forward
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : receive a pvm object and forward it without unpacking
@MD                  object
@MU  USAGE         : bool forward(int msgTag);
@MC  CONSTRAINTS   : the stream must be initialized by 'set' or the
@MC                  send/rec-stream constructor
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag    the message tag of the class, from which an object
@MP                      should be forwarded
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : bool  true:  object received and forwarded
@MP                  false: not received and forwarded 
@MP                         (nonblocking or timeout receive)
******************************************************************************/
bool
cppvmForwardStream::forward(int msgTag)
{
  bool forwardRes;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    forwardRes = forwardTry(msgTag);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
    return false;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return forwardRes;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : forwardTry
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : receive a pvm object and forward it without unpacking
@MD                  object
@MU  USAGE         : bool forwardTry(int msgTag);
@MC  CONSTRAINTS   : the stream must be initialized by 'set' or the
@MC                  send/rec-stream constructor
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag    the message tag of the class, from which an object
@MP                      should be forwarded
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : bool  true:  object received and forwarded
@MP                  false: not received and forwarded 
@MP                         (nonblocking or timeout receive)
******************************************************************************/
bool
cppvmForwardStream::forwardTry(int msgTag)
{
  if (initialized == false){
    CPPvmInternalThrow( cppvmExceptionForbiddenCall(THIS_MODULE,
                        "cppvmForwardStream",
                        "forwardTry",
                        "forward stream was not initialized, cannot forward object. Call 'set' before using the cppvmForwardStream"));
    lastForwResult=false;
    return false;
  }

  if(sendStrm.getSendMode() != CPPvmNonblockingSend){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                      "cppvmForwardStream",   
                      "forwardTry",
                      "can only forward with nonblocking send"));
    lastForwResult=false;
    return false;
  }

  // allow only the first receiving object to initialize a receiving
  if(pvmInitializedRec){
    CPPvmInternalThrow( cppvmExceptionForbiddenCall(THIS_MODULE,
                        "cppvmForwardStream",
                        "forwardTry",
                        "forward cannot be called within a forward command!"));
    lastForwResult=false;
    return false;
  }

  // set context
  pvmInitRecContext = pvm_setcontext(recStrm.getContext());
  
  pvmInitializedRec = true;

  int recBufId=-1;
  int dummy2, dummy3, source;
  
  if(recStrm.getRecMode() == CPPvmBlockingRec){
    recBufId = pvm_recv(recStrm.getRecFromTid(), msgTag);
  }
  else if(recStrm.getRecMode() == CPPvmNonblockingRec){
    recBufId = pvm_nrecv(recStrm.getRecFromTid(), msgTag);
    if(recBufId == 0){
      pvmInitializedRec = false;
      // reset context
      pvm_setcontext(pvmInitRecContext);
      lastForwResult=false;
      return false;
    }
  }
  else if(recStrm.getRecMode() == CPPvmTimeoutRec){
    recBufId = pvm_trecv(recStrm.getRecFromTid(), msgTag, recStrm.getTimeoutDirect() );
    if(recBufId == 0){
      pvmInitializedRec = false;
      // reset context
      pvm_setcontext(pvmInitRecContext);
      lastForwResult=false;
      return false;
    }
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmForwardStream",   
                                                   "forwardTry",
                                                   "unknown receive mode"));
  }

  if(recBufId < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmForwardStream",   
                                                "forwardTry",
                                                "cannot receive a message"));
    lastForwResult=false;
    return false;
  }

  pvm_bufinfo(recBufId, &dummy2, &dummy3, &source);
  
  // store the information of the sending process
  // this is useful to obtain information about the sending process
  // in case of 'any' receive (tid = -1)
  recStrm.setLastRecFromTid(source);

  // reset context
  pvm_setcontext(pvmInitRecContext);

  // begin send (forward) message

  // save old send message buffer and set new context
  pvmInitSendMsgBuf = pvm_getsbuf();
  pvmInitSendContext = pvm_setcontext(sendStrm.getContext(0));

  // set receive buffer to send buffer
  pvm_setsbuf(recBufId);

  int info;
  
  // send to other task(s)
  if(sendStrm.getSendConnection() == CPPvmOneToOne){
    info = pvm_send(sendStrm.getTid(), msgTag);
  }
  else if(sendStrm.getSendConnection() == CPPvmMulticast){
    info = pvm_mcast(sendStrm.getTidsDirect(), sendStrm.getNumTid(), msgTag);
  }
  else if(sendStrm.getSendConnection() == CPPvmGroupBroadcast){
    info = pvm_bcast(sendStrm.getGroup(), msgTag);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmForwardStream",   
                                                   "forwardTry",
                                                   "unknown send connection type"));
    lastForwResult=false;
    return false;
  }

  // send error ?
  if(info < 0){
    if(info==PvmNoGroup){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmObject",   
                                                  "exitSend",
                                                  "cannot send data to group: no such group"));
    }
    else if(info==PvmNoInst){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmObject",   
                                                  "exitSend",
                                                  "cannot send data to group: no other group members"));
    }
    else{
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmForwardStream",   
                                                  "forwardTry",
                                                  "cannot send data properly"));
    }
    lastForwResult=false;
    return false;
  }

  // reset context
  pvm_setcontext(pvmInitSendContext);
  // set old send buffer again
  pvm_setsbuf(pvmInitSendMsgBuf);
  // reset values
  pvmInitializedRec = false;

  lastForwResult=true;
  return true;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getRecStrm
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the receive descriptor
@MD                  (from witch the data should was received)
@MU  USAGE         : void getRecStrm();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : cppvmReceiveStream&  receive descriptor
******************************************************************************/
cppvmReceiveStream&
cppvmForwardStream::getRecStrm()
{
  if (initialized == false){
    CPPvmInternalThrow( cppvmExceptionForbiddenCall(THIS_MODULE,
                        "cppvmForwardStream",
                        "getRecStrm",
                        "forward stream was not initialized, cannot return cppvmReceiveStream. Call 'set' before using the cppvmForwardStream"));
  }
  return recStrm;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getSendStrm
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the send descriptor
@MD                  (to witch the data should be forwarded)
@MU  USAGE         : void getSendStrm();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : cppvmSendStream&  send descriptor
******************************************************************************/
cppvmSendStream&
cppvmForwardStream::getSendStrm()
{
  if (initialized == false){
    CPPvmInternalThrow( cppvmExceptionForbiddenCall(THIS_MODULE,
                        "cppvmForwardStream",
                        "getSendStrm",
                        "forward stream was not initialized, cannot return cppvmSendStream. Call 'set' before using the cppvmForwardStream"));
  }
  return sendStrm;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmForwardStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmForwardStream::cppvmPrintOn(ostream& strm)
{
  strm << recStrm;
  strm << sendStrm;
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

