/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_inmu.cpp 
@IE  CREATED     : 16.10.2001
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : internal mutex class; the implementation depends on:
@ID                1) compiler switch 'CPPvmThread'
@ID                2) operating system (UNIX or Windows)
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_inmu.h"

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmInternalMutex
@MK  CLASS NAME    : cppvmInternalMutex
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor 
@MU  USAGE         : 
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : for UNIX systems the method "setUnixMutex" MUST
@MR                  be used after the constructor
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmInternalMutex::cppvmInternalMutex ()
{
#ifdef CPPvmThread
#ifdef WIN32
  winLock = new CSingleLock(&winMutex);
#endif
#endif
}

#ifdef CPPvmThread
#ifndef WIN32
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmInternalMutex
@MK  CLASS NAME    : cppvmInternalMutex
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for UNIX systems
@MU  USAGE         : 
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the mutex MUST be initialized externally.
@MR                  Otherwise the initializer macro
@MR                  "PTHREAD_MUTEX_INITIALIZER" can not be used
@M#  --------------------------------------------------------------------------
@MP  IN      : mutexPtr pointer to initialized mutex variable
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmInternalMutex::cppvmInternalMutex(pthread_mutex_t* mutexPtr)
{
  unixMutexPtr = mutexPtr;
  // For POSIX portability of statically allocated mutexes 
  // use the macro PTHREAD_MUTEX_INITIALIZER for the
  // extern mutex variable, e.g.
  // mutex = PTHREAD_MUTEX_INITIALIZER;
  // Otherwise the initialization depends on the
  // operating system:
  // Synopsis SOLOARIS:
  // pthread_mutex_init(&communicationMutex, NULL);
  // Synopsis HP:
  //         int pthread_mutex_init(
  //          pthread_mutex_t *mutex,
  //          pthread_mutexattr_t attr);
  // pthread_mutex_init(&communicationMutex, pthread_mutexattr_default);
}
#endif
#endif

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmInternalMutex
@MK  CLASS NAME    : cppvmInternalMutex
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor 
@MU  USAGE         : ~cppvmInternalMutex
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : cppvmInternalMutex
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmInternalMutex::~cppvmInternalMutex() 
{
#ifdef CPPvmThread
#ifdef WIN32
  delete winLock;
#else
  pthread_mutex_destroy(unixMutexPtr);
#endif
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : lock
@MK  CLASS NAME    : cppvmInternalMutex
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : lock mutual exclusion
@MU  USAGE         : lock()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmInternalMutex::lock()
{
#ifdef CPPvmThread
#ifdef WIN32
  winLock->Lock();
#else
  pthread_mutex_lock(unixMutexPtr);
#endif
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : unlock
@MK  CLASS NAME    : cppvmInternalMutex
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unlock mutual exclusion
@MU  USAGE         : unlock()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmInternalMutex::unlock()
{
#ifdef CPPvmThread
#ifdef WIN32
  winLock->Unlock();
#else
  pthread_mutex_unlock(unixMutexPtr);
#endif
#endif
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

