/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_mutexcoll.cpp 
@IE  CREATED     : 29.07.2001
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : semaphore collection for cppvmMutex
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

// is only used by 'cppvmMutex' ('pvm_mutex.h')
// and 'cppvmMutex' is using templates
#ifndef CPPvmDoNotUseTemplates

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// These types are needed for 'pvm_sema.h'
// This was made to avoid cyclic includes
#include "pvm_str.h"
#include "pvm_int.h"
#include "pvm_spaw.h"
#include "pvm_rec.h"
#include "pvm_send.h"
#include "pvm_conn.h"
// semaphore
#include "pvm_sema.h"
// own header
#include "pvm_mutexcoll.h"
#ifndef CPPvmDoNotUseStl
#include "pvm_stl.h"
#endif


/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_mutexcoll.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                           VARIABLES                             *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : semaphoreObj
@VD  DESCRIPTION   : global semaphore
@VT  TYPE          : cppvmSeam
***
******************************************************************************/
cppvmSemaphore cppvmMutexCollection::semaphoreObj; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmMutexCollectionSemaphoreMap
@VD  DESCRIPTION   : global semaphore map
@VT  TYPE          : cppvmSeam
***
******************************************************************************/
#ifndef CPPvmDoNotUseStl
static map<string, cppvmSemaphore> cppvmMutexCollectionSemaphoreMap; 
#endif

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmMutexCollection
@MK  CLASS NAME    : cppvmMutexCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : cppvmMutexCollection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmMutexCollection::cppvmMutexCollection()
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmMutexCollection
@MK  CLASS NAME    : cppvmMutexCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmMutexCollection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmMutexCollection::~cppvmMutexCollection()
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : enter
@MK  CLASS NAME    : cppvmMutexCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set semaphore value to 1 and enter semaphore
@MU  USAGE         : void enter(char *semaphoreName);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : semphoreName name of the semaphore
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : -
******************************************************************************/
void
cppvmMutexCollection::enter(char *semaphoreName)
{
#ifndef CPPvmDoNotUseStl
  // STL version
  
  // 'internalTryToSetNumber' communicates with the semaphore server
  // To avoid this communication, already created semaphores
  // are stored and recalled.
  if(cppvmMutexCollectionSemaphoreMap.empty()==true){
    // start semaphore server
    // (the semaphore server is NOT started when the
    //  default constructor is used!)
    cppvmSemaphore tmpSemaphore(semaphoreName);
    semaphoreObj = tmpSemaphore;
    semaphoreObj.internalTryToSetNumber(1);                               
    semaphoreObj.enter();
    // add semaphore to map
    string key = semaphoreName;
#ifdef CPPvmDoNotUseStdNamespace
    cppvmMutexCollectionSemaphoreMap.insert(
        map<string,cppvmSemaphore>::value_type(key, semaphoreObj));
#else
    cppvmMutexCollectionSemaphoreMap.insert(
        std::map<std::string,cppvmSemaphore>::value_type(key, semaphoreObj));
#endif
  }
  else{
#ifdef CPPvmDoNotUseStdNamespace
    map<string, cppvmSemaphore>::iterator i;
#else
    std::map<std::string, cppvmSemaphore>::iterator i;
#endif
    string key = semaphoreName;
    i = cppvmMutexCollectionSemaphoreMap.find(key);
    if(i==cppvmMutexCollectionSemaphoreMap.end()){
      // not found
      semaphoreObj.setName(semaphoreName);
      semaphoreObj.internalTryToSetNumber(1);                               
      semaphoreObj.enter();
      if(cppvmMutexCollectionSemaphoreMap.size()<=CPPvmMutexCollMapMaxSize){
        // add semaphore only to map if the map has 
        // not reached a maximal number
        // it is probably faster to set the semaphore value
        // via the server than searching the semphore in
        // the list
#ifdef CPPvmDoNotUseStdNamespace
        cppvmMutexCollectionSemaphoreMap.insert(
            map<string,cppvmSemaphore>::value_type(key, semaphoreObj));
#else
        cppvmMutexCollectionSemaphoreMap.insert(
            std::map<std::string,cppvmSemaphore>::value_type(key, semaphoreObj));
#endif
      }
    }
    else{
      // reuse semaphore in map
      semaphoreObj = i->second;
      semaphoreObj.enter();
    }
  }
#else
  // version without STL
  
  static bool firstCall=true;
  
  if(firstCall==true){
    // start semaphore server
    // (the semaphore server is NOT started when the
    //  default constructor is used!)
    cppvmSemaphore tmpSemaphore(semaphoreName);
    firstCall = false;
  }
  
  semaphoreObj.setName(semaphoreName);
  semaphoreObj.internalTryToSetNumber(1);                               
  semaphoreObj.enter();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : leave
@MK  CLASS NAME    : cppvmMutexCollection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : leave semaphore
@MU  USAGE         : void leave();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      :
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : -
******************************************************************************/
void
cppvmMutexCollection::leave()
{
  semaphoreObj.leave();
}

//CPPvmDoNotUseTemplates
#endif

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

