/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_noti.cpp 
@IE  CREATED     : 24.07.2000
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : notification class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

// class header
#include "pvm_noti.h" 

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_noti.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmMaxId
@VD  DESCRIPTION   : unique id number for all framework objects
@VD                  highest actual ID
@VT  TYPE          : int
***
******************************************************************************/
int cppvmNotification::cppvmMaxId = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmTotalObjNum
@VD  DESCRIPTION   : total number of cppvmObjects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmNotification::cppvmTotalObjNum = 0; 

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmNotification
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : cppvmNotification
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmNotification::cppvmNotification()
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  initialized=false;
  notificationEvent = -1;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmNotification
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmNotification
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmNotification::cppvmNotification(int event, int msgTag, 
                                     int number, int* tids)
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  set(event, msgTag, number, tids);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmNotification
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmNotification::cppvmNotification (const cppvmNotification& obj)
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  initialized = obj.initialized;
  
  notificationEvent = obj.notificationEvent;
  notificationMsgTag = obj.notificationMsgTag;
  notificationTids = obj.notificationTids;
  notificationNumber = obj.notificationNumber;

  receivedTids = obj.receivedTids;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmNotification
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmNotification()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      :
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
cppvmNotification::~cppvmNotification()
{
  // dec total number of objects
  cppvmTotalObjNum--;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set passive method
@MU  USAGE         : void set(int event, int msgTag);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmNotification::set(int event, int msgTag)
{
  notificationEvent=event;
  notificationMsgTag=msgTag;
  notificationNumber=0;

  notificationTids.resize(0);
  
  initialized=true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method
@MU  USAGE         : void set(int event, int msgTag, 
@MU                           int number, int* tids);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmNotification::set(int event, int msgTag, int number, int* tids)
{
  notificationEvent=event;
  notificationMsgTag=msgTag;
  notificationNumber=number;

  // the 'tids' parameter is not used for the 'PvmHostAdd' event
  if(event==PvmTaskExit || event==PvmHostDelete){
    notificationTids.resize(number);
    int i;
    for(i=0; i<number; i++){
      notificationTids[i]=tids[i];
    }
  }
  else{
    notificationTids.resize(0);
  }
  
  int info;
  info=pvm_notify(notificationEvent, notificationMsgTag,
                  notificationNumber, notificationTids.getPtr());
  if(info<0)
  {
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmNotification",   
                                                   "set",
                                                   "could not start notification request"));
  }

  initialized=true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  info for configuration object attributes 
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmNotification&
cppvmNotification::operator= (const cppvmNotification& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }

  initialized = obj.initialized;
  
  notificationEvent = obj.notificationEvent;
  notificationMsgTag = obj.notificationMsgTag;
  notificationTids = obj.notificationTids;
  notificationNumber = obj.notificationNumber;

  receivedTids = obj.receivedTids;
  
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to compare
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmNotification::operator== (const cppvmNotification& obj)
{
  return (initialized==obj.initialized &&
          notificationEvent==obj.notificationEvent &&
          notificationMsgTag==obj.notificationMsgTag &&
          notificationTids==obj.notificationTids &&
          notificationNumber==obj.notificationNumber);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to compare
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmNotification::operator== (const cppvmNotification& obj) const
{
  return (initialized==obj.initialized &&
          notificationEvent==obj.notificationEvent &&
          notificationMsgTag==obj.notificationMsgTag &&
          notificationTids==obj.notificationTids &&
          notificationNumber==obj.notificationNumber);
}
 

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmReceive
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : receive notification 
@MU  USAGE         : bool cppvmReceive(cppvmReceiveStream& recStrm);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm out stream
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
bool
cppvmNotification::cppvmReceive(cppvmReceiveStream& recStrm)
{
  CPPvmNotificationThrowIfNotInit

  int bufid=-1;
  int dummy2, dummy3, source;
  int oldContext;

  // set context
  oldContext = pvm_setcontext(recStrm.getContext());
  
  if(recStrm.getRecMode() == CPPvmBlockingRec){
    bufid = pvm_recv(recStrm.getRecFromTid(), notificationMsgTag);
  }
  else if(recStrm.getRecMode() == CPPvmNonblockingRec){
    bufid = pvm_nrecv(recStrm.getRecFromTid(), notificationMsgTag);
    if(bufid == 0){
      // reset context
      pvm_setcontext(oldContext);
      return false;
    }
  }
  else if(recStrm.getRecMode() == CPPvmTimeoutRec){
    bufid = pvm_trecv(recStrm.getRecFromTid(), notificationMsgTag, 
                      recStrm.getTimeoutDirect() );
    if(bufid == 0){
      // reset context
      pvm_setcontext(oldContext);
      return false;
    }
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmNotification",   
                                                   "initRec",
                                                   "unknown receive mode"));
  }
  

  if(bufid < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmNotification",   
                                                "initRec",
                                                "cannot receive a message"));
  }

  pvm_bufinfo(bufid, &dummy2, &dummy3, &source);
  
  // store the information of the sending process
  // this is useful to obtain information about the sending process
  // in case of 'any' receive (tid = -1)
  recStrm.setLastRecFromTid(source);

  // unpack message data
  unpackMsg();
  
  return true;
  
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : unpackMsg
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unpack notification message
@MU  USAGE         : void unpackMsg();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmNotification::unpackMsg()
{

  if(notificationEvent==PvmHostAdd){
    unpackInt(&numberResults, 1, 1);
    receivedTids.resize(numberResults);
    unpackInt(receivedTids.getPtr(), numberResults, 1);
  }
  else{
    receivedTids.resize(1);
    // PvmHostDelete or PvmTaskExit events
    unpackInt(receivedTids.getPtr(), 1, 1);
  }
}
  
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : unpackInt
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : unpack integer values
@MU  USAGE         : void unpackInt(int *ptr, int nitme, int stride);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmNotification::unpackInt(int *ptr, int nitem, int stride)
{
  int unpackResult;
  
  unpackResult=pvm_upkint(ptr, nitem, stride);
  
  if(unpackResult < 0) {
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmNotification",   
                                                "unpackInt",
                                                "cannot unpack notification data"));
  }                                                      
}
 
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print on stream method
@MU  USAGE         : void cppvmPrintOn(ostream& strm);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm out stream
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmNotification::cppvmPrintOn(ostream& strm)
{
  if(initialized==false){
    strm << "notification object is not initialized" << endl;
  }
  else{
    strm << "Notification for: ";
    if(notificationEvent==PvmTaskExit){
      strm << "exiting tasks";
    }
    else if(notificationEvent==PvmHostDelete){
      strm << "deleted hosts";
    }
    else if(notificationEvent==PvmHostAdd){
      strm << "added hosts";
    }
    else{
      strm << "unknown event";
    }
    strm << endl;
    strm << "Message Tag: " << notificationMsgTag;
    strm << " number: " << notificationNumber;
    strm << " task ids: " << notificationTids;
    strm << endl;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getHostDaemonTid
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the host daemon task id for a task or host name
@MU  USAGE         : int  getHostDaemonTid(char* taskOrHostName);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskOrHostName  name of a task or a host
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int pvm daemon task id
******************************************************************************/
int
cppvmNotification::getHostDaemonTid(char* taskOrHostName)
{
  int tid = getTidOfTask(taskOrHostName);
  if(tid>0){
    int dtid = pvm_tidtohost(tid);
    if(dtid == PvmSysErr){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmNotification",
                                                  "getHostDaemonTid",
                                                  "pvmd is not responding."));
    }
    else if(dtid>0){
      return dtid;
    }
  }
  
  int i=0;
  int nhost;
  int narch;
  struct pvmhostinfo *hostp;
  bool found = false;
  int position=0;
  
  hostInfo(&nhost, &narch, &hostp);
  
  while(i<nhost && found==false){
    if(strcmp(hostp[i].hi_name,taskOrHostName)==0){
      found = true;
      position = i;
    }
    i++;
  }
  
  if(found==true){
    return (hostp[position].hi_tid);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                              "cppvmNotification",
                              "getHostDaemonTid",
                              "no pvm daemon tid found for the given task/host name."));
    return -1;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : hostInfo
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : configuration of the virtual machine
@MU  USAGE         : void hostInfo(int *nhost,int *narch,
@MU                                struct pvmhostinfo **hostp)
@MC  CONSTRAINTS   : the space for the struct pvmhostinfo is allocated by
@MC                  this method
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : nhost  number of hosts on the virtual machine
@MP            narch  number of different data formats being used
@MP            pvmhostinfo  pointer to an array of structures containing
@MP                         information about each host
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmNotification::hostInfo(int *nhost,
                            int *narch,
                            struct pvmhostinfo **hostp)
{
  int info;

  info = pvm_config(nhost, narch, hostp);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                   "cppvmNotification",
                   "hostInfo",
                   "could not get configuration informations of the virtual machine."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : taskInfo
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : information about tasks running on the virtual machine
@MU  USAGE         : void taskInfo(int where,int *ntask,
@MU                                struct pvmtaskinfo **taskp);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : the space for the struct pvmtaskinfo array is given
@MR                  by this method
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : where   information specifier:
@MP                    0:        all tasks on the virtual machine
@MP                    pvmd tid: all task on this host (tid of the pvm daemon, 
@MP                              use getHostDaemonTid to get this id)
@MP                    tid:      specific task
@MP            ntask   number of task (depending on 'where')
@MP            taskp   pointer to an array of structures which contain
@MP                    information about each task
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmNotification::taskInfo(int where,
                            int *ntask,
                            struct pvmtaskinfo **taskp)
{
  int info;

  if(ntask == NULL ||
     taskp == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmNotification",
                                                   "taskInfo",
                                                   "a parameter is a NULL pointer."));
  }
  
  info = pvm_tasks(where, ntask, taskp);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                           "cppvmNotification",
                           "taskInfo",
                           "could not get task informations of the virtual machine."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTidOfTask
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the task id of a task running in pvm specified
@MD                  by its name
@MU  USAGE         : int getTidOfTask(char *taskName)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskName  name of the pvm task
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : int > 0: task id of the process
@MP                <=0: task does not exists in pvm
******************************************************************************/
int
cppvmNotification::getTidOfTask(char *taskName)
{
  int i;
  int ntask;
  struct pvmtaskinfo *taskp;

  if(taskName == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmNotification",
                                                   "getTidOfTask",
                                                   "name is a NULL pointer"));
  }

  taskInfo(0, &ntask, &taskp);

  for(i=0;i<ntask;i++){
    if(strcmp(taskp[i].ti_a_out, taskName)==0){
      return taskp[i].ti_tid;
    }
  }
  
  return -1;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getReceivedTids
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the received task/host ids
@MU  USAGE         : cppvmArrayInt getReceivedTids()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : cppvmArrayInt task or host ids
******************************************************************************/
cppvmArrayInt
cppvmNotification::getReceivedTids()
{
  CPPvmNotificationThrowIfNotInit 
  return receivedTids;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getEvent
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the notification event set in the 
@MD                  constructor/set method
@MU  USAGE         : cppvmArrayInt getEvent()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : int  notification event (PvmTaskExit, PvmHostDelete, PvmHostAdd)
******************************************************************************/
int
cppvmNotification::getEvent()
{
  CPPvmNotificationThrowIfNotInit 
  return notificationEvent;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getMsgTag
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the message tag of the notification message set in
@MD                  the constructor/set method
@MU  USAGE         : cppvmArrayInt getMsgTag()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : int  message tag
******************************************************************************/
int
cppvmNotification::getMsgTag()
{
  CPPvmNotificationThrowIfNotInit 
  return notificationMsgTag;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTids
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the tasks/hosts ids set in the constructor/set method
@MU  USAGE         : cppvmArrayInt getTids()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : cppvmArrayInt task or host ids
******************************************************************************/
cppvmArrayInt
cppvmNotification::getTids()
{
  CPPvmNotificationThrowIfNotInit 
  return notificationTids;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getNumber
@MK  CLASS NAME    : cppvmNotification
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the number of tasks/hosts ids set in the 
@MD                  constructor/set method
@MU  USAGE         : cppvmArrayInt getNumber()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : int  number of task/host ids
******************************************************************************/
int
cppvmNotification::getNumber()
{
  CPPvmNotificationThrowIfNotInit 
  return notificationNumber;
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

