/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_obj.cpp 
@IE  CREATED     : 5.12.1997
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm object base class, defines all standard methods for the
@ID                objects of the framework
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_obj.h"
#include "pvm_conn.h"


/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_obj.cpp"

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmExchangeBlockingSend
@CD  DESCRIPTION   : indicates a blocking send operation
***
******************************************************************************/
#define CPPvmExchangeBlockingSend    0

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmExchangeNonblockingSend
@CD  DESCRIPTION   : indicates a nonblocking send operation
***
******************************************************************************/
#define CPPvmExchangeNonblockingSend 1

/*@@***************************   CONSTANT   **********************************
***
@CN  CONSTANT NAME : CPPvmHandshakeArrayInitLen
@CD  DESCRIPTION   : initial length of the 'sendblocking' handshake tid array
***
******************************************************************************/
#define CPPvmHandshakeArrayInitLen 1024

/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmMaxId
@VD  DESCRIPTION   : unique id number for all framework objects
@VD                  highest actual ID
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::cppvmMaxId = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmTotalObjNum
@VD  DESCRIPTION   : total number of cppvmObjects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::cppvmTotalObjNum = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmTransferRec
@VD  DESCRIPTION   : indicates if the transfer is a receive
@VT  TYPE          : bool
***
******************************************************************************/
bool cppvmObject::cppvmTransferRec;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmTransferSend
@VD  DESCRIPTION   : indicates if the transfer is a send
@VT  TYPE          : bool
***
******************************************************************************/
bool cppvmObject::cppvmTransferSend; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitializedSend
@VD  DESCRIPTION   : indicates if pvm sending was initialized
@VT  TYPE          : bool
***
******************************************************************************/
bool cppvmObject::pvmInitializedSend; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmNewSendMsgBuf
@VD  DESCRIPTION   : new created message buffer id for sending
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmNewSendMsgBuf; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitSendMsgBuf
@VD  DESCRIPTION   : current message buffer id before sending
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmInitSendMsgBuf; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitSendContext
@VD  DESCRIPTION   : current context id before sending
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmInitSendContext; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmObjIdInitSend
@VD  DESCRIPTION   : id of the object which has initialized
@VD                  pvm sending
@VD                  this makes the initial object unique among other objects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmObjIdInitSend;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmObjTagInitSend
@VD  DESCRIPTION   : message tag of the object which has initialized
@VD                  pvm sending
@VD                  this makes the initial class unique compared to its
@VD                  base class(es)
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmObjTagInitSend;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitializedRec
@VD  DESCRIPTION   : indicates if pvm receiving was initialized
@VT  TYPE          : bool
***
******************************************************************************/
bool cppvmObject::pvmInitializedRec; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmObjIdInitRec
@VD  DESCRIPTION   : id of the object which has initialized
@VD                  pvm receiving
@VD                  this makes the initial object unique among other objects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmObjIdInitRec;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmInitRecContext
@VD  DESCRIPTION   : current context id before receiving
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmInitRecContext; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmObjTagInitRec
@VD  DESCRIPTION   : message tag of the object which has initialized
@VD                  pvm receiving
@VD                  this makes the initial class unique compared to its
@VD                  base class(es)
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmObjTagInitRec;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmRecHandshake
@VD  DESCRIPTION   : handshake delivered by the sending taks
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmRecHandshake;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmSendHandshake
@VD  DESCRIPTION   : handshake send to the receiving taks
@VT  TYPE          : int
***
******************************************************************************/
int cppvmObject::pvmSendHandshake;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmSendHandshakeTids
@VD  DESCRIPTION   : array of handshake tids
@VD                  receive handshakes from these tids during blocking send
@VT  TYPE          : int*
***
******************************************************************************/
int* cppvmObject::pvmSendHandshakeTids;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmSendHandshakeTidsNum
@VD  DESCRIPTION   : current number of handshake tids in the array
@VT  TYPE          : int
***
******************************************************************************/
unsigned int cppvmObject::pvmSendHandshakeTidsNum;

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : pvmSendHandshakeTidsSize
@VD  DESCRIPTION   : total size of the handshake tid array
@VT  TYPE          : int
***
******************************************************************************/
unsigned int cppvmObject::pvmSendHandshakeTidsSize;

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmObject
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmObject
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmObject::cppvmObject(int objectMsgTag)
{
  if(objectMsgTag==CPPvmNoObjectMsgTag){
    // do not use local message tag
    cppvmDeactivateObjectMsgTag();
  }
  else{
    checkMsgTagRange(objectMsgTag);
    setCPPvmObjectMsgTag(objectMsgTag);
  }

  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;
  
  if(cppvmObjId==0) {
    // initialize global variables only once
    // (for the first object)
    // if e.g. a new 'cppvmObject' is created during an receive
    // ('initializedSend' == true), the value would
    // be set to false (what is fatal)
    pvmInitializedSend = false;
    pvmInitializedRec = false;
    cppvmTransferRec = false;
    cppvmTransferSend = false;
    pvmSendHandshakeTids = new int[CPPvmHandshakeArrayInitLen];
    pvmSendHandshakeTidsNum = 0;
    pvmSendHandshakeTidsSize = CPPvmHandshakeArrayInitLen;
    
  }

  // set mbox values
  mboxKeyName = NULL;
  mboxInitRead = false;
  mboxInitWrite = false;
  mboxInitDel = false;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmObject
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmObject::cppvmObject (const cppvmObject& obj)
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  // set initial values
  mboxKeyName = NULL;
  mboxInitRead = false;
  mboxInitWrite = false;
  mboxInitDel = false;

  // is the mbox (or distributed obj.) initialized?
  if(obj.mboxInitRead==true){
    cppvmSetMboxRead(obj.mboxKeyName, obj.mboxFlagsRead);
  }
  if(obj.mboxInitWrite==true){
    cppvmSetMboxWrite(obj.mboxKeyName, obj.mboxFlagsWrite,
                      obj.mboxSendEncoding, obj.mboxOwnSendBuf);
  }
  if(obj.mboxInitDel==true){
    cppvmSetMboxDel(obj.mboxKeyName, obj.mboxFlagsDel);
  }

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmObject
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmObject()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      :
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
cppvmObject::~cppvmObject()
{
  cppvmTotalObjNum--;

  if(mboxKeyName!=NULL){
    // delete allocated string
    delete [] mboxKeyName;
  }

}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  info for configuration object attributes 
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmObject&
cppvmObject::operator= (const cppvmObject& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }
  return *this;
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : initSend
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : initialize sending if this was not done before
@MD                  this must be called at first position in every send method
@MD                  of the cppvmObjects
@MU  USAGE         : initSend(cppvmSendStream& send, int msgTag)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : send   description of the pvm sending process
@MP            msgTag  message tag (message identifier) of the object
@MP                    (maybe needed in further enhancements)
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::initSend(cppvmSendStream& send, int msgTag)
{

  if(pvmInitializedSend){
    return;
  }

  // save old send message buffer and set new context
  pvmInitSendMsgBuf = pvm_getsbuf();
  pvmInitSendContext = pvm_setcontext(send.getContext(0));

  cppvmTransferSend = true;
  pvmInitializedSend = true;
  // save the unique (!) object id of the initializing object
  // this will be used later to recognize the object within 'exitSend'
  pvmObjTagInitSend = msgTag;
  pvmObjIdInitSend = getCPPvmObjId();

  int msgEncodingExistingBuf;
  int msgEncodingNewBuf;
  int msgEncodingMode = cppvmConnection::getGlobalMsgEncoding();

  // preset local message encoding
  msgEncodingNewBuf = send.getMsgBufEncoding();
  msgEncodingExistingBuf = send.getInitSend();

  // use global settings if wanted by user
  if(msgEncodingMode==CPPvmAllPvmDataDefault){
    msgEncodingNewBuf = PvmDataDefault;
    msgEncodingExistingBuf = PvmDataDefault;
  }
  else if(msgEncodingMode==CPPvmAllPvmDataInPlace){
    msgEncodingNewBuf = PvmDataInPlace;
    msgEncodingExistingBuf = PvmDataInPlace;
  } 
  else if(msgEncodingMode==CPPvmAllPvmDataRaw){
    msgEncodingNewBuf = PvmDataRaw;
    msgEncodingExistingBuf = PvmDataRaw;
  } 


  if(send.getMsgBuf()==true){
    // use a new created message buffer for sending
    pvmNewSendMsgBuf = pvm_mkbuf(msgEncodingNewBuf);
    if(pvmNewSendMsgBuf < 0){
      if(pvmNewSendMsgBuf == PvmBadParam){
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                  "cppvmObject",
                                  "initSend",
                                  "giving an invalid encoding value"));
      }
      else if(pvmNewSendMsgBuf == PvmNoMem){	
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                  "cppvmObject",
                                  "initSend",
                                  "not enough memory to create the buffer"));
      }
      else{
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                  "cppvmObject",
                                  "initSend",
                                  "cannot create new buffer"));
      }
    }
    pvm_setsbuf(pvmNewSendMsgBuf);
  }
  else{
    pvm_initsend(msgEncodingExistingBuf);
  }
  if(send.getSendMode() == CPPvmBlockingSend){
    pvmSendHandshake = CPPvmExchangeBlockingSend;
  }
  else if(send.getSendMode() == CPPvmNonblockingSend){
    pvmSendHandshake = CPPvmExchangeNonblockingSend;
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                        "cppvmObject",   
                        "initSend",
                        "the 'sendMode' in 'send' parameter has a bad value"));
  }

  if((pvm_pkint(&pvmSendHandshake, 1, 1)) < 0) {
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "initSend",
                                                "cannot pack send handshake"));
  }                                                      
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : exitSend
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : sending the sending buffer if the object was the
@MD                  initial sending object
@MD                  this must be called at last position in every send method
@MD                  of the cppvmObjects
@MU  USAGE         : exitSend(cppvmSendStream& send, int msgTag)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag  message tag (message identifier) of the object
@MP  INOUT   : send   description of the pvm sending process
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::exitSend(cppvmSendStream& send, int msgTag)
{
  // is this object not the initial send object
  if(msgTag != pvmObjTagInitSend ||
     getCPPvmObjId() != pvmObjIdInitSend ){
    return;
  }

  // reset values
  cppvmTransferSend = false;
  pvmInitializedSend = false;
  
  int info=-1;
  int sendTypeInfo;
  unsigned int i;

  // send to other task(s)
  if(send.getSendConnection() == CPPvmOneToOne){
    for(i=0; i<(unsigned int)send.getNumTid(); i++){
      pvm_setcontext(send.getContext(i));
      info = pvm_send(send.getTid(i), msgTag);
    }
  }
  else if(send.getSendConnection() == CPPvmMulticast){
    pvm_setcontext(send.getContext(0));
    info = pvm_mcast(send.getTidsDirect(), send.getNumTid(), msgTag);
  }
  else if(send.getSendConnection() == CPPvmGroupBroadcast){
    info = pvm_bcast(send.getGroup(), msgTag);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                        "cppvmObject",   
                        "exitSend",
                        "unknown send connection type"));
  }

  // send error ?
  if(info < 0){
    if(info==PvmNoGroup){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmObject",   
                                                  "exitSend",
                                                  "cannot send data to group: no such group"));
    }
    else if(info==PvmNoInst){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmObject",   
                                                  "exitSend",
                                                  "cannot send data to group: no other group members"));
    }
    else{
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmObject",   
                                                  "exitSend",
                                                  "cannot send data properly"));
    }
  }
  // wait for handshake(s) if this was a blocking send
  if(send.getSendMode() == CPPvmBlockingSend){
    // init the 'pvmSendHandshakeTids' array
    if(send.getSendConnection() == CPPvmGroupBroadcast){
      // init the handshake array for group broadcast
      cppvmSetHandshakeTidArrayGroupBroadcast(send);
    }
    else {
      // init the handshake array for one to one connection or multicast
      // CPPvmOneToOne or CPPvmMulticast
      // IMPORTANT: Multicasted (array of tids) and broadcasted
      //            (to group members) messages are NOT send to
      //            the caller even if the caller is in the
      //            array of tids resp. member of the group
      // So for blocking send, do NOT wait for an answer
      // of the caller process (this process), decrease the
      // number of answers
      cppvmSetHandshakeTidArrayOneOrMulticast(send);
    }
    
    // wait for a handshake from all task specified in 'pvmSendHandshakeTids'
    for(i=0; i<pvmSendHandshakeTidsNum; i++){
      pvm_recv(pvmSendHandshakeTids[i], CPPvmHandshakeMsgTag);
      
      if((pvm_upkint(&sendTypeInfo, 1, 1)) < 0) {
        CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                 "cppvmObject",   
                                                 "exitSend",
                                                 "cannot unpack send type information"));
      }                                                      

      if(sendTypeInfo!=CPPvmExchangeBlockingSend){
        CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                    "cppvmObject",   
                                                    "exitSend",
                                                    "got the wrong handshake"));
      }
    }
  }
  else if(send.getSendMode()!=CPPvmNonblockingSend){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "exitSend",
                                                "unknown send mode"));
  }

  // reset context
  pvm_setcontext(pvmInitSendContext);
  if(send.getMsgBuf()==true){
    // a new created message buffer was used for sending
    pvm_setsbuf(pvmInitSendMsgBuf);
    pvm_freebuf(pvmNewSendMsgBuf);
  }

  // this is needed to delete the allocated memory after sending
  // allocated by transfer functions
  cppvmCleanMemoryAfterSend();

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : exitRec
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : exit receiving if this was not done before
@MD                  this must be called at last position in every 
@MD                  receive method of the cppvmObjects
@MD                  if the sending mode was 'BLOCKING', send a handshake
@MD                  for successful receiving back
@MU  USAGE         : exitRec(cppvmReceiveStream& rec, int msgTag);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : rec     description of the pvm receiving process
@MP            msgTag  message tag (message identifier) of the object
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::exitRec(cppvmReceiveStream& rec, int msgTag)
{
  // is this object not the initial send object
  if(msgTag != pvmObjTagInitRec ||
     getCPPvmObjId()  != pvmObjIdInitRec){
    return;
  }

  // reset values
  pvmInitializedRec = false;
  cppvmTransferRec = false;

  int oldSendBuf;
  int newSendBuf;
  int info;
  
  // tid of this process
  int thisProcessTid = pvm_mytid();
  if(thisProcessTid < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                        "cppvmObject",
                        "exitRec",
                        "cannot get my tid (pvm enrolled?)"));
  }

  // blocking receive was indicated by the sent handshake
  // send handshake back only for other tid's not for the own one!
  if((pvmRecHandshake == CPPvmExchangeBlockingSend) &&
     (thisProcessTid != rec.getLastRecFromTid())){
    oldSendBuf = pvm_getsbuf();
    // use a new created message buffer for sending
    newSendBuf = pvm_mkbuf(PvmDataDefault);
    if(newSendBuf < 0){
      if(newSendBuf == PvmBadParam){
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                  "cppvmObject",
                                  "exitRec",
                                  "giving an invalid encoding value"));
      }
      else if(newSendBuf == PvmNoMem){	
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                  "cppvmObject",
                                  "exitRec",
                                  "not enough memory to create the buffer"));
      }
      else{
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                          "cppvmObject",
                                                          "exitRec",
                                                          "cannot create new buffer"));
      }
    }
    pvm_setsbuf(newSendBuf);

    if((pvm_pkint(&pvmRecHandshake, 1, 1)) < 0) {
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmObject",   
                                                  "exitRec",
                                                  "cannot pack receive handshake"));
    }                                                      

    info = pvm_send(rec.getLastRecFromTid(), CPPvmHandshakeMsgTag);

    // send error ?
    if(info < 0){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmObject",   
                                                  "exitRec",
                                                  "cannot send data properly"));
    }
    // reset send buffers
    pvm_setsbuf(oldSendBuf);
    pvm_freebuf(newSendBuf);
  }
  else if(pvmRecHandshake!=CPPvmExchangeNonblockingSend){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "exitRec",
                                                "unknown send handshake"));
  }
  // reset context
  pvm_setcontext(pvmInitRecContext);

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : initRec
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : initialize the receiving buffer if the object was the
@MD                  initial receiving object
@MD                  this must be called at the first position in every
@MD                  receive method of the cppvmObjects
@MU  USAGE         : initRec(cppvmSendStream& send, int msgTag)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag  message tag (message identifier) of the object
@MP  INOUT   : rec     description of the pvm receiving process
@MP  OUT     : 
@MP  RETURNS : bool    message received or not (nonblocking, timeout)
@MP                    always true for blocking receive
@MP                    always true if the object was not the initial
@MP                           receiving object
******************************************************************************/
bool
cppvmObject::initRec(cppvmReceiveStream& rec, int msgTag)
{
  // allow only the first receiving object to initialize a receiving
  if(pvmInitializedRec){
    return true;
  }
  // set context
  pvmInitRecContext = pvm_setcontext(rec.getContext());
  
  cppvmTransferRec = true;
  pvmInitializedRec = true;
  // save the unique (!) object id of the initializing object
  // this will be used later to recognize the object within 'exitSend'
  pvmObjIdInitRec = getCPPvmObjId();
  pvmObjTagInitRec = msgTag;

  int bufid=-1;
  int dummy2, dummy3, source;
  
  if(rec.getRecMode() == CPPvmBlockingRec){
    bufid = pvm_recv(rec.getRecFromTid(), msgTag);
  }
  else if(rec.getRecMode() == CPPvmNonblockingRec){
    bufid = pvm_nrecv(rec.getRecFromTid(), msgTag);
    if(bufid == 0){
      pvmInitializedRec = false;
      cppvmTransferRec = false;
      // reset context
      pvm_setcontext(pvmInitRecContext);
      return false;
    }
  }
  else if(rec.getRecMode() == CPPvmTimeoutRec){
    bufid = pvm_trecv(rec.getRecFromTid(), msgTag, rec.getTimeoutDirect() );
    if(bufid == 0){
      cppvmTransferRec = false;
      pvmInitializedRec = false;
      // reset context
      pvm_setcontext(pvmInitRecContext);
      return false;
    }
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmObject",   
                                                   "initRec",
                                                   "unknown receive mode"));
  }
  

  if(bufid < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "initRec",
                                                "cannot receive a message"));
  }

  pvm_bufinfo(bufid, &dummy2, &dummy3, &source);
  
  // store the information of the sending process
  // this is useful to obtain information about the sending process
  // in case of 'any' receive (tid = -1)
  rec.setLastRecFromTid(source);
  // save the information, weather the send wants a handshake back or not
  if((pvm_upkint(&pvmRecHandshake, 1, 1)) < 0) {
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "initRec",
                                                "cannot unpack receive handshake"));
  }                                                      

  return true;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : pvmSend
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : send a pvm object
@MU  USAGE         : cppvmSend(cppvmSendStream& sendStrm);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : sendStrm   description of the pvm "send to" process
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSend(cppvmSendStream& sendStrm)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    cppvmSendTry(sendStrm);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : pvmSendTry
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : send a pvm object
@MU  USAGE         : cppvmSendTry(cppvmSendStream& sendStrm);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : sendStrm   description of the pvm "send to" process
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSendTry(cppvmSendStream& sendStrm)
{
  // initialize sending (if this object is the initializing object)
  initSend(sendStrm, cppvmMsgTag());

  // pack object data
  cppvmTransfer();

  // send data and reset values (if this object is the initializing object)
  exitSend(sendStrm, cppvmMsgTag());

}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : pvmReceive
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : receive a pvm object
@MU  USAGE         : cppvmReceive(cppvmReceiveStream& recStrm);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : recStrm   description of the pvm "receive from" process
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : bool  true:  object received
@MP                  false: not received (nonblocking receive or timeout)
******************************************************************************/
bool
cppvmObject::cppvmReceive(cppvmReceiveStream& recStrm)
{
  bool recRes;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    recRes = cppvmReceiveTry(recStrm);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
    return false;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return recRes;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : pvmReceiveTry
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : receive a pvm object
@MU  USAGE         : cppvmReceiveTry(cppvmReceiveStream& recStrm);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : recStrm   description of the pvm "receive from" process
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : bool  true:  object received
@MP                  false: not received (nonblocking receive or timeout)
******************************************************************************/
bool
cppvmObject::cppvmReceiveTry(cppvmReceiveStream& recStrm)
{
  // initialize sending (if this object is the initializing object)
  // return 'false' if the message has not arrived yes
  // (nonblocking receive or timout receive)
  if(initRec(recStrm, cppvmMsgTag())==false){
    return false;
  }

  // unpack object data
  cppvmTransfer();
  
  // exit receive and reset values (if this object is the initializing object)
  exitRec(recStrm, cppvmMsgTag());

  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmDistObj
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : distributed object (special kind of message box object)
@MD                  set all message mailbox variables for read/write
@MD                  objects from/to the message mailbox
@MU  USAGE         : cppvmDistObj(char *keyName,
@MU                             bool persistent,
@MU                             int sendEncoding,
@MU                             bool useOwnSendBuffer);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyName    database key, any null-terminated string
@MP                       must not contain regular expression chars (*, ?)
@MR            persistent true: distributed object
@MR                         entry remains in the database when the owner task
@MR                         exits. Entries are removed from the database when
@MR                         PVM is halted or a reset is issued from the console
@MR                       false: distributed object is removed from database if
@MR                         the task has ended
@MR                         produce the same results as using PvmMboxDefault
@MR            sendEncoding PvmDataDefault: XDR encoding
@MP                         PvmDataRaw:     no encoding
@MP                                         only for homogenous architectures
@MP                         PvmDataInPlace: data left in place during packing;
@MP                                         not allowed for shared memory arch.
@MP            useOwnSendBuffer false: default send buffer is used
@MP                             true: a new send buffer is created and deleted
@MP                                   after sending the message
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmDistObj(char *keyName,
                          bool persistent,
                          int sendEncoding,
                          bool useOwnSendBuffer)
{
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    cppvmDistObjTry(keyName, persistent,
                    sendEncoding, useOwnSendBuffer);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmDistObjTry
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : distributed object (special kind of message box object)
@MD                  set all message mailbox variables for read/write
@MD                  objects from/to the message mailbox
@MU  USAGE         : cppvmDistObjTry(char *keyName,
@MU                             bool persistent,
@MU                             int sendEncoding,
@MU                             bool useOwnSendBuffer);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyName    database key, any null-terminated string
@MP                       must not contain regular expression chars (*, ?)
@MR            persistent true: distributed object
@MR                         entry remains in the database when the owner task
@MR                         exits. Entries are removed from the database when
@MR                         PVM is halted or a reset is issued from the console
@MR                       false: distributed object is removed from database if
@MR                         the task has ended
@MR                         produce the same results as using PvmMboxDefault
@MR            sendEncoding PvmDataDefault: XDR encoding
@MP                         PvmDataRaw:     no encoding
@MP                                         only for homogenous architectures
@MP                         PvmDataInPlace: data left in place during packing;
@MP                                         not allowed for shared memory arch.
@MP            useOwnSendBuffer false: default send buffer is used
@MP                             true: a new send buffer is created and deleted
@MP                                   after sending the message
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmDistObjTry(char *keyName,
                             bool persistent,
                             int sendEncoding,
                             bool useOwnSendBuffer)
{
  int writeFlags = PvmMboxDefault + PvmMboxOverWritable;

  if(persistent==true){
    writeFlags = writeFlags + PvmMboxPersistent;
  }
  
  mboxInitWrite=true;
  mboxInitRead=true;
  cppvmSetMboxKey(keyName, 0);
  cppvmSetMboxSend(sendEncoding, useOwnSendBuffer);
  mboxFlagsRead = PvmMboxDefault;
  mboxFlagsWrite = writeFlags;

  ///////////////////////////////////////////////////////////////////////////
  // the next lines would be the more beautyful solution
  // unfortunately, gcc 2.7.2.3.x does not support REAL exception handling
  //
  // is there already a distributed object with this key in the database?
  //CPPvmInternalTry {
  //  // try to read the object from database
  //  cppvmReadMboxInfoTry();
  //
  //}
  //CPPvmInternalCatch(cppvmException& exception) {
  //  // there was no such entry in the database (or other error)
  //  // anyway, try to send this object to the database
  //  cppvmWriteMboxInfoTry();
  //}
  ///////////////////////////////////////////////////////////////////////////


  int newRBuf;	

  // read data from message box
  // the receive buffer is already set to newRBuf after calling this method
  // is is not necessary to call pvm_setrbuf()
  newRBuf = cppvmReadMboxInfoBase();

  if(newRBuf==-1){
    
    // there was no such entry in the database
    // try to send this object to the database
    cppvmWriteMboxInfoTry();
  }
  else{
    // indicate 'cppvmTransfer' to unpack data
    cppvmTransferRec = true;

    // unpack data	
    cppvmTransfer();
    
    // reset values
    cppvmTransferRec = false;
    
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetMbox
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set all message mailbox variables for read/delete/write
@MD                  objects from/to the message mailbox
@MU  USAGE         : cppvmSetMbox(char *keyName,
@MU                             int writeFlags,
@MU                             int readFlags,
@MU                             int delFlags,
@MU                             int sendEncoding,
@MU                             bool useOwnSendBuffer);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyName    database key, any null-terminated string
@MP                       must not contain regular expression chars (*, ?)
@MR            writeFlags options for write into database, the options can be
@MR                       combined by adding them together (+)
@MR                       PvmMboxDefault: 
@MR                         inserts entry as the only named instance for a 
@MR                         given name. This entry may only be modified and 
@MR                         deleted by its owner. It is auto matically deleted 
@MR                         when its owner exits
@MR                       PvmMboxPersistent:
@MR                         entry remains in the database when the owner task
@MR                         exits. Entries are removed from the database when
@MR                         PVM is halted or a reset is issued from the console
@MR                       PvmMboxMultiInstance:
@MR                         permits multiple entry instances of the same name.
@MR                         PVM will assign an index key to each instance.
@MR                       PvmMboxOverWritable:  
@MR                         permits other tasks to overwrite and delete this
@MR                         database entry
@MR            readFlags  options for read from the database, the options can 
@MR                       be combined by adding them together (+)
@MR                       PvmMboxDefault:
@MR                         exact match on key <name, index> is returned. 
@MR                       PvmMboxFirstAvail:  
@MR                         the first entry in <name> with index greater than
@MR                         or equal to the specified index parameter is
@MR                         returned. PvmMboxFirstAvail with index=0 will
@MR                         produce the same results as using PvmMboxDefault
@MR                       PvmMboxReadAndDelete:  
@MR                         return entry and delete from database. Task must
@MR                         be permitted to do both read and delete, otherwise
@MR                         an error will occur. 
@MR            delFlags  options for delete entries from the database
@MR                      no flags presently specified   
@MR            sendEncoding PvmDataDefault: XDR encoding
@MP                         PvmDataRaw:     no encoding
@MP                                         only for homogenous architectures
@MP                         PvmDataInPlace: data left in place during packing;
@MP                                         not allowd for shared memory arch.
@MP            useOwnSendBuffer false: default send buffer is used
@MP                             true: a new send buffer is created and deleted
@MP                                   after sending the message
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetMbox(char *keyName,
                          int writeFlags,
                          int readFlags,
                          int delFlags,
                          int sendEncoding,
                          bool useOwnSendBuffer)
{
  mboxInitDel=true;
  mboxInitWrite=true;
  mboxInitRead=true;
  cppvmSetMboxKey(keyName, 0);
  cppvmSetMboxSend(sendEncoding, useOwnSendBuffer);
  mboxFlagsRead = readFlags;
  mboxFlagsWrite = writeFlags;
  mboxFlagsDel = delFlags;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetMboxRead
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set all message mailbox variables for read
@MD                  objects from the message mailbox
@MU  USAGE         : cppvmSetMboxRead(char *keyName,
@MU                                 int readFlags);
@MU                             
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyName    database key, any null-terminated string
@MP                       must not contain regular expression chars (*, ?)
@MR            readFlags  options for read from the database, the options can 
@MR                       be combined by adding them together (+)
@MR                       PvmMboxDefault:
@MR                         exact match on key <name, index> is returned. 
@MR                       PvmMboxFirstAvail:  
@MR                         the first entry in <name, index> with index 
@MR                         greater than
@MR                         or equal to the specified index parameter is
@MR                         returned. PvmMboxFirstAvail with index=0 will
@MR                         produce the same results as using PvmMboxDefault
@MR                       PvmMboxReadAndDelete:  
@MR                         return entry and delete from database. Task must
@MR                         be permitted to do both read and delete, otherwise
@MR                         an error will occur. 
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetMboxRead(char *keyName,
                              int readFlags)
{
  mboxInitRead=true;
  cppvmSetMboxKey(keyName, 0);
  mboxFlagsRead = readFlags;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetMboxWrite
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set all message mailbox variables for write
@MD                  objects to the message mailbox
@MU  USAGE         : cppvmSetMboxWrite(char *keyName,
@MU                                  int writeFlags,
@MU                                  int sendEncoding);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyName    database key, any null-terminated string
@MP                       must not contain regular expression chars (*, ?)
@MR            writeFlags options for write into database, the options can be
@MR                       combined by adding them together (+)
@MR                       PvmMboxDefault: 
@MR                         inserts entry as the only named instance for a 
@MR                         given name. This entry may only be modified and 
@MR                         deleted by its owner. It is auto matically deleted 
@MR                         when its owner exits
@MR                       PvmMboxPersistent:
@MR                         entry remains in the database when the owner task
@MR                         exits. Entries are removed from the database when
@MR                         PVM is halted or a reset is issued from the console
@MR                       PvmMboxMultiInstance:
@MR                         permits multiple entry instances of the same name.
@MR                         PVM will assign an index key to each instance.
@MR                       PvmMboxOverWritable:  
@MR                         permits other tasks to overwrite and delete this
@MR                         database entry
@MR            sendEncoding PvmDataDefault: XDR encoding
@MP                         PvmDataRaw:     no encoding
@MP                                         only for homogenous architectures
@MP                         PvmDataInPlace: data left in place during packing;
@MP                                         not allowd for shared memory arch.
@MP            useOwnSendBuffer false: default send buffer is used
@MP                             true: a new send buffer is created and deleted
@MP                                   after sending the message
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetMboxWrite(char *keyName,
                               int writeFlags,
                               int sendEncoding,
                               bool useOwnSendBuffer)
{
  mboxInitWrite=true;
  cppvmSetMboxKeyName(keyName);
  cppvmSetMboxSend(sendEncoding, useOwnSendBuffer);
  mboxFlagsWrite = writeFlags;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetMboxDel
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set all message mailbox variables for delete
@MD                  objects from the message mailbox
@MU  USAGE         : cppvmSetMboxDel(char *keyName,
@MU                                int delFlags);
@MU                             
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyName    database key, any null-terminated string
@MP                       must not contain regular expression chars (*, ?)
@MR            delFlags  options for delete entries from the database
@MR                      no flags presently specified   
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetMboxDel(char *keyName,
                             int delFlags)
{
  mboxInitDel=true;
  cppvmSetMboxKey(keyName, 0);
  mboxFlagsDel = delFlags;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetMboxKey
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set all message mailbox keys
@MU  USAGE         : cppvmSetMboxKey(char *keyName,
@MU                                 int keyIndex);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyName    database key, any null-terminated string
@MP                       must not contain regular expression chars (*, ?)
@MR            keyIndex   database key, default index = 0
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetMboxKey(char *keyName,
		          int keyIndex)
{
  cppvmSetMboxKeyName(keyName);
  cppvmSetMboxKeyIndex(keyIndex);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetMboxKeyName
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set message mailbox name key
@MU  USAGE         : cppvmSetMboxKeyName(char *keyName);
@MU                             
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyName    database key, any null-terminated string
@MP                       must not contain regular expression chars (*, ?)
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetMboxKeyName(char *keyName)
{
  if(keyName == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmObject",   
                                                   "cppvmSetMboxKeyName",
                                                   "keyName is NULL"));
  }
  if(mboxKeyName!=NULL){
    // delete allocated string
    delete [] mboxKeyName;
  }
  // allocate memory new string (char number plus '\0')
  mboxKeyName = new char [(strlen(keyName) + 1)];
  // copy string
  strcpy(mboxKeyName, keyName);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetMboxKeyIndex
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set message mailbox index key
@MU  USAGE         : cppvmSetMboxKeyIndex(int keyIndex);
@MU                             
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : keyIndex   database key, default index = 0
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetMboxKeyIndex(int keyIndex)
{
  mboxKeyIndex = keyIndex;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetMboxSend
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set message mailbox send encoding and message buffer
@MU  USAGE         : cppvmSetMboxSend(int sendEncoding,
@MU                                 bool useOwnSendBuffer);
@MU                             
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : sendEncoding PvmDataDefault: XDR encoding
@MP                         PvmDataRaw:     no encoding
@MP                                         only for homogenous architectures
@MP                         PvmDataInPlace: data left in place during packing;
@MP                                         not allowd for shared memory arch.
@MP            useOwnSendBuffer false: default send buffer is used
@MP                             true: a new send buffer is created and deleted
@MP                                   after sending the message
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetMboxSend(int sendEncoding,
			  bool useOwnSendBuffer)
{
  if(sendEncoding != PvmDataDefault &&
     sendEncoding != PvmDataRaw     &&
     sendEncoding != PvmDataInPlace){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmObject",   
                                                   "cppvmSetMboxSendEncoding",
                                                   "invalid sendEncoding argument"));
  }    
  mboxSendEncoding = sendEncoding;  
  mboxOwnSendBuf = useOwnSendBuffer;
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmWriteMboxInfoBase
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : write information in the database
@MD                  with the current 'keyName' and 'writeFlags'
@MU  USAGE         : cppvmWriteMboxInfoBase();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmWriteMboxInfoBase()
{
  int info;

  info = pvm_putinfo(mboxKeyName, 
                     pvm_getsbuf(),
	             mboxFlagsWrite);
  if(info == PvmBadParam ||
     info == PvmNoSuchBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmWriteMboxInfoBase",
                       "invalid or non existant send buffer. Maybe you have set a new message buffer with 'cppvmSetMboxMsgBufSend(buffer)' and deleted 'buffer' afterwards."));
  }
  else if(info == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmWriteMboxInfoBase",
                       "unable to allocate memory to pack data"));
  }
  else if(info == PvmExists){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmWriteMboxInfoBase",
                       "the requested key already exist"));
  }
  else if(info == PvmDenied){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmWriteMboxInfoBase",
                       "the key is locked by another task and cannot be replaced or deleted"));
  }
  else if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "cppvmWriteMboxInfoBase",
                                                "error writing in database"));
  }
  

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmReadMboxInfoBase
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : read information from the database
@MD                  with the current 'keyName', 'keyIndex' and 'readFlags'
@MU  USAGE         : int cppvmReadMboxInfoBase();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : int  -1:   the requested <name, index> entry does not exist
@MP                 else: receive buffer id
******************************************************************************/
int
cppvmObject::cppvmReadMboxInfoBase()
{
  int bufid;

  bufid = pvm_recvinfo(mboxKeyName, 
                       mboxKeyIndex,
	               mboxFlagsRead);
  if(bufid == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmReadMboxInfoBase",
                       "unable to allocate memory to pack data"));
  }
  else if(bufid == PvmNotFound){
    return -1;
  }
  else if(bufid == PvmDenied){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmReadMboxInfoBase",
                       "the key is locked by another task and cannot be replaced or deleted"));
  }
  else if(bufid < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "cppvmReadMboxInfoBase",
                                                "error reading from database"));
  }
  return bufid;

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmWriteMboxInfo
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : write (pack+send) information in the database
@MD                  with the current 'keyName' and 'writeFlags'
@MU  USAGE         : cppvmWriteMboxInfo();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmWriteMboxInfo()
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    cppvmWriteMboxInfoTry();
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmWriteMboxInfoTry
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : write (pack+send) information in the database
@MD                  with the current 'keyName' and 'writeFlags'
@MU  USAGE         : cppvmWriteMboxInfoTry();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmWriteMboxInfoTry()
{
  if(mboxInitWrite==false){
    // mbox was not initialized, do nothing
    return;
  }

  int oldSBuf=0;
  int newSBuf=0;
  //int info;

  // set send buffer if initialized
  if(mboxOwnSendBuf==true){
    oldSBuf = pvm_getsbuf();
    newSBuf = pvm_mkbuf(mboxSendEncoding);
    if(newSBuf < 0){
      if(newSBuf == PvmBadParam){
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                           "cppvmObject",
                           "cppvmWriteMboxInfoTry",
                           "giving an invalid encoding value"));
      }
      else if(newSBuf == PvmNoMem){	
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                           "cppvmObject",
                           "cppvmWriteMboxInfoTry",
                           "not enough memory to create the send buffer"));
      }
      else{
	      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                           "cppvmObject",
                           "cppvmWriteMboxInfoTry",
                           "cannot create new send buffer"));
      }
    }
    pvm_setsbuf(newSBuf);
  }
  else{
    pvm_initsend(mboxSendEncoding);
  }
  // indicate 'cppvmTransfer' to pack data
  cppvmTransferSend = true;

  // pack data	
  cppvmTransfer();

  // put data into message box
  cppvmWriteMboxInfoBase();

  // reset values
  cppvmTransferSend = false;
  if(mboxOwnSendBuf==true){
    pvm_setsbuf(oldSBuf);
    pvm_freebuf(newSBuf);
  }
  // this is needed to delete the allocated memory after sending
  // allocated by transfer functions
  cppvmCleanMemoryAfterSend();
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmReadMboxInfo
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : read (receive + unpack) information from the database
@MD                  with the current 'keyName', 'keyIndex' and 'readFlags'
@MU  USAGE         : cppvmReadMboxInfo(int keyIndex);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MR  IN      : keyIndex   database key index, default index = 0
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmReadMboxInfo(int keyIndex)
{ 
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    cppvmReadMboxInfoTry(keyIndex);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmReadMboxInfoTry
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : read (receive + unpack) information from the database
@MD                  with the current 'keyName', 'keyIndex' and 'readFlags'
@MU  USAGE         : cppvmReadMboxInfoTry(int keyIndex);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MR  IN      : keyIndex   database key index, default index = 0
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmReadMboxInfoTry(int keyIndex)
{ 
  if(mboxInitRead==false){
    // mbox was not initialized, do nothing
    return;
  }

  // set key index
  cppvmSetMboxKeyIndex(keyIndex);
  
  int newRBuf;	

  // indicate 'cppvmTransfer' to unpack data
  cppvmTransferRec = true;

  // read data from message box
  // the receive buffer is already set to newRBuf after calling this method
  // is is not necessary to call pvm_setrbuf()
  newRBuf = cppvmReadMboxInfoBase();

  if(newRBuf==-1){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                "cppvmObject",   
                                "cppvmReadMboxInfoTry",
                                "the requested <name, index> entry does not exist"));
  }

  // unpack data	
  cppvmTransfer();

  // reset values
  cppvmTransferRec = false;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmDelMboxInfo
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : delete information from the database
@MD                  with the current 'keyName', 'keyIndex' and 'delFlags'
@MU  USAGE         : cppvmDelMboxInfo(int keyIndex);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MR  IN      : keyIndex   database key index, default index = 0
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmDelMboxInfo(int keyIndex)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    cppvmDelMboxInfoTry(keyIndex);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmDelMboxInfoTry
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : delete information from the database
@MD                  with the current 'keyName', 'keyIndex' and 'delFlags'
@MU  USAGE         : cppvmDelMboxInfoTry(int keyIndex);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MR  IN      : keyIndex   database key index, default index = 0
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmDelMboxInfoTry(int keyIndex)
{
  if(mboxInitDel==false){
    // mbox was not initialized, do nothing
    return;
  }

  // set key index
  cppvmSetMboxKeyIndex(keyIndex);

  int info;

  info = pvm_delinfo(mboxKeyName, 
                     mboxKeyIndex,
		     mboxFlagsDel);
  if(info == PvmNotFound){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmDelMboxInfoTry",
                       "requested key does not exist"));
  }
  else if(info == PvmDenied){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                "cppvmObject",   
                "cppvmDelMboxInfoTry",
                "the key is locked by another task and cannot be replaced or deleted"));
  }
  else if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmDelMboxInfoTry",
                       "error deleting from database"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmGetMboxInfo
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get information about the database entry
@MD                  with the current 'keyName' 
@MU  USAGE         : cppvmGetMboxInfo(struct pvmmboxinfo *classInfo);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 'classes' is allocated by this method
@MR                  and freed on the next call of this method or the
@MR                  call of the other method 'cppvmGetMboxInfo'
@MR                  struct pvmmboxinfo{
@MR                    char *mi_name;      // class name
@MR                    int   mi_nentries;  // # of entries for this class
@MR                    int  *mi_indices;   // mbox entry indices
@MR                    int  *mi_owners;    // mbox entry owner tids
@MR                    int  *mi_flags;     // mbox entry flags
@MR                  }  
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : classInfo  pvmmboxinfo mailbox entry matching current 'keyName'
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmGetMboxInfo(struct pvmmboxinfo *classInfo)
{
  int nclasses;
  struct pvmmboxinfo *classes=NULL;

  cppvmGetMboxInfo(mboxKeyName, &nclasses, &classes);

  cout << "pvm_obj.cc : nclasses: " << nclasses << endl;

  if(nclasses <= 0){
    CPPvmInternalThrow( cppvmExceptionForbiddenCall(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmGetMboxInfo",
                       "An object with the key 'mboxKeyName' is not registered in the database."));
  }
  if(nclasses > 1){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmObject",   
                       "cppvmGetMboxInfo",
                       "more than one entry for 'mboxKeyName'. 'mboxKeyName' must contain unallowed regular expression patterns "));
  }

  // copy the first (and only!) result 
  classInfo->mi_name = classes[0].mi_name;
  classInfo->mi_nentries = classes[0].mi_nentries;
  classInfo->mi_indices = classes[0].mi_indices;
  classInfo->mi_owners = classes[0].mi_owners;
  classInfo->mi_flags = classes[0].mi_flags;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmGetMboxInfo
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get information about database entries
@MU  USAGE         : cppvmGetMboxInfo(char *pattern,
@MU                                 int *nclasses,
@MU                                 struct pvmmboxinfo **classes);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 'classes' is allocated by this method
@MR                  and freed on the next call of this method or the
@MR                  call of the other method 'cppvmGetMboxInfo'
@MR                  struct pvmmboxinfo{
@MR                    char *mi_name;      // class name
@MR                    int   mi_nentries;  // # of entries for this class
@MR                    int  *mi_indices;   // mbox entry indices
@MR                    int  *mi_owners;    // mbox entry owner tids
@MR                    int  *mi_flags;     // mbox entry flags
@MR                  }  
@M#  --------------------------------------------------------------------------
@MP  IN      : pattern  GNU regular expression to match on names in mailbox
@MP                     database. Additionally, the singular '*' will match
@MP                     all names
@MP  INOUT   : 
@MP  OUT     : nclasses number of classes matching pattern
@MP            classes  array of pvmmboxinfo mailbox entries matching pattern
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmGetMboxInfo(char *pattern,
		          int *nclasses,
		          struct pvmmboxinfo **classes)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    cppvmGetMboxInfoTry(pattern, nclasses, classes);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmGetMboxInfoTry
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get information about database entries
@MU  USAGE         : cppvmGetMboxInfoTry(char *pattern,
@MU                                 int *nclasses,
@MU                                 struct pvmmboxinfo **classes);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 'classes' is allocated by this method
@MR                  and freed on the next call of this method or the
@MR                  call of the other method 'cppvmGetMboxInfo'
@MR                  struct pvmmboxinfo{
@MR                    char *mi_name;      // class name
@MR                    int   mi_nentries;  // # of entries for this class
@MR                    int  *mi_indices;   // mbox entry indices
@MR                    int  *mi_owners;    // mbox entry owner tids
@MR                    int  *mi_flags;     // mbox entry flags
@MR                  }  
@M#  --------------------------------------------------------------------------
@MP  IN      : pattern  GNU regular expression to match on names in mailbox
@MP                     database. Additionally, the singular '*' will match
@MP                     all names
@MP  INOUT   : 
@MP  OUT     : nclasses number of classes matching pattern
@MP            classes  array of pvmmboxinfo mailbox entries matching pattern
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmGetMboxInfoTry(char *pattern,
                                 int *nclasses,
                                 struct pvmmboxinfo **classes)
{
  int info;
  info = pvm_getmboxinfo(pattern, nclasses, classes);
  if(info == PvmBadParam){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "cppvmGetMboxInfoTry",
                                                "invalid argument"));
  }    
  else if(info == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmObject",   
                                                "cppvmGetMboxInfoTry",
                                                "unable to allocate memory"));
  }    
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmResizeHandshakeArray
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : resize the handshake array; copy old entries,
@MD                  leave the current entries number unchanged
@MD                  do nothing if the new size is smaller than the old size
@MU  USAGE         : cppvmResizeHandshakeArray(unsigned int newLen);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : The size of the array can only get bigger, not smaller
@M#  --------------------------------------------------------------------------
@MP  IN      : newLen  minimal new length of the handshake array
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmResizeHandshakeArray(unsigned int newLen)
{
  // it is not possible to make the size smaller
  if(newLen <= pvmSendHandshakeTidsSize){
    return;
  }
  unsigned int i;
  unsigned int newArrayLen;
  int *newArr;

  // * 2 of the current array size
  // or more (if needed)
  if(newLen > (pvmSendHandshakeTidsSize * 2)){
    newArrayLen = newLen;
  }
  else{
    newArrayLen = pvmSendHandshakeTidsSize * 2;
  }
  
  // allocate new array
  newArr = new int[newArrayLen];

  // copy current tids
  for(i=0; i<pvmSendHandshakeTidsNum; i++){
    newArr[i] = pvmSendHandshakeTids[i];
  }

  // replace old array with new array
  pvmSendHandshakeTidsSize = newArrayLen;
  delete [] pvmSendHandshakeTids;
  pvmSendHandshakeTids = newArr;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmAddToHandshakeArray
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add the tid to the handshake tid array
@MD                  resize the array if necessary
@MD                  tid's which are already in the array are NOT added
@MU  USAGE         : cppvmAddToHandshakeArray(int tid);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : tid  tid to add to the handshake array
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmAddToHandshakeArray(int tid)
{
  unsigned int i;
  for(i=0;i<pvmSendHandshakeTidsNum;i++){
    if(pvmSendHandshakeTids[i] == tid){
      return;
    }
  }

  // resize array (if necessary)
  cppvmResizeHandshakeArray(pvmSendHandshakeTidsNum+1);

  // add element
  pvmSendHandshakeTids[pvmSendHandshakeTidsNum]=tid;
  // increment number of elements
  pvmSendHandshakeTidsNum++;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetHandshakeTidArrayGroupBroadcast
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the handshake array for blocking send using
@MD                  group broadcasting
@MU  USAGE         : cppvmSetHandshakeTidArrayGroupBroadcast(cppvmSendStream& send);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : send   description of the pvm sending process
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetHandshakeTidArrayGroupBroadcast(cppvmSendStream& send){
  struct pvmtaskinfo *taskp;
  int ntask;
  int i;
  int info;
  
  // reset old handshake array
  cppvmResetHandshakeArray();

  // tid of this process
  int thisProcessTid = pvm_mytid();
  if(thisProcessTid < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                "cppvmObject",
                                "cppvmSetHandshakeTidArrayGroupBroadcast",
                                "cannot get my tid (pvm enrolled?)"));
  }
  // information for all tasks on the virtual machine
  info = pvm_tasks(0, &ntask, &taskp);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                        "cppvmConnection",
                        "cppvmSetHandshakeTidArrayGroupBroadcast",
                        "could not get task informations of the virtual machine."));
  }
  // loop for all tasks in the virtual machine
  for(i=0;i<ntask;i++){
    if((thisProcessTid != taskp[i].ti_tid) &&
       (pvm_getinst(send.getGroup(), taskp[i].ti_tid) >= 0)){
      // the task is not equal to this task and
      // it belongs to the specified group
      // so wait for a handshake from this task
      cppvmAddToHandshakeArray(taskp[i].ti_tid);
    }
  }
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSetHandshakeTidArrayOneOrMulticast
@MK  CLASS NAME    : cppvmObject
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the handshake array for blocking send using
@MD                  one to one or muticast sending
@MU  USAGE         : cppvmSetHandshakeTidArrayOneOrMulticast(cppvmSendStream& send);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : send   description of the pvm sending process
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmObject::cppvmSetHandshakeTidArrayOneOrMulticast(cppvmSendStream& send)
{
  int i;
  
  // reset old handshake array
  cppvmResetHandshakeArray();

  // tid of this process
  int thisProcessTid = pvm_mytid();
  if(thisProcessTid < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                "cppvmObject",
                                "cppvmSetHandshakeTidArrayOneOrMulticast",
                                "cannot get my tid (pvm enrolled?)"));
  }
  
  // loop for all contacted tids
  for(i=0; i<send.getNumTid(); i++){
    if(send[i] != thisProcessTid){
      // do not allow a handshake for sending to the own task
      cppvmAddToHandshakeArray(send[i]);
    }
  }
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
