/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_proh.cpp 
@IE  CREATED     : 26.6.1998
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm process handling base class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_proh.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_proh.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmMaxId
@VD  DESCRIPTION   : unique id number for all framework objects
@VD                  highest actual ID
@VT  TYPE          : int
***
******************************************************************************/
int cppvmProcessHandling::cppvmMaxId = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmTotalObjNum
@VD  DESCRIPTION   : total number of cppvmObjects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmProcessHandling::cppvmTotalObjNum = 0; 

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmProcessHandling
@MK  CLASS NAME    : cppvmProcessHandling
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmProcessHandling
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmProcessHandling::cppvmProcessHandling()
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmProcessHandling
@MK  CLASS NAME    : cppvmProcessHandling
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmProcessHandling::cppvmProcessHandling (const cppvmProcessHandling& /*obj*/)
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmProcessHandling
@MK  CLASS NAME    : cppvmProcessHandling
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmProcessHandling()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      :
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
cppvmProcessHandling::~cppvmProcessHandling()
{
  // dec total number of objects
  cppvmTotalObjNum--;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmProcessHandling
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  info for configuration object attributes 
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmProcessHandling&
cppvmProcessHandling::operator= (const cppvmProcessHandling& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : startPvmDaemon
@MK  CLASS NAME    : cppvmProcessHandling
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : start a pvm daemon on this node if the daemon
@MD                  is not already running
@MU  USAGE         : void startPvmDaemon(int argc, char **argv, int block);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : argc   number of arguments for the daemon
@MP            argv   arguments for the daemon
@MP            block  specifying whether to block until startup of all hosts
@MP                   complete or return immediately.
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool: true:  done
@MP                  false: daemon is already started
******************************************************************************/
bool
cppvmProcessHandling::startPvmDaemon(int argc, char **argv, int block)
{
  int info;
  int oldval;

  // suppress PVM error messages
  // (e.g. "libpvm [t40002]: pvm_start_pvmd(): Duplicate host") 
  oldval = pvm_setopt(PvmAutoErr, 0);
  
  info = pvm_start_pvmd(argc, argv, block);

  // allow PVM error messages again
  pvm_setopt(PvmAutoErr, oldval);

  if(info == PvmSysErr){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmProcessHandling",
                                                "startPvmDaemon",
                                                "local pvmd is not responding"));
  }
  if(info == PvmDupHost){
    return false;
  }
  return true;
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
