/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_rec.cpp 
@IE  CREATED     : 5.12.1997
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : description of a pvm receive command
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// includes for predeclared classes to avoid cyclic includes
#include "pvm_task.h"
#include "pvm_conn.h"
#include "pvm_spaw.h"
#include "pvm_spmu.h"
// include own class
#include "pvm_rec.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_rec.cpp"


/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmReceiveStream
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : only use it if it is later on initialized by a
@MU                  "set" method
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream()
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set receive stream for 'cppvmTask' objects
@MU  USAGE         : cppvmReceiveStream(cppvmTask& task,
@MU                           cppvmRecTask mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task     description of the task to receive from
@MP            mode     receive mode:
@MP                     * CPPvmRBtask:  receive blocking
@MP                     * CPPvmRNBtask: receive nonblocking
@MP                     * CPPvmRTtask:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTtask
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTtask
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(cppvmTask& task,
                                       cppvmRecTask mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(task, mode, seconds, useconds);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set receive stream for 'cppvmConnection' objects
@MU  USAGE         : cppvmReceiveStream(cppvmConnection& conn,
@MU                           cppvmRecConnection mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : conn     description of the process to receive from
@MP            mode     receive mode:
@MP                     * CPPvmRBparent:  receive blocking
@MP                     * CPPvmRNBparent: receive nonblocking
@MP                     * CPPvmRTparent:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTparent
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTparent
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(cppvmConnection& conn,
                                       cppvmRecConnection mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(conn, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set receive stream for 'cppvmSpawnConnection' objects
@MU  USAGE         : cppvmReceiveStream(cppvmSpawnConnection& sconn,
@MU                           cppvmRecSpawnConnection mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : conn     description of the process to receive from
@MP            mode     receive mode:
@MP                     * CPPvmRBchild:  receive blocking
@MP                     * CPPvmRNBchild: receive nonblocking
@MP                     * CPPvmRTchild:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTchild
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTchild
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(cppvmSpawnConnection& sconn,
                                       cppvmRecSpawnConnection mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(sconn, mode, seconds, useconds);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for processes specified by 'taskId'
@MD                  with current context
@MU  USAGE         : cppvmReceiveStream(int taskId,
@MU                           cppvmRecTid mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskId   task id of the process to receive from
@MP            mode     receive mode:
@MP                     * CPPvmRBtid:  receive blocking
@MP                     * CPPvmRNBtid: receive nonblocking
@MP                     * CPPvmRTtid:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTtid
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTtid
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(int taskId,
                                       cppvmRecTid mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(taskId, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for processes specified by 'taskId'
@MD                  with current context
@MU  USAGE         : cppvmReceiveStream(int taskId, int context_,
@MU                           cppvmRecTid mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskId   task id of the process to receive from
@MP            context_ receive context
@MP            mode     receive mode:
@MP                     * CPPvmRBtid:  receive blocking
@MP                     * CPPvmRNBtid: receive nonblocking
@MP                     * CPPvmRTtid:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTtid
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTtid
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(int taskId, int context_,
                                       cppvmRecTid mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(taskId, context_, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for receive from any processes 
@MD                  with current context
@MU  USAGE         : cppvmReceiveStream(cppvmRecAny mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : mode     receive mode:
@MP                     * CPPvmRBany:  receive blocking
@MP                     * CPPvmRNBany: receive nonblocking
@MP                     * CPPvmRTany:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTany
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTany
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(cppvmRecAny mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for processes specified by 'taskId'
@MD                  with current context
@MU  USAGE         : cppvmReceiveStream(int context_,
@MU                           cppvmRecAny mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : context_ receive context
@MP            mode     receive mode:
@MP                     * CPPvmRBany:  receive blocking
@MP                     * CPPvmRNBany: receive nonblocking
@MP                     * CPPvmRTany:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTany
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTany
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(int context_,
                                       cppvmRecAny mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(context_, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for receive from the same processes, from which
@MD                  a receive was done by 'oldStrm'
@MD                  with current context
@MU  USAGE         : cppvmReceiveStream(cppvmReceiveStream& oldStrm,
@MU                           cppvmRecSame mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : oldStrm  receive from the same process as the last receive
@MD                     was done in this stream
@MD            mode     receive mode:
@MP                     * CPPvmRBsame:  receive blocking
@MP                     * CPPvmRNBsame: receive nonblocking
@MP                     * CPPvmRTsame:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTsame
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTsame
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(cppvmReceiveStream& oldStrm,
                                       cppvmRecSame mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(oldStrm, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for receive from the same processes, from which
@MD                  a receive was done by 'oldStrm'
@MD                  with current context
@MU  USAGE         : cppvmReceiveStream(cppvmReceiveStream& oldStrm,
@MU                           cppvmRecSame mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : oldStrm  receive from the same process as the last receive
@MD                     was done in this stream
@MD            mode     receive mode:
@MP                     * CPPvmRBsame:  receive blocking
@MP                     * CPPvmRNBsame: receive nonblocking
@MP                     * CPPvmRTsame:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTsame
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTsame
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream(cppvmReceiveStream& oldStrm,
                                       int context_,
                                       cppvmRecSame mode,
                                       int seconds, int useconds)
    : cppvmStream(), 
      initialized(false), lastRecResult(false)
{
  set(oldStrm, context_, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmReceiveStream
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmReceiveStream
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmReceiveStream::~cppvmReceiveStream()
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmReceiveStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream& 
cppvmReceiveStream::operator= (const cppvmReceiveStream& obj) 
{
  initialized = obj.initialized;
  //CPPvmRecThrowIfNotInit

  if(cppvmIsSame(obj)){
    return *this;
  }
  cppvmStream::operator=(obj);
  lastRecResult = obj.lastRecResult;
  recFromTid = obj.recFromTid;
  lastRecFromTid = obj.lastRecFromTid;
  timeout.tv_sec = obj.timeout.tv_sec;
  timeout.tv_usec = obj.timeout.tv_usec;
  recMode = obj.recMode;
  context = obj.context;
  
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmReceiveStream
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmReceiveStream::cppvmReceiveStream (const cppvmReceiveStream& obj)
    : cppvmStream(obj), 
      context(obj.context), initialized(obj.initialized),
      recFromTid(obj.recFromTid), lastRecFromTid(obj.lastRecFromTid),
      recMode(obj.recMode),
      lastRecResult(obj.lastRecResult)
      
{
  //CPPvmRecThrowIfNotInit
  timeout.tv_sec = obj.timeout.tv_sec;
  timeout.tv_usec = obj.timeout.tv_usec;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmReceiveStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmReceiveStream::operator== (const cppvmReceiveStream& obj) 
{
  CPPvmRecThrowIfNotInit
  return ( recFromTid        == obj.recFromTid      &&
           lastRecFromTid    == obj.lastRecFromTid  &&
           timeout.tv_sec    == obj.timeout.tv_sec  &&
           timeout.tv_usec   == obj.timeout.tv_usec &&
           recMode           == obj.recMode         &&
           context           == obj.context);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmReceiveStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmReceiveStream::operator== (const cppvmReceiveStream& obj) const
{
  CPPvmRecThrowIfNotInit
  return ( recFromTid        == obj.recFromTid      &&
           lastRecFromTid    == obj.lastRecFromTid  &&
           timeout.tv_sec    == obj.timeout.tv_sec  &&
           timeout.tv_usec   == obj.timeout.tv_usec &&
           recMode           == obj.recMode         &&
           context           == obj.context);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator!=
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are not equal
@MU  USAGE         : !=
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmReceiveStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmReceiveStream::operator!= (const cppvmReceiveStream& obj) 
{
  CPPvmRecThrowIfNotInit
    return ! (operator==(obj));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probe
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message of a certain tag has arrived
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MU  USAGE         : bool probe(int msgtag)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag message tag (for example cppvmFloat::cppvmClassMsgTag())
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : bool  true: message with this message tag has arrived
@MP                  false: no message with this message tag has arrived
******************************************************************************/
bool
cppvmReceiveStream::probe(int msgTag)
{
  CPPvmRecThrowIfNotInit

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  bool info;
  
  info = probeTry(msgTag);

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return info;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probeTry
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message of a certain tag has arrived
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MU  USAGE         : bool probeTry(int msgtag)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag message tag (for example cppvmFloat::cppvmClassMsgTag())
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : bool  true: message with this message tag has arrived
@MP                  false: no message with this message tag has arrived
******************************************************************************/
bool
cppvmReceiveStream::probeTry(int msgTag)
{
  CPPvmRecThrowIfNotInit

  int bufid;
  
  bufid = pvm_probe(recFromTid, msgTag);

  if(bufid==0){
    return false;
  }
  else{
    return true;
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probeAny
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if any message has arrived
@MU  USAGE         : bool probeAny()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : bool  true: message with this message tag has arrived
@MP                  false: no message with this message tag has arrived
******************************************************************************/
bool
cppvmReceiveStream::probeAny()
{
  bool info;
  int msgTag;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  info = probeAnyTry(msgTag);

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return info;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probeAny
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message has arrived
@MD                  if yes, the message tag is returned
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MU  USAGE         : bool probeAny(int& msgtag)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : msgTag message tag
@MP  RETURNS : bool  true: message with this message tag has arrived
@MP                  false: no message with this message tag has arrived
******************************************************************************/
bool
cppvmReceiveStream::probeAny(int& msgTag)
{
  bool info;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  info = probeAnyTry(msgTag);

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return info;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probeAnyTry
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message has arrived
@MD                  if yes, the message tag is returned
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MU  USAGE         : bool probeAnyTry(int& msgtag)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : msgTag message tag
@MP  RETURNS : bool  true: message with this message tag has arrived
@MP                  false: no message with this message tag has arrived
******************************************************************************/
bool
cppvmReceiveStream::probeAnyTry(int& msgTag)
{
  CPPvmRecThrowIfNotInit

  int bufid;
    
  bufid = pvm_probe(recFromTid, -1);

  if(bufid==0){
    return false;
  }
  else{
    int info;
    int bytes;
    int tid;
    
    info = pvm_bufinfo(bufid, &bytes, &msgTag, &tid);
    if(info < 0){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmReceiveStream",
                                                  "probe(int&)",
                                                  "could not get buffer informations."));
    }
    return true;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probe
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message of a certain tag has arrived
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MD                  return some informations about the arrived message
@MU  USAGE         : bool probe(int msgtag, int& bufid, int& bytes, int& tid)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag message tag (for example cppvmFloat::cppvmClassMsgTag())
@MP  INOUT   : -
@MP  OUT     : bufid  message buffer, in which the msg has arrived
@MP            bytes  message body length in bytes
@MP            tid    source of the message
@MP  RETURNS : bool  true: message with this message tag has arrived;
@MP                        bufid, bytes and tid are filled with values
@MP                  false: no message with this message tag has arrived;
@MP                         bufid, bytes and tid = 0
******************************************************************************/
bool
cppvmReceiveStream::probe(int msgTag, int& bufid, int& bytes, int& tid)
{
  bool info;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  info = probeTry(msgTag, bufid, bytes, tid);

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return info;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probeTry
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message of a certain tag has arrived
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MD                  return some informations about the arrived message
@MU  USAGE         : bool probeTry(int msgtag, int& bufid, int& bytes, int& tid)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : msgTag message tag (for example cppvmFloat::cppvmClassMsgTag())
@MP  INOUT   : -
@MP  OUT     : bufid  message buffer, in which the msg has arrived
@MP            bytes  message body length in bytes
@MP            tid    source of the message
@MP  RETURNS : bool  true: message with this message tag has arrived;
@MP                        bufid, bytes and tid are filled with values
@MP                  false: no message with this message tag has arrived;
@MP                         bufid, bytes and tid = 0
******************************************************************************/
bool
cppvmReceiveStream::probeTry(int msgTag, int& bufid, int& bytes, int& tid)
{
  CPPvmRecThrowIfNotInit

  bytes = 0;
  tid = 0;
  
  bufid = pvm_probe(recFromTid, msgTag);
  if(bufid==0){
    return false;
  }
  else{
    int dummyTag;
    int info;
    
    info = pvm_bufinfo(bufid, &bytes, &dummyTag, &tid);
    if(info < 0){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmReceiveStream",
                                                  "probeTry(int msgTag, ...",
                                                  "could not get buffer informations."));
    }
    return true;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probeAny
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message has arrived
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MD                  return some informations about the arrived message
@MU  USAGE         : bool probeAny(int& msgtag, int& bufid,
@MU                                int& bytes, int& tid)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : msgTag message tag (for example cppvmFloat::cppvmClassMsgTag())
@MP            bufid  message buffer, in which the msg has arrived
@MP            bytes  message body length in bytes
@MP            tid    source of the message
@MP  RETURNS : bool  true: message with this message tag has arrived;
@MP                        bufid, bytes and tid are filled with values
@MP                  false: no message with this message tag has arrived;
@MP                         bufid, bytes and tid = 0
******************************************************************************/
bool
cppvmReceiveStream::probeAny(int& msgTag, int& bufid, int& bytes, int& tid)
{
  bool info;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  info = probeAnyTry(msgTag, bufid, bytes, tid);

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return info;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : probeAnyTry
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if a message has arrived
@MD                  the message is expected from the node specified
@MD                  whithin this descriptor
@MD                  return some informations about the arrived message
@MU  USAGE         : bool probeAnyTry(int& msgtag, int& bufid,
@MU                                int& bytes, int& tid)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : msgTag message tag (for example cppvmFloat::cppvmClassMsgTag())
@MP            bufid  message buffer, in which the msg has arrived
@MP            bytes  message body length in bytes
@MP            tid    source of the message
@MP  RETURNS : bool  true: message with this message tag has arrived;
@MP                        bufid, bytes and tid are filled with values
@MP                  false: no message with this message tag has arrived;
@MP                         bufid, bytes and tid = 0
******************************************************************************/
bool
cppvmReceiveStream::probeAnyTry(int& msgTag, int& bufid, int& bytes, int& tid)
{
  CPPvmRecThrowIfNotInit

  bytes = 0;
  tid = 0;
  
  bufid = pvm_probe(recFromTid, -1);
  if(bufid==0){
    return false;
  }
  else{
    int info;
    
    info = pvm_bufinfo(bufid, &bytes, &msgTag, &tid);
    if(info < 0){
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmReceiveStream",
                                                  "probeAnyTry(int& msgTag, ...",
                                                  "could not get buffer informations."));
    }
    return true;
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setRecFromTid
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the task id of the process which sends the data
@MD                  (from witch the data should be received)
@MU  USAGE         : void setRecFromTid(int newTid)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newTid  task id of the sending process
@MP                    -1: receive from any process (any tid)
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::setRecFromTid(int newTid)
{
  CPPvmRecThrowIfNotInit
  if (newTid==PvmNoParent){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmReceiveStream",
                       "setRecFromTid",
                       "cannot receive from parent - No parent process"));
  }

  // negative tids can occur on NT machines!!!
  // So I will not throw an exception any more...
  /*if (newTid<-1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmReceiveStream",
                                                 "setRecFromTid",
                                                 "tid cannot be < -1"));
                                                 }*/
  recFromTid = newTid;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setLastRecFromTid
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the task id of the process which sends the data
@MD                  (from witch the data should was received)
@MU  USAGE         : void setLastRecFromTid(int newTid)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newTid  task id of the sending process
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::setLastRecFromTid(int newTid)
{
  CPPvmRecThrowIfNotInit
  // the source of notification messages can be negative!
  /*if (newTid<0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
    "cppvmReceiveStream",
    "setLastRecFromTid",
    "tid cannot be < 0"));
  }*/
  lastRecFromTid = newTid;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setRecMode
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the receive mode
@MU  USAGE         : void setRecMode(cppvmRecMode newMode)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newMode new receiving mode
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::setRecMode(cppvmRecMode newMode)
{
  CPPvmRecThrowIfNotInit
  if(newMode != CPPvmBlockingRec    &&
     newMode != CPPvmNonblockingRec &&
     newMode != CPPvmTimeoutRec){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmReceiveStream",
                       "setRecMode",
                       "the given 'cppvmRecMode' does not exist. "));
  }
  recMode = newMode;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setTimeout
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the time for timeout receive
@MU  USAGE         : void setTimeout(struct timeout& newTimeout)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newTimeout   new receiving time for timeout
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::setTimeout(struct timeval& newTimeout)
{
  CPPvmRecThrowIfNotInit
  if(recMode != CPPvmTimeoutRec){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmReceiveStream",
                       "setTimeout",
                       "cannot set timeout, cppvmReceiveStream object is not in timeout mode (use setRecMode(CPPvmTimeoutRec) fist)."));
  }
  if(newTimeout.tv_sec < 0 ||
     newTimeout.tv_usec < 0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmReceiveStream",
                       "setTimeout",
                       "the given 'newTimeout' contains negative times. "));
  }
  if(newTimeout.tv_usec >= 1000000){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmReceiveStream",
                       "setTimeout",
                       "'newTimeout.tv_usec' > 1000000. Seconds must be set with the 'newTimeout.tv_sec' variable"));
  }
  timeout = newTimeout;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setTimeout
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the time for timeout receive
@MU  USAGE         : void setTimeout(int seconds, int useconds)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : seconds  seconds until timeout
@MP            useconds microseconds until timeout
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::setTimeout(int seconds, int useconds)
{
  CPPvmRecThrowIfNotInit
  if(recMode != CPPvmTimeoutRec){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmReceiveStream",
                       "setTimeout",
                       "cannot set timeout, cppvmReceiveStream object is not in timeout mode (use setRecMode(CPPvmTimeoutRec) fist)."));
  }
  if(seconds < 0 ||
     useconds < 0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                   "cppvmReceiveStream",
                                   "setTimeout",
                                   "the given timouts contain negative times."));
  }
  if(useconds >= 1000000){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
              "cppvmReceiveStream",
              "setTimeout",
              "'useconds' > 1000000. Seconds must be set with the 'seconds' variable"));
  }
  timeout.tv_sec = seconds;
  timeout.tv_usec = useconds;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmReceiveStream::cppvmPrintOn(ostream& strm)
{
  CPPvmRecThrowIfNotInit

  if(recFromTid == -1) {
    strm << " Receive Data from any process";
  }
  else {
    strm << " Receive Data from process with TID: " << recFromTid;
  }
  cout << endl;
  
  if(lastRecFromTid == CPPvmRecNoLastReceive) {
    strm << "No receive until now." << endl;
  }
  else {
    strm << "Last receive from tid " << lastRecFromTid << "." << endl;
  }
  if(recMode==CPPvmBlockingRec){
    strm << "With blocking receive." << endl;
  }
  else if(recMode==CPPvmNonblockingRec){
    strm << "With nonblocking receive." << endl;
  }
  else{
    strm << " with timout receive." << endl;
    strm << "Timout seconds: " << timeout.tv_sec
         << " useconds: " << timeout.tv_usec << endl;
    }
  if(lastRecResult){
    strm << "Last receive result: true.";
  }
  else{
    strm << "Last receive result: false.";
  }
  
  strm << " Context: " << context << endl;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resetObject
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : reset all object variables
@MU  USAGE         : void resetObject()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::resetObject()
{
  if(initialized){
    // clean up
  }
  else{
    initialized = true;
  }
  
  lastRecFromTid = CPPvmRecNoLastReceive;
  lastRecResult = false;
  
  context = pvm_getcontext();
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setContext
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the context variable
@MU  USAGE         : void setContext(int context_)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : context_ new context
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::setContext(int context_)
{
  context = context_;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set receive stream for 'cppvmTask' objects
@MU  USAGE         : void set(cppvmTask& task,
@MU                           cppvmRecTask mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task     description of the task to receive from
@MP            mode     receive mode:
@MP                     * CPPvmRBtask:  receive blocking
@MP                     * CPPvmRNBtask: receive nonblocking
@MP                     * CPPvmRTtask:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTtask
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTtask
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(cppvmTask& task,
                        cppvmRecTask mode,
                        int seconds, int useconds)
{
  // reset/init object
  resetObject();

  context = task.getContextId();
  recFromTid = task.getTid();
  
  // negative tids can occur on NT machines!!!
  // So I will not throw an exception any more...
  /*if (recFromTid < -1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmReceiveStream",
                                                 "set(cppvmTask&...)",
                                                 "tid of given 'cppvmTask' is < -1."));
                                                 }*/
  
  if(mode==CPPvmRBtask){
    recMode = CPPvmBlockingRec;
  }
  else if(mode==CPPvmRNBtask){
    recMode = CPPvmNonblockingRec;
  }
  else if(mode==CPPvmRTtask){
    recMode = CPPvmTimeoutRec;
    setTimeout(seconds, useconds);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmReceiveStream",
                                                   "set(cppvmTask& task,...)",
                                                   "receive mode does not exist."));
  }
  
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set receive stream for 'cppvmConnection' objects
@MU  USAGE         : void set(cppvmConnection& conn,
@MU                           cppvmRecConnection mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : conn     description of the process to receive from
@MP            mode     receive mode:
@MP                     * CPPvmRBparent:  receive blocking
@MP                     * CPPvmRNBparent: receive nonblocking
@MP                     * CPPvmRTparent:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTparent
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTparent
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(cppvmConnection& conn,
                        cppvmRecConnection mode,
                        int seconds, int useconds)
{
  // reset/init object
  resetObject();

  context = conn.getContextId();
  recFromTid = conn.getParentTid();
  
  // negative tids can occur on NT machines!!!
  // So I will not throw an exception any more...
  /*if (recFromTid < -1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                   "cppvmReceiveStream",
                                   "set(cppvmConnection&...)",
                                   "tid of given 'cppvmConnection' is < -1."));
                                   }*/
  
  if(mode==CPPvmRBparent){
    recMode = CPPvmBlockingRec;
  }
  else if(mode==CPPvmRNBparent){
    recMode = CPPvmNonblockingRec;
  }
  else if(mode==CPPvmRTparent){
    recMode = CPPvmTimeoutRec;
    setTimeout(seconds, useconds);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmReceiveStream",
                                                   "set(cppvmConnection&...)",
                                                   "receive mode does not exist."));
  }
  
  
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set receive stream for 'cppvmSpawnConnection' objects
@MU  USAGE         : void set(cppvmSpawnConnection& sconn,
@MU                           cppvmRecSpawnConnection mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : conn     description of the process to receive from
@MP            mode     receive mode:
@MP                     * CPPvmRBchild:  receive blocking
@MP                     * CPPvmRNBchild: receive nonblocking
@MP                     * CPPvmRTchild:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTchild
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTchild
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(cppvmSpawnConnection& sconn,
                        cppvmRecSpawnConnection mode,
                        int seconds, int useconds)
{
  // reset/init object
  resetObject();

  context = sconn.getContextId();
  recFromTid = sconn.getChildTid();
  
  // negative tids can occur on NT machines!!!
  // So I will not throw an exception any more...
  /*if (recFromTid < -1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                            "cppvmReceiveStream",
                            "set(cppvmSpawnConnection&...)",
                            "tid of given 'cppvmSpawnConnection' is < -1."));
                            }*/
  
  if(mode==CPPvmRBchild){
    recMode = CPPvmBlockingRec;
  }
  else if(mode==CPPvmRNBchild){
    recMode = CPPvmNonblockingRec;
  }
  else if(mode==CPPvmRTchild){
    recMode = CPPvmTimeoutRec;
    setTimeout(seconds, useconds);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmReceiveStream",
                                                   "set(cppvmSpawnConnection&...)",
                                                   "receive mode does not exist."));
  }
  
  
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for processes specified by 'taskId'
@MD                  with current context
@MU  USAGE         : void set(int taskId,
@MU                           cppvmRecTid mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskId   task id of the process to receive from
@MP            mode     receive mode:
@MP                     * CPPvmRBtid:  receive blocking
@MP                     * CPPvmRNBtid: receive nonblocking
@MP                     * CPPvmRTtid:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTtid
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTtid
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(int taskId,
                        cppvmRecTid mode,
                        int seconds, int useconds)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  set(taskId, context, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for processes specified by 'taskId'
@MD                  with current context
@MU  USAGE         : void set(int taskId, int context_,
@MU                           cppvmRecTid mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskId   task id of the process to receive from
@MP            context_ receive context
@MP            mode     receive mode:
@MP                     * CPPvmRBtid:  receive blocking
@MP                     * CPPvmRNBtid: receive nonblocking
@MP                     * CPPvmRTtid:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTtid
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTtid
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(int taskId, int context_,
                        cppvmRecTid mode,
                        int seconds, int useconds)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  context = context_;
  recFromTid = taskId;
  
  // negative tids can occur on NT machines!!!
  // So I will not throw an exception any more...
  /*if (recFromTid < -1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmReceiveStream",
                                                 "set(int taskId,...)",
                                                 "given 'taskId' is < -1."));
                                                 }*/
  
  if(mode==CPPvmRBtid){
    recMode = CPPvmBlockingRec;
  }
  else if(mode==CPPvmRNBtid){
    recMode = CPPvmNonblockingRec;
  }
  else if(mode==CPPvmRTtid){
    recMode = CPPvmTimeoutRec;
    setTimeout(seconds, useconds);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmReceiveStream",
                                                   "set(int tid...)",
                                                   "receive mode does not exist."));
  }
  
  
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for receive from any processes 
@MD                  with current context
@MU  USAGE         : void set(cppvmRecAny mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : mode     receive mode:
@MP                     * CPPvmRBany:  receive blocking
@MP                     * CPPvmRNBany: receive nonblocking
@MP                     * CPPvmRTany:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTany
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTany
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(cppvmRecAny mode,
                        int seconds, int useconds)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  set(context, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for processes specified by 'taskId'
@MD                  with current context
@MU  USAGE         : void set(int context_,
@MU                           cppvmRecAny mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : context_ receive context
@MP            mode     receive mode:
@MP                     * CPPvmRBany:  receive blocking
@MP                     * CPPvmRNBany: receive nonblocking
@MP                     * CPPvmRTany:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTany
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTany
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(int context_,
                        cppvmRecAny mode,
                        int seconds, int useconds)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  context = context_;
  recFromTid = -1;
  
  if(mode==CPPvmRBany){
    recMode = CPPvmBlockingRec;
  }
  else if(mode==CPPvmRNBany){
    recMode = CPPvmNonblockingRec;
  }
  else if(mode==CPPvmRTany){
    recMode = CPPvmTimeoutRec;
    setTimeout(seconds, useconds);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmReceiveStream",
                                                   "set(int context_...)",
                                                   "receive mode does not exist."));
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for receive from the same processes, from which
@MD                  a receive was done by 'oldStrm'
@MD                  with current context
@MU  USAGE         : void set(cppvmReceiveStream& oldStrm,
@MU                           cppvmRecSame mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : oldStrm  receive from the same process as the last receive
@MD                     was done in this stream
@MD            mode     receive mode:
@MP                     * CPPvmRBsame:  receive blocking
@MP                     * CPPvmRNBsame: receive nonblocking
@MP                     * CPPvmRTsame:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTsame
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTsame
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(cppvmReceiveStream& oldStrm,
                        cppvmRecSame mode,
                        int seconds, int useconds)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  set(oldStrm, context, mode, seconds, useconds);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmReceiveStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for receive from the same processes, from which
@MD                  a receive was done by 'oldStrm'
@MD                  with current context
@MU  USAGE         : void set(cppvmReceiveStream& oldStrm,
@MU                           cppvmRecSame mode,
@MU                           int seconds, int useconds);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : oldStrm  receive from the same process as the last receive
@MD                     was done in this stream
@MD            mode     receive mode:
@MP                     * CPPvmRBsame:  receive blocking
@MP                     * CPPvmRNBsame: receive nonblocking
@MP                     * CPPvmRTsame:  timeout receive
@MP            seconds: timeout seconds; only used for mode=CPPvmRTsame
@MP            ueconds: timeout microseconds; only used for mode=CPPvmRTsame
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmReceiveStream::set(cppvmReceiveStream& oldStrm,
                        int context_,
                        cppvmRecSame mode,
                        int seconds, int useconds)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  context = context_;
  recFromTid = oldStrm.getLastRecFromTid();
  
  // negative tids can occur on NT machines!!!
  // So I will not throw an exception any more...
  /*if (recFromTid < -1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmReceiveStream",
                                                 "set(cppvmReceiveStream& oldStrm...)",
                                                 "given id in 'oldStrm' is < -1."));
                                                 }*/
  
  if(mode==CPPvmRBsame){
    recMode = CPPvmBlockingRec;
  }
  else if(mode==CPPvmRNBsame){
    recMode = CPPvmNonblockingRec;
  }
  else if(mode==CPPvmRTsame){
    recMode = CPPvmTimeoutRec;
    setTimeout(seconds, useconds);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmReceiveStream",
                                                   "set(cppvmReceiveStream& oldStrm...)",
                                                   "receive mode does not exist."));
  }
}




/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/
