/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_sead.cpp 
@IE  CREATED     : 18.2.1999
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : semaphore administration class for semaphore server
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// include standard CPPvm classes
#include "pvm_str.h"
#include "pvm_int.h"
#include "pvm_coll.h"
// cppvm classes
#include "pvm_cose.h"
// class header
#include "pvm_sead.h" 

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_sead.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmMaxId
@VD  DESCRIPTION   : unique id number for all framework objects
@VD                  highest actual ID
@VT  TYPE          : int
***
******************************************************************************/
int cppvmSemaphoreAdministrator::cppvmMaxId = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmTotalObjNum
@VD  DESCRIPTION   : total number of cppvmObjects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmSemaphoreAdministrator::cppvmTotalObjNum = 0; 

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSemaphoreAdministrator
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmSemaphoreAdministrator
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName      key name of the semaphore
@MP            number        total size of allowed entries for this semaphore
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSemaphoreAdministrator::cppvmSemaphoreAdministrator()
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSemaphoreAdministrator
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSemaphoreAdministrator::cppvmSemaphoreAdministrator (const cppvmSemaphoreAdministrator& /* obj */)
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmSemaphoreAdministrator
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmSemaphoreAdministrator()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      :
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
cppvmSemaphoreAdministrator::~cppvmSemaphoreAdministrator()
{
  cppvmTotalObjNum--;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  info for configuration object attributes 
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSemaphoreAdministrator&
cppvmSemaphoreAdministrator::operator= (const cppvmSemaphoreAdministrator& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }
  
  return *this;
}
  
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print on stream method
@MU  USAGE         : void cppvmPrintOn(ostream& strm);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm out stream
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphoreAdministrator::cppvmPrintOn(ostream& strm)
{
  strm << "semaphore administrator: list of semaphores:"
       << endl << complexSemaphoreList << endl;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : enter
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : call enter operation for a semaphore
@MU  USAGE         : bool enter(cppvmString& semaName, int tid);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName key name of the semaphore
@MP            tid      task id of the calling process
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool: true: enter allowed, process can continue
@MP                  false: enter not allowed, process has to sleep
******************************************************************************/
bool
cppvmSemaphoreAdministrator::enter(cppvmString& semaName, int tid)
{
  int index;
  bool result;
  
  index = getIndexOfSemaphoreInList(semaName);
  if(index < 0){
    // there is no semaphore with this name in the list
    addNewSemaphoreToList(semaName);
    // get index again
    index = getIndexOfSemaphoreInList(semaName);
  }
  // call enter operation of this semaphore
  result = ((cppvmComplexSemaphore&)(complexSemaphoreList[index])).enter(tid);
  
  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : enterNB
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : call enterNB operation for a semaphore
@MU  USAGE         : bool enterNB(cppvmString& semaName);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName key name of the semaphore
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool: true: enter allowed, process can continue
@MP                  false: enter failed
******************************************************************************/
bool
cppvmSemaphoreAdministrator::enterNB(cppvmString& semaName)
{
  int index;
  bool result;
  
  index = getIndexOfSemaphoreInList(semaName);
  if(index < 0){
    // there is no semaphore with this name in the list
    addNewSemaphoreToList(semaName);
    // get index again
    index = getIndexOfSemaphoreInList(semaName);
  }
  // call enter operation of this semaphore
  result = ((cppvmComplexSemaphore&)(complexSemaphoreList[index])).enterNB();
  
  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : leave
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : call leave operation for a semaphore
@MU  USAGE         : int leave(cppvmString& semaName);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName key name of the semaphore
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int < 0: no waiting process
@MP                >=0: task id of a waiting process
******************************************************************************/
int
cppvmSemaphoreAdministrator::leave(cppvmString& semaName)
{
  int index;
  int result;
  
  index = getIndexOfSemaphoreInList(semaName);
  if(index < 0){
    // there is no semaphore with this name in the list
    addNewSemaphoreToList(semaName);
    // get index again
    index = getIndexOfSemaphoreInList(semaName);
  }
  // call enter operation of this semaphore
  result = ((cppvmComplexSemaphore&)(complexSemaphoreList[index])).leave();
  
  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setNumber
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : call setNumber for semaphore
@MD                  the method can only be called once
@MD                  use reset() to set the number again
@MU  USAGE         : bool setNumber(cppvmString& semaName, 
@MU                                 int newNumber);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName  key name of the semaphore
@MP            newNumber new maximal number for semaphore
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool true:  number was set
@MP                 false: number was not set
******************************************************************************/
bool
cppvmSemaphoreAdministrator::setNumber(cppvmString& semaName,
                                       int newNumber)
{
  int index;
  bool result;

  index = getIndexOfSemaphoreInList(semaName);
  if(index < 0){
    // there is no semaphore with this name in the list
    addNewSemaphoreToList(semaName);
    // get index again
    index = getIndexOfSemaphoreInList(semaName);
  }
  // call setNumber operation of this semaphore
  result = (bool)(
      ((cppvmComplexSemaphore&)(complexSemaphoreList[index])).
      setNumber((unsigned int) newNumber));
  
  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getNumber
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : call getMaxNumber operation for a semaphore
@MU  USAGE         : int getMaxNumber(cppvmString& semaName);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName  key name of the semaphore
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int number of semaphore entries
******************************************************************************/
int
cppvmSemaphoreAdministrator::getNumber(cppvmString& semaName)
{
  int index;
  int result;
  
  index = getIndexOfSemaphoreInList(semaName);
  if(index < 0){
    // there is no semaphore with this name in the list
    addNewSemaphoreToList(semaName);
    // get index again
    index = getIndexOfSemaphoreInList(semaName);
  }
  // call enter operation of this semaphore
  result = (int)(
      ((cppvmComplexSemaphore&)(complexSemaphoreList[index])).
      getNumber());
  
  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resetNumber
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : delete semaphore in list; create new element and set 
@MD                  number
@MU  USAGE         : void resetNumber(cppvmString& semaName,
@MU                                   int number);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName  key name of the semaphore
@MP            number    new entry number
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphoreAdministrator::resetNumber(cppvmString& semaName, 
                                         int number)
{
  int index;
  bool result;
  
  // delete old entry
  index = getIndexOfSemaphoreInList(semaName);
  if(index >= 0){
    // semaphore was found
    // delete element
    complexSemaphoreList.remove(index);
  }
  // create new entry
  addNewSemaphoreToList(semaName);
  index = getIndexOfSemaphoreInList(semaName);
  // call setNumber operation of this semaphore
  result = (bool)(
      ((cppvmComplexSemaphore&)(complexSemaphoreList[index])).
      setNumber((unsigned int) number));

  if(result==false){
    cerr << "pvm_sead: resetNumber: set returned false" << endl;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : addNewSemaphoreToList
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add a new semaphore to the list
@MU  USAGE         : void addNewSemaphoreToList(cppvmString& semaName);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName key name of the semaphore
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphoreAdministrator::addNewSemaphoreToList(cppvmString& semaName)
{
  complexSemaphoreList.add(new cppvmComplexSemaphore(semaName));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getIndexOfSemaphoreInList
@MK  CLASS NAME    : cppvmSemaphoreAdministrator
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : return index of a semaphore in list
@MU  USAGE         : int getIndexOfSemaphoreInList(cppvmString& semaName)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName key name of the semaphore
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int >=0 index
@MP                < 0 not in list
******************************************************************************/
int
cppvmSemaphoreAdministrator::getIndexOfSemaphoreInList(cppvmString& semaName)
{
  bool found=false;
  unsigned int i=0;
  int index=-1;
  
  while(i<complexSemaphoreList.numElems() && found==false){
    if(((cppvmComplexSemaphore&)(complexSemaphoreList[i])).compareSemaName(semaName)==
       true){
      found = true;
      index = (int)i;
    }
    i++;
  }
  
  return index;
}

  

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

