/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_sema.cpp 
@IE  CREATED     : 18.2.1999
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : semaphore class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// cppvm classes
#include "pvm_str.h"
#include "pvm_int.h"
#include "pvm_spaw.h"
#include "pvm_rec.h"
#include "pvm_send.h"
#include "pvm_conn.h"
// class header
#include "pvm_sema.h" 
// defined for protocol to server
#include "pvm_sepr.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_sema.cpp"

#define CPPvmSemaphoreKeyName    "cppvmSemaphoreServer"

#define CPPvmSemaphoreServerSpawnName "cppvms"
#define CPPvmSemaphoreServerNameWin  "cppvms.exe"
#define CPPvmSemaphoreServerNameUnix "cppvms"

/*@@@**************************************************************************
**********                                                           **********
*******                            VARIABLES                            *******
**********                                                           **********
******************************************************************************/

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmMaxId
@VD  DESCRIPTION   : unique id number for all framework objects
@VD                  highest actual ID
@VT  TYPE          : int
***
******************************************************************************/
int cppvmSemaphore::cppvmMaxId = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : cppvmTotalObjNum
@VD  DESCRIPTION   : total number of cppvmObjects
@VT  TYPE          : int
***
******************************************************************************/
int cppvmSemaphore::cppvmTotalObjNum = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : semaphoreServerConnected
@VD  DESCRIPTION   : indicate if the server was already connected
@VT  TYPE          : bool
***
******************************************************************************/
bool cppvmSemaphore::semaphoreServerConnected = false; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : serverTid
@VD  DESCRIPTION   : task id of the semaphore server
@VT  TYPE          : int
***
******************************************************************************/
int cppvmSemaphore::serverTid = 0; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : receiveBStream
@VD  DESCRIPTION   : blocking receive stream from semaphore server
@VT  TYPE          : cppvmReceiveStream
***
******************************************************************************/
cppvmReceiveStream cppvmSemaphore::receiveBStream; 

/*@@***************************   VARIABLE   **********************************
***
@VN  VARIABLE NAME : sendNBStream
@VD  DESCRIPTION   : non-blocking send stream to semaphore server
@VT  TYPE          : cppvmSendStream
***
******************************************************************************/
cppvmSendStream cppvmSemaphore::sendNBStream; 


/*@@@**************************************************************************
**********                                                           **********
*******                            MACROS                               *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmSemaThrowIfNotInit
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : throws an exception if the key name of the object was not
@MD                 initialized 
@MU  USAGE        : only for internal use!
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmSemaThrowIfNotInit                                                    \
if(semaKeyName.getLength()==0){                                                    \
  CPPvmInternalThrow (cppvmExceptionForbiddenCall("pvm_sema.cpp",                  \
                  "cppvmSemaphore",                                                \
                  "Macro: CPPvmSemaThrowIfNotInit",                                \
                  "semaphore key name was not set. Must be set before using it")); \
}

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSemaphore
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : cppvmSemaphore
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSemaphore::cppvmSemaphore()
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  // write this dummy value into
  // the message mailbox
  mboxEntry = 1;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSemaphore
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmSemaphore
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName      key name of the semaphore
@MP            hostName      host of the semaphore server
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSemaphore::cppvmSemaphore(char *semaName, char *hostName)
{
  if(semaName==NULL && hostName==NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                        "cppvmSemaphore",
                                        "constructor",
                                        "semaName and hostName are NULL pointer."));
  }

  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  // write this dummy value into
  // the message mailbox
  mboxEntry = 1;
  
  // start (if not already running)
  // and connect to semaphore server
  connectSemaphoreServer(hostName);

  if(semaName!=NULL){
    setName(semaName);
  }
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSemaphore
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSemaphore::cppvmSemaphore (const cppvmSemaphore& obj)
{
  // give unique id
  cppvmObjId = cppvmMaxId;
  ++cppvmMaxId;

  // inc total number of objects
  cppvmTotalObjNum++;

  semaKeyName = obj.semaKeyName;

  // start (if not already running)
  // and connect to semaphore server
  connectSemaphoreServer(NULL);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmSemaphore
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmSemaphore()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      :
@MP  INOUT   :
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
cppvmSemaphore::~cppvmSemaphore()
{
  // dec total number of objects
  cppvmTotalObjNum--;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  info for configuration object attributes 
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSemaphore&
cppvmSemaphore::operator= (const cppvmSemaphore& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }
  semaKeyName = obj.semaKeyName;
  
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : connectSemaphoreServer
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : start the semaphore server if it was not already started
@MD                  get task id of running server
@MU  USAGE         : void connectSemaphoreServer(char *hostName)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : hostName      host of the semaphore server
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSemaphore::connectSemaphoreServer(char *hostName)
{
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    connectSemaphoreServerTry(hostName);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : connectSemaphoreServerTry
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : start the semaphore server if it was not already started
@MD                  get task id of running server
@MU  USAGE         : void connectSemaphoreServerTry(char *hostName)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : hostName      host of the semaphore server
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSemaphore::connectSemaphoreServerTry(char *hostName)
{
  if(semaphoreServerConnected==false){
    if(checkSemaphoreServerRunning()==true){
      // get task id of running server
      getSemaphoreServerTid();
    }
    else{
      // start server and store task id
      startSemaphoreServer(hostName);
    }
    semaphoreServerConnected=true;
    receiveBStream.set(serverTid, PvmBaseContext, CPPvmRBtid);
    sendNBStream.set(serverTid, PvmBaseContext, CPPvmSNBtid);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : startSemaphoreServer
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : start the semaphore server if it was not already started
@MU  USAGE         : void startSemaphoreServer(char *hostName)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : hostName      host of the semaphore server
@MP  INOUT   : -
@MP  OUT     : - 
@MP  RETURNS : -
******************************************************************************/
void
cppvmSemaphore::startSemaphoreServer(char *hostName)
{
  cppvmSpawnConnection *semaphoreServer;
  
  // spawn semaphore server 
  // spawn in PvmBaseContext to be reachable from every process in every
  // context 

  if(hostName==NULL){
    // let pvm choose the host
    semaphoreServer = new cppvmSpawnConnection(CPPvmSemaphoreServerSpawnName, "",
                                               PvmTaskDefault, "", 
                                               CPPvmGivenContext, PvmBaseContext);
  }
  else{
    // user defined host
    semaphoreServer = new cppvmSpawnConnection(CPPvmSemaphoreServerSpawnName, "",
                                               PvmTaskHost, hostName, 
                                               CPPvmGivenContext, PvmBaseContext);
  }
  

  // get server task id
  serverTid = semaphoreServer->getChildTid();

  // the object can be deleted after the task id is read
  // this is only possible because the destructor of 
  // cppvmSpawnConnection does not kill the child process
  delete semaphoreServer;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : checkSemaphoreServerRunning
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : check if the semaphore server is already running
@MU  USAGE         : bool checkSemaphoreServerRunning()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : - 
@MP  RETURNS : bool true:  is already running
@MP                 false: is not running, must be started
******************************************************************************/
bool
cppvmSemaphore::checkSemaphoreServerRunning()
{

  int info;
  int oldSBuf;
  int newSBuf;
  
  oldSBuf = pvm_getsbuf();
  newSBuf = pvm_mkbuf(PvmDataDefault);
  if(newSBuf < 0){
    if(newSBuf == PvmBadParam){
     CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                 "cppvmSemaphore",
                                                 "checkSemaphoreServerRunning",
                                                 "giving an invalid encoding value"));
     return false;
    }
    else if(newSBuf == PvmNoMem){	
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                            "cppvmSemaphore",
                                            "checkSemaphoreServerRunning",
                                            "not enough memory to create the buffer"));
     return false;
    }
    else{
      CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                  "cppvmSemaphore",
                                                  "checkSemaphoreServerRunning",
                                                  "cannot create new buffer"));
     return false;
    }
  }

  
  pvm_setsbuf(newSBuf);

  // pack mboxEntry
  pvm_pkint(&mboxEntry,1,1);

  info = pvm_putinfo(CPPvmSemaphoreKeyName,
                     pvm_getsbuf(),
	             PvmMboxDefault + PvmMboxPersistent);

  // reset values
  pvm_setsbuf(oldSBuf);
  pvm_freebuf(newSBuf);

  if(info == PvmExists || info == PvmDenied){
    // message mailbox entry exists
    return true;
  }
  else if(info == PvmBadParam ||
     info == PvmNoSuchBuf){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                       "cppvmSemaphore",   
                       "checkSemaphoreServerRunning",
                       "invalid or non existant send buffer. Maybe you have set a new message buffer with 'cppvmSetMboxMsgBufSend(buffer)' and deleted 'buffer' afterwards."));
    return false;
  }
  else if(info == PvmNoMem){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                            "cppvmSemaphore",   
                                            "checkSemaphoreServerRunning",
                                            "unable to allocate memory to pack data"));
    return false;
  }
  else if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmSemaphore",   
                                                "checkSemaphoreServerRunning",
                                                "error writing in database"));
    return false;
  }

  // server does not exist
  return false;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getSemaphoreServerTid
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get task id of a running server 
@MU  USAGE         : void getSemaphoreServerTid()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSemaphore::getSemaphoreServerTid()
{
  int i;
  int ntask;
  struct pvmtaskinfo *taskp;
  bool processExists=false;

  while(processExists==false){
    
    taskInfo(0, &ntask, &taskp);

    for(i=0;i<ntask;i++){
      if((strcmp(taskp[i].ti_a_out, CPPvmSemaphoreServerNameWin)==0) ||
         (strcmp(taskp[i].ti_a_out, CPPvmSemaphoreServerNameUnix)==0)){
        serverTid = taskp[i].ti_tid;
        processExists = true;
      }
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : taskInfo
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : information about tasks running on the virtual machine
@MU  USAGE         : void taskInfo(int where,int *ntask,
@MU                                struct pvmtaskinfo **taskp);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : the space for the struct pvmtaskinfo array is given
@MR                  by this method
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : where   information specifier:
@MP                    0:        all tasks on the virtual machine
@MP                    pvmd tid: all task on this host (tid of the pvm daemon, 
@MP                              use getHostDaemonTid to get this id
@MP                    tid:      specific task
@MP            ntask   number of task (depending on 'where')
@MP            taskp   pointer to an array of structures which contain
@MP                    information about each task
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::taskInfo(int where,
                          int *ntask,
                          struct pvmtaskinfo **taskp)
{
  int info;

  if(ntask == NULL ||
     taskp == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSemaphore",
                                                   "taskInfo",
                                                   "a parameter is a NULL pointer."));
  }
  
  info = pvm_tasks(where, ntask, taskp);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                      "cppvmSemaphore",
                      "taskInfo",
                      "could not get task informations of the virtual machine."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : enter
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : enter semaphore (passeer), this operation locks the 
@MD                  process if more than the total size for this semaphore
@MP                  have entered
@MU  USAGE         : void enter();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::enter()
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    enterTry();
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : enterTry
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : enter semaphore (passeer), this operation locks the 
@MD                  process if more than the total size for this semaphore
@MP                  have entered
@MU  USAGE         : void enterTry();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::enterTry()
{
  CPPvmSemaThrowIfNotInit 

  bool result;
  
  // connect to server
  operation = CPPvmToSemaphoreServerEnter;
  result = serverTransfer();
  if(result==false){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                        "cppvmSemaphore",
                                        "enterTry",
                                        "got bad answer from semaphore server. Probably an old server is still alive from former runs or protocol error."));
  }
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : enterNB
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : non-blocking enter semaphore (passeer), the return
@MD                  value indicates if the enter was successful
@MU  USAGE         : bool enterNB();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
bool
cppvmSemaphore::enterNB()
{
  bool result;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    result = enterNBTry();
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
    return false;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : enterNBTry
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : non-blocking enter semaphore (passeer), the return
@MD                  value indicates if the enter was successful
@MU  USAGE         : bool enterNBTry();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
bool
cppvmSemaphore::enterNBTry()
{
  CPPvmSemaThrowIfNotInit 

  bool result;
  
  // connect to server
  operation = CPPvmToSemaphoreServerEnterNB;
  result = serverTransfer();

  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : leave
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : leave semaphore (verlaat), this operation unlocks the 
@MD                  former enter operation
@MU  USAGE         : void leave();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::leave()
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    leaveTry();
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : leaveTry
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : leave semaphore (verlaat), this operation unlocks the 
@MD                  former enter operation
@MU  USAGE         : void leaveTry();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::leaveTry()
{
  CPPvmSemaThrowIfNotInit 

  bool result;
  
  // connect to server
  operation = CPPvmToSemaphoreServerLeave;
  result = serverTransfer();

  if(result==false){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                        "cppvmSemaphore",
                                        "leaveTry",
                                        "got bad answer from semaphore server. Probably an old server is still alive from former runs or protocol error."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setNumber
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the number of allowed entries for this
@MD                  semaphore
@MU  USAGE         : void setNumber(unsigned int number);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : number wanted number of allowed entries
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::setNumber(unsigned int number)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    setNumberTry(number);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setNumberTry
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the allowed number of entries for this
@MD                  semaphore
@MU  USAGE         : void setNumberTry(unsigned int number);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : number wanted number of allowed entries
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::setNumberTry(unsigned int number)
{
  CPPvmSemaThrowIfNotInit 

  bool result;
  
  // connect to server
  operation = CPPvmToSemaphoreServerSetNumber;
  // set wanted number
  numSemaSet = (int) number;
  
  result = serverTransfer();
  if(result==false){
    CPPvmInternalThrow( cppvmExceptionForbiddenCall(THIS_MODULE,
                                            "cppvmSemaphore",
                                            "setNumberTry",
                                            "the number can only set once from all processes. Maybe another process has called setNumber() for this semaphore. Use resetNumber() to modify the number."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : tryToSetNumber
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : Try to set the allowed number of entries for this
@MD                  semaphore. If the number is already set, nothing
@MD                  happens, otherwise the number is set.
@MU  USAGE         : void tryToSetNumber(unsigned int number);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : number wanted number of allowed entries
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool true: value set 
@MP                 false: value was already set
******************************************************************************/
bool
cppvmSemaphore::tryToSetNumber(unsigned int number)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    return tryToSetNumberTry(number);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
  return false;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : tryToSetNumberTry
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : Try to set the allowed number of entries for this
@MD                  semaphore. If the number is already set, nothing
@MD                  happens, otherwise the number is set.
@MU  USAGE         : void tryToSetNumberTry(unsigned int number);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : number wanted number of allowed entries
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool true: value set 
@MP                 false: value was already set
******************************************************************************/
bool
cppvmSemaphore::tryToSetNumberTry(unsigned int number)
{
  CPPvmSemaThrowIfNotInit 

  bool result;
  
  // connect to server
  operation = CPPvmToSemaphoreServerSetNumber;
  // set wanted number
  numSemaSet = (int) number;
  
  result = serverTransfer();

  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : internalTryToSetNumber
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : Try to set the allowed number of entries for this
@MD                  semaphore. If the number is already set, nothing
@MD                  happens, otherwise the number is set.
@MU  USAGE         : void internalTryToSetNumber(unsigned int number);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : This method is NOT threadsafe!!! This method is 
@MR                  (and should) only be used by the class 
@MR                  'cppvmMutexCollection' ('pvm_mutexcoll.h/.cpp').
@MR                  The mutex collection is only used by 'cppvmMutex'
@MR                  ('pvm_mutex.h'). 'cppvmMutex' calls already the
@MR                  functions 'lock/unlock'. To avoid
@MR                  duplicated locks (which might not work) this
@MR                  method does not use locks!
@M#  --------------------------------------------------------------------------
@MP  IN      : number wanted number of allowed entries
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
bool
cppvmSemaphore::internalTryToSetNumber(unsigned int number)
{
  CPPvmSemaThrowIfNotInit 

  bool result;
  
  // connect to server
  operation = CPPvmToSemaphoreServerSetNumber;
  // set wanted number
  numSemaSet = (int) number;
  
  result = serverTransfer();

  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getNumber
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the current number of entries for this semaphore
@MU  USAGE         : unsigned int getNumber();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : unsigned int   number of entries
******************************************************************************/
unsigned int
cppvmSemaphore::getNumber()
{
  unsigned int result;
  
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    result = getNumberTry();
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
    return 0;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif

  return result;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getNumberTry
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the current number of entries for this semaphore
@MU  USAGE         : unsigned int getNumberTry();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : unsigned int   real number of allowed entries
******************************************************************************/
unsigned int
cppvmSemaphore::getNumberTry()
{
  CPPvmSemaThrowIfNotInit 

  bool result;
  
  // connect to server
  operation = CPPvmToSemaphoreServerGetNumber;
  
  result = serverTransfer();
  if(result==false){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                        "cppvmSemaphore",
                                        "getNumberTry",
                                        "got bad answer from semaphore server. Probably an old server is still alive from former runs or protocol error."));
  }
  // return current allowed entries
  return (unsigned int) ((int) numSemaGet);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resetNumber
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : reset semaphore
@MU  USAGE         : void resetNumber(unsigned int number);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::resetNumber(unsigned int number)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif

  CPPvmInternalTry{
    resetNumberTry(number);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resetNumberTry
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : reset semaphore
@MU  USAGE         : void resetNumberTry(unsigned int number);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::resetNumberTry(unsigned int number)
{
  CPPvmSemaThrowIfNotInit 

  bool result;
  
  // connect to server
  operation = CPPvmToSemaphoreServerResetNumber;
  // set wanted number
  numSemaSet = (int) number;

  result = serverTransfer();
  if(result==false){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                        "cppvmSemaphore",
                                        "resetTry",
                                        "got bad answer from semaphore server. Probably an old server is still alive from former runs or protocol error."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : serverTransfer
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : transfer information to/from the server
@MU  USAGE         : bool serverTransfer();
@MC  CONSTRAINTS   : the transfer operation must be specified in variable
@MC                  operation
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool true:  received ok     from server
@MP                 false: received failed from server
******************************************************************************/
bool
cppvmSemaphore::serverTransfer()
{
  // cppvmString and cppvmInt objects are transfered
  // to ensure the original message tags, set 
  // the values explicit
  int oldStringMsgTag;
  int oldIntMsgTag;
  
  // remind old values
  oldStringMsgTag = cppvmString::cppvmClassMsgTag();
  oldIntMsgTag = cppvmInt::cppvmClassMsgTag();
  
  // set standard values
  cppvmString::setCPPvmClassMsgTag(CPPvmMsgTagClassString);
  cppvmInt::setCPPvmClassMsgTag(CPPvmMsgTagClassInt);
  

  // send operation to server
  sendNBStream << operation;
  // send semaphore name to server
  sendNBStream << semaKeyName;

  if((int)operation == CPPvmToSemaphoreServerSetNumber ||
     (int)operation == CPPvmToSemaphoreServerResetNumber){
    // send number if needed
    sendNBStream << numSemaSet;
  }

  if((int)operation == CPPvmToSemaphoreServerGetNumber){
    // number was sended from server
    receiveBStream >> numSemaGet;
  }
  
  // receive result from server
  receiveBStream >> operationResult;

  // reset class message tags
  cppvmString::setCPPvmClassMsgTag(oldStringMsgTag);
  cppvmInt::setCPPvmClassMsgTag(oldIntMsgTag);

  // return result from server
  if((int)operationResult == CPPvmFromSemaphoreServerOK){
    return true;
  }
  else{
    return false;
  }
}

  
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print on stream method
@MU  USAGE         : void cppvmPrintOn(ostream& strm);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm out stream
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::cppvmPrintOn(ostream& strm)
{
  strm << "semaphore key name: " << semaKeyName
       << endl;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setName
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the semaphore name
@MU  USAGE         : void setName(char *semaName);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : semaName  name of the semaphore
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::setName(char *semaName)
{
  if(semaName==NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSemaphore",
                                                   "setName",
                                                   "semaName is a NULL pointer."));
  }
  
  // assign name
  semaKeyName=semaName;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : killServer
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : kill the semaphore server
@MU  USAGE         : void killServer();
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::killServer()
{
  int serverTid, info;
  
  cppvmConnection pvmConn;
  serverTid = pvmConn.getTidOfTask(CPPvmSemaphoreServerNameUnix);
  if(serverTid==-1){
    serverTid = pvmConn.getTidOfTask(CPPvmSemaphoreServerNameWin);
  }
  info = pvm_kill(serverTid);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmSemaphore",
                                                "killServer",
                                                "cannot kill server process."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : startServer
@MK  CLASS NAME    : cppvmSemaphore
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : start the semaphore server
@MU  USAGE         : void startServer(char *hostName);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         : Nothing happens if the semaphore server is already 
@MR                  running
@M#  --------------------------------------------------------------------------
@MP  IN      : hostName  NULL: PVM starts the server on a host of its choice
@MP                      else: hostname of the host to start the server on
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSemaphore::startServer(char *hostName)
{
  connectSemaphoreServer(hostName);
}


/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



