/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_send.cpp 
@IE  CREATED     : 5.12.1997
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : description of a pvm send command
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/

// includes for predeclared classes to avoid cyclic includes
#include "pvm_task.h"
#include "pvm_conn.h"
#include "pvm_spaw.h"
#include "pvm_spmu.h"
#include "pvm_rec.h"
// include own class
#include "pvm_send.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_send.cpp"


/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSendStream
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : only use it if it is later on initialized by a
@MU                  set method
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSendStream::cppvmSendStream() 
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     sendMode(CPPvmNonblockingSend), sendConnection(CPPvmOneToOne),
     numTid(0), 
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for cppvmTask objects
@MU  USAGE         : cppvmSendStream(cppvmTask& task,
@MU                           cppvmSendTask mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task     description of the process to send to
@MP            mode     send mode:
@MP                     * CPPvmSBtask:  send blocking
@MP                     * CPPvmSNBtask: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(cppvmTask& task,
                                 cppvmSendTask mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0), 
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(task, mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for cppvmConnection objects
@MU  USAGE         : cppvmSendStream(cppvmConnection& conn,
@MU                           cppvmSendConnection mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : conn     description of the parent process to send to
@MP            mode     send mode:
@MP                     * CPPvmSBparent:  send blocking
@MP                     * CPPvmSNBparent: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(cppvmConnection& conn,
                                 cppvmSendConnection mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(conn, mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for cppvmSpawnConnection objects
@MU  USAGE         : cppvmSendStream(cppvmSpawnConnection& sconn,
@MU                           cppvmSendSpawnConnection mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : sconn     description of the child process to send to
@MP            mode     send mode:
@MP                     * CPPvmSBchild:  send blocking
@MP                     * CPPvmSNBchild: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(cppvmSpawnConnection& sconn,
                                 cppvmSendSpawnConnection mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(sconn, mode, initSend_);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for cppvmSpawnMultiConnection objects
@MU  USAGE         : cppvmSendStream(cppvmSpawnMultiConnection& smconn,
@MU                           cppvmSendSpawnMultiConnection mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : smconn   description of the child processes to send to
@MP            mode     send mode:
@MP                     * CPPvmSBtaggedChildren: send blocking
@MP                                              to tagged children
@MP                     * CPPvmSBtaggedChildren: send nonblocking
@MP                                              to tagged children
@MP                     * CPPvmSBallChildren:  send blocking to all children
@MP                     * CPPvmSNBallChildren: send nonblocking to all children
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(cppvmSpawnMultiConnection& smconn,
                                 cppvmSendSpawnMultiConnection mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0), 
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(smconn, mode, initSend_);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for group broadcasting
@MD                  with current context
@MU  USAGE         : cppvmSendStream(char *group_,
@MU                           cppvmSendGroup mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group_   group name
@MP            mode     send mode:
@MP                     * CPPvmSBgroup:  send blocking
@MP                     * CPPvmSNBgroup: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(char *group_,
                                 cppvmSendGroup mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(group_, mode, initSend_);
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for group broadcasting
@MD                  with given context
@MU  USAGE         : cppvmSendStream(char *group_, int context_, 
@MU                           cppvmSendGroup mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group_   group name
@MP            context_ context to send with
@MP            mode     send mode:
@MP                     * CPPvmSBgroup:  send blocking
@MP                     * CPPvmSNBgroup: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(char *group_,
                                 int context_,
                                 cppvmSendGroup mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(group_, context_, mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for send to a specific task
@MD                  with current context
@MU  USAGE         : cppvmSendStream(int taskId,
@MU                           cppvmSendTid mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskId   task id of the process to send to
@MP            mode     send mode:
@MP                     * CPPvmSBtid:  send blocking
@MP                     * CPPvmSNBtid: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(int taskId,
                                 cppvmSendTid mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(taskId, mode, initSend_);
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for send to a specific task
@MD                  with given context
@MU  USAGE         : cppvmSendStream(int taskId, int context_, 
@MU                           cppvmSendTid mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskId   task id of the process to send to
@MP            context_ context to send with
@MP            mode     send mode:
@MP                     * CPPvmSBtid:  send blocking
@MP                     * CPPvmSNBtid: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(int taskId,
                                 int context_,
                                 cppvmSendTid mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(taskId, context_, mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for send to specific tasks
@MD                  with current context
@MU  USAGE         : cppvmSendStream(int* taskIds,
@MU                           unsigned int numTasks,
@MU                           cppvmSendTids mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskIds   array of task ids of the processes to send to
@MP            numTasks number of tasks in 'taskIds' (array length)
@MP            mode     send mode:
@MP                     * CPPvmSBtids:  send blocking
@MP                     * CPPvmSNBtids: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(int* taskIds,
                                 unsigned int numTasks,
                                 cppvmSendTids mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(taskIds, numTasks, mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for send to specific tasks
@MD                  with given context
@MU  USAGE         : cppvmSendStream(int* taskIds, unsigned int numTasks,
@MU                           int context_, 
@MU                           cppvmSendTids mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskIds  array of  task ids of the processes to send to
@MP            numTasks number of tasks in 'taskIds' (array length)
@MP            context_ context to send with
@MP            mode     send mode:
@MP                     * CPPvmSBtids:  send blocking
@MP                     * CPPvmSNBtids: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(int* taskIds,
                                 unsigned int numTask,
                                 int context_,
                                 cppvmSendTids mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(taskIds, numTask, context_, mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for send to specific tasks
@MD                  with given context array
@MU  USAGE         : cppvmSendStream(int* taskIds, unsigned int numTasks,
@MU                           int* contextArr, 
@MU                           cppvmSendTids mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskIds    array of  task ids of the processes to send to
@MP            numTasks   number of tasks in 'taskIds' (array length)
@MP            contextArr context array to send with
@MP            mode     send mode:
@MP                     * CPPvmSBtids:  send blocking
@MP                     * CPPvmSNBtids: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(int* taskIds,
                                 unsigned int numTask,
                                 int* contextArr,
                                 cppvmSendTids mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(taskIds, numTask, contextArr, mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor to send back to a process, from which
@MD                  a message was received with 'recStrm'
@MD                  with current context
@MU  USAGE         : cppvmSendStream(cppvmReceiveStream& recStrm,
@MU                           cppvmSendBack mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : recStrm  send back to the process from which a message
@MP                     was received with this stream
@MP            mode     send mode:
@MP                     * CPPvmSBback:  send blocking
@MP                     * CPPvmSNBback: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(cppvmReceiveStream& recStrm,
                                 cppvmSendBack mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(recStrm, mode, initSend_);
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : constructor
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor for send back to a process, from which
@MD                  a message was received with 'recStrm'
@MD                  with given context
@MU  USAGE         : cppvmSendStream(cppvmReceiveStream& recStrm,
@MU                                  int context_, 
@MU                                  cppvmSendBack mode,
@MU                                  int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : recStrm  send back to the process from which a message
@MP                     was received with this stream
@MP            context_ context to send with
@MP            mode     send mode:
@MP                     * CPPvmSBtid:  send blocking
@MP                     * CPPvmSNBtid: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream(cppvmReceiveStream& recStrm,
                                 int context_,
                                 cppvmSendBack mode,
                                 int initSend_)
    :cppvmStream(),
     initialized(false),
     msgBufSend(false), sendBufEncoding(PvmDataDefault),
     numTid(0),
     initSend(PvmDataDefault), group(NULL), groupAllocated(false)
{
  set(recStrm, context_, mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmSendStream
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmSendStream
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSendStream::~cppvmSendStream()
{
  if(initialized){
    delete [] sendToTids;
    delete [] context;
    if(groupAllocated){
      delete [] group;
    }
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmSendStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream& 
cppvmSendStream::operator= (const cppvmSendStream& obj) 
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  initialized = obj.initialized;
  //CPPvmSendThrowIfNotInit
  int i;
  
  if(cppvmIsSame(obj)){
    return *this;
  }
  cppvmStream::operator=(obj);
  numTid = obj.numTid;
  sendMode = obj.sendMode;
  sendConnection = obj.sendConnection;
  initSend = obj.initSend;
  msgBufSend = obj.msgBufSend;
  sendBufEncoding = obj.sendBufEncoding;
  
  delete [] sendToTids;
  delete [] context;
  sendToTids = new int[numTid];
  context = new int[numTid];

  for(i=0; i<numTid; i++){
    sendToTids[i] = obj.sendToTids[i];
    context[i] = obj.context[i];
  }

  if(groupAllocated){
    delete [] group;
  }
  if(obj.groupAllocated){
    setGroup(obj.group);
  }

  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSendStream
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSendStream::cppvmSendStream (const cppvmSendStream& obj)
    : cppvmStream(obj),
      initialized(obj.initialized),
      msgBufSend(obj.msgBufSend),
      sendBufEncoding(obj.sendBufEncoding),
      sendMode(obj.sendMode), sendConnection(obj.sendConnection),
      numTid(obj.numTid), 
      initSend(obj.initSend), 
      group(NULL),
      groupAllocated(false)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  int i;

  if (numTid<0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSendStream",
                                                 "copy constructor",
                                                 "obj.numTid cannot be negative"));
  }
  delete [] sendToTids;
  delete [] context;
  sendToTids = new int[numTid];
  context = new int[numTid];

  for(i=0; i<numTid; i++){
    sendToTids[i] = obj.sendToTids[i];
    context[i] = obj.context[i];
  }

  if(obj.groupAllocated){
    setGroup(obj.group);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmSendStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmSendStream::operator== (const cppvmSendStream& obj) 
{
  CPPvmSendThrowIfNotInit
  int i;
  
  if(sendMode != obj.sendMode ||
     sendConnection != obj.sendConnection  ||
     initSend != obj.initSend  ||
     numTid   != obj.numTid ||
     msgBufSend != obj.msgBufSend ||
     sendBufEncoding != obj.sendBufEncoding){
    return false;
  }
  for(i=0; i<numTid; i++){
    if(sendToTids[i] != obj.sendToTids[i] ||
       context[i] != obj.context[i]){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are equal
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  cppvmSendStream object
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmSendStream::operator== (const cppvmSendStream& obj) const
{
  CPPvmSendThrowIfNotInit
  int i;
  
  if(sendMode != obj.sendMode ||
     sendConnection != obj.sendConnection  ||
     initSend != obj.initSend  ||
     numTid   != obj.numTid ||
     msgBufSend != obj.msgBufSend ||
     sendBufEncoding != obj.sendBufEncoding){
    return false;
  }
  for(i=0; i<numTid; i++){
    if(sendToTids[i] != obj.sendToTids[i] ||
       context[i] != obj.context[i]){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator!=
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : test if values are not equal
@MU  USAGE         : !=
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : i  cppvmSendStream value
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmSendStream::operator!= (const cppvmSendStream& obj) 
{
  CPPvmSendThrowIfNotInit
  return ! (operator==(obj));
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : addTask
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add a tasks id and context to the receiving processes
@MD                  for multicasting
@MU  USAGE         : void addTask(int tid, int context_)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid       task id to add
@MP            context_  context to add
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::addTask(int tid, int context_)
{
  CPPvmSendThrowIfNotInit
  int i;
  int *tmpTids;
  int *tmpContext;
  bool found;

  if (tid<0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSendStream",
                                                 "addTask",
                                                 "tid cannot be negative"));
  }

  found = false;
  i = 0;
  while(found==false && i < numTid){
    if(sendToTids[i] == tid){
      found = true;
    }
    else{
      i++;
    }
  }

  if(found){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                "cppvmSendStream",
                                                "addTask",
                                                "tid is already in the sending array"));
  }

  numTid = numTid + 1;
  tmpTids = new int[numTid];
  tmpContext = new int[numTid];

  for(i=0; i<numTid-1; i++){
    tmpTids[i] = sendToTids[i];
    tmpContext[i] = context[i];
  }
  tmpTids[numTid-1] = tid;
  tmpContext[numTid-1] = context_;
  
  delete [] sendToTids;
  delete [] context;
  sendToTids = tmpTids;
  context = tmpContext;

  // multicast or oneToOne connection?
  bool differ = false;
  int referenceContext = context[0];
  
  for(i=1; i<numTid; i++){
    if(context[i]!=referenceContext){
      differ = true;
    }
  }
  if(differ){
    sendConnection = CPPvmMulticast;
  }
  else{
    sendConnection = CPPvmOneToOne;
  }
  
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : remTask
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove a tasks id and context from the receiving processes
@MU  USAGE         : void remTask(int tid)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : there must be at least one tid left
@M#  --------------------------------------------------------------------------
@MP  IN      : tid  task id to remove
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::remTask(int tid)
{
  CPPvmSendThrowIfNotInit
  int i;
  int foundPos=-1;
  bool found;
  int *tmpTids;
  int *tmpContext;

  if(tid<0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSendStream",
                                                 "remTask",
                                                 "tid cannot be negative"));
  }
  if(numTid == 1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmSendStream",
                       "remTask",
                       "cannot remove tid. There must be at least one tid left"));
  }

  found = false;
  i = 0;
  while(found==false && i < numTid){
    if(sendToTids[i] == tid){
      foundPos = i;
      found = true;
    }
    else{
      i++;
    }
  }

  if(!found){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "remTask",
                                                   "tid is not in the sending array"));
  }

  sendToTids[foundPos] = sendToTids[numTid-1];
  context[foundPos] = context[numTid-1];
  
  numTid = numTid - 1;
  tmpTids = new int[numTid];
  tmpContext = new int[numTid];

  for(i=0; i<numTid; i++){
    tmpTids[i] = sendToTids[i];
    tmpContext[i] = context[i];
  }

  delete [] sendToTids;
  delete [] context;

  sendToTids = tmpTids;
  context = tmpContext;

  if(numTid==1){
    sendConnection = CPPvmOneToOne;
  }

}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setTid
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set task id of the receiving process
@MU  USAGE         : void setTid(int)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid  id of the receiving process
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::setTid(int tid)
{
  CPPvmSendThrowIfNotInit
  if (tid<0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSendStream",
                                                 "setTid",
                                                 "tid cannot be negative"));
  }
  if (tid==PvmNoParent){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmSendStream",
                       "setTid",
                       "cannot send to parent tid - No parent process"));
  }
  if (tid<0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSendStream",
                                                 "setTid",
                                                 "tid cannot be < 0"));
  }
  sendToTids[0] = tid;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setTid
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set task id of the receiving process
@MU  USAGE         : void setTid(int)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : tid      id of the receiving process
@MP            indexTid index in tid array
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::setTid(int tid, int indexTid)
{
  CPPvmSendThrowIfNotInit
  if ( indexTid>= numTid){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                   "cppvmSendStream",
                                   "setTid(int tid, int indexTid)",
                                   "Index is not in the array limits"));
  }
  if (tid==PvmNoParent){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                   "cppvmSendStream",
                                   "setTid(int tid, int indexTid)",
                                   "cannot send to parent tid - No parent process"));
  }
  if (tid<0){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSendStream",
                                                 "setTid(int tid, int indexTid)",
                                                 "tid cannot be < 0"));
  }
  sendToTids[indexTid] = tid;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setSendConnection
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the send connection 
@MU  USAGE         : void setSendConnection(cppvmSendConnectionType newConnection)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newConnection new connection
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::setSendConnection(cppvmSendConnectionType newConnection)
{
  CPPvmSendThrowIfNotInit

  if(newConnection != CPPvmOneToOne  &&
     newConnection != CPPvmMulticast   &&
     newConnection != CPPvmGroupBroadcast){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                            "cppvmSendStream",
                                            "setSendConnection",
                                            "the given connection does not exist. "));
  }
  sendConnection = newConnection;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setSendMode
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the sending mode
@MU  USAGE         : void setSendMode(cppvmSendMode newMode)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : newMode new sending mode
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::setSendMode(cppvmSendMode newMode)
{
  CPPvmSendThrowIfNotInit
  if(newMode != CPPvmBlockingSend  &&
     newMode != CPPvmNonblockingSend){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSendStream",
                                                 "setSendMode",
                                                 "the given mode does not exist. "));
  }
  sendMode = newMode;
}
 
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setInitSend
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the 'pvm_initsend' encoding scheme
@MU  USAGE         : void setInitSend(int initSend_)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : initSend_  new 'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::setInitSend(int initSend_)
{
  CPPvmSendThrowIfNotInit
  if(initSend_ != PvmDataDefault  &&
     initSend_ != PvmDataRaw      &&
     initSend_ != PvmDataInPlace){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                       "cppvmSendStream",
                       "setInitSend",
                       "the given encoding scheme does not exist. "));
  }
  initSend = initSend_;
}
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setGroup
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the group name
@MU  USAGE         : void setGroup()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group_ new group name
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::setGroup(char *group_)
{
  CPPvmSendThrowIfNotInit
  int len, size;
    
  if (group_ == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "setGroup",
                                                   "group name is NULL"));
  }

  if(groupAllocated){
    delete [] group; // delete old string
  }
  len = strlen(group_);       // char number
  size = len + 1;             // old size plus '\0'
  group = new char [size];    // allocate memory
  strcpy(group, group_);      // copy string
  groupAllocated = true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTids
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get tasks ids of the receiving processes
@MD                  for multicasting
@MU  USAGE         : void getTids(int *tids, int *tid)
@MC  CONSTRAINTS   : the space for the tids must be allocated by the caller
@MC                  (use getNumTid() for the number of tids)
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : num   number of ids
@MP  INOUT   : tids  array of the ids of the receiving processes
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::getTids(int *tids, int *num)
{
  CPPvmSendThrowIfNotInit
  int i;
  
  (*num) = numTid;
  for(i=0; i<numTid; i++){
    tids[i] = sendToTids[i];
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getInitSend
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the 'pvm_initsend' encoding scheme
@MU  USAGE         : int getInitSend()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : int  'pvm_initsend' encoding scheme
******************************************************************************/
int
cppvmSendStream::getInitSend()
{
  CPPvmSendThrowIfNotInit

  return initSend;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getGroup
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the group name
@MU  USAGE         : char* getGroup()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : char *  NULL: no group name
@MP                    else: group name
******************************************************************************/
char *
cppvmSendStream::getGroup()
{
  CPPvmSendThrowIfNotInit

  return group;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setMsgBuf
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set if a new send message buffer is wanted or not, and
@MD                  set also the encoding of this message buffer
@MU  USAGE         : void setMsgBuf(bool ownMsgBuffer,
@MU                                 int encoding);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : ownMsgBuffer true: new message buffer for each sent message
@MP                         false: use default message buffer
@MP            encoding:    PvmDataDefault: XDR encoding
@MP                         PvmDataRaw:     no encoding
@MP                                         only for homogenous architectures
@MP                         PvmDataInPlace: data left in place during packing;
@MP                                         not allowd for shared memory arch.
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::setMsgBuf(bool ownMsgBuffer,
		       int encoding)
{
  msgBufSend = ownMsgBuffer;
  if(encoding != PvmDataDefault &&
     encoding != PvmDataRaw     &&
     encoding != PvmDataInPlace){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmObject",   
                                                   "setMsgBuf",
                                                   "invalid sendEncoding argument"));
  }    
  sendBufEncoding = encoding;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSendStream::cppvmPrintOn(ostream& strm)
{
  CPPvmSendThrowIfNotInit
  int i;
  
  if(sendConnection==CPPvmOneToOne){
    strm << " One to one connection " << endl;
  }
  else if(sendConnection==CPPvmMulticast){
    strm << " Multicast " << endl;
  }
  else{
    strm << " Group-Broadcast " << endl;
  }
  if(sendMode==CPPvmBlockingSend){
    strm << " with blocking send. " << endl;
  }
  else{
    strm << " with nonblocking send. " << endl;
  }
  for(i=0; i<numTid; i++){
    strm << "Send to process with task id: " << sendToTids[i] 
         << " with context: " << context[i] << endl;
  }
  strm << "send with own message buffer: ";
  if(msgBufSend==true){
    strm << "true. encoding: ";
    if(sendBufEncoding == PvmDataDefault){
      strm << "PvmDataDefault. ";
    }
    else if(sendBufEncoding == PvmDataRaw){
      strm << "PvmDataRaw. ";
    }
    else {
      strm << "PvmDataInPlace. ";
    }
  } 
  else{
    strm << "false. ";
  } 
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resetObject
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : reset all object variables
@MU  USAGE         : void resetObject()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::resetObject()
{
  // set some default values
  sendMode = CPPvmNonblockingSend;
  sendConnection = CPPvmOneToOne;
  initSend = PvmDataDefault;
  msgBufSend = false;
  sendBufEncoding = PvmDataDefault;

  if(initialized){
    delete [] sendToTids;
    delete [] context;
    if(groupAllocated){
      delete [] group;
    }
  }
  numTid = 1;
  sendToTids = new int[numTid];
  context = new int[numTid];
  group = NULL;
  
  sendToTids[0] = 0;
  context[0] = pvm_getcontext();

  initialized = true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setContext
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the context variable
@MU  USAGE         : void setContext(int context_)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : context_ new context
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::setContext(int context_)
{
  CPPvmSendThrowIfNotInit
  context[0] = context_;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for cppvmTask objects
@MU  USAGE         : void set(cppvmTask& task,
@MU                           cppvmSendTask mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task     description of the task to send to
@MP            mode     send mode:
@MP                     * CPPvmSBtask:  send blocking
@MP                     * CPPvmSNBtask: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(cppvmTask& task,
                     cppvmSendTask mode,
                     int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmOneToOne);

  setContext(task.getContextId());
  setTid(task.getTid());

  if(mode==CPPvmSBtask){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBtask){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(cppvmTask& task...)",
                                                   "send mode does not exist."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for cppvmConnection objects
@MU  USAGE         : void set(cppvmConnection& conn,
@MU                           cppvmSendConnection mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : conn     description of the parent process to send to
@MP            mode     send mode:
@MP                     * CPPvmSBparent:  send blocking
@MP                     * CPPvmSNBparent: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(cppvmConnection& conn,
                     cppvmSendConnection mode,
                     int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmOneToOne);

  setContext(conn.getContextId());
  setTid(conn.getParentTid());

  if(mode==CPPvmSBparent){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBparent){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(cppvmConnection& conn...)",
                                                   "send mode does not exist."));
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for cppvmSpawnConnection objects
@MU  USAGE         : void set(cppvmSpawnConnection& sconn,
@MU                           cppvmSendSpawnConnection mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : sconn     description of the child process to send to
@MP            mode     send mode:
@MP                     * CPPvmSBchild:  send blocking
@MP                     * CPPvmSNBchild: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(cppvmSpawnConnection& sconn,
                        cppvmSendSpawnConnection mode,
                        int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmOneToOne);

  setContext(sconn.getContextId());
  setTid(sconn.getChildTid());

  if(mode==CPPvmSBchild){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBchild){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                     "cppvmSendStream",
                                     "set(cppvmSpawnConnection& sconn...)",
                                     "send mode does not exist."));
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for cppvmSpawnMultiConnection objects
@MU  USAGE         : void set(cppvmSpawnMultiConnection& smconn,
@MU                           cppvmSendSpawnMultiConnection mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : smconn   description of the child processes to send to
@MP            mode     send mode:
@MP                     * CPPvmSBtaggedChildren: send blocking
@MP                                              to tagged children
@MP                     * CPPvmSBtaggedChildren: send nonblocking
@MP                                              to tagged children
@MP                     * CPPvmSBallChildren:  send blocking to all children
@MP                     * CPPvmSNBallChildren: send nonblocking to all children
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(cppvmSpawnMultiConnection& smconn,
                        cppvmSendSpawnMultiConnection mode,
                        int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmOneToOne);

  if(mode==CPPvmSBtaggedChildren ||
     mode==CPPvmSBallChildren){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBtaggedChildren ||
          mode==CPPvmSNBallChildren){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                             "cppvmSendStream",
                             "set(cppvmSpawnMultiConnection& smconn...)",
                             "send mode does not exist."));
  }


  bool mustBeTagged = false;
  bool firstSet = true;
  unsigned int i;
  
  if(mode==CPPvmSBtaggedChildren ||
     mode==CPPvmSNBtaggedChildren){
    mustBeTagged = true;
  }

  for(i=0; i<smconn.numElems(); i++){
    // add only tagged elements
    if(mustBeTagged){
      if(smconn.isElemTagged(i)){
        if(firstSet){
          setTid(smconn[i].getChildTid());
          setContext(smconn[i].getContextId());
          firstSet=false;
        }
        else{
          addTask(smconn[i].getChildTid(),
                  smconn[i].getContextId());
        }
      }
    }
    // add all elements
    else{
      if(firstSet){
        setTid(smconn[i].getChildTid());
        setContext(smconn[i].getContextId());
        firstSet=false;
      }
      else{
        addTask(smconn[i].getChildTid(),
                smconn[i].getContextId());
      }
    }
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for group broadcasting
@MD                  with current context
@MU  USAGE         : void set(char *group_,
@MU                           cppvmSendGroup mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group_   group name
@MP            mode     send mode:
@MP                     * CPPvmSBgroup:  send blocking
@MP                     * CPPvmSNBgroup: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(char *group_,
                     cppvmSendGroup mode,
                     int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  set(group_, context[0], mode, initSend_);
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for group broadcasting
@MD                  with given context
@MU  USAGE         : void set(char *group_, int context_, 
@MU                           cppvmSendGroup mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : group_   group name
@MP            context_ context to send with
@MP            mode     send mode:
@MP                     * CPPvmSBgroup:  send blocking
@MP                     * CPPvmSNBgroup: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(char *group_,
                     int context_,
                     cppvmSendGroup mode,
                     int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmGroupBroadcast);
  setGroup(group_);
  setContext(context_);
  
  if(mode==CPPvmSBgroup){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBgroup){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(char *group_ ...)",
                                                   "send mode does not exist."));
  }
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for send to a specific task
@MD                  with current context
@MU  USAGE         : void set(int taskId,
@MU                           cppvmSendTid mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskId   task id of the process to send to
@MP            mode     send mode:
@MP                     * CPPvmSBtid:  send blocking
@MP                     * CPPvmSNBtid: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(int taskId,
                        cppvmSendTid mode,
                        int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  set(taskId, context[0], mode, initSend_);
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for send to a specific task
@MD                  with given context
@MU  USAGE         : void set(int taskId, int context_, 
@MU                           cppvmSendTid mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskId   task id of the process to send to
@MP            context_ context to send with
@MP            mode     send mode:
@MP                     * CPPvmSBtid:  send blocking
@MP                     * CPPvmSNBtid: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(int taskId,
                     int context_,
                     cppvmSendTid mode,
                     int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmOneToOne);

  setContext(context_);
  setTid(taskId);
  
  if(mode==CPPvmSBtid){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBtid){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(int taskId...)",
                                                   "send mode does not exist."));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for send to specific tasks
@MD                  with current context
@MU  USAGE         : cppvmSendStream(int* taskIds,
@MU                           unsigned int numTasks,
@MU                           cppvmSendTids mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskIds   array of task ids of the processes to send to
@MP            numTasks number of tasks in 'taskIds' (array length)
@MP            mode     send mode:
@MP                     * CPPvmSBtids:  send blocking
@MP                     * CPPvmSNBtids: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(int* taskIds,
                     unsigned int numTasks,
                     cppvmSendTids mode,
                     int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  set(taskIds, numTasks, context[0], mode, initSend_);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for send to specific tasks
@MD                  with given context
@MU  USAGE         : send(int* taskIds, unsigned int numTasks,
@MU                           int context_, 
@MU                           cppvmSendTids mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskIds  array of  task ids of the processes to send to
@MP            numTasks number of tasks in 'taskIds' (array length)
@MP            context_ context to send with
@MP            mode     send mode:
@MP                     * CPPvmSBtids:  send blocking
@MP                     * CPPvmSNBtids: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(int* taskIds,
                     unsigned int numTasks,
                     int context_,
                     cppvmSendTids mode,
                     int initSend_)
{
  if(taskIds==NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(int* tids,...)",
                                                   "taskIds pointer is NULL."));
  }

  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmOneToOne);

  if(mode==CPPvmSBtids){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBtids){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(int* tids,...)",
                                                   "send mode does not exist."));
  }

  // set first task...
  if(numTasks>0){
    setTid(taskIds[0]);
    setContext(context_);
  }
  // ... and the rest
  unsigned int i;
  for(i=1;i<numTasks;i++){
    addTask(taskIds[i], context_);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for send to specific tasks
@MD                  with given context array
@MU  USAGE         : send(int* taskIds, unsigned int numTasks,
@MU                           int* contextArr, 
@MU                           cppvmSendTids mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : taskIds    array of  task ids of the processes to send to
@MP            numTasks   number of tasks in 'taskIds' (array length)
@MP            contextArr array of context ids to send with
@MP            mode       send mode:
@MP                       * CPPvmSBtids:  send blocking
@MP                       * CPPvmSNBtids: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(int* taskIds,
                     unsigned int numTasks,
                     int* contextArr,
                     cppvmSendTids mode,
                     int initSend_)
{
  if(taskIds==NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(int* tids,...)",
                                                   "taskIds pointer is NULL."));
  }
  if(contextArr==NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(int* tids,...)",
                                                   "contextArr pointer is NULL."));
  }
  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmOneToOne);

  if(mode==CPPvmSBtids){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBtids){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(int* tids,...)",
                                                   "send mode does not exist."));
  }

  // set first task...
  if(numTasks>0){
    setTid(taskIds[0]);
    setContext(contextArr[0]);
  }
  // ...and the rest
  unsigned int i;
  for(i=1;i<numTasks;i++){
    addTask(taskIds[i], contextArr[i]);
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for send back to a process, from which
@MD                  a message was received with 'recStrm'
@MD                  with current context
@MU  USAGE         : void set(cppvmReceiveStream& recStrm,
@MU                           cppvmSendBack mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : recStrm  send back to the process from which a message
@MP                     was received with this stream
@MP            mode     send mode:
@MP                     * CPPvmSBback:  send blocking
@MP                     * CPPvmSNBback: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(cppvmReceiveStream& recStrm,
                     cppvmSendBack mode,
                     int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  set(recStrm, context[0], mode, initSend_);
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : set
@MK  CLASS NAME    : cppvmSendStream
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set method for send back to a process, from which
@MD                  a message was received with 'recStrm'
@MD                  with given context
@MU  USAGE         : void set(cppvmReceiveStream& recStrm, int context_, 
@MU                           cppvmSendBack mode,
@MU                           int initSend_);
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : recStrm  send back to the process from which a message
@MP                     was received with this stream
@MP            context_ context to send with
@MP            mode     send mode:
@MP                     * CPPvmSBtid:  send blocking
@MP                     * CPPvmSNBtid: send nonblocking
@MP            initSend_  'pvm_initsend' scheme
@MP                       PvmDataDefault: XDR encoding (works everytime)
@MP                       PvmDataRaw: do NOT encode information
@MP                                   (only for send/receive between task on
@MP                                    homogeneous architectures)
@MP                       PvmDataInPlace: do not copy data in sending buffer,
@MP                                       use the data memory instead
@MP                                       ATTENTION: the data memory MUST not
@MP                                       deleted between packing and sending!
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSendStream::set(cppvmReceiveStream& recStrm,
                     int context_,
                     cppvmSendBack mode,
                     int initSend_)
{
  // reset/init object
  // set 'context' to current context
  resetObject();

  // set values
  setMsgBuf(false, PvmDataDefault);
  setInitSend(initSend_);
  setSendConnection(CPPvmOneToOne);

  setContext(context_);
  setTid(recStrm.getLastRecFromTid());
  
  if(mode==CPPvmSBback){
    setSendMode(CPPvmBlockingSend);
  }
  else if(mode==CPPvmSNBback){
    setSendMode(CPPvmNonblockingSend);
  }
  else{
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSendStream",
                                                   "set(cppvmRecieveStream recStrm...)",
                                                   "send mode does not exist."));
  }
}



/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

