/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_spaw.cpp 
@IE  CREATED     : 5.12.1997
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm spawn class 
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// includes for predeclared classes to avoid cyclic includes
#include "pvm_str.h"
#include "pvm_task.h"
// include own class
#include "pvm_spaw.h"

/*@@@**************************************************************************
**********                                                           **********
*******                 CONSTANTS FOR TYPE DEFINITIONS                  *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_spaw.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : spawnError
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : put error condition of 'pvm_spawn' in an string
@MU  USAGE         : bool spawnError(int infoErr, int tidErr, char errMsg[]);
@MC  CONSTRAINTS   : 'errMsg' array must be given by caller
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : infoErr   info (error code) returned by 'pvm_spawn'
@MP            tidErr    tids (error code) returned by 'pvm_spawn'
@MP  INOUT   : errMsg    error message
@MP  OUT     :
@MP  RETURNS : bool  true:  error occurred
@MP                  false: no error
******************************************************************************/
bool
cppvmSpawnConnection::spawnError(int infoErr, int tidErr, char errMsg[])
{
  if(! (infoErr < 0 || tidErr < 0)){
    // both not below zero so no error occurred
    return false;
  }
  
  if(infoErr == PvmBadParam ||
     tidErr  == PvmBadParam){
    strcpy(errMsg, "invalid argument value");
  }
  if(infoErr == PvmNoHost ||
     tidErr  == PvmNoHost){
    strcpy(errMsg, "specified host is not in the virtual machine");
  }
  if(infoErr == PvmNoFile ||
     tidErr  == PvmNoFile){
    strcpy(errMsg, "specified executable cannot be found. The default location is in $PVM_ROOT/bin/$PVM_ARCH");
  }
  if(infoErr == PvmNoMem ||
     tidErr  == PvmNoMem){
    strcpy(errMsg, "malloc failed. Not enough memory on host.");
  }
  if(infoErr == PvmSysErr ||
     tidErr  == PvmSysErr){
    strcpy(errMsg, "pvmd not responding");
  }
  if(infoErr == PvmOutOfRes ||
     tidErr  == PvmOutOfRes){
    strcpy(errMsg, "out of resources");
  }
  return true;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : spawnIt
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : spawn the process
@MD                  All stderr and stdout of spawned tasks are loged in a
@MD                  file '/tmp/pvml.<uid>'.
@MU  USAGE         : void spawnIt(char *task, char **argv, char *where,
@MU                               cppvmContextMode contextFlag,
@MU                               int   contextId_,
@MU                               cppvmOptionMode optionFlag);
@MU                               
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task name (program) to start
@MP            argv  array of command line arguments for task
@MP            optionFlag specifies special PVM options:
@MP               * 'PvmTaskDefault' PVM can choose any machine
@MP               * 'PvmTaskHost'    'where' specifies a particular host
@MP               * 'PVMTaskArch'    'where' specifies a type of architecture 
@MP               * 'PVMTaskDebug'   start up processes under debugger
@MP               * 'PVMTaskTrace'   process will generate PVM trace data
@MP               * 'PVMMppFront'    tart process on MPP front-end
@MP               * 'PVMHostCompl'   use complement host set
@MP            where specifies where to start the task
@MP                  'where' depends on the value of 'optionFlag'
@MP                   if 'optionFlag' is 0, then 'where' is ignored and PVM
@MP                   will select the most appropriate host
@MP            contextFlag specifies the context of the child process:
@MP               * 'CPPvmCurrentContext': spawn within the context of the
@MP                 parent process
@MP               * 'CPPvmNewContext':  create a new context and spawn the
@MP                 child process within this new context
@MP               * 'CPPvmGivenContext': spawn the child process in the context
@MP                 given by 'contextId_'
@MP            contextId_ specifies a context; only used if 'contextFlag'
@MP               is set to 'CPPvmGivenContext'
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnConnection::spawnIt(char *task,
                              char **argv,
                              int optionFlag,
                              char *where,
                              cppvmContextMode contextFlag,
                              int   contextId_)
{
  int numSpawnedTasks;
  char errMsg[100];

  // remind child data
  childName = task;
  childParams.clean();
  int i=0;
  while(argv[i]!=NULL){
    childParams.add(new cppvmString(argv[i]));
    i++;
  }
  
  if(task == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSpawnConnection",
                                                   "spawnIt",
                                                   "task name is NULL"));
    return;
  }
  if(argv == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSpawnConnection",
                                                   "spawnIt",
                                                   "argv is NULL"));
    return;
  }
  if(where == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSpawnConnection",
                                                   "spawnIt",
                                                   "where is NULL"));
    return;
  }

  // set context parameter
  setContextMode(contextFlag, contextId_);
  // add host if specified in 'optionFlag' and 'where'
  setHost(optionFlag, where);
  
  // spawn the application in this context
  setOwnContext();
  
  // spawn one child process
  numSpawnedTasks = pvm_spawn(task, argv,
                              optionFlag, where, 1, childTid);

  if(spawnError(numSpawnedTasks, childTid[0], errMsg)){
    // error while spawning
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmSpawnConnection",
                                                "spawnIt",
                                                errMsg));
  }
  
  if(numSpawnedTasks != 1){
    // error while spawning
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmSpawnConnection",
                                                "spawnIt",
                                                "cannot spawn process."));
  }

  enrolledChild = clock();

  setHostEnvironment();
  
  // reset the old context
  resetContext();

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setContextMode
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the child context 
@MU  USAGE         : void setContextMode(cppvmContextMode contextFlag,
@MU                                      int   contextId_); 
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : contextFlag specifies the context of the child process:
@MP               * 'CPPvmCurrentContext': spawn within the context of the
@MP                 parent process
@MP               * 'CPPvmNewContext':  create a new context and spawn the
@MP                 child process within this new context
@MP               * 'CPPvmGivenContext': spawn the child process in the context
@MP                 given by 'contextId_'
@MP            contextId_ specifies a context; only used if 'contextFlag'
@MP               is set to 'CPPvmGivenContext'
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnConnection::setContextMode(cppvmContextMode contextFlag,
                                   int   contextId_)
{
  if(contextFlag==CPPvmCurrentContext){
    contextId = pvm_getcontext();
  }
  else if(contextFlag==CPPvmGivenContext){
    contextId = contextId_;
  }
  else if(contextFlag==CPPvmNewContext){
    contextId = pvm_newcontext();
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setHost
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add host if specified in 'optionFlag' and 'where'
@MU  USAGE         : void setHost(int optionFlag, char * where)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : optionFlag specifies special PVM options:
@MP               * 'PvmTaskDefault' PVM can choose any machine
@MP               * 'PvmTaskHost'    'where' specifies a particular host
@MP               * 'PVMTaskArch'    'where' specifies a type of architecture 
@MP               * 'PVMTaskDebug'   start up processes under debugger
@MP               * 'PVMTaskTrace'   process will generate PVM trace data
@MP               * 'PVMMppFront'    tart process on MPP front-end
@MP               * 'PVMHostCompl'   use complement host set
@MP            where: specifies the host on which the child is spawned
@MP               * "" use the same host as the parent process
@MP               * else: add the host to the virtual machine
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnConnection::setHost(int optionFlag,
                              char *where)
{
  if(!(flagMemberOfSet(optionFlag,PvmTaskHost))){
    // the process should not be spawned on a
    // specific host
    return;
    
  }
  if(where == NULL || (strcmp(where, "")==0)){
    // error while spawning
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                     "cppvmSpawnConnection",
                                     "setHost",
                                     "'PvmTaskHost' is set and the host specifier 'where' is NULL or an empty string"));
  }
  // a host was specified
  // try to add a new host
  addHostTry(where);
}



/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSpawnConnection
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MD                  All stderr and stdout of spawned tasks are loged in a
@MD                  file '/tmp/pvml.<uid>'.
@MU  USAGE         : void cppvmSpawnConnection(char *task, char *args,
@MU                               int optionFlag,
@MU                               char *where,
@MU                               cppvmContextMode contextFlag,
@MU                               int   contextId_);
@MU                               
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task name (program) to start
@MP            args  arguments for task
@MP            optionFlag specifies special PVM options:
@MP               * 'PvmTaskDefault' PVM can choose any machine
@MP               * 'PvmTaskHost'    'where' specifies a particular host
@MP               * 'PVMTaskArch'    'where' specifies a type of architecture 
@MP               * 'PVMTaskDebug'   start up processes under debugger
@MP               * 'PVMTaskTrace'   process will generate PVM trace data
@MP               * 'PVMMppFront'    start process on MPP front-end
@MP               * 'PVMHostCompl'   use complement host set
@MP            where specifies where to start the task
@MP                  'where' depends on the value of 'optionFlag'
@MP                   if 'optionFlag' is 0, then 'where' is ignored and PVM
@MP                   will select the most appropriate host
@MP            contextFlag specifies the context of the child process:
@MP               * 'CPPvmCurrentContext': spawn within the context of the
@MP                 parent process
@MP               * 'CPPvmNewContext':  create a new context and spawn the
@MP                 child process within this new context
@MP               * 'CPPvmGivenContext': spawn the child process in the context
@MP                 given by 'contextId_'
@MP            contextId_ specifies a context; only used if 'contextFlag'
@MP               is set to 'CPPvmGivenContext'
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSpawnConnection::cppvmSpawnConnection(char *task,
                                           char *args,
                                           int optionFlag,
                                           char *where,
                                           cppvmContextMode contextFlag,
                                           int   contextId_)
    : cppvmConnection(), childKilled(false)
{
  char *argsArray[2];

  // initialize program argument(s)
  // the end of the array is indicated by a 'NULL' value
  argsArray[0] = args;
  argsArray[1] = NULL;

#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif
  
  CPPvmInternalTry{
    spawnIt(task, argsArray, optionFlag, where, contextFlag, contextId_);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSpawnConnection
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MD                  All stderr and stdout of spawned tasks are loged in a
@MD                  file '/tmp/pvml.<uid>'.
@MU  USAGE         : void cppvmSpawnConnection(char *task, char **argv,
@MU                               int optionFlag,
@MU                               char *where,
@MU                               cppvmContextMode contextFlag,
@MU                               int   contextId_);
@MU                               
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task name (program) to start
@MP            argv  array of command line arguments for task
@MP            optionFlag specifies special PVM options:
@MP               * 'PvmTaskDefault' PVM can choose any machine
@MP               * 'PvmTaskHost'    'where' specifies a particular host
@MP               * 'PVMTaskArch'    'where' specifies a type of architecture 
@MP               * 'PVMTaskDebug'   start up processes under debugger
@MP               * 'PVMTaskTrace'   process will generate PVM trace data
@MP               * 'PVMMppFront'    start process on MPP front-end
@MP               * 'PVMHostCompl'   use complement host set
@MP            where specifies where to start the task
@MP                  'where' depends on the value of 'optionFlag'
@MP                   if 'optionFlag' is 0, then 'where' is ignored and PVM
@MP                   will select the most appropriate host
@MP            contextFlag specifies the context of the child process:
@MP               * 'CPPvmCurrentContext': spawn within the context of the
@MP                 parent process
@MP               * 'CPPvmNewContext':  create a new context and spawn the
@MP                 child process within this new context
@MP               * 'CPPvmGivenContext': spawn the child process in the context
@MP                 given by 'contextId_'
@MP            contextId_ specifies a context; only used if 'contextFlag'
@MP               is set to 'CPPvmGivenContext'
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSpawnConnection::cppvmSpawnConnection(char *task,
                                           char **argv,
                                           int optionFlag,
                                           char *where,
                                           cppvmContextMode contextFlag,
                                           int   contextId_)
    : cppvmConnection(), childKilled(false)
{
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.lock();
#endif
  
  CPPvmInternalTry{
    spawnIt(task, argv, optionFlag, where, contextFlag, contextId_);
  }
  CPPvmInternalCatchDots{
#ifdef CPPvmThread
    cppvmGlobalVariables::communicationMutex.unlock();
#endif
    CPPvmInternalThrowAgain;
  }
#ifdef CPPvmThread
  cppvmGlobalVariables::communicationMutex.unlock();
#endif
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmSpawnConnection
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmSpawnConnection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSpawnConnection::~cppvmSpawnConnection()
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to assign
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmSpawnConnection&
cppvmSpawnConnection::operator= (const cppvmSpawnConnection& obj)
{
  // no self-assignment
  if(cppvmIsSame(obj)){
    return *this;
  }
  
  childTid[0] = obj.childTid[0];

  enrolledChild = obj.enrolledChild;
  childKilled = obj.childKilled;
  childName = obj.childName;
  childParams.clean();
  unsigned int i;
  for(i=0; i<obj.childParams.numElems(); i++){
    cppvmString* tmpStr = new cppvmString;
    (*tmpStr) = (*((cppvmString*)(obj.childParams.getConstObjPtr(i))));
    childParams.add(tmpStr);
    i++;
  }
  
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmSpawnConnection::operator== (const cppvmSpawnConnection& obj)
{
  return (cppvmConnection::operator==(obj) &&
          childTid[0] == obj.childTid[0] &&
          childName == obj.childName &&
          childParams == obj.childParams &&
          childKilled == obj.childKilled);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmSpawnConnection::operator== (const cppvmSpawnConnection& obj) const
{
  return (cppvmConnection::operator==(obj) &&
          childTid[0] == obj.childTid[0] &&
          childName == obj.childName &&
          childParams == obj.childParams &&
          childKilled == obj.childKilled);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnConnection::cppvmPrintOn(ostream& strm)
{
  if(childKilled){
    strm << "Child process was killed" << endl;
  }
  else{
    strm << "Child TID: " << childTid[0] << endl;
    strm << "Parent TID: " << myTid << endl;
    strm << "Child name: " << childName << endl;
    strm << "Child parameter: " << childParams << endl;
    strm << "Context id: " << contextId << endl;
    strm << "Host information: name: " << childHostName
         << " architecture: " << childHostArchitecture
         << " speed: " << childHostSpeed
         << " daemon task id: " << childHostDaemonTid << endl;
  }
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getChildTask
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get the description of the childt task as cppvmTask
@MD                  object
@MU  USAGE         : getChildTask();
@MC  CONSTRAINTS   : if there is no child task, an exception is thrown
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : cppvmTask task description
******************************************************************************/
cppvmTask
cppvmSpawnConnection::getChildTask()
{
  PVM_THROW_IF_CHILD_KILLED
  return cppvmTask(childTid[0], contextId, childName);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : killChild
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : kill the child(s) of this parent process 
@MU  USAGE         : void killChild()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnConnection::killChild()
{
  int info;
  
  if(childKilled){
    CPPvmInternalThrow( cppvmExceptionForbiddenCall(THIS_MODULE,
                                                    "cppvmSpawnConnection",
                                                    "killChild",
                                                    "child process was already killed"));
  }

  // killing spawned processes 
  info = pvm_kill(childTid[0]);
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmSpawnConnection",
                                                "killChild",
                                                "cannot kill child process."));
  }
  
  childKilled = true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : flagMemberOfSet
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : check if a flag is part of a set of flags or not
@MU  USAGE         : bool flagMemberOfSet(int flagSet, int flag)
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : flagSet set of flags
@MP            flag    test if this flag is part of the set of flags
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool  true:  is part of set
@MP                  false: is not part of set
******************************************************************************/
bool
cppvmSpawnConnection::flagMemberOfSet(int flagSet, int flag)
{
  if((flagSet & flag)==0){
    return false;
  }
  else{
    return true;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setHostEnvironment
@MK  CLASS NAME    : cppvmSpawnConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the variables "childHostDaemonTid", "childHostName",
@MD                  "childHostArchitecture", and "childHostSpeed"  
@MU  USAGE         : void setHostEnvironment()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnConnection::setHostEnvironment()
{
  childHostDaemonTid = getHostDaemonTid(childTid[0]);
  childHostName = getHostName(childTid[0]);
  childHostArchitecture = getHostArchitecture(childTid[0]);
  childHostSpeed = getHostSpeed(childTid[0]);
}


/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

