/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_spmu.cpp 
@IE  CREATED     : 4.2.1998
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm multiple spawn class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// includes for predeclared classes to avoid cyclic includes
#include "pvm_task.h"
#include "pvm_coll.h"
// include own class
#include "pvm_spmu.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/
#define THIS_MODULE "pvm_spmu.cpp"

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : initTag
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : init the tag array
@MU  USAGE         : initTag(unsigned int numElem);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the former 'tag' will be not deleted!
@MR                  
@MR                  
@M#  --------------------------------------------------------------------------
@MP  IN      : numElems  number of tag elements
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::initTag(unsigned int numElems)
{
  unsigned int i;
  
  tags = new char[numElems];

  if(numElems<CPPvmSpawnMultiConnectionTagCleanFieldSize*2){
    // assign single values
    for(i=0;i<numElems;i++){
      tags[i] = (char) 0;
    }
  }
  else{
    // for large arrays use memcopy
    unsigned int numMemCopy;
    unsigned int restSingleCopyIndex;
    
    // clean memory to copy
    char tmpArray[CPPvmSpawnMultiConnectionTagCleanFieldSize];  
    for(i=0; i< CPPvmSpawnMultiConnectionTagCleanFieldSize; i++){
      tmpArray[i]=(char) 0;
    }
    
    numMemCopy = numElems / CPPvmSpawnMultiConnectionTagCleanFieldSize;
    restSingleCopyIndex = numMemCopy * CPPvmSpawnMultiConnectionTagCleanFieldSize;
    
    unsigned int index = 0;
    // use memcpy
    for(i=0; i< numMemCopy; i++){
      memcpy((void*)(&tags[index]), (void*)(&tmpArray[0]), 
             sizeof(char) *  CPPvmSpawnMultiConnectionTagCleanFieldSize);
      index = index +  CPPvmSpawnMultiConnectionTagCleanFieldSize;
    }
    
    // copy the rest as single values
    for(i=restSingleCopyIndex; i<numElems; i++){
      tags[i]=(char) 0;
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resizeTag
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : resize the tag array
@MU  USAGE         : resizeTag(unsigned int numElem, unsigned int oldElem);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@MR                  
@MR                  
@M#  --------------------------------------------------------------------------
@MP  IN      : numElems  new number of tag elements
@MP            oldElems  old number of tag elements
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::resizeTag(unsigned int numElems,
                                     unsigned int oldElems)
{
  char *tmp;
  unsigned int i;
  unsigned int maxLen;
  
  tmp = new char[numElems];

  if(numElems>oldElems){
    maxLen=oldElems;
  }
  else{
    maxLen=numElems;
  }
  // copy old values
  if(maxLen>0){
    memcpy((void*)tmp, (void*)tags, sizeof(char) *  maxLen);
  }

  delete [] tags;
  tags = tmp;

  if(numElems<=maxLen){
    // done
    return;
  }
  
  // clean all (possible) new elements

  if(numElems<CPPvmSpawnMultiConnectionTagCleanFieldSize*2){
    // assign single values
    for(i=maxLen;i<numElems;i++){
      tags[i] = (char) 0;
    }
  }
  else{
    // for large arrays use memcopy
    unsigned int numMemCopy;
    unsigned int restSingleCopyIndex;
    
    // clean memory to copy
    char tmpArray[CPPvmSpawnMultiConnectionTagCleanFieldSize];  
    for(i=0; i< CPPvmSpawnMultiConnectionTagCleanFieldSize; i++){
      tmpArray[i]=(char) 0;
    }
    
    numMemCopy = (numElems-maxLen) / CPPvmSpawnMultiConnectionTagCleanFieldSize;
    restSingleCopyIndex = maxLen + (numMemCopy * 
                                    CPPvmSpawnMultiConnectionTagCleanFieldSize);
    
    unsigned int index = maxLen;
    // use memcpy
    for(i=0; i< numMemCopy; i++){
      memcpy((void*)(&tags[index]), (void*)(&tmpArray[0]), 
             sizeof(char) *  CPPvmSpawnMultiConnectionTagCleanFieldSize);
      index = index +  CPPvmSpawnMultiConnectionTagCleanFieldSize;
    }
    
    // copy the rest as single values
    for(i=restSingleCopyIndex; i<numElems; i++){
      tags[i]=(char) 0;
    }
  }

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSpawnMultiConnection
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MD                  All stderr and stdout of spawned tasks are loged in a
@MD                  file '/tmp/pvml.<uid>'.
@MU  USAGE         : void cppvmSpawnMultiConnection(char *task,
@MU                               unsigned int numTask,
@MU                               char *args,
@MU                               int optionFlag,
@MU                               char *where,
@MU                               cppvmContextMode contextFlag,
@MU                               int   contextId_,
@MU                               cppvmMultiContextMode contextFlag);
@MU                               
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task name (program) to start
@MP            numTasks number of task to spawn
@MP            args  arguments for task
@MP            optionFlag specifies special PVM options:
@MP               * 'PvmTaskDefault' PVM can choose any machine
@MP               * 'PvmTaskHost'    'where' specifies a particular host
@MP               * 'PVMTaskArch'    'where' specifies a type of architecture 
@MP               * 'PVMTaskDebug'   start up processes under debugger
@MP               * 'PVMTaskTrace'   process will generate PVM trace data
@MP               * 'PVMMppFront'    tart process on MPP front-end
@MP               * 'PVMHostCompl'   use complement host set
@MP            where specifies where to start the task
@MP                  'where' depends on the value of 'optionFlag'
@MP                   if 'optionFlag' is 0, then 'where' is ignored and PVM
@MP                   will select the most appropriate host
@MP            contextFlag specifies the context of one child process:
@MP               * 'CPPvmCurrentContext': spawn within the context of the
@MP                 parent process
@MP               * 'CPPvmNewContext':  create a new context and spawn the
@MP                 child within this new context
@MP               * 'CPPvmGivenContext': spawn the children in the context
@MP                 given by 'contextId_'
@MP            contextId_ specifies a context; only used if 'contextFlag'
@MP               is set to 'CPPvmGivenContext'
@MP            multiContextFlag specifies the context of all child process:
@MP               this parameter is only used if 'contextFlag' is set to
@MP               'CPPvmNewContext'
@MP               * 'CPPvmOneNewContext':  create one new context and spawn all
@MP                 children within this new context
@MP               * 'CPPvmAllNewContext':  create a new context for every
@MP                  child and spawn the children within these contexts
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSpawnMultiConnection::cppvmSpawnMultiConnection(char *task,
                                                     unsigned int numTasks,
                                                     char *args,
                                                     int optionFlag,
                                                     char *where,
                                                     cppvmContextMode
                                                       contextFlag,
                                                     int   contextId_,
                                                     cppvmMultiContextMode
                                                       multiContextFlag)
    : cppvmProcessHandling()
{
  char *argsArray[2];

  // initialize program argument(s)
  // the end of the array is indicated by a 'NULL' value
  argsArray[0] = args;
  argsArray[1] = NULL;

  spawnIt(task, numTasks, argsArray, optionFlag, where, 
          contextFlag, contextId_, multiContextFlag);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSpawnMultiConnection
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MD                  All stderr and stdout of spawned tasks are loged in a
@MD                  file '/tmp/pvml.<uid>'.
@MU  USAGE         : void cppvmSpawnMultiConnection(char *task,
@MU                               unsigned int numTask,
@MU                               char **argv,
@MU                               int optionFlag,
@MU                               char *where,
@MU                               cppvmContextMode contextFlag,
@MU                               int   contextId_,
@MU                               cppvmMultiContextMode contextFlag);
@MU                               
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task name (program) to start
@MP            numTasks number of task to spawn
@MP            argv vector of command line arguments for a spawned task
@MP            optionFlag specifies special PVM options:
@MP               * 'PvmTaskDefault' PVM can choose any machine
@MP               * 'PvmTaskHost'    'where' specifies a particular host
@MP               * 'PVMTaskArch'    'where' specifies a type of architecture 
@MP               * 'PVMTaskDebug'   start up processes under debugger
@MP               * 'PVMTaskTrace'   process will generate PVM trace data
@MP               * 'PVMMppFront'    tart process on MPP front-end
@MP               * 'PVMHostCompl'   use complement host set
@MP            where specifies where to start the task
@MP                  'where' depends on the value of 'optionFlag'
@MP                   if 'optionFlag' is 0, then 'where' is ignored and PVM
@MP                   will select the most appropriate host
@MP            contextFlag specifies the context of one child process:
@MP               * 'CPPvmCurrentContext': spawn within the context of the
@MP                 parent process
@MP               * 'CPPvmNewContext':  create a new context and spawn the
@MP                 child within this new context
@MP               * 'CPPvmGivenContext': spawn the children in the context
@MP                 given by 'contextId_'
@MP            contextId_ specifies a context; only used if 'contextFlag'
@MP               is set to 'CPPvmGivenContext'
@MP            multiContextFlag specifies the context of all child process:
@MP               this parameter is only used if 'contextFlag' is set to
@MP               'CPPvmNewContext'
@MP               * 'CPPvmOneNewContext':  create one new context and spawn all
@MP                 children within this new context
@MP               * 'CPPvmAllNewContext':  create a new context for every
@MP                  child and spawn the children within these contexts
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSpawnMultiConnection::cppvmSpawnMultiConnection(char *task,
                                                     unsigned int numTasks,
                                                     char **argv,
                                                     int optionFlag,
                                                     char *where,
                                                     cppvmContextMode
                                                       contextFlag,
                                                     int   contextId_,
                                                     cppvmMultiContextMode
                                                       multiContextFlag)
    : cppvmProcessHandling()
{
  spawnIt(task, numTasks, argv, optionFlag, where, 
          contextFlag, contextId_, multiContextFlag);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSpawnMultiConnection
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmSpawnMultiConnection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSpawnMultiConnection::cppvmSpawnMultiConnection()
    : cppvmProcessHandling()
{
  // init pvm daemon
  initDaemon();

  initTag(0);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmSpawnMultiConnection
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmSpawnMultiConnection
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmSpawnMultiConnection::~cppvmSpawnMultiConnection()
{
  clean();
  delete [] tags;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : spawnIt
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : do the spawning
@MD                  All stderr and stdout of spawned tasks are loged in a
@MD                  file '/tmp/pvml.<uid>'.
@MU  USAGE         : void spawnIt(char *task,
@MU                               unsigned int numTask,
@MU                               char **argv,
@MU                               int optionFlag,
@MU                               char *where,
@MU                               cppvmContextMode contextFlag,
@MU                               int   contextId_,
@MU                               cppvmMultiContextMode contextFlag);
@MU                               
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : task  task name (program) to start
@MP            numTasks number of task to spawn
@MP            argv vector of command line arguments for a spawned task
@MP            optionFlag specifies special PVM options:
@MP               * 'PvmTaskDefault' PVM can choose any machine
@MP               * 'PvmTaskHost'    'where' specifies a particular host
@MP               * 'PVMTaskArch'    'where' specifies a type of architecture 
@MP               * 'PVMTaskDebug'   start up processes under debugger
@MP               * 'PVMTaskTrace'   process will generate PVM trace data
@MP               * 'PVMMppFront'    tart process on MPP front-end
@MP               * 'PVMHostCompl'   use complement host set
@MP            where specifies where to start the task
@MP                  'where' depends on the value of 'optionFlag'
@MP                   if 'optionFlag' is 0, then 'where' is ignored and PVM
@MP                   will select the most appropriate host
@MP            contextFlag specifies the context of one child process:
@MP               * 'CPPvmCurrentContext': spawn within the context of the
@MP                 parent process
@MP               * 'CPPvmNewContext':  create a new context and spawn the
@MP                 child within this new context
@MP               * 'CPPvmGivenContext': spawn the children in the context
@MP                 given by 'contextId_'
@MP            contextId_ specifies a context; only used if 'contextFlag'
@MP               is set to 'CPPvmGivenContext'
@MP            multiContextFlag specifies the context of all child process:
@MP               this parameter is only used if 'contextFlag' is set to
@MP               'CPPvmNewContext'
@MP               * 'CPPvmOneNewContext':  create one new context and spawn all
@MP                 children within this new context
@MP               * 'CPPvmAllNewContext':  create a new context for every
@MP                  child and spawn the children within these contexts
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::spawnIt(char *task,
                                   unsigned int numTasks,
                                   char **argv,
                                   int optionFlag,
                                   char *where,
                                   cppvmContextMode
                                   contextFlag,
                                   int   contextId_,
                                   cppvmMultiContextMode
                                   multiContextFlag)
{
  // init pvm daemon
  initDaemon();

  unsigned int i;
  cppvmSpawnConnection *spawnObjPtr;
  cppvmContextMode newFlag=contextFlag;
  int startIndex = 0;
  int localContext = contextId_;
  
  // well, why calling the constructor with
  // zero spawns, tststs
  if(numTasks==0){
    return;
  }
  
  // init tag array, the array will be enlarged by the following
  // 'add' method
  initTag(0);

  // spawn all children within only ONE context
  if(newFlag==CPPvmNewContext){
    if(multiContextFlag==CPPvmMultiOneNewContext){
      spawnObjPtr = new cppvmSpawnConnection(task, argv, optionFlag, where,
                                             newFlag, localContext);
      add(spawnObjPtr);
      localContext=spawnObjPtr->getContextId();
      newFlag = CPPvmGivenContext;
      startIndex = 1;
    }
  }

  for(i=startIndex;i<numTasks;i++){
    spawnObjPtr = new cppvmSpawnConnection(task, argv, optionFlag, where,
                                           newFlag, localContext);
    add(spawnObjPtr);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmSpawnMultiConnection::operator== (cppvmSpawnMultiConnection& obj)
{
  unsigned int i;
  
  if(cppvmIsSame(obj)){
    return true;
  }
  if(spawnCollection.numElems() != obj.spawnCollection.numElems()){
    return false;
  }
  for (i=0; i<spawnCollection.numElems(); i++){
    if (((cppvmSpawnConnection&)(spawnCollection[i])) !=
        ((cppvmSpawnConnection&)(obj.spawnCollection[i]))){
      return false;
    }
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::cppvmPrintOn(ostream& strm)
{
  unsigned int i;

  strm << "Spawned tasks collection: " << endl;
  for(i=0; i<spawnCollection.numElems(); i++){
    strm << ((cppvmSpawnConnection&)spawnCollection[i]);
    if(tags[i] == (char)0){
      strm << "untagged";
    }
    else{
      strm << "tagged";
    }
    strm << endl;
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmScanFrom
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : scan method for << operator
@MU  USAGE         : cppvmScanFrom()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to scan object from
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::cppvmScanFrom(istream& /*strm*/)
{
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : remove
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove element from collection
@MU  USAGE         : remove(cppvmSpawnConnection& obj)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   cppvmSpawnConnection to remove
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::remove(cppvmSpawnConnection& obj)
{
  unsigned int i;
  int index=-1;

  for(i=0; i<spawnCollection.numElems(); i++){
    if(((cppvmSpawnConnection&)spawnCollection[i]).cppvmIsSame(obj)){
      index = (int)i;
    }
  }
  if(index == -1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSpawnMultiConnection",
                                                 "remove(cppvmSpawnConnection&)",
                                                 "Object not in spawn collection"));
    
  }
  // remove the tag of this object
  shiftTagsLeft((unsigned int) index, (unsigned int) spawnCollection.numElems());
  // resize the tag collection
  resizeTag(spawnCollection.numElems()-1, spawnCollection.numElems());
  // remove the object from the spawnCollection
  spawnCollection.remove(obj);

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : add
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : add pointer to an element to collection
@MU  USAGE         : add(cppvmSpawnConnection *objPtr)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   pointer to cppvmSpawnConnection to add
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::add(cppvmSpawnConnection *objPtr)
{
  if(objPtr == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSpawnMultiConnection",
                                                   "add(cppvmSpawnConnection*)",
                                                   "cannot add NULL pointer"));
    
  }
  resizeTag(spawnCollection.numElems()+1, spawnCollection.numElems());
  spawnCollection.add(objPtr);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : remove
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove element (cppvmSpawnConnection pointer) from collection
@MU  USAGE         : remove(cppvmSpawnConnection *objPtr)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   pointer to cppvmSpawnConnection to remove
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::remove(cppvmSpawnConnection *objPtr)
{
  unsigned int i;
  int index=-1;

  if(objPtr == NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmSpawnMultiConnection",
                                                   "remove(cppvmSpawnConnection *)",
                                                   "cannot remove NULL pointer"));
    
  }

  for(i=0; i<spawnCollection.numElems(); i++){
    if(((cppvmSpawnConnection&)spawnCollection[i]).cppvmIsSame(*objPtr)){
      index = (int)i;
    }
  }
  if(index == -1){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSpawnMultiConnection",
                                                 "remove(cppvmSpawnConnection*)",
                                                 "Object not in spawn collection"));
    
  }
  // remove the tag of this object
  shiftTagsLeft((unsigned int) index, (unsigned int) spawnCollection.numElems());
  // resize the tag collection
  resizeTag(spawnCollection.numElems()-1, spawnCollection.numElems());
  // remove the object from the spawnCollection
  spawnCollection.remove(objPtr);

}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : remove
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove element (cppvmSpawnConnection pointer) from collection
@MU  USAGE         : remove(unsigned int)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : index  index of object to remove
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::remove(unsigned int index)
{
  if(index >=spawnCollection.numElems()){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSpawnMultiConnection",
                                                 "remove(unsigned int)",
                                                 "Bad index"));
  }

  // remove the tag of this object
  shiftTagsLeft((unsigned int) index, (unsigned int) spawnCollection.numElems());
  // resize the tag collection
  resizeTag(spawnCollection.numElems()-1, spawnCollection.numElems());
  // remove the object from the spawnCollection
  spawnCollection.remove(index);
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : clean
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : remove all elements (cppvmSpawnConnection pointer) from collection
@MD                  and remove references
@MU  USAGE         : clean()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::clean()
{
  // resize the tag collection
  resizeTag(0, spawnCollection.numElems());
  // remove the object from the spawnCollection
  spawnCollection.clean();
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmSpawnMultiConnection
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : append another collection to this collection; the appended collection
@MD                  is deleted
@MU  USAGE         : void append(cppvmSpawnMultiConnection *objColl);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to append
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
void
cppvmSpawnMultiConnection::append(cppvmSpawnMultiConnection& obj)
{
  int oldNum = obj.spawnCollection.numElems();
  
  spawnCollection.append(&(obj.spawnCollection));
  // the obj.spawnCollection is now empty, resize its tags
  obj.resizeTag(0, oldNum);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : shiftTagsLeft
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : shift all tags chars to the left to keep
@MD                  the collection ordered
@MU  USAGE         : shiftTagsLeft(unsigned int indexMin,
@MU                                unsigned int indexMax)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : example: collection(a,b,c,d,e)
@MR                  collection.shiftTagsLeft(2,4) -> collection(a,b,d,e,?)
@MR                  
@MR                  
@M#  --------------------------------------------------------------------------
@MP  IN      : indexMin  the collection is shifted left up to this element
@MP            indexMax  all objects up to this index are shifted
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::shiftTagsLeft(unsigned int indexMin,
                             unsigned int indexMax)
{
  if(indexMin > indexMax){
    return;
  }
  
  unsigned int numCopy = indexMax-indexMin;
  if(numCopy>0){
    // copy all object pointer
    memmove((void*) &tags[indexMin], (void*) &tags[indexMin+1], 
           sizeof(char) * numCopy);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : isElemTagged
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : is a specific element tagged?
@MU  USAGE         : bool isElemTagged(unsigned int index)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : index     index of the spawnCollection element to test
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool      indicates whether an element is tagged or not
******************************************************************************/
bool
cppvmSpawnMultiConnection::isElemTagged(unsigned int index)
{
  if(index >=spawnCollection.numElems()){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSpawnMultiConnection",
                                                 "isElemTagged(unsigned int)",
                                                 "Bad index"));
  }
  if(tags[index]==char(1)){
    return true;
  }
  else{
    return false;
  }
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : tagElem
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : tag a spawnCollection element
@MU  USAGE         : tagElem(unsigned int index)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : index     index of the spawnCollection element to tag
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::tagElem(unsigned int index)
{
  if(index >=spawnCollection.numElems()){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSpawnMultiConnection",
                                                 "tagElem(unsigned int)",
                                                 "Bad index"));
  }
  tags[index]=char(1);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : untagElem
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : untag a spawnCollection element
@MU  USAGE         : untagElem(unsigned int index)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : index     index of the spawnCollection element to untag
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::untagElem(unsigned int index)
{
  if(index >=spawnCollection.numElems()){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmSpawnMultiConnection",
                                                 "untagElem(unsigned int)",
                                                 "Bad index"));
  }
  tags[index]=char(0);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : tagAllElems
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : tag all elements of the spawnCollection
@MU  USAGE         : tagAllElems()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : index     index of the spawnCollection element to tag
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::tagAllElems()
{
  unsigned int i;

  for(i=0;i<spawnCollection.numElems();i++){
    tags[i]=char(1);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : untagAllElems
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : untag all elements of the spawnCollection
@MU  USAGE         : untagAllElems()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : index     index of the spawnCollection element to untag
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::untagAllElems()
{
  unsigned int i;

  for(i=0;i<spawnCollection.numElems();i++){
    tags[i]=char(0);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getTaggedChildTasks
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get all child tasks of the elements in the spawnCollection
@MD                  if they are tagged in the tags collection
@MU  USAGE         : cppvmObjectCollection *getTaggedChildTasks()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : cppvmObjectCollection * the collection filled with cppvmTask objects is allocated
@MP                         by this method
******************************************************************************/
cppvmObjectCollection *
cppvmSpawnMultiConnection::getTaggedChildTasks()
{
  unsigned int i;
  cppvmObjectCollection *taskColl;
  cppvmTask *task;

  taskColl = new cppvmObjectCollection();
  for(i=0;i<spawnCollection.numElems();i++){
    if(tags[i]==(char)1){
      task = new cppvmTask;
      (*task) = ((cppvmSpawnConnection&)spawnCollection[i]).getChildTask();
      taskColl->add(task);
    }
  }
  return taskColl;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : getAllChildTasks
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : get all child tasks of the elements in the spawnCollection
@MU  USAGE         : cppvmObjectCollection *getAllChildTasks()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : cppvmObjectCollection * the collection filled with cppvmTask object is allocated
@MP                         by this method
******************************************************************************/
cppvmObjectCollection *
cppvmSpawnMultiConnection::getAllChildTasks()
{
  unsigned int i;
  cppvmObjectCollection *taskColl;
  cppvmTask *task;

  taskColl = new cppvmObjectCollection();
  for(i=0;i<spawnCollection.numElems();i++){
    task = new cppvmTask;
    (*task) = ((cppvmSpawnConnection&)spawnCollection[i]).getChildTask();
    taskColl->add(task);
  }
  return taskColl;
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : halt
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : shut down the entire \pvm{} system
@MU  USAGE         : void halt()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::halt()
{
  int info;
  info = pvm_halt();
  
  if(info < 0){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
                                                "cppvmSpawnMultiConnection",
                                                "halt",
                                                "local pvmd is not responding"));
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : initDaemon
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : start the pvmdaemon (if not allready started)
@MU  USAGE         : void initDaemon()
@MC  CONSTRAINTS   : 
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
void
cppvmSpawnMultiConnection::initDaemon()
{
  char *argsArray[1];
  // dummy arguments
  argsArray[0] = NULL;

  startPvmDaemon(0, argsArray, 1);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : killTaggedChildren
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : kill all tagged children processes
@MU  USAGE         : void killTaggedChildren();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : - 
******************************************************************************/
void
cppvmSpawnMultiConnection::killTaggedChildren()
{
  unsigned int i;

  for(i=0; i<spawnCollection.numElems(); i++){
    if(tags[i] != (char)0){
      ((cppvmSpawnConnection&)spawnCollection[i]).killChild();
    }
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : killAllChildren
@MK  CLASS NAME    : cppvmSpawnMultiConnection
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : kill all children processes
@MU  USAGE         : void killAllChildren();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : -
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : - 
******************************************************************************/
void
cppvmSpawnMultiConnection::killAllChildren()
{
  unsigned int i;

  for(i=0; i<spawnCollection.numElems(); i++){
    ((cppvmSpawnConnection&)spawnCollection[i]).killChild();
  }
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

