/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_stl.cpp 
@IE  CREATED     : 3.2.2000
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : specialized template transfer function for the stl library
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

#ifndef CPPvmDoNotUseTemplates
#ifndef CPPvmDoNotUseStl

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// include classes for collection handling
#include "pvm_str.h"
#include "pvm_arbo.h"
#include "pvm_mem.h"
// include own header
#include "pvm_stl.h"

/*@@@**************************************************************************
**********                                                           **********
*******                           VARIABLES                             *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                            FUNCTIONS                            *******
**********                                                           **********
******************************************************************************/



/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : cppvmTransferStl*
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : specified template transfer functions for stl objects
@FU  USAGE         : 
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : 
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void cppvmTransferStlString (string& data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmString> tmpStr(1);
    tmpStr.data[0] = data.c_str();
    tmpStr.data[0].cppvmTransfer();
  }
  else{
    cppvmString tmpStr;
    tmpStr.cppvmTransfer();
    data.assign(tmpStr.getPtr());
  }
}

void cppvmTransferStlString (string * data, int nitem, int stride ) {
  int i;
  for (i=0; i <nitem ; i++) {
    cppvmTransferStlString( data[i], stride);
  }
}

template <>
void cppvmTransferStlDeque (deque<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

#ifdef CPPvmDoNotUseStdNamespace
  deque<bool>::iterator i;
#else
  std::deque<bool>::iterator i;
#endif
  int j=0;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      for (i=data.begin(); i != data.end() ; i++) {
        (tmpData.data[0])[j] = *i;
        j++;
      }
      tmpData.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      for (i=data.begin(); i != data.end() ; i++) {
        *i = tmpData[j];
        j++;
      }
    }
  }
}

template <>
void cppvmTransferStlList (list<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

#ifdef CPPvmDoNotUseStdNamespace
  list<bool>::iterator i;
#else
  std::list<bool>::iterator i;
#endif
  int j=0;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      for (i=data.begin(); i != data.end() ; i++) {
        (tmpData.data[0])[j] = *i;
        j++;
      }
      tmpData.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      for (i=data.begin(); i != data.end() ; i++) {
        *i = tmpData[j];
        j++;
      }
    }
  }
}

#ifdef CPPvmUseStlSlist
template <>
void cppvmTransferStlSlist (slist<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

#ifdef CPPvmDoNotUseStdNamespace
  slist<bool>::iterator i;
#else

#ifdef __GNUC__
// gcc version < 3.1.0 ?
#if __GNUC__ < 3 || \
    (__GNUC__ == 3 && __GNUC_MINOR__ < 1)
  std::slist<bool>::iterator i;
#else
  __gnu_cxx::slist<bool>::iterator i;
#endif // GCC_VERSION
#else
  std::slist<bool>::iterator i;
#endif // __GNUC__

#endif // CPPvmDoNotUseStdNamespace
  int j=0;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      for (i=data.begin(); i != data.end() ; i++) {
        (tmpData.data[0])[j] = *i;
        j++;
      }
      tmpData.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      for (i=data.begin(); i != data.end() ; i++) {
        *i = tmpData[j];
        j++;
      }
    }
  }
}
#endif // CPPvmUseStlSlist

template <>
void cppvmTransferStlMultiset (multiset<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

#ifdef CPPvmDoNotUseStdNamespace
  multiset<bool>::iterator i;
#else
  std::multiset<bool>::iterator i;
#endif
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if(data.size()>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      int j=0;
      for (i=data.begin(); i != data.end() ; i++) {
        (tmpData.data[0])[j]=(*i);
        j++;
      }
      tmpData.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.clear();
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      int j;
      for (j=0; j<(int)(tmpSize); j++) {
        data.insert(tmpData[j]);
      }
    }
  }
}

template <>
void cppvmTransferStlQueue (queue<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

  int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      i=0;
      while((int)(data.size())>0){
        (tmpData.data[0])[i] = data.front();
        data.pop();
        i++;
      }
      tmpData.data[0].cppvmTransfer();
      // copy data back
      for(i=0; i< (int)(tmpSize.data[0]); i++){
        data.push((tmpData.data[0])[i]);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    // delete all data
    while(data.size()>0){
      data.pop();
    }
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      for (i=0; i<(int)(tmpSize);i++){
        data.push(tmpData[i]);
      }
    }
  }
}

#ifdef CPPVM_NODEADCODE
// The next specialized template function causes an internal error in Visual C++
template <>
void cppvmTransferStlPriorityqueue (priority_queue<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

  int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      i=0;
      while(data.size()>0){
        (tmpData.data[0])[i] = data.top();
        data.pop();
        i++;
      }
      tmpData.data[0].cppvmTransfer();
      // copy data back
      for(i=0; i< (int)(tmpSize.data[0]); i++){
        data.push((tmpData.data[0])[i]);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    // delete all data
    while(data.size()>0){
      data.pop();
    }
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      for (i=0; i<(int)(tmpSize);i++){
        data.push(tmpData[i]);
      }
    }
  }
}
#endif // CPPVM_NODEADCODE

template <>
void cppvmTransferStlStack (stack<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

  int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      i=0;
      while(data.size()>0){
        (tmpData.data[0])[i] = data.top();
        data.pop();
        i++;
      }
      // send data in correct order
      tmpData.data[0].cppvmTransfer();
      // copy data back
      for(i=0; i< (int)(tmpSize.data[0]); i++){
        data.push((tmpData.data[0])[i]);
      }
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    // delete all data
    while(data.size()>0){
      data.pop();
    }
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      // change order
      for (i=((int)(tmpSize))-1; i>=0; i--){
        data.push(tmpData[i]);
      }
    }
  }
}

#ifdef CPPvmUseStlValarray
template <>
void cppvmTransferStlValarray (valarray<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

  unsigned int i;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      for (i=0; i<data.size(); i++){
        (tmpData.data[0])[i] = data[i];
      }
      tmpData.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      for (i=0; i<data.size();i++){
        data[i]=tmpData[i];
      }
    }
  }
}
#endif // CPPvmUseStlValarray


template <>
void cppvmTransferStlVector (vector<bool> & data, int stride ) {
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;

#ifdef CPPvmDoNotUseStdNamespace
  vector<bool>::iterator i;
#else
  std::vector<bool>::iterator i;
#endif
  int j=0;
  if(cppvmIsTransferSend()==true){
    cppvmNewSendStackData<cppvmInt> tmpSize(1);
    tmpSize.data[0] = (int)(data.size());
    tmpSize.data[0].cppvmTransfer();
    if((int)(data.size())>0){
      cppvmNewSendStackData<cppvmArrayBool> tmpData(1);
      tmpData.data[0].resize((int)(data.size()));
      for (i=data.begin(); i != data.end() ; i++) {
        (tmpData.data[0])[j] = *i;
        j++;
      }
      tmpData.data[0].cppvmTransfer();
    }
  }
  else{
    cppvmInt tmpSize;
    tmpSize.cppvmTransfer();
    data.resize((int)(tmpSize));
    if((int)(tmpSize)>0){
      cppvmArrayBool tmpData;
      tmpData.cppvmTransfer();
      for (i=data.begin(); i != data.end() ; i++) {
        *i = tmpData[j];
        j++;
      }
    }
  }
}



#endif // CPPvmDoNotUseStl
#endif // CPPvmDoNotUseTemplates

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/



