/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_str.cpp 
@IE  CREATED     : 5.12.1997
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : pvm string class
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
#include "pvm_str.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

/* On Windows systems 'pvm3.h' includes 'pvmwin.h'. 
   In 'pvmwin.h' the standard C defines 'va_start' 
   and 'va_end' are undefined! This is why these 
   defines must be redefined again according 
   to the Windows compiler version.
*/

#ifdef VISUALCPP
#ifndef va_start
#define va_start(ap,v)  ( ap = (va_list)&v + _INTSIZEOF(v) )
#endif // va_start
#ifndef va_end
#define va_end(ap)      ( ap = (va_list)0 )
#endif // va_end
#endif // VISUALCPP

#ifdef WINBORLAND
#ifndef va_start
#define va_start(ap, parmN) ((void)((ap) = (va_list)((char _FAR *)(&parmN)+__size(parmN))))
#endif // va_start
#ifndef va_end
#define va_end(ap)          ((void)0)
#endif // va_end
#endif // WINBORLAND

#define THIS_MODULE "pvm_str.cpp"

#define CPPvmMaxStringInputLength 1024

/*@@@**************************************************************************
**********                                                           **********
*******                           VARIABLES                             *******
**********                                                           **********
******************************************************************************/
CPPvmBuildInMsgTagMethodsDefinition(cppvmString,
                                    CPPvmMsgTagClassString)


/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmString
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : default constructor
@MU  USAGE         : cppvmString
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmString::cppvmString()
    : cppvmSingleData(), 
      cstring(NULL),
      cppvmTransferMode(CPPvmTransferModeClean),
      fitSizeToLength(false), 
      initEmptyString(false)
{
  len = 0;            // char number
  size = len + 1;             // old size plus '\0'
  resizeClean(len);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmString
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmString
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : len_        string length
@MP            cleanString clean string after allocation or not
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmString::cppvmString(unsigned int len_, bool cleanString)
    : cppvmSingleData(), 
      cstring(NULL),
      cppvmTransferMode(CPPvmTransferModeClean),
      fitSizeToLength(false),
      initEmptyString(false)
{
  if(cleanString==true){
    resizeClean(len_, false);
  }
  else{
    resize(len_);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmString
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmString
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : s  string to copy in object
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmString::cppvmString(char* s)
    : cppvmSingleData(), 
      cstring(NULL),
      cppvmTransferMode(CPPvmTransferModeClean),
      fitSizeToLength(false),
      initEmptyString(false)
{
  if(s == (char*)NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmString",
                                                   "constructor",
                                                   "string is NULL"));
  }

  len = strlen(s);            // char number
  size = len + 1;             // old size plus '\0'
  resizeWithoutFillAndCopy(len);
  strcpy(cstring, s);         // copy sting
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmString
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : constructor
@MU  USAGE         : cppvmString
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : s  string to copy in object
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmString::cppvmString(const char* s)
    : cppvmSingleData(), 
      cstring(NULL),
      cppvmTransferMode(CPPvmTransferModeClean),
      fitSizeToLength(false),
      initEmptyString(false)
{
  if(s == (char*)NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmString",
                                                   "constructor",
                                                   "string is NULL"));
  }

  len = strlen(s);            // char number
  size = len + 1;             // old size plus '\0'
  resizeWithoutFillAndCopy(len);
  strcpy(cstring, s);         // copy sting
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmString
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : copy constructor
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj   object to copy
@MP  INOUT   : - 
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmString::cppvmString (const cppvmString& obj)
    : cppvmSingleData(obj), 
      cstring(NULL),
      cppvmTransferMode(CPPvmTransferModeClean),
      fitSizeToLength(false),
      initEmptyString(false)
{
  // get char number
  len = obj.len;
  size = len+1;
  resizeWithoutFillAndCopy(len);
  // copy string
  strcpy(cstring, obj.cstring); // copy sting
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : ~cppvmString
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : destructor
@MU  USAGE         : ~cppvmString
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
cppvmString::~cppvmString()
{
  delete [] cstring;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator=
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : assignment
@MU  USAGE         : =
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to assign
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
cppvmString&
cppvmString::operator= (const cppvmString& obj)
{
  if(cppvmIsSame(obj)){
    return *this;
  }
  cppvmSingleData::operator=(obj);
  
  resizeWithoutFillAndCopy(obj.len);

  // copy string
  strcpy(cstring, obj.cstring);
    
  return *this;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmString::operator== (const cppvmString& obj)
{
  return (strcmp(cstring, obj.cstring) == 0);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : operator==
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : equality
@MU  USAGE         : ==
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : obj  object to test
@MP  INOUT   : -
@MP  OUT     : -
@MP  RETURNS : -
******************************************************************************/
bool
cppvmString::operator== (const cppvmString& obj) const
{
  return (strcmp(cstring, obj.cstring) == 0);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmTransfer
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : transfer object from/to buffer
@MU  USAGE         : cppvmTransfer()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::cppvmTransfer()
{
  // transfer base class variables
  cppvmSingleData::cppvmTransfer();

  if(cppvmIsTransferReceive()==true){
    // receive array
    // remind old length
    int oldLen=len;
    // transfer array length
    cppvmTransferInt(&strLength);

    if(cppvmTransferMode==CPPvmTransferModeAppend){
      // received array length plus old length
      resize((unsigned int)(oldLen+strLength));
      // add the incomming array
      cppvmTransferStr(&(cstring[oldLen]));
    }
    else{
      resizeWithoutFillAndCopy((unsigned int)strLength);
      cppvmTransferStr(cstring);
    }
  }
  else{
    // send array
    // transfer array length
    strLength = (int) len;
    cppvmTransferInt(&strLength);
    cppvmTransferStr(cstring);
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmPrintOn
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : print method for << operator
@MU  USAGE         : cppvmPrintOn()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to print object on
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::cppvmPrintOn(ostream& strm)
{
  strm << cstring;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cppvmScanFrom
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : scan method for << operator
@MU  USAGE         : cppvmScanFrom()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : strm  stream to scan object from
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::cppvmScanFrom(istream& strm)
{
  char tmp[CPPvmMaxStringInputLength];
  
  // get chars until line feet
  strm.get(tmp, CPPvmMaxStringInputLength-1, '\n');

  char c;
  // the next char must be '\n', otherwise there was
  // a read error
  if(strm.get(c) && c!='\n'){
    CPPvmInternalThrow( cppvmExceptionBadResult(THIS_MODULE,
			"cppvmString",
			"cppvmScanFrom",
			"string read error (string to long or not terminated by '\n')"));
  }
  
  // read successful
  resizeWithoutFillAndCopy((unsigned int)strlen(tmp));
  strcpy(cstring, tmp);       // copy string
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : append
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : append a char to the string
@MU  USAGE         : void append(char c)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::append(char c)
{
  // remind old length
  int oldLen=len;

  // resize string
  resize(len+1);

  // copy char
  cstring[oldLen] = c;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : append
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : append another string to the string
@MU  USAGE         : void append(char *str)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::append(char *str)
{
  // remind old length
  int oldLen=len;
  
  if(str == (char*)NULL){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmString",
                                                   "append(char *str)",
                                                   "str is NULL"));
  }

  // resize string
  resize((unsigned int)(len+strlen(str)));
  // set old string end once again
  cstring[oldLen] = '\0';
  // append new string
  strcat(cstring, str);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : append
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : append another cppvmString object to the string
@MU  USAGE         : void append(cppvmString& obj)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::append(cppvmString& obj)
{
  append(obj.cstring);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : clean
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : clean string (do not free memory!) 
@MU  USAGE         : void clean()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::clean()
{
  cstring[0] = '\0' ;
  len = 0;                   // no length at all
}

  
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : countWords
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : count the number of words
@MD                  words are:
@MD                  every sequence of chars separated by whitespaces
@MU  USAGE         : int countWords()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
int
cppvmString::countWords()
{
  int i = 0;

  while(setOnWord(i)>=0){
    i++;
  }
  return i;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : copyWord
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : append the word with 'index' of this string
@MD                  to the 'target' string
@MD                  Example: Src:    " Hi out there!" 
@MD                           Target: "Let me "
@MD                           src.copyWord(1, target);
@MD                           Target: "Let me out"
@MD                  words are:
@MD                  every sequence of chars separated by whitespaces
@MU  USAGE         : bool copyWord(unsigned int index, cppvmString& target);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : bool  copy successful or not (wrong index?)
******************************************************************************/
bool
cppvmString::copyWord(unsigned int index, cppvmString& target)
{
  int i;

  i = setOnWord(index);
  if(i<0){
    // wrong index, do nothing
    return false;
  }
  while(!isspace(cstring[i]) && cstring[i] != '\0'){
    target.append(cstring[i]);
    i++;
  }
  return true;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cycleShiftLeft
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : cycle shift left
@MD                  "hi out there" -> "i out thereh"
@MU  USAGE         : void cycleShiftLeft();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::cycleShiftLeft()
{
  char tmp;

  if(len <= 1){
    return;
  }
  
  tmp = cstring[0];
  memmove((void*) &cstring[0], (void*) &cstring[1], sizeof(char) * (len-1));
  cstring[len-1]=tmp;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cycleShiftRight
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : cycle shift right
@MD                  "hi out there" -> "ehi out ther"
@MU  USAGE         : void cycleShiftRight();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::cycleShiftRight()
{
  char tmp;

  if(len <= 1){
    return;
  }
  
  tmp = cstring[len-1];
  memmove((void*) &cstring[1], (void*) &cstring[0], sizeof(char) * (len-1));
  cstring[0] = tmp;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setOnWord
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : give the index of the first char of the i'st
@MD                  word
@MD                  words are:
@MD                  every sequence of chars separated by whitespaces
@MU  USAGE         : int setOnWord(unsigned int i)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : i    number of the desired word in the string
@MP                 the first word is i = 0
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  >= 0: index of the first char of word number i
@MP                 <  0: there is no word number i
@MP                       (empty string or less number of words in string)
@MP                        
@MP                        
@MP                        
******************************************************************************/
int
cppvmString::setOnWord(unsigned int i)
{
  int index = 0;
  unsigned int wordNumber = 0;

  if(isspace(cstring[(unsigned int)index])){
    // first char is a whitespace, overread it
    index = setOnNextWordBegin((unsigned int)index);    
    if(index < 0){
      // there was no next word 
      return index;
    }
  }
  while(!(wordNumber==i)){
    // get index of next word
    index = setOnNextWordBegin((unsigned int)index);    
    if(index < 0){
      // there was no next word 
      return index;
    }
    wordNumber++;
  }
  // word number i was found
  return index;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setOnNextWordBegin
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : give the index of the first char of the next word
@MD                  Example: String: "first second   third"
@MD                           Index:     ^
@MD                           Result:        ^
@MD                           Index:                ^
@MD                           Result:                 ^
@MD                  words are:
@MD                  every sequence of chars separated by whitespaces
@MU  USAGE         : int setOnWord(unsigned int index)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : index  index (begin the search with this index)
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : int  >= 0: index of the first char of next word 
@MP                 <  0: there is no word number i
@MP                       (wrong index)
******************************************************************************/
int
cppvmString::setOnNextWordBegin(unsigned int index)
{
  unsigned int newIndex = (int)index;
  
  if( newIndex>=len || cstring[newIndex]=='\0'){
    return -1;
  }
  
  // overread all whitespaces (if there)
  if(isspace(cstring[newIndex])){
    while(isspace(cstring[newIndex])){
      newIndex++;
      if( newIndex>=len || cstring[newIndex]=='\0'){
        return -1;
      }
    }
  }
  else{
    // no whitespaces, overread word and following whitespaces
    while(!(isspace(cstring[newIndex]))){
      newIndex++;
      if( newIndex>=len || cstring[newIndex]=='\0'){
        return -1;
      }
    }
    while(isspace(cstring[newIndex])){
      newIndex++;
      if( newIndex>=len || cstring[newIndex]=='\0'){
        return -1;
      }
    }
  }
  return newIndex;
}
 
/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : printFormat
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : clean string, transform other data types into strings
@MD                  example:
@MD                  int i = 5;
@MD                  float f = 8.8;
@MD                  string.printFormat("%d %f",i,f)
@MU  USAGE         : void printFormat(const char *format, ...);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::printFormat(const char *format, ...)
{
  char erg[CPPvmMaxStringInputLength];

  clean();

  va_list argList;
  va_start(argList, format);
  vsprintf(erg, format, argList);
  va_end(argList);

  append(erg);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : scanFormat
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : transform the string into other data types 
@MD                  example:
@MD                  int i;
@MD                  float f;
@MD                  cppvmString string("1 3.3");
@MD                  string.scanFormat("%d %f",&i,&f);
@MU  USAGE         : void scanFormat(const char *format, ...);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
// NOTE:
// the function 'vsscanf' is not available on every operating system
// so I decided to remove this method for compatibiliy reason
//void
//cppvmString::scanFormat(const char *format, ...)
//{
//  va_list argList;
//  va_start(argList, format);
//  vsscanf(cstring, format, argList);
//  va_end(argList);
//}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : read
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : return current string
@MU  USAGE         : char *read();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
char *
cppvmString::read()
{
  // get object from the mailbox (if the mailbox was initialized)
  cppvmReadMboxInfo();

  return cstring;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : write
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set current string
@MU  USAGE         : void write(char *value_)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::write(char *value_)
{
  resizeWithoutFillAndCopy((unsigned int)(strlen(value_)));
  strcpy(cstring, value_);    // copy string

  // send it to the mailbox (if the mailbox was initialized)
  cppvmWriteMboxInfo();
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : readValDouble
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : return the current string as double
@MU  USAGE         : double readValDouble();
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
double
cppvmString::readValDouble()
{
  int result;
  float floatVal;

  // sometimes double values cannot be converted correctly
  // from strings so:
  // read as float value, convert to double
  result = sscanf(cstring, "%f", &floatVal);
  if(result<=0 || result==EOF){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmString",
                                                   "readValDouble",
                                                   "cannot convert string into double"));
  }

  return (double)floatVal;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : writeValDouble
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : clean string; write a double value into the string
@MU  USAGE         : void writeValDouble(double value_);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::writeValDouble(double value_)
{
  int result;
  char valStr[CPPvmMaxStringInputLength];

  clean();
  result = sprintf(valStr, "%f", value_);
  if(result < 0){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmString",
                                                   "writeValDouble",
                                                   "cannot convert double into string"));
  }
  append(valStr);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setCPPvmTransferMode
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : set the pvm transfer mode of this object
@MU  USAGE         : setCPPvmTransferMode()
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         :
@M#  --------------------------------------------------------------------------
@MP  IN      : 
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::setCPPvmTransferMode(int mode)
{
  if(mode!=CPPvmTransferModeClean &&
     mode!=CPPvmTransferModeAppend){
    CPPvmInternalThrow( cppvmExceptionBadParameter(THIS_MODULE,
                                                   "cppvmString",
                                                   "setCPPvmTransferMode",
                                                   "unknown transfer mode"));
  }
  cppvmTransferMode = mode;
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resize
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : resize the array to a new length
@MD                  copy old values if possible
@MD                  'fitSizeToLength' == true: allocate new array if
@MD                                             if newLen!=size+1
@MD                  'fitSizeToLength' == false: allocate new array if
@MD                                              newLen+1>size
@@MU  USAGE         : void resize(unsigned int newLen)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the class variables 'len' and 'size' must have
@MR                  correct values, 'array' can be NULL
@M#  --------------------------------------------------------------------------
@MP  IN      : newLen  new length of the array
@MP  INOUT   : 
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::resize(unsigned int newLen)
{
  resizeWithoutFill(newLen);
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resizeClean
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : resize string size; clean string
@MD                  'fitSizeToLength' == true: allocate new array if
@MD                                             if newLen!=size+1
@MD                  'fitSizeToLength' == false: allocate new array if
@MD                                              newLen+1>size
@MU  USAGE         : resizeClean(unsigned int newLen,
@MU                              bool cleanOnlyNewElements)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the class variables 'len' and 'size' must have
@MR                  correct values, 'cstring' can be NULL
@M#  --------------------------------------------------------------------------
@MP  IN      : newLen              new array length
@MP            cleanOnlyNewElements do not clean the complete array
@MP                                 but only new elements if the array
@MP                                 gets larger
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::resizeClean(unsigned int newLen,
                         bool cleanOnlyNewElements)
{
  unsigned int oldLen = (unsigned int)len;

  if(cleanOnlyNewElements){
    resizeWithoutFill(newLen);
    if(oldLen<newLen){
      // clean only new elements
      cleanRange(oldLen, newLen-1);
      cstring[newLen] = '\0';
    }
  }
  else{
    resizeWithoutFillAndCopy(newLen);
    // 'clean' string
    if(newLen>0){
      cleanRange(0, newLen-1);
    }
    cstring[newLen] = '\0';
  }
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resizeWithoutFill
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : resize array length without filling
@MD                  new allocated longs with ' '
@MD                  copy old values if possible
@MD                  'fitSizeToLength' == true: allocate new array if
@MD                                             if newLen!=size+1
@MD                  'fitSizeToLength' == false: allocate new array if
@MD                                              newLen+1>size
@MU  USAGE         : void resizeWithoutFill(unsignd int newLen)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the class variables 'len' and 'size' must have
@MR                  correct values, 'array' can be NULL
@M#  --------------------------------------------------------------------------
@MP  IN      : newLen  new array length
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::resizeWithoutFill(unsigned int newLen)
{
  if(arrayIsNULL()==true){
    len = newLen;
    size = len+1;
    // allocate memory if there is no string
    newArray((unsigned int)size);
    cstring[len] = '\0';
    return;
  }

  if(newLen==len && (!fitSizeToLength)){
    // the array has the correct length
    // and must not be fitted to the
    // current size
    // do nothing
    return;
  }
  if(newLen==len && len==size-1){
    // the array has the correct length
    // and the length is fitted to the
    // current size
    // do nothing
    return;
  }

  if((!fitSizeToLength) && newLen < size){
    // do not fit the size of the array
    // and the old size is big enough
    // just cut the string
    len = newLen;
    cstring[len] = '\0';
    return;
  }
  
  // new array is needed
  int maxLen;
    
  if(len>=newLen){
    maxLen=newLen;
  }
  else{
    maxLen=len;
  }
  
  len = newLen;
  size = len+1;
  newArrayCopy((unsigned int)size, maxLen);
  // empty string
  cstring[len] = '\0';
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : resizeWithoutFillAndCopy
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : resize string size without filling
@MD                  new allocated chars with ' ' and without
@MD                  copying old values
@MD                  'fitSizeToLength' == true: allocate new array if
@MD                                             if newLen!=size+1
@MD                  'fitSizeToLength' == false: allocate new array if
@MD                                              newLen+1>size
@MU  USAGE         : void resizeWithoutFillAndCopy(unsignd int newLen)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the class variables 'len' and 'size' must have
@MR                  correct values, 'cstring' can be NULL
@M#  --------------------------------------------------------------------------
@MP  IN      : newLen  new string size
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::resizeWithoutFillAndCopy(unsigned int newLen)
{
  if(arrayIsNULL()==true){
    len = newLen;
    size = len+1;
    // allocate memory if there is no string
    newArray((unsigned int)size);
    return;
  }

  if(newLen==len && (!fitSizeToLength)){
    // the array has the correct length
    // and must not be fitted to the
    // current size
    // do nothing
    return;
  }
  if(newLen==len && len==size-1){
    // the array has the correct length
    // and the length is fitted to the
    // current size
    // do nothing
    return;
  }

  if((!fitSizeToLength) && newLen < size){
    // do not fit the size of the array
    // and the old size is big enough
    // just cut the string
    len = newLen;
    cstring[len] = '\0';
    return;
  }
  
  // the former array is not needed any more
  len = newLen;
  size = len+1;
  // allocate memory
  newArray((unsigned int)size);
  // empty string
  cstring[len] = '\0';
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : newArray
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : allocate new array
@MU  USAGE         : void newArray(unsignd int newSize)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : newSize  new array size
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::newArray(unsigned int newSize)
{
  if(cstring!=(char*)NULL){
    // the former array is not needed any more
    delete [] cstring;
  }
  // allocate memory
  cstring = new char [newSize];  // allocate memory
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : newArrayCopy
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : allocate new array and copy elements from old array
@MU  USAGE         : void newArrayCopy(unsigned int newSize, 
@MU                                    unsigned int numCopyElements)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : newSize         new array size
@MP            numCopyElements number of element to copy from the old into the
@MP                            new array
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::newArrayCopy(unsigned int newSize, unsigned int numCopyElements)
{
  if(cstring==(char*)NULL){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmString",
                                                 "newArrayCopy",
                                                 "string to copy from is NULL!"));
  }
  
  // new array is needed
  char *tmpString = cstring;

  // allocate memory  
  cstring = new char [newSize];  

  if(numCopyElements>0){
    // copy memory
    memcpy((void*)cstring, (void*)tmpString, sizeof(char) * numCopyElements);
  }

  // the former array is not needed any more
  delete [] tmpString;
}


/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : setFitSizeToLength
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : fit the array size to the current length
@MU  USAGE         : void setFitSizeToLength(bol value);
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : 
@M#  --------------------------------------------------------------------------
@MP  IN      : value  true: whenever a new or delete of a the string
@MP                         is needed, the string size is fitted to the
@MP                         string length
@MP                   false: the size can only get bigger, not smaller     
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::setFitSizeToLength(bool value)
{
  fitSizeToLength = value;
  if(fitSizeToLength){
    // fit the size to the length
    resize(len);
  }
  
}

/*@@*****************************   METHOD   **********************************
***
@MN  METHOD NAME   : cleanRange
@MK  CLASS NAME    : cppvmString
@MA  AUTHOR        : goerzig
@M#  
@MD  DESCRIPTION   : clean a range in the array
@MU  USAGE         : cleanRange(unsigned int fromIndex, unsigned int toIndex)
@MC  CONSTRAINTS   :
@M#  
@MR  NOTES         : the class variables 'len' and 'size' must have
@MR                  correct values, 'array' must not be NULL
@M#  --------------------------------------------------------------------------
@MP  IN      : fromIndex    index of first element to clean
@MP            toIndex      index of last element to clean
@MP  INOUT   :
@MP  OUT     :
@MP  RETURNS : 
******************************************************************************/
void
cppvmString::cleanRange(unsigned int fromIndex,
                        unsigned int toIndex)
{

  if(fromIndex>toIndex){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmString",
                                                 "cleanRange",
                                                 "'fromIndex' is greater than 'toIndex'"));
  }
  if(fromIndex>=(unsigned int)len){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmString",
                                                 "cleanRange",
                                                 "'fromIndex' is greater than array length"));
  }
  if(toIndex>=(unsigned int)len){
    CPPvmInternalThrow( cppvmExceptionOutOfRange(THIS_MODULE,
                                                 "cppvmString",
                                                 "cleanRange",
                                                 "'to' is greater than array length"));
  }

  unsigned int i;
  unsigned int numMemCopy;
  unsigned int restSingleCopyIndex;
  
  // clean memory to copy
  if(initEmptyString==false){
    for(i=0; i< CPPvmStringCleanFieldSize; i++){
      emptyString[i]=(char)0;
    }
    initEmptyString=true;
  }
  
  numMemCopy = ((toIndex-fromIndex)+1) / CPPvmStringCleanFieldSize;
  restSingleCopyIndex = fromIndex + (numMemCopy * CPPvmStringCleanFieldSize);
  
  unsigned int index = fromIndex;
  // use memcpy
  for(i=0; i< numMemCopy; i++){
    memcpy((void*)(&cstring[index]), (void*)(&emptyString[0]), 
           sizeof(char) *  CPPvmStringCleanFieldSize);
    index = index +  CPPvmStringCleanFieldSize;
  }
  
  // copy the rest as single values
  for(i=restSingleCopyIndex; i<=toIndex; i++){
    cstring[i]=(char)0;
  }
}

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

