/*@@@**************************************************************************
**********                                                           **********
*******                      IMPLEMENTATION FILE                        *******
**********                                                           **********
*******************************************************************************
@IF  FILE        : pvm_tran.cpp 
@IE  CREATED     : 10.12.1999
@IA  AUTHOR      : goerzig
@IJ  PROJECT     : cppvm
@I#  
@ID  DESCRIPTION : CPPvm transfer functions
@IU  USAGE       : 
@IC  CONSTRAINTS : 
@I#  
@IR  NOTES       : 
@I#  --------------------------------------------------------------------------
@I#  Technical Information:
@I#  
@II  LANGUAGE        : C++
@II  HARDWARE        : PC 386
@II  COMPILER        : gcc version 2.7.2.3
@II  OPERATINGSYSTEM : Linux
@I#  --------------------------------------------------------------------------
@I#  Version Manager Informations:
@I#  
@IV  FILENAME    : $Source: $
@IV  AUTHOR      : $Author: $
@IV  REVISION    : $Revision: $
@IV  DATE        : $Date: $
@IV  STATE       : $State: $
@IV  LOCKER      : $Locker: $
@I#  
@IV  VERSION HISTORY : $Log: $
*******************************************************************************
*****       (C) COPYRIGHT Steffen Goerzig - All Rights Reserved           *****
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                           INCLUDES                              *******
**********                                                           **********
******************************************************************************/
// include own header
#include "pvm_tran.h"
// cppvm includes
#include "pvm_cincl.h"
#include "pvm_arbo.h"
#include "pvm_nssdata.h"

/*@@@**************************************************************************
**********                                                           **********
*******                            CONSTANTS                            *******
**********                                                           **********
******************************************************************************/

/* On Windows systems 'pvm3.h' includes 'pvmwin.h'. 
   In 'pvmwin.h' the standard C defines 'va_start' 
   and 'va_end' are undefined! This is why these 
   defines must be redefined again according 
   to the Windows compiler version.
*/

#ifdef VISUALCPP
#ifndef va_start
#define va_start(ap,v)  ( ap = (va_list)&v + _INTSIZEOF(v) )
#endif // va_start
#ifndef va_end
#define va_end(ap)      ( ap = (va_list)0 )
#endif // va_end
#endif // VISUALCPP

#ifdef WINBORLAND
#ifndef va_start
#define va_start(ap, parmN) ((void)((ap) = (va_list)((char _FAR *)(&parmN)+__size(parmN))))
#endif // va_start
#ifndef va_end
#define va_end(ap)          ((void)0)
#endif // va_end
#endif // WINBORLAND


/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmThrowIfPackError
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : generate an exception when PVM data packing 
@MD                 returns an error
@MU  USAGE        : CPPvmThrowIfPackError(packResult)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : packResult  result (return value) of a PVM packing function
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmThrowIfPackError(packResult)                                \
  if((packResult) < 0) {                                                 \
    CPPvmInternalThrow( cppvmExceptionBadResult(PVM_OBJ_HEADER,          \
                                                "cppvmTransfer*",        \
                                                "CPPvmThrowIfPackError", \
                                                "cannot pack data"));    \
  }                                                                      \

/*@@*****************************   MACRO   ***********************************
***
@MN  MACRO NAME   : CPPvmThrowIfUnpackError
@MA  AUTHOR       : goerzig
@M#  
@MD  DESCRIPTION  : generate an exception when PVM data unpacking 
@MD                 returns an error
@MU  USAGE        : CPPvmThrowIfUnpackError(unpackResult)
@MC  CONSTRAINTS  : 
@MS  SIDE EFFECTS : 
@M#  
@MR  NOTES        : 
@M#  --------------------------------------------------------------------------
@M#  Macro Parameter:
@M#  
@MP  IN      : unpackResult  result (return value) of a PVM unpacking function
@MP  INOUT   : 
@MP  OUT     : 
@MP  RETURNS : 
******************************************************************************/
#define CPPvmThrowIfUnpackError(unpackResult)                              \
  if((unpackResult) < 0) {                                                 \
    CPPvmInternalThrow( cppvmExceptionBadResult(PVM_OBJ_HEADER,            \
                                                "cppvmTransfer*",          \
                                                "CPPvmThrowIfUnpackError", \
                                                "cannot unpack data"));    \
  }                                                                        \

/*@@@**************************************************************************
**********                                                           **********
*******                           VARIABLES                             *******
**********                                                           **********
******************************************************************************/

/*@@@**************************************************************************
**********                                                           **********
*******                             METHODS                             *******
**********                                                           **********
******************************************************************************/


/*@@@**************************************************************************
**********                                                           **********
*******                            FUNCTIONS                            *******
**********                                                           **********
******************************************************************************/


/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : cppvmTransferSend/Receive
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : send or receive transfer
@FU  USAGE         : 
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : 
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : 
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
bool
cppvmIsTransferSend()
{
  return cppvmObject::cppvmClassIsTransferSend();
}
bool
cppvmIsTransferReceive()
{
  return cppvmObject::cppvmClassIsTransferReceive();
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : cppvmTransfer*
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : transfer data types to/from send/receive buffer
@FU  USAGE         : 
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : a cppvmExceptionBadResult exception is thrown if
@FR                  transfer fails
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : 
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
void
cppvmTransferBool(bool *ptr, int nitem, int stride)
{
  // dummy use of 'stride' to avoid compiler warnings
  stride=0;
  
  int i;
  if(cppvmIsTransferSend()==true){
    cppvmArrayBool *tmpArrayPtr = new cppvmArrayBool(nitem, false);
    cppvmAddToSendStack(tmpArrayPtr);
    for(i=0; i<nitem; i++){
      (*tmpArrayPtr)[i] = ptr[i];
    }
    tmpArrayPtr->cppvmTransfer();
  }
  else{
    cppvmArrayBool tmpArray(nitem);
    tmpArray.cppvmTransfer();
    for(i=0; i<nitem; i++){
      ptr[i] = tmpArray[i];
    }
  }
}

void
cppvmTransferFSend(char *ptr, ...)
{
  va_list ap;
  va_start(ap, ptr);
  CPPvmThrowIfPackError(pvm_vpackf(ptr, ap));
  va_end(ap);
}

void
cppvmTransferFReceive(char *ptr, ...)
{
  va_list ap;
  va_start(ap, ptr);
  CPPvmThrowIfUnpackError(pvm_vunpackf(ptr, ap));
  va_end(ap);
}

void
cppvmTransferByte(char *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkbyte(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkbyte(ptr, nitem, stride));
  }
}

void
cppvmTransferCplx(float *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkcplx(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkcplx(ptr, nitem, stride));
  }
}

void
cppvmTransferDcplx(double *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkdcplx(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkdcplx(ptr, nitem, stride));
  }
}

void
cppvmTransferDouble(double *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkdouble(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkdouble(ptr, nitem, stride));
  }
}
void
cppvmTransferFloat(float *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkfloat(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkfloat(ptr, nitem, stride));
  }
}

void
cppvmTransferInt(int *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkint(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkint(ptr, nitem, stride));
  }
}

void
cppvmTransferUint(unsigned int *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkuint(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkuint(ptr, nitem, stride));
  }
}

void
cppvmTransferUshort(unsigned short *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkushort(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkushort(ptr, nitem, stride));
  }
}

void
cppvmTransferUlong(unsigned long *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkulong(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkulong(ptr, nitem, stride));
  }
}

void
cppvmTransferLong(long *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pklong(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upklong(ptr, nitem, stride));
  }
}

void
cppvmTransferShort(short *ptr, int nitem, int stride)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkshort(ptr, nitem, stride));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkshort(ptr, nitem, stride));
  }
}

void
cppvmTransferStr(char *ptr)
{
  if(cppvmIsTransferSend()==true){
    CPPvmThrowIfPackError(pvm_pkstr(ptr));
  }
  else{
    CPPvmThrowIfUnpackError(pvm_upkstr(ptr));
  }
}

/*@@****************************   FUNCTION   *********************************
***
@FN  FUNCTION NAME : cppvmTransferTemplates
@FA  AUTHOR        : goerzig
@F#  
@FD  DESCRIPTION   : transfer template data types to/from send/receive buffer
@FU  USAGE         : 
@FC  CONSTRAINTS   : 
@FS  SIDE EFFECTS  : 
@F#  
@FR  NOTES         : a cppvmExceptionBadResult exception is thrown if
@FR                  transfer fails
@F#  --------------------------------------------------------------------------
@F#  Function Parameter:
@F#  
@FP  IN      : 
@FP  INOUT   : 
@FP  OUT     : 
@FP  RETURNS : 
******************************************************************************/
#ifndef CPPvmDoNotUseTemplates
// C++ type transfer
template<>
void
cppvmTransferTemplate(bool* data, int nitem, int stride)
{
  cppvmTransferBool(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(bool& data, int stride)
{
  cppvmTransferBool(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(char* data, int nitem, int stride)
{
  cppvmTransferByte(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(char& data, int stride)
{
  cppvmTransferByte(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(int* data, int nitem, int stride)
{
  cppvmTransferInt(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(int& data, int stride)
{
  cppvmTransferInt(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(short* data, int nitem, int stride)
{
  cppvmTransferShort(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(short& data, int stride)
{
  cppvmTransferShort(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(long* data, int nitem, int stride)
{
  cppvmTransferLong(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(long& data, int stride)
{
  cppvmTransferLong(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(unsigned int* data, int nitem, int stride)
{
  cppvmTransferUint(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(unsigned int& data, int stride)
{
  cppvmTransferUint(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(unsigned short* data, int nitem, int stride)
{
  cppvmTransferUshort(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(unsigned short& data, int stride)
{
  cppvmTransferUshort(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(unsigned long* data, int nitem, int stride)
{
  cppvmTransferUlong(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(unsigned long& data, int stride)
{
  cppvmTransferUlong(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(float* data, int nitem, int stride)
{
  cppvmTransferFloat(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(float& data, int stride)
{
  cppvmTransferFloat(&data, 1, stride);
}
template<>
void
cppvmTransferTemplate(double* data, int nitem, int stride)
{
  cppvmTransferDouble(data, nitem, stride);
}
template<>
void
cppvmTransferTemplate(double& data, int stride)
{
  cppvmTransferDouble(&data, 1, stride);
}

#endif // CPPvmDoNotUseTemplates

/*@@@**************************************************************************
**********                                                           **********
*******                           END OF FILE                           *******
**********                                                           **********
******************************************************************************/

